/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 *     National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 *     Operator of Los Alamos National Laboratory.
 * EPICS BASE Versions 3.13.7
 * and higher are distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*
 *      Author: Jeffrey O. Hill
 */

#include <string>
#include <stdexcept>

#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"

#include "epicsThrowTrace.h"
#include "errlog.h"

#define epicsExportSharedSymbols
#include "iocinf.h"
#include "syncGroup.h"
#include "oldAccess.h"
#include "cac.h"
#include "sgAutoPtr.h"

CASG :: CASG ( Guard & guard, ca_client_context & client ) :
    _client ( client ), 
    _magic ( CASG_MAGIC ),
    _cleanupCompleted ( false )
{
    client.installCASG ( guard, *this );
}

CASG :: ~CASG ()
{
    if ( ! _cleanupCompleted ) {
        Guard guard ( _client.mutexRef() );
        this-> cleanup ( guard );
    }
}

void CASG :: cleanup ( Guard & guard )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    if ( ! _cleanupCompleted ) {
        if ( this->verify ( guard ) ) {
            this->reset ( guard );
            _client.uninstallCASG ( guard, *this );
            _magic = 0;
        }
        else {
            this->printFormated ( 
                "cac: attempt to destroy invalid "
                "sync group ignored\n" );
        }
        _cleanupCompleted = true;
    }
}

bool CASG::verify ( Guard & ) const
{
    return ( _magic == CASG_MAGIC );
}

/*
 * CASG::block ()
 */
int CASG::block ( 
    Guard * pcbGuard, 
    Guard & guard, 
    double timeout )
{
    epicsTime cur_time;
    epicsTime beg_time;
    double delay;
    double remaining;
    int status;

    guard.assertIdenticalMutex ( _client.mutexRef() );

    // prevent recursion nightmares by disabling blocking
    // for IO from within a CA callback. 
    if ( epicsThreadPrivateGet ( caClientCallbackThreadId ) ) {
        return ECA_EVDISALLOW;
    }

    if ( timeout < 0.0 ) {
        return ECA_TIMEOUT;
    }

    cur_time = epicsTime::getCurrent ();

    _client.flush ( guard );

    beg_time = cur_time;
    delay = 0.0;

    while ( 1 ) {
        if ( _ioPendingList.count() == 0u ) {
            status = ECA_NORMAL;
            break;
        }

        remaining = timeout - delay;
        if ( remaining <= CAC_SIGNIFICANT_DELAY ) {
            /*
             * Make sure that we take care of
             * recv backlog at least once
             */
            status = ECA_TIMEOUT;
            break;
        }

        if ( pcbGuard ) {
            Guard :: Release unguard ( guard );
            {
                Guard :: Release uncbGuard ( *pcbGuard );
                _sem.wait ( remaining );
            }
        }
        else {
            Guard :: Release unguard ( guard );
            _sem.wait ( remaining );
        }

        /*
         * force a time update 
         */
        cur_time = epicsTime::getCurrent ();

        delay = cur_time - beg_time;
    }

    this->reset ( guard );

    return status;
}

void CASG::reset ( 
    Guard & guard )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    this->destroyCompletedIO ( guard );
    this->destroyPendingIO ( guard );
}

// lock must be applied
void CASG::destroyCompletedIO ( 
    Guard & guard )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    syncGroupNotify * pNotify;
    while ( ( pNotify = _ioCompletedList.get () ) ) {
        pNotify->destroy ( guard, * this );
    }
}

void CASG::destroyPendingIO ( 
    Guard & guard )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    while ( syncGroupNotify * pNotify = _ioPendingList.first () ) {
        pNotify->cancel ( guard );
        // cancel must release the guard while 
        // canceling put callbacks so we
        // must double check list membership
        if ( pNotify->ioPending ( guard ) ) {
            _ioPendingList.remove ( *pNotify );
        }
        else {
            _ioCompletedList.remove ( *pNotify );
        }
        pNotify->destroy ( guard, *this );
    }
}

void CASG::show ( unsigned level ) const
{
    Guard guard ( _client.mutexRef () );
    this->show ( guard, level );
}

void CASG::show ( 
    Guard & guard, unsigned level ) const
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    ::printf ( "Sync Group: id=%u, magic=%u, opPend=%u\n",
        this->getId (), _magic, _ioPendingList.count () );
    if ( level ) {
        ::printf ( "\tPending" );
        tsDLIterConst < syncGroupNotify > notifyPending = 
            _ioPendingList.firstIter ();
        while ( notifyPending.valid () ) {
            notifyPending->show ( guard, level - 1u );
            notifyPending++;
        }
        ::printf ( "\tCompleted" );
        tsDLIterConst < syncGroupNotify > notifyCompleted = 
            _ioCompletedList.firstIter ();
        while ( notifyCompleted.valid () ) {
            notifyCompleted->show ( guard, level - 1u );
            notifyCompleted++;
        }
    }
}

bool CASG::ioComplete (
    Guard & guard )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    this->destroyCompletedIO ( guard );
    return _ioPendingList.count () == 0u;
}

void CASG::put ( Guard & guard, chid pChan, 
    unsigned type, arrayElementCount count, const void * pValue )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    sgAutoPtr < syncGroupWriteNotify > pNotify ( 
        guard, *this, _ioPendingList );
    pNotify = new ( _freeListWriteOP )
                    syncGroupWriteNotify ( *this, pChan );
    pNotify->begin ( guard, type, count, pValue );
    pNotify.release ();
}

void CASG::get ( Guard & guard, chid pChan, 
                unsigned type, arrayElementCount count, void *pValue )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    sgAutoPtr < syncGroupReadNotify > 
        pNotify ( guard, *this, _ioPendingList );
    pNotify = new ( _freeListReadOP )
                syncGroupReadNotify ( *this, pChan, pValue );
    pNotify->begin ( guard, type, count );
    pNotify.release ();
}

void CASG::completionNotify ( 
    Guard & guard, syncGroupNotify & notify )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    _ioPendingList.remove ( notify );
    _ioCompletedList.add ( notify );
    if ( _ioPendingList.count () == 0u ) {
        _sem.signal ();
    }
}

void CASG::recycleSyncGroupWriteNotify ( 
    Guard & guard, syncGroupWriteNotify & io )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    :: operator delete ( & io, _freeListWriteOP );
}

void CASG::recycleSyncGroupReadNotify ( 
    Guard & guard, syncGroupReadNotify & io )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    :: operator delete ( & io, _freeListReadOP );
}

int CASG :: printFormated ( const char *pformat, ... )
{
    va_list theArgs;
    int status;

    va_start ( theArgs, pformat );
    
    status = _client.varArgsPrintFormated ( pformat, theArgs );
    
    va_end ( theArgs );
    
    return status;
}

void CASG::exception ( 
    Guard & guard, 
    int status, const char * pContext, 
    const char * pFileName, unsigned lineNo )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    if ( status != ECA_CHANDESTROY ) {
        _client.exception ( 
            guard, status, pContext, pFileName, lineNo );
    }
}

void CASG::exception ( 
    Guard & guard, 
    int status, const char * pContext,
    const char * pFileName, unsigned lineNo, oldChannelNotify & chan, 
    unsigned type, arrayElementCount count, unsigned op )
{
    guard.assertIdenticalMutex ( _client.mutexRef() );
    if ( status != ECA_CHANDESTROY ) {
        _client.exception ( 
            guard, status, pContext, pFileName, 
            lineNo, chan, type, count, op );
    }
}

