/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#ifndef comBufh
#define comBufh

#include <new>
#include <cstring>

#ifdef epicsExportSharedSymbols
#   define comBufh_epicsExportSharedSymbols
#   undef epicsExportSharedSymbols
#endif

#include "epicsAssert.h"
#include "epicsMutex.h"
#include "epicsTypes.h"
#include "tsFreeList.h"
#include "tsDLList.h"
#include "osiWireFormat.h"
#include "compilerDependencies.h"

#ifdef comBufh_epicsExportSharedSymbols
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

static const unsigned comBufSize = 0x4000;

typedef epics :: Mutex :: Guard Guard;

typedef tsFreeList < class comBuf, 0x10 > comBufMemoryManager;

class wireSendAdapter { // X aCC 655
public:
    virtual size_t sendBytes ( const void * pBuf, 
        size_t nBytesInBuf, 
        const class epicsTime & currentTime ) = 0;
protected:
    virtual ~wireSendAdapter() {}
};

enum swioCircuitState { 
        swioConnected, 
        swioPeerHangup, 
        swioPeerAbort, 
        swioLinkFailure,
        swioLocalAbort
};
struct statusWireIO {
    size_t bytesCopied;
    swioCircuitState circuitState;
};

class wireRecvAdapter { // X aCC 655
public:
    virtual void recvBytes ( void * pBuf, 
        size_t nBytesInBuf, statusWireIO & ) = 0;
protected:
    virtual ~wireRecvAdapter() {}
};

class comBuf :
    public tsDLNode < comBuf > {
public:
    comBuf ();
    size_t unoccupiedBytes () const;
    size_t occupiedBytes () const;
    size_t uncommittedBytes () const;
    static size_t capacityBytes ();
    void clear ();
    size_t copyInBytes ( const void *pBuf, size_t nBytes );
    size_t push ( comBuf & );
    template < class T >
    bool push ( const T & value );
    template < class T >
    size_t push ( const T * pValue, size_t nElem );
    size_t push ( const epicsInt8 * pValue, size_t nElem );
    size_t push ( const epicsUInt8 * pValue, size_t nElem );
    size_t push ( const epicsOldString * pValue, size_t nElem );
    void commitIncomming ();
    void clearUncommittedIncomming ();
    bool copyInAllBytes ( const void *pBuf, size_t nBytes );
    size_t copyOutBytes ( void *pBuf, size_t nBytes );
    bool copyOutAllBytes ( void *pBuf, size_t nBytes );
    size_t removeBytes ( size_t nBytes );
    bool flushToWire ( wireSendAdapter &, const epicsTime & currentTime );
    void fillFromWire ( wireRecvAdapter &, statusWireIO & );
    struct popStatus {
        bool success;
        bool nowEmpty;
    };
    template < class T >
    popStatus pop ( T & );
    static void throwInsufficentBytesException ();

private:
    epicsUInt8 _buf [ comBufSize ];
    comBuf * _pNext;
    size_t _commitIndex;
    size_t _nextWriteIndex;
    size_t _nextReadIndex;
    size_t _refCount;
    void _reference ( Guard & );
    void _unreference ( Guard & );
    template < class T >
    bool push ( const T * ); // disabled
};

inline comBuf :: comBuf () : 
    _pNext ( 0 ),
    _commitIndex ( 0u ),
    _nextWriteIndex ( 0u ), 
    _nextReadIndex ( 0u ),
    _refCount ( 0u )
{
}

inline void comBuf :: clear ()
{
    _commitIndex = 0u;
    _nextWriteIndex = 0u;
    _nextReadIndex = 0u;
}

inline size_t comBuf :: unoccupiedBytes () const
{
    return sizeof ( _buf ) - _nextWriteIndex;
}

inline size_t comBuf :: occupiedBytes () const
{
    return _commitIndex - _nextReadIndex;
}

inline size_t comBuf :: uncommittedBytes () const
{
    return _nextWriteIndex - _commitIndex;
}

inline size_t comBuf :: push ( comBuf & bufIn )
{
    size_t nBytes = this->copyInBytes ( 
        & bufIn._buf[ bufIn._nextReadIndex ], 
        bufIn._commitIndex - bufIn._nextReadIndex );
    bufIn._nextReadIndex += nBytes;
    return nBytes;
}

inline size_t comBuf :: capacityBytes ()
{
    return comBufSize;
}

inline void comBuf :: fillFromWire ( 
    wireRecvAdapter & wire, statusWireIO & stat )
{
    wire.recvBytes ( 
        & _buf[_nextWriteIndex], 
        sizeof ( _buf ) - _nextWriteIndex, stat );
    if ( stat.circuitState == swioConnected ) {
        _nextWriteIndex += stat.bytesCopied;
    }
}

template < class T >
inline bool comBuf :: push ( const T & value )
{
    unsigned index = _nextWriteIndex;
    unsigned available = sizeof ( _buf ) - index;
    if ( sizeof ( value ) > available ) {
        return false;
    }
    WireSet ( value, & _buf[index] );
    _nextWriteIndex = index + sizeof ( value );
    return true;
}

inline size_t comBuf :: push ( const epicsInt8 *pValue, size_t nElem )
{
    return copyInBytes ( pValue, nElem );
}

inline size_t comBuf :: push ( const epicsUInt8 *pValue, size_t nElem )
{
    return copyInBytes ( pValue, nElem );
}

inline size_t comBuf :: push ( const epicsOldString * pValue, size_t nElem )
{
    size_t index = _nextWriteIndex;
    size_t available = sizeof ( _buf ) - index;
    size_t nBytes = sizeof ( *pValue ) * nElem;
    if ( nBytes > available ) {
        nElem = available / sizeof ( *pValue );
        nBytes = nElem * sizeof ( *pValue );
    }
    memcpy ( & _buf[ index ], pValue, nBytes );
    _nextWriteIndex = index + nBytes;
    return nElem;
}

template < class T >
size_t comBuf :: push ( const T * pValue, size_t nElem )
{
    size_t index = _nextWriteIndex;
    size_t available = sizeof ( _buf ) - index;
    size_t nBytes = sizeof ( *pValue ) * nElem;
    if ( nBytes > available ) {
        nElem = available / sizeof ( *pValue );
    }
    for ( unsigned i = 0u; i < nElem; i++ ) {
        // allow native floating point formats to be converted to IEEE
        WireSet( pValue[i], & _buf[index] );
        index += sizeof ( *pValue );
    }
    _nextWriteIndex = index;
    return nElem;
}

inline void comBuf :: commitIncomming ()
{
    _commitIndex = _nextWriteIndex;
}

inline void comBuf :: clearUncommittedIncomming ()
{
    _nextWriteIndex = _commitIndex;
}

inline bool comBuf :: copyInAllBytes ( const void *pBuf, size_t nBytes )
{
    size_t index = _nextWriteIndex;
    size_t available = sizeof ( _buf ) - index;
    if ( nBytes <= available ) {
        memcpy ( & _buf[index], pBuf, nBytes );
        _nextWriteIndex = index + nBytes;
        return true;
    }
    return false;
}

inline size_t comBuf :: copyInBytes ( const void * pBuf, size_t nBytes )
{
    unsigned index = _nextWriteIndex;
    unsigned available = sizeof ( _buf ) - index;
    if ( nBytes > available ) {
        nBytes = available;
    }
    memcpy ( & _buf[index], pBuf, nBytes );
    _nextWriteIndex = index + nBytes;
    return nBytes;
}

inline bool comBuf :: copyOutAllBytes ( void * pBuf, size_t nBytes )
{
    size_t index = _nextReadIndex;
    size_t occupied = _commitIndex - index;
    if ( nBytes <= occupied ) {
        memcpy ( pBuf, &_buf[index], nBytes);
        _nextReadIndex = index + nBytes;
        return true;
    }
    return false;
}

inline size_t comBuf :: copyOutBytes ( void *pBuf, size_t nBytes )
{
    size_t index = _nextReadIndex;
    size_t occupied = _commitIndex - index;
    if ( nBytes > occupied ) {
        nBytes = occupied;
    }
    memcpy ( pBuf, &_buf[index], nBytes);
    _nextReadIndex = index + nBytes;
    return nBytes;
}

inline size_t comBuf :: removeBytes ( size_t nBytes )
{
    size_t index = _nextReadIndex;
    size_t occupied = _commitIndex - index;
    if ( nBytes > occupied ) {
        nBytes = occupied;
    }
    _nextReadIndex = index + nBytes;
    return nBytes;
}

template < class T >
comBuf :: popStatus comBuf :: pop ( T & returnVal )
{
    size_t nrIndex = _nextReadIndex;
    size_t popIndex = nrIndex + sizeof ( returnVal );
    size_t cIndex = _commitIndex;
    popStatus status;
    status.success = true;
    status.nowEmpty = false;
    if ( popIndex >= cIndex ) {
        if ( popIndex == cIndex ) {
            status.nowEmpty = true;
        }
        else {
            status.success = false;
            return status;
        }
    }
    returnVal = WireGet < T > ( & _buf[ nrIndex ] );
    _nextReadIndex = popIndex;
    return status;
}

#endif // ifndef comBufh
