/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*  
 *
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *
 *  Author Jeffrey O. Hill
 *  johill@lanl.gov
 */

#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"

#include "iocinf.h"
#include "virtualCircuit.h"

comQueRecv :: comQueRecv ( 
    epics :: Mutex & mutex, 
        comBufMemoryManager & comBufMemoryManagerIn ): 
    _mutex ( mutex ),
    _comBufMemMgr ( comBufMemoryManagerIn ), 
    _nBytesPending ( 0u ) 
{
}

comQueRecv :: ~comQueRecv ()
{
    Guard guard ( _mutex );
    this->clear ( guard );
}

void comQueRecv :: clear ( Guard & guard )
{
    guard.assertIdenticalMutex ( _mutex );
    comBuf *pBuf;
    while ( ( pBuf = _bufs.get () ) ) {
        _comBufMemMgr.deleteProduct ( *pBuf );
    }
    _nBytesPending = 0u;
}

size_t comQueRecv :: copyOutBytes ( 
     Guard & guard, epicsInt8 *pBuf, unsigned nBytes )
{
    guard.assertIdenticalMutex ( _mutex );
    size_t totalBytes = 0u;
    do {
        comBuf * pComBuf = _bufs.first ();
        if ( ! pComBuf ) {
            _nBytesPending -= totalBytes;
            return totalBytes;
        }
        totalBytes += pComBuf->copyOutBytes ( 
            &pBuf[totalBytes], nBytes - totalBytes );
        if ( pComBuf->occupiedBytes () == 0u ) {
            _bufs.remove ( *pComBuf );
            _comBufMemMgr.deleteProduct ( *pComBuf );
        }
    }
    while ( totalBytes < nBytes );
    _nBytesPending -= totalBytes;
    return totalBytes;
}

size_t comQueRecv :: removeBytes ( 
     Guard & guard, unsigned nBytes )
{
    guard.assertIdenticalMutex ( _mutex );
    size_t totalBytes = 0u;
    size_t bytesLeft = nBytes;
    while ( bytesLeft ) {
        comBuf * pComBuf = _bufs.first ();
        if ( ! pComBuf ) {
            _nBytesPending -= totalBytes;
            return totalBytes;
        }
        size_t nBytesThisTime = 
            pComBuf->removeBytes ( bytesLeft );
        if ( pComBuf->occupiedBytes () == 0u ) {
            _bufs.remove ( *pComBuf );
            _comBufMemMgr.deleteProduct ( *pComBuf );
        }
        if ( nBytesThisTime == 0u) {
            break;
        }
        totalBytes += nBytesThisTime;
        bytesLeft = nBytes - totalBytes;
    }
    _nBytesPending -= totalBytes;
    return totalBytes;
}

void comQueRecv :: popString ( 
     Guard & guard, epicsOldString *pStr )
{
    guard.assertIdenticalMutex ( _mutex );
    for ( unsigned i = 0u; i < sizeof ( *pStr ); i++ ) {
        pStr[0][i] = this->popInt8 ( guard );
    }
}

void comQueRecv :: pushLastComBufReceived ( 
     Guard & guard, comBuf & bufIn )
   
{
    guard.assertIdenticalMutex ( _mutex );
    bufIn.commitIncomming ();
    comBuf * pComBuf = _bufs.last ();
    if ( pComBuf ) {
        if ( pComBuf->unoccupiedBytes() ) {
            _nBytesPending += pComBuf->push ( bufIn );
            pComBuf->commitIncomming ();
        }
    }
    size_t bufBytes = bufIn.occupiedBytes();
    if ( bufBytes ) {
        _nBytesPending += bufBytes;
        _bufs.add ( bufIn );
    }
    else {
        _comBufMemMgr.deleteProduct ( bufIn );
    }
}

// 1) split between buffers expected to run slower
// 2) using canonical unsigned tmp avoids ANSI C conversions to int
// 3) cast required because sizeof(unsigned) >= sizeof(epicsUInt32)
epicsUInt16 comQueRecv :: _multiBufferPopUInt16 ( 
    Guard & guard )
{
    guard.assertIdenticalMutex ( _mutex );
    epicsUInt16 tmp;
    if ( this->occupiedBytes ( guard ) >= sizeof ( tmp ) ) {
        unsigned byte1 = this->popUInt8 ( guard );
        unsigned byte2 = this->popUInt8 ( guard );
        tmp = static_cast < epicsUInt16 > ( ( byte1 << 8u ) | byte2 );
    }
    else {
        comBuf :: throwInsufficentBytesException ();
        tmp = 0u;
    }
    return tmp;
}

// 1) split between buffers expected to run slower
// 2) using canonical unsigned temporary avoids ANSI C conversions to int
// 3) cast required because sizeof(unsigned) >= sizeof(epicsUInt32)
epicsUInt32 comQueRecv :: _multiBufferPopUInt32 (
    Guard & guard )
{
    guard.assertIdenticalMutex ( _mutex );
    epicsUInt32 tmp;
    if ( this->occupiedBytes ( guard ) >= sizeof (tmp) ) {
        // 1) split between buffers expected to run slower
        // 2) using canonical unsigned temporary avoids 
        // ANSI C conversions to int
        // 3) cast required because sizeof(unsigned) 
        // >= sizeof(epicsUInt32)
        unsigned byte1 = this->popUInt8 ( guard );
        unsigned byte2 = this->popUInt8 ( guard );
        unsigned byte3 = this->popUInt8 ( guard );
        unsigned byte4 = this->popUInt8 ( guard );
        tmp = static_cast <epicsUInt32>
            ( ( byte1 << 24u ) | ( byte2 << 16u ) | //X aCC 392
              ( byte3 << 8u ) | byte4 );
    }
    else {
        comBuf :: throwInsufficentBytesException ();
        tmp = 0u; // avoid compiler warnings
    }
    return tmp;
}

void comQueRecv :: _removeAndDestroyBuf ( 
    Guard & guard, comBuf & buf )
{
    guard.assertIdenticalMutex ( _mutex );
    _bufs.remove ( buf );
    _comBufMemMgr.deleteProduct ( buf );
}

epicsUInt8 comQueRecv :: popUInt8 ( Guard & guard ) 
{
    guard.assertIdenticalMutex ( _mutex );
    comBuf * pComBuf = _bufs.first ();
    if ( ! pComBuf ) {
        comBuf::throwInsufficentBytesException ();
    }
    epicsUInt8 tmp = '\0';
    comBuf::popStatus status = pComBuf->pop ( tmp );
    if ( ! status.success ) {
        comBuf::throwInsufficentBytesException ();
    }
    if ( status.nowEmpty ) {
        _removeAndDestroyBuf ( guard, *pComBuf );
    }
    _nBytesPending--;
    return tmp;
}

epicsUInt16 comQueRecv :: popUInt16 ( Guard & guard )
{
    guard.assertIdenticalMutex ( _mutex );
    comBuf * pComBuf = _bufs.first ();
    if ( ! pComBuf ) {
        comBuf::throwInsufficentBytesException ();
    }
    // try first for all in one buffer efficent version
    epicsUInt16 tmp = 0;
    comBuf::popStatus status = pComBuf->pop ( tmp );
    if ( status.success ) {
        _nBytesPending -= sizeof ( epicsUInt16 );
        if ( status.nowEmpty ) {
            _removeAndDestroyBuf ( guard, *pComBuf );
        }
        return tmp;
    }
    return _multiBufferPopUInt16 ( guard );
}

epicsUInt32 comQueRecv :: popUInt32 ( Guard & guard )
{
    guard.assertIdenticalMutex ( _mutex );
    comBuf *pComBuf = _bufs.first ();
    if ( ! pComBuf ) {
        comBuf::throwInsufficentBytesException ();
    }
    // try first for all in one buffer efficent version
    epicsUInt32 tmp = 0;
    comBuf::popStatus status = pComBuf->pop ( tmp );
    if ( status.success ) {
        _nBytesPending -= sizeof ( epicsUInt32 );
        if ( status.nowEmpty ) {
            _removeAndDestroyBuf ( guard, *pComBuf );
        }
        return tmp;
    }
    return _multiBufferPopUInt32 ( guard );
}

bool comQueRecv :: popOldMsgHeader ( 
    Guard & guard, caHdrLargeArray & msg )
{
    guard.assertIdenticalMutex ( _mutex );
    // try first for all in one buffer efficent version
    comBuf * pComBuf = _bufs.first ();
    if ( ! pComBuf ) {
        return false;
    }
    unsigned avail = pComBuf->occupiedBytes ();
    if ( avail >= sizeof ( caHdr ) ) {
        pComBuf->pop ( msg.m_cmmd );
        ca_uint16_t smallPostsize = 0;
        pComBuf->pop ( smallPostsize );
        msg.m_postsize = smallPostsize;
        pComBuf->pop ( msg.m_dataType );
        ca_uint16_t smallCount = 0;
        pComBuf->pop ( smallCount );
        msg.m_count = smallCount;
        pComBuf->pop ( msg.m_cid );
        pComBuf->pop ( msg.m_available );
        _nBytesPending -= sizeof ( caHdr );
        if ( avail == sizeof ( caHdr ) ) {
            _removeAndDestroyBuf ( guard, *pComBuf );
        }
        return true;
    }
    else if ( this->occupiedBytes ( guard ) >= sizeof ( caHdr ) ) {
        msg.m_cmmd = this->popUInt16 ( guard );
        msg.m_postsize = this->popUInt16 ( guard );
        msg.m_dataType = this->popUInt16 ( guard );
        msg.m_count = this->popUInt16 ( guard );
        msg.m_cid = this->popUInt32 ( guard );
        msg.m_available = this->popUInt32 ( guard );
        return true;
    }
    else {
        return false;
    }
}

