/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *  Author: Jeff Hill
 */

#include <string>
#include <stdexcept>

#include "errlog.h"

#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"

#include "epicsThrowTrace.h"

#define epicsExportSharedSymbols
#include "iocinf.h"
#include "nciu.h"
#include "cac.h"
#include "db_access.h" // for dbf_type_to_text
#include "caerr.h"

netSubscription::netSubscription ( 
        privateInterfaceForIO & chanIn, 
        unsigned typeIn, arrayElementCount countIn, 
        unsigned maskIn, cacStateNotify & notifyIn ) :
    count ( countIn ), privateChanForIO ( chanIn ),
    notify ( notifyIn ), type ( typeIn ), mask ( maskIn ),
    subscribed ( false )
{
    if ( ! dbr_type_is_valid ( typeIn ) ) {
        epicsThrowTrace cacChannel::badType ();
    }
    if ( this->mask == 0u ) {
        epicsThrowTrace cacChannel::badEventSelection ();
    }
}

void netSubscription::destroy ( 
    Guard & guard, cacRecycle & recycle )
{
    recycle.destroySubscription ( guard, *this );
}

class netSubscription * netSubscription::isSubscription ()
{
    return this;
}

void netSubscription::show ( unsigned /* level */ ) const
{
    ::printf ( "event subscription IO at %p, type %s, element count %lu, mask %u\n", 
        static_cast < const void * > ( this ), 
        dbf_type_to_text ( static_cast < int > ( this->type ) ), 
        this->count, this->mask );
}

void netSubscription::show ( 
    Guard &, unsigned level ) const
{
    this->show ( level );
}

void netSubscription::completion ( 
    Guard & guard, cacRecycle & )
{
    errlogPrintf ( "subscription update w/o data ?\n" );
}

void netSubscription::exception ( 
    Guard & guard, cacRecycle & recycle, 
    int status, const char * pContext )
{
    if ( status == ECA_DISCONN ) {
        this->subscribed = false;
    }
    if ( status == ECA_CHANDESTROY ) {
        this->privateChanForIO.ioCompletionNotify ( guard, *this );
        this->notify.exception ( 
            guard, status, pContext, UINT_MAX, 0 );
        recycle.destroySubscription ( guard, *this );
    }
    else {
        // guard.assertIdenticalMutex ( this->mutex );
        if ( this->privateChanForIO.connected ( guard )  ) {
            this->notify.exception ( 
                guard, status, pContext, UINT_MAX, 0 );
        }
    }
}

void netSubscription::exception ( 
    Guard & guard, 
    cacRecycle & recycle, int status, const char * pContext, 
    unsigned typeIn, arrayElementCount countIn )
{
    if ( status == ECA_DISCONN ) {
        this->subscribed = false;
    }
    if ( status == ECA_CHANDESTROY ) {
        this->privateChanForIO.ioCompletionNotify ( guard, *this );
        this->notify.exception ( 
            guard, status, pContext, UINT_MAX, 0 );
        recycle.destroySubscription ( guard, *this );
    }
    else {
        //guard.assertIdenticalMutex ( this->mutex );
        if ( this->privateChanForIO.connected ( guard ) ) {
            this->notify.exception ( 
                guard, status, pContext, typeIn, countIn );
        }
    }
}

void netSubscription::completion ( 
    Guard & guard, cacRecycle &,
    unsigned typeIn, arrayElementCount countIn, 
    const void * pDataIn )
{
    // guard.assertIdenticalMutex ( this->mutex );
    if ( this->privateChanForIO.connected ( guard )  ) {
        this->notify.current ( 
            guard, typeIn, countIn, pDataIn );
    }
}

void netSubscription::subscribeIfRequired (
    Guard & guard, nciu & chan )
{
    if ( ! this->subscribed ) {
        chan.getPIIU(guard)->subscriptionRequest ( 
            guard, chan, *this );
        this->subscribed = true;
    }
}

void netSubscription::unsubscribeIfRequired ( 
    Guard & guard, nciu & chan )
{
    if ( this->subscribed ) {
        chan.getPIIU(guard)->subscriptionCancelRequest ( 
            guard, chan, *this );
        this->subscribed = false;
    }
}

void netSubscription::forceSubscriptionUpdate (
    Guard & guard, nciu & chan )
{
    chan.getPIIU(guard)->subscriptionUpdateRequest ( 
        guard, chan, *this );
}






