/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*  
 *
 *                              
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *                                  
 *  Copyright, 1986, The Regents of the University of California.
 *                                  
 *           
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 *	505 665 1831
 */

#ifndef syncGrouph  
#define syncGrouph

#ifdef epicsExportSharedSymbols
#   define syncGrouph_restore_epicsExportSharedSymbols
#   undef epicsExportSharedSymbols
#endif

#include "tsDLList.h"
#include "tsFreeList.h"
#include "resourceLib.h"
#include "epicsEvent.h"
#include "compilerDependencies.h"

#ifdef syncGrouph_restore_epicsExportSharedSymbols
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "cadef.h"
#include "cacIO.h"

static const unsigned CASG_MAGIC = 0xFAB4CAFE;

// used to control access to CASG's recycle routines which
// should only be indirectly invoked by CASG when its lock
// is applied
class casgRecycle {             // X aCC 655
public:
    virtual void recycleSyncGroupWriteNotify ( 
        Guard &, class syncGroupWriteNotify & io ) = 0;
    virtual void recycleSyncGroupReadNotify ( 
        Guard &, class syncGroupReadNotify & io ) = 0;
protected:
    virtual ~casgRecycle () {}
};

class syncGroupNotify : 
    public tsDLNode < syncGroupNotify > {
public:
    virtual void destroy (  
        Guard & guard, 
        casgRecycle & ) = 0;
    virtual bool ioPending (  
        Guard & guard ) = 0;
    virtual void cancel (
        Guard & guard ) = 0;
    virtual void show ( 
        Guard &, 
        unsigned level ) const = 0;
protected:
    virtual ~syncGroupNotify () {}
};

class syncGroupReadNotify : 
    public syncGroupNotify, public cacReadNotify {
public:
    syncGroupReadNotify ( struct CASG & sgIn, chid, void * pValueIn );
    void destroy (  
        Guard & guard, 
        casgRecycle & );
    bool ioPending (  
        Guard & guard );
    void begin (  Guard &, 
        unsigned type, arrayElementCount count );
    void cancel (
        Guard & guard );
    void show (  Guard &, unsigned level ) const;
protected:
    ~syncGroupReadNotify ();
private:
    chid chan;
    struct CASG & sg;
    void * pValue;
    const unsigned _magic;
    cacChannel::ioid id;
    bool idIsValid;
    bool ioComplete;
    void completion (
        Guard &, unsigned type, 
        arrayElementCount count, const void * pData );
    void exception (
        Guard &, int status, 
        const char * pContext, unsigned type, 
        arrayElementCount count );
	syncGroupReadNotify ( const syncGroupReadNotify & );
	syncGroupReadNotify & operator = ( const syncGroupReadNotify & );
};

class syncGroupWriteNotify : 
    public syncGroupNotify, public cacWriteNotify {
public:
    syncGroupWriteNotify  ( struct CASG &, chid );
    void destroy (  
        Guard & guard, casgRecycle & );
    bool ioPending (  
        Guard & guard );
    void begin (  Guard &, unsigned type, 
        arrayElementCount count, const void * pValueIn );
    void cancel (
        Guard & guard );
    void show (  Guard &, unsigned level ) const;
protected:
    ~syncGroupWriteNotify ();
private:
    chid chan;
    struct CASG & sg;
    const unsigned magic;
    cacChannel::ioid id;
    bool idIsValid;
    bool ioComplete;
    void completion ( Guard & );
    void exception ( 
        Guard &, int status, const char *pContext, 
		unsigned type, arrayElementCount count );
	syncGroupWriteNotify ( const syncGroupWriteNotify & );
	syncGroupWriteNotify & operator = ( const syncGroupWriteNotify & );
};

struct ca_client_context;

template < class T > class sgAutoPtr;

struct CASG : 
    public chronIntIdRes < CASG >, private casgRecycle {
public:
    CASG ( Guard &, ca_client_context & cacIn );
    void cleanup ( Guard & guard );
    ~CASG ();
    bool ioComplete ( Guard & guard );
    bool verify ( Guard & ) const;
    int block ( Guard * pcbGuard, 
        Guard & guard, double timeout );
    void reset ( Guard & guard );
    void show ( Guard &, unsigned level ) const;
    void show ( unsigned level ) const;
    void get ( Guard &, chid pChan, 
        unsigned type, arrayElementCount count, void * pValue );
    void put ( Guard &, chid pChan, 
        unsigned type, arrayElementCount count, 
        const void * pValue );
    void completionNotify ( 
        Guard &, syncGroupNotify & );
    int printFormated ( const char * pFormat, ... );
    void exception ( 
         Guard &, int status, const char * pContext, 
        const char * pFileName, unsigned lineNo );
    void exception ( 
         Guard &, int status, const char * pContext,
        const char * pFileName, unsigned lineNo,
        oldChannelNotify & chan, unsigned type, 
        arrayElementCount count, unsigned op );
private:
    typedef tsFreeList < class syncGroupReadNotify, 128 > 
                FreeListReadNotify;
    typedef tsFreeList < class syncGroupWriteNotify, 128 > 
                FreeListWriteNotify;
    FreeListReadNotify _freeListReadOP;
    FreeListWriteNotify _freeListWriteOP;
    tsDLList < syncGroupNotify > _ioPendingList;
    tsDLList < syncGroupNotify > _ioCompletedList;
    epicsEvent _sem;
    ca_client_context & _client;
    unsigned _magic;
    bool _cleanupCompleted;
    void recycleSyncGroupWriteNotify (  
        Guard &, syncGroupWriteNotify & io );
    void recycleSyncGroupReadNotify (  
        Guard &, syncGroupReadNotify & io );

    void destroyPendingIO ( 
        Guard & guard );
    void destroyCompletedIO ( 
        Guard & guard );

	CASG ( const CASG & ); //disabled
	CASG & operator = ( const CASG & ); // disabled

    friend class sgAutoPtr < syncGroupWriteNotify >;
    friend class sgAutoPtr < syncGroupReadNotify >;
};

class boolFlagManager {
public:
    boolFlagManager ( bool & flag );
    ~boolFlagManager ();
    void release ();
private:
    bool * pBool;
};

inline boolFlagManager::boolFlagManager ( bool & flagIn ) :
    pBool ( & flagIn )
{
    *this->pBool = true;
}

inline boolFlagManager::~boolFlagManager ()
{
    if ( this->pBool ) {
        *this->pBool = false;
    }
}

inline void boolFlagManager::release ()
{
    this->pBool = 0;
}

inline bool syncGroupWriteNotify::ioPending (  
    Guard & /* guard */ )
{
    return ! this->ioComplete;
}

inline bool syncGroupReadNotify::ioPending (  
    Guard & /* guard */ )
{
    return ! this->ioComplete;
}

#endif // ifdef syncGrouph
