/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*
 *      Author: Jeffrey O. Hill
 *              hill@luke.lanl.gov
 *              (505) 665 1831
 */

#include <string>
#include <stdexcept>

#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"

#include "epicsThrowTrace.h"
#include "errlog.h"

#define epicsExportSharedSymbols
#include "iocinf.h"
#include "syncGroup.h"
#include "oldAccess.h"

syncGroupReadNotify::syncGroupReadNotify ( 
    CASG & sgIn, chid pChan, void * pValueIn ) :
    chan ( pChan ), sg ( sgIn ), pValue ( pValueIn ), 
    _magic ( CASG_MAGIC ), id ( 0u ), 
    idIsValid ( false ), ioComplete ( false )
{
}

void syncGroupReadNotify::begin ( 
    Guard & guard, 
    unsigned type, arrayElementCount count )
{
    this->chan->eliminateExcessiveSendBacklog ( guard );
    this->ioComplete = false;
    boolFlagManager mgr ( this->idIsValid );
    this->chan->read ( guard, type, count, *this, &this->id );
    mgr.release ();
}

void syncGroupReadNotify::cancel ( 
    Guard & guard )
{
    if ( this->idIsValid ) {
        this->chan->ioCancel ( guard, this->id );
        this->idIsValid = false;
    }
}

void syncGroupReadNotify :: destroy ( 
    Guard & guard, casgRecycle & recycle )
{
    this->~syncGroupReadNotify ();
    recycle.recycleSyncGroupReadNotify ( guard, *this );
}

syncGroupReadNotify :: ~syncGroupReadNotify ()
{
    assert ( ! this->idIsValid );
}

void syncGroupReadNotify :: completion (
    Guard & guard, unsigned type, 
    arrayElementCount count, const void * pData )
{
    if ( _magic != CASG_MAGIC ) {
        this->sg.printFormated ( 
            "cac: sync group io_complete(): "
            "bad sync grp op magic number?\n" );
        return;
    }

    if ( this->pValue ) {
        size_t size = dbr_size_n ( type, count );
        memcpy ( this->pValue, pData, size );
    }
    this->sg.completionNotify ( guard, *this );
    this->idIsValid = false;
    this->ioComplete = true;
}

void syncGroupReadNotify::exception (
    Guard & guard, int status, const char * pContext, 
    unsigned type, arrayElementCount count )
{
    if ( _magic != CASG_MAGIC ) {
        this->sg.printFormated ( 
            "cac: sync group io_complete(): "
            "bad sync grp op magic number?\n" );
        return;
    }
    this->idIsValid = false;
    this->sg.exception ( guard, status, pContext, 
        __FILE__, __LINE__, *this->chan, type, count, CA_OP_GET );
    //
    // This notify is left installed at this point as a place holder indicating that
    // all requests have not been completed. This notify is not uninstalled until
    // CASG::block() times out or unit they call CASG::reset().
    //
}

void syncGroupReadNotify::show ( 
    Guard & guard, unsigned level ) const
{
    ::printf ( "pending sg read op: pVal=%p\n", this->pValue );
    if ( level > 0u ) {
        ::printf ( "pending sg op: magic=%u sg=%p\n",
            _magic, static_cast < void * > ( & this->sg ) );
    }
}



