/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 *     National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 *     Operator of Los Alamos National Laboratory.
 * EPICS BASE Versions 3.13.7
 * and higher are distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*  
 *
 *                              
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *                                  
 *  Copyright, 1986, The Regents of the University of California.
 *                                  
 *           
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 *	505 665 1831
 */

#ifndef udpiiuh
#define udpiiuh

#ifdef epicsExportSharedSymbols
#   define udpiiuh_epicsExportSharedSymbols
#   undef epicsExportSharedSymbols
#endif

#include "osiSock.h"
#include "epicsThread.h"
#include "epicsMemory.h"
#include "epicsTime.h"
#include "tsDLList.h"

#ifdef udpiiuh_epicsExportSharedSymbols
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "netiiu.h"
#include "searchTimer.h"
#include "disconnectGovernorTimer.h"
#include "repeaterSubscribeTimer.h"
#include "SearchDest.h"

extern "C" void cacRecvThreadUDP ( void *pParam );

epicsShareFunc void epicsShareAPI caStartRepeaterIfNotInstalled ( 
    unsigned repeaterPort );
epicsShareFunc void epicsShareAPI caRepeaterRegistrationMessage ( 
    SOCKET sock, const osiSockAddr & local, unsigned attemptNumber );
extern "C" epicsShareFunc void caRepeaterThread ( 
    void * pDummy );
epicsShareFunc void ca_repeater ( void );

class cac;
class cacContextNotify;

class udpRecvThread : 
        private epicsThreadRunable {
public:
    udpRecvThread ( 
        class udpiiu & iiuIn, cacContextNotify &, epics :: Mutex &,
        const char * pName, unsigned stackSize, unsigned priority );
    virtual ~udpRecvThread ();
    void start ();
    bool exitWait ( double delay );
    void show ( unsigned level ) const;
private:
    class udpiiu & iiu;
    epics :: Mutex & cbMutex;
    cacContextNotify & ctxNotify;
    epicsThread thread;
    void run();
};

static const double minRoundTripEstimate = 32e-3; // seconds
static const double maxRoundTripEstimate = 30; // seconds
static const double maxSearchPeriodDefault = 5.0 * 60.0; // seconds
static const double maxSearchPeriodLowerLimit = 60.0; // seconds
static const double beaconAnomalySearchPeriod = 5.0; // seconds

class udpiiu : 
    private netiiu, 
    private searchTimerNotify, 
    private disconnectGovernorNotify,
    private repeaterTimerNotify {
public:
    udpiiu ( 
        Guard & cacGuard,
        class epicsTimerQueueActive &, 
        epics :: Mutex & callbackControl, 
        epics :: Mutex & mutualExclusion, 
        cacContextNotify &,
        class cac &,
        unsigned port,
        tsDLList < SearchDest > & );
    virtual ~udpiiu ();
    void installNewChannel ( 
        Guard &, nciu &, netiiu * & );
    void installDisconnectedChannel ( 
        Guard &, nciu & );
    void beaconAnomalyNotify ( 
        Guard & guard );
    void shutdown ( Guard & cbGuard, 
        Guard & guard );
    void show ( unsigned level ) const;
    
    // exceptions
    class noSocket {};

private:
    class SearchDestUDP :
        public SearchDest {
    public:
        SearchDestUDP ( const osiSockAddr &, udpiiu & );
        void searchRequest ( 
            Guard &, const char * pBuf, size_t bufLen );
        void show ( 
            Guard &, unsigned level ) const;
    private:
        osiSockAddr _destAddr;
        udpiiu & _udpiiu;
    };
    class SearchRespCallback : 
        public SearchDest :: Callback {
    public:
        SearchRespCallback ( udpiiu & );
        void notify (
            const caHdr &, const void * pPayload,
            const osiSockAddr &, const epicsTime & );
        void show ( 
            Guard &, unsigned level ) const;
    private:
        udpiiu & _udpiiu;
    };
    char xmitBuf [MAX_UDP_SEND];   
    char recvBuf [MAX_UDP_RECV];
    udpRecvThread recvThread;
    repeaterSubscribeTimer repeaterSubscribeTmr;
    disconnectGovernorTimer govTmr;
    tsDLList < SearchDest > _searchDestList;
    osiSockAddr repeaterAddress;
    double maxPeriod;
    double rtteMean;
    double rtteMeanDev;
    cac & cacRef;
    mutable epics :: Mutex & cbMutex;
    mutable epics :: Mutex & cacMutex;
    epics_auto_ptr < epics_auto_ptr < class searchTimer >, 
				eapt_array > ppSearchTmr;
    size_t nBytesInXmitBuf;
    size_t nTimers;
    size_t beaconAnomalyTimerIndex;
    ca_uint32_t sequenceNumber;
    ca_uint32_t lastReceivedSeqNo;
    SOCKET sock;
    ca_uint16_t repeaterPort;
    ca_uint16_t serverPort;
    ca_uint16_t localPort;
    bool shutdownCmd;
    bool lastReceivedSeqNoIsValid;

    bool wakeupMsg ();

    void postMsg ( 
            const osiSockAddr & net_addr, 
            char *pInBuf, arrayElementCount blockSize,
            const epicsTime & currenTime );

    bool pushDatagramMsg ( Guard &, 
        const caHdr & hdr, const void * pExt, 
        ca_uint16_t extsize);

    typedef bool ( udpiiu::*pProtoStubUDP ) ( 
        const caHdr &, 
        const osiSockAddr &, const epicsTime & );

    // UDP protocol dispatch table
    static const pProtoStubUDP udpJumpTableCAC[];

    // UDP protocol stubs
    bool versionAction ( 
        const caHdr &, 
        const osiSockAddr &, const epicsTime & );
    bool badUDPRespAction ( 
        const caHdr &msg, 
        const osiSockAddr & netAddr, const epicsTime & );
    bool searchRespAction ( 
        const caHdr &msg, 
        const osiSockAddr & net_addr, const epicsTime & );
    bool exceptionRespAction ( 
        const caHdr &msg, 
        const osiSockAddr & net_addr, const epicsTime & );
    bool beaconAction ( 
        const caHdr &msg, 
        const osiSockAddr & net_addr, const epicsTime & );
    bool notHereRespAction ( 
        const caHdr &msg, 
        const osiSockAddr & net_addr, const epicsTime & );
    bool repeaterAckAction ( 
        const caHdr &msg, 
        const osiSockAddr & net_addr, const epicsTime & );

    // netiiu stubs
    unsigned getHostName ( 
        Guard &, char * pBuf, 
        unsigned bufLength ) const throw ();
    const char * pHostName (
        Guard & ) const throw (); 
        bool ca_v41_ok (
        Guard & ) const;
    bool ca_v42_ok (
        Guard & ) const;
    size_t requestMessageBytesPending ( 
        Guard & mutualExclusionGuard );
    void flush ( 
        Guard & mutualExclusionGuard );
    void writeRequest ( 
        Guard &, nciu &, 
        unsigned type, arrayElementCount nElem, 
        const void * pValue );
    void writeNotifyRequest ( 
        Guard &, 
        nciu &, netWriteNotifyIO &, 
        unsigned type, arrayElementCount nElem, 
        const void * pValue );
    void readNotifyRequest ( 
        Guard &, nciu &, 
        netReadNotifyIO &, unsigned type, 
        arrayElementCount nElem );
    void clearChannelRequest ( 
        Guard &, 
        ca_uint32_t sid, ca_uint32_t cid );
    void subscriptionRequest ( 
        Guard &, 
        nciu &, netSubscription & );
    void subscriptionUpdateRequest ( 
        Guard &, 
        nciu &, netSubscription & );
    void subscriptionCancelRequest ( 
        Guard &, 
        nciu & chan, netSubscription & subscr );
    void flushRequest ( 
        Guard & );
    void requestRecvProcessPostponedFlush (
        Guard & );
        osiSockAddr getNetworkAddress (
        Guard & ) const;
    void uninstallChan ( 
        Guard &, nciu & );
    void uninstallChanDueToSuccessfulSearchResponse ( 
        Guard &, nciu &, 
    const class epicsTime & currentTime );
        double receiveWatchdogDelay (
        Guard & ) const;
    bool searchMsg (
        Guard &, ca_uint32_t id, 
            const char * pName, unsigned nameLength );

    // searchTimerNotify stubs
    double getRTTE ( Guard & ) const;
    void updateRTTE ( Guard &, double rtte );
    bool pushVersionMsg ();
    void boostChannel ( 
        Guard & guard, nciu & chan );
    void noSearchRespNotify ( 
        Guard &, nciu & chan, unsigned index );
    bool datagramFlush ( 
        Guard &, const epicsTime & currentTime );
    ca_uint32_t datagramSeqNumber ( 
        Guard & ) const;

    // disconnectGovernorNotify
    void govExpireNotify ( 
        Guard &, nciu & );

    // repeaterTimerNotify
    void repeaterRegistrationMessage ( 
        unsigned attemptNumber );
    int printFormated ( 
        Guard & callbackControl, 
        const char * pformat, ... );

	udpiiu ( const udpiiu & );
	udpiiu & operator = ( const udpiiu & );

    friend class udpRecvThread;
#   if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
        friend class SearchDestUDP;
        friend class SearchRespCallback;
#   endif
};

#endif // udpiiuh

