
// Author: Jeff HIll

#include <climits>

#include "BaseString.h"

#define epicsExportSharedSymbols
#include "caHdrLargeArray.h"
#include "ByteSeqOut.h"
#include "caIntfIO.h"

namespace ca {

ByteSeqOut :: 
    ByteSeqOut ( const Ptr < VecAllocator > & pVecAllocator ) :
    m_pVecAllocator ( pVecAllocator ),
    m_iteratorActive ( false )
{
    if ( m_pVecAllocator.isNill () ) {
        throw std :: invalid_argument ( 
            "ByteSeqOut ctor: nill vec allocator supplied" );
    }
}

ByteSeqOut :: 
    ~ByteSeqOut ()
{
    assert ( ! m_iteratorActive );
    m_clearAll ();
}

#if 0
Vec < const epicsUInt8 > 
    ByteSeqOut :: 
        next ()
{
    if ( ! m_pCurrent.isValid () ) {
        m_pCurrent = m_extend ();
    }
    tsDLIter < ListNodeComBuf > pCur = m_pCurrent;
    m_pCurrent++;
    return pCur->getRef ();
}
#endif

void ByteSeqOut :: 
    m_clearAll ()
{
    while ( ListNodeComBuf * pOut = m_bufList.get () ) {
        m_freeListListNodeComBuf.deleteProduct ( *pOut );
    }
}

void ByteSeqOut :: 
    flush ( Writer & writer )
{
    while ( ListNodeComBuf * pBuf = m_bufList.get () ) {
        if ( pBuf == m_pFirstNodeUncommitted.pointer () ) {
            size_t nElemCur = pBuf->nElem ();
            size_t nElemCurUncommitted = 
                m_vecFirstUncommitted.nElem ();
            assert ( nElemCur >= nElemCurUncommitted );
            pBuf->flush ( writer, nElemCur - nElemCurUncommitted );
            m_freeListListNodeComBuf.deleteProduct ( *pBuf );
            break;
        }
        pBuf->flush ( writer );
        m_freeListListNodeComBuf.deleteProduct ( *pBuf );
    }
}

tsDLIter < ByteSeqOut :: ListNodeComBuf > 
    ByteSeqOut :: m_extend ()
{
    ListNodeComBuf * pListNodeComBuf = 
        new ( m_freeListListNodeComBuf ) 
            ListNodeComBuf ( m_pVecAllocator->alloc () );
    m_bufList.add ( *pListNodeComBuf );
    return m_bufList.lastIter ();
}

void ByteSeqOut :: 
    m_tidyUp ()
{
    tsDLIter < ListNodeComBuf > pBuf = m_pFirstNodeUncommitted;
    if ( m_vecFirstUncommitted.nElem () > 0 ) {
        pBuf++;
    }
    while ( pBuf.valid () ) {
        tsDLIter < ListNodeComBuf > pBufTmp = pBuf;
        pBuf++;
        m_bufList.remove ( *pBufTmp );
        m_freeListListNodeComBuf.deleteProduct ( *pBufTmp );
    }
}

void ByteSeqOut :: 
    show ( unsigned level ) const
{
    printf ( "ByteSeqOut\n" );
    if ( level > 0 ) {
        m_bufList.show ( level - 1 );
        m_vecFirstUncommitted.show ( level - 1 );
        m_pFirstNodeUncommitted.show ( level - 1 );
        m_freeListListNodeComBuf.show ( level - 1 );
        printf ( "m_iteratorActive = %c\n", 
            m_iteratorActive ? 'T' : 'F' );
    }
}

size_t ByteSeqOut :: 
    nBytesBuffered () const
{
    size_t nBufs = m_bufList.count ();
    size_t nBytes = nBufs * m_pVecAllocator->blockSize ();
    const size_t uncommittedBytes = 
        m_vecFirstUncommitted.nElem ();
    assert ( nBytes >= uncommittedBytes );
    nBytes -= uncommittedBytes;
    return nBytes;
}

ByteSeqOut :: ListNodeComBuf ::
    ListNodeComBuf ( const Vec < epicsUInt8 > & vec ) :
    m_vec ( vec )
{
}

void ByteSeqOut :: ListNodeComBuf ::
    show ( unsigned level ) const
{
    printf ( "ListNodeComBuf\n" );
    if ( level > 0 ) {
        m_vec.show ( level - 1u );
    }
}

void ByteSeqOut :: ListNodeComBuf ::
    flush ( Writer & writer )
{
    writer.write ( m_vec );
    m_vec.clear ();
}

void ByteSeqOut :: ListNodeComBuf ::
    flush ( Writer & writer, size_t nBytes )
{
    writer.write ( m_vec.popVec ( nBytes ) );
}

void ByteSeqOut :: Iter :: 
    m_moveToNextBuf ()
{
    while ( m_vecCurrentNode.nElem () == 0u ) {
        m_pCurNode++;
        if ( ! m_pCurNode.valid () ) {
            m_pCurNode = m_byteSeqOut.m_extend ();
            m_vecCurrentNode = m_pCurNode->getRef ();
            assert ( m_vecCurrentNode.nElem () > 0u );
            return;
        }
        m_vecCurrentNode = m_pCurNode->getRef ();
    }
}

void ByteSeqOut :: Iter ::
    m_pushToNextBufUInt8 ( const epicsUInt8 & val )
{
    m_moveToNextBuf ();
    bool success = m_vecCurrentNode.pushNet ( val );
    assert ( success );
}

void ByteSeqOut :: Iter ::
    m_multiBufferPushUint16 ( const epicsUInt16 & val )
{
    // preserve big endian wire push semantics
    // when crossing buffer boundaries
    this->pushUInt8 ( static_cast < epicsUInt8 > ( val >> 8 ) );
    this->pushUInt8 ( static_cast < epicsUInt8 > ( val ) );
}

void ByteSeqOut :: Iter ::
    m_multiBufferPushUInt32 ( const epicsUInt32 & val )
{
    // preserve big endian wire push semantics
    // when crossing buffer boundaries
    m_multiBufferPushUint16 ( static_cast < epicsUInt16 > ( val >> 16 ) );
    m_multiBufferPushUint16 ( static_cast < epicsUInt16 > ( val ) );
}

void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsFloat64 & val )
{
#   ifdef ULLONG_MAX
        const epicsUInt64 & ullVal ( 
            reinterpret_cast < const epicsUInt64 & > ( val ) );
        // preserve big endian wire push semantics
        // when crossing buffer boundaries
        m_multiBufferPushUInt32 ( static_cast < epicsUInt32 > ( ullVal >> 32 ) );
        m_multiBufferPushUInt32 ( static_cast < epicsUInt32 > ( ullVal ) );
#   else
#       error The baseline was that we nolonger use compilers that 
#       error dont conform to c99, but that will probably be a wrong
#       error guess, and I will need to implement a less efficent 
#       error version switching on a byte order macro
#   endif
}

void ByteSeqOut :: Iter ::
    pushString ( const StringReader & val )
{
    size_t len = val.size ();
    size_t pos = 0u;
    while ( len ) {
        if ( m_vecCurrentNode.nElem () == 0 ) {
            m_moveToNextBuf ();
        }
        size_t nActual = m_vecCurrentNode.pushNetString ( val, pos );
        if ( nActual > len ) {
            throw std :: range_error ( 
                "caPtr :: pushString: copied more than was requested?" );
        }
        len -= nActual;
        pos += nActual;
    }
    if ( m_vecCurrentNode.nElem () == 0 ) {
        m_moveToNextBuf ();
    }
    bool success = m_vecCurrentNode.pushNet ( '0' );
    assert ( success );
}

void ByteSeqOut :: Iter ::
    m_pushZeros ( size_t len )
{
    while ( len ) {
        if ( m_vecCurrentNode.nElem () == 0 ) {
            m_moveToNextBuf ();
        }
        size_t nActual = m_vecCurrentNode.pushZeros ( len );
        if ( nActual > len ) {
            throw std :: range_error ( 
                "caPtr :: pushZeros: "
                "copied more than was requested?" );
        }
        len -= nActual;
    }
}


#if 0

class ByteSeqOut :: IterBuf {
public:
    IterBuf ( Guard &, ByteSeqOut & );
    // there can be only one iterator outstanding at a time
    // so the copy constructor disconnects the source
    // iterator from the stream
    IterBuf ( IterBuf & );
    ~IterBuf ();
    void next ( Vec < const epicsUInt8 > :: Locker &, size_t nBytesToSkip = 0 );
    void commit ( Vec < const epicsUInt8 > :: Locker & );
    template < class Elem >
    void allocBufPtr ( Ptr < Elem > & pBuf );
    template < class Elem >
    void releaseBufPtr ( Ptr < Elem > & pBuf );
private:
    // guarantees that the ByteSeqOut is protected 
    // for the full duration of this object's lifespan
    Guard m_guard;
    ByteSeqOut * m_pByteSeqOut;
    tsDLIter < ListNodeComBuf > m_pCurrent;
    size_t m_nextWriteIndex;
    IterBuf & operator = ( const IterBuf & );
};

template < class Elem >
void ByteSeqOut :: IterBuf ::
    releaseBufPtr ( Ptr < Elem > & pBuf )
{
    pBuf.release ( m_guard );
}

template < class Elem >
void ByteSeqOut :: IterBuf ::
    allocBufPtr ( Ptr < Elem > & pBuf )
{
    if ( ! m_pByteSeqOut ) {
        throw std :: runtime_error ( 
            "ByteSeqOut :: IterBuf :: allocBufPtr (): "
            "iterator is disconnected from its out stream" );
    }
    ComBuf * pComBuf = new ( m_guard, m_pByteSeqOut->m_comBufMemMgr ) 
        ComBuf ( m_guard, m_pByteSeqOut->m_comBufDestructor );
    pComBuf->assignPtr ( m_guard, pBuf );
}

ByteSeqOut :: IterBuf ::
    IterBuf ( Guard & guard, ByteSeqOut & byteSeqOut ) :
    m_guard ( guard ), 
    m_pByteSeqOut ( & byteSeqOut ),
    m_pCurrent ( byteSeqOut.m_pCurrent ), 
    m_nextWriteIndex ( byteSeqOut.m_nextWriteIndex )
{
    if ( byteSeqOut.m_iteratorActive ) {
        throw std :: runtime_error ( 
            "there currently can be only one "
            "output iterator active at a time" );
    }
    byteSeqOut.m_iteratorActive = true;
}

// there can be only one iterator outstanding at a time
// so the copy constructor disconnects the source
// iterator from the stream
ByteSeqOut :: IterBuf ::
    IterBuf ( IterBuf & iterBuf ) :
    m_guard ( iterBuf.m_guard ), 
    m_pByteSeqOut ( iterBuf.m_pByteSeqOut ),
    m_nextWriteIndex ( 0 )
{
    if ( m_pByteSeqOut ) {
        m_pCurrent = m_pByteSeqOut->m_pCurrent;
        m_nextWriteIndex = m_pByteSeqOut->m_nextWriteIndex;
        iterBuf.m_pByteSeqOut = 0;
    }
}

ByteSeqOut :: IterBuf ::
    ~IterBuf ()
{
    if ( m_pByteSeqOut ) {
        assert ( m_pByteSeqOut->m_iteratorActive );
        m_pByteSeqOut->m_tidyUp ( m_guard );
        m_pByteSeqOut->m_iteratorActive = false;
    }
}

void ByteSeqOut :: IterBuf ::
    next ( Vec < const epicsUInt8 > :: Locker & vecLocker, size_t nBytesToSkip )
{
    if ( ! m_pByteSeqOut ) {
        throw std :: runtime_error ( 
            "ByteSeqOut :: IterBuf :: next (): "
            "iterator is disconnected from its out stream" );
    }
    while ( nBytesToSkip ) {
        if ( ! m_pCurrent.valid () ) {
            m_pCurrent = m_pByteSeqOut->m_extend ( m_guard );
            assert ( m_pCurrent.valid () );
            m_nextWriteIndex = 0;
        }
        if  ( m_nextWriteIndex < vecLocker.nElem () ) {
            size_t nBytes = ComBuf :: size () - m_nextWriteIndex;
            if ( nBytesToSkip >= nBytes ) {
                nBytesToSkip -= nBytes;
                m_pCurrent++;
                m_nextWriteIndex = 0;
            }
            else {
                m_nextWriteIndex += nBytesToSkip;
                nBytesToSkip = 0u;
            }
        }
    }
    while ( true ) {
        if ( ! m_pCurrent.valid () ) {
            m_pCurrent = m_pByteSeqOut->m_extend ( m_guard );
            assert ( m_pCurrent.valid () );
            m_nextWriteIndex = 0;
        }
        if  ( m_nextWriteIndex < ComBuf :: size () ) {
            m_pCurrent->assignPtr ( m_guard, vecLocker );
            vecLocker.advance ( m_guard, m_nextWriteIndex );
            // its essential here that we position at the 
            // end of the frame, but leave the iterator 
            // pointing at the last frame so we can rewind
            m_nextWriteIndex = ComBuf :: size ();
            return;
        }
        m_pCurrent++;
        m_nextWriteIndex = 0; 
    }
}

void ByteSeqOut :: IterBuf ::
    commit ( VecGuard < const epicsUInt8 > & pBuf )
{
    if ( ! m_pByteSeqOut ) {
        throw std :: runtime_error ( 
            "ByteSeqOut :: IterBuf :: commit (): "
            "iterator is disconnected from its out stream" );
    }
    if ( pBuf.nElem () ) {
        if ( ! m_pCurrent.valid () ) {
            throw std :: runtime_error ( 
                "ByteSeqOut :: IterBuf :: commit: "
                "attempt to rewind w/o a current buf?" );
        }
        if ( pBuf.nElem () > m_nextWriteIndex ) {
            throw std :: runtime_error ( 
                "ByteSeqOut :: IterBuf :: commit: "
                "attempt to rewind more than a current buf?" );
        }
#error must be wrong
        m_nextWriteIndex -= pBuf.nElem ();
    }
    pBuf.release ( m_guard );
    m_pByteSeqOut->m_pCurrent = m_pCurrent;
    m_pByteSeqOut->m_nextWriteIndex = m_nextWriteIndex;
}
#endif

} // end of namespace ca
