
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#ifndef ByteSeqOuth
#define ByteSeqOuth

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_ByteSeqOuth
#   undef epicsExportSharedSymbols
#endif

#include "tsDLList.h"
#include "tsFreeList.h"

#ifdef epicsExportSharedSymbols_ByteSeqOuth
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caVec.h"
#include "caPtr.h"

class StringReader;

namespace ca {

using namespace da;

struct Writer;
struct VecAllocator;

class ByteSeqOut {
public:
    class Iter;
    ByteSeqOut ( const Ptr < VecAllocator > & );
    ~ByteSeqOut ();
	void show ( unsigned level ) const;
    size_t nBytesBuffered () const;
    void flush ( Writer & );
private:
    class ListNodeComBuf;
    tsDLList < ListNodeComBuf > m_bufList;
    Vec < epicsUInt8 > m_vecFirstUncommitted;
    tsDLIter < ListNodeComBuf > m_pFirstNodeUncommitted;
    Ptr < VecAllocator > m_pVecAllocator;
    tsFreeList < ListNodeComBuf > m_freeListListNodeComBuf;
    bool m_iteratorActive;
    void m_clearAll ();
    tsDLIter < ListNodeComBuf > m_extend ();
    void m_tidyUp ();
    ByteSeqOut ( const ByteSeqOut & ); // disabled
    ByteSeqOut & operator = ( const ByteSeqOut & ); // disabled
};

class ByteSeqOut :: ListNodeComBuf : 
    public tsDLNode < ListNodeComBuf > {
public:
    ListNodeComBuf ( const Vec < epicsUInt8 > & );
    //template < class Elem >
    //void assignPtr ( Guard &, Ptr < Elem * > & );
    void flush ( Writer & );
    void flush ( Writer &, size_t nBytes );
    void show ( unsigned level ) const;
    Vec < epicsUInt8 > getRef ();
    size_t nElem ();
private:
    Vec < epicsUInt8 > m_vec;
};

class ByteSeqOut :: Iter {
public:
    Iter ( ByteSeqOut & );
     ~Iter ();

    // the push functions may m_not_ be used after
    // calling commit
    void pushInt8 ( const epicsInt8 & );
    void pushUInt8 ( const epicsUInt8 & );
    void pushInt16 ( const epicsInt16 & );
    void pushUInt16 ( const epicsUInt16 & );
    void pushInt32 ( const epicsInt32 & );
    void pushUInt32 ( const epicsUInt32 & );
    void pushFloat32 ( const epicsFloat32 & );
    void pushFloat64 ( const epicsFloat64 & );
    void pushString ( const StringReader & ); 
    template < class T >
    void pushArray ( ElemSeq & elemSeq, size_t nElem );
    void pushZeros ( size_t nOctets );
    void commit ();
protected:
    // private data 
    ByteSeqOut & m_byteSeqOut;
    Vec < epicsUInt8 > m_vecCurrentNode;
    tsDLIter < ListNodeComBuf > m_pCurNode;

    // private functions
    void m_pushZeros ( size_t nBytes );
    void m_multiBufferPush ( const epicsInt16 & );
    void m_multiBufferPush ( const epicsUInt16 & );
    void m_multiBufferPush ( const epicsInt32 & );
    void m_multiBufferPush ( const epicsUInt32 & );
    void m_multiBufferPush ( const epicsFloat32 & );
    void m_multiBufferPush ( const epicsFloat64 & );
    void m_multiBufferPushUint16 ( const epicsUInt16 & );
    void m_multiBufferPushUInt32 ( const epicsUInt32 & );
    template < class T >
    void m_push ( const T & val );
    void m_pushToNextBufUInt8 ( const epicsUInt8 & );
    void m_moveToNextBuf ();
    // IterOut does _not_ have copy semantics due to its
    // being based on abstract buffer iterator, IterBuf,
    // who's internal state isnt easily duplicated.
    Iter ( const Iter & ); // disabled
    Iter & operator = ( const Iter & ); // disabled
};

inline Vec < epicsUInt8 > 
    ByteSeqOut :: ListNodeComBuf :: 
        getRef ()
{
    return m_vec;
}

inline size_t ByteSeqOut :: ListNodeComBuf ::
    nElem ()
{
    return m_vec.nElem ();
}

inline ByteSeqOut :: Iter ::
    Iter ( ByteSeqOut & byteSeqOut ) :
    m_byteSeqOut ( byteSeqOut ),
    m_vecCurrentNode ( byteSeqOut.m_vecFirstUncommitted ),
    m_pCurNode ( byteSeqOut.m_pFirstNodeUncommitted )
{
    assert ( m_byteSeqOut.m_iteratorActive == false );
    m_byteSeqOut.m_iteratorActive = true;
}

inline ByteSeqOut :: Iter ::
    ~Iter () 
{
    assert ( m_byteSeqOut.m_iteratorActive == true );
    m_byteSeqOut.m_iteratorActive = false;
}

inline void ByteSeqOut :: Iter ::
    commit () 
{
    m_byteSeqOut.m_vecFirstUncommitted = m_vecCurrentNode;
    m_byteSeqOut.m_pFirstNodeUncommitted = m_pCurNode;
}

inline void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsUInt16 & val )
{
    m_multiBufferPushUint16 ( val );
}

inline void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsInt16 & val )
{
    m_multiBufferPushUint16 ( 
        reinterpret_cast < const epicsUInt16 & > ( val ) );
}

inline void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsUInt32 & val )
{
    m_multiBufferPushUInt32 ( val );
}

inline void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsInt32 & val )
{
    m_multiBufferPushUInt32 ( 
        reinterpret_cast < const epicsUInt32 & > ( val ) );
}

inline void ByteSeqOut :: Iter ::
    m_multiBufferPush ( const epicsFloat32 & val )
{
    m_multiBufferPushUInt32 ( 
        reinterpret_cast < const epicsUInt32 & > ( val ) );
}

template < class T >
inline void ByteSeqOut :: Iter ::
    m_push ( const T & val )
{
    if ( ! m_vecCurrentNode.pushNet ( val ) ) {
        m_multiBufferPush ( val );
    }
}

inline void ByteSeqOut :: Iter ::
    pushUInt8 ( const epicsUInt8 & val )
{
    if ( ! m_vecCurrentNode.pushNet ( val ) ) {
        m_pushToNextBufUInt8 ( val );
    }
}

inline void ByteSeqOut :: Iter ::
    pushInt8 ( const epicsInt8 & val )
{
    pushUInt8 ( static_cast < epicsUInt8 > ( val ) );
}

inline void ByteSeqOut :: Iter ::
    pushInt16 ( const epicsInt16 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushUInt16 ( const epicsUInt16 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushInt32 ( const epicsInt32 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushUInt32 ( const epicsUInt32 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushFloat32 ( const epicsFloat32 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushFloat64 ( const epicsFloat64 & val )
{
    m_push ( val );
}

inline void ByteSeqOut :: Iter ::
    pushZeros ( size_t nBytes )
{
    if ( nBytes ) {
        m_pushZeros ( nBytes );
    }
}

template < class T >
void ByteSeqOut :: Iter ::
    pushArray ( ElemSeq & elemSeq, size_t nElem )
{
    while ( nElem ) {
        size_t nBytes = m_vecCurrentNode.nElem ();
        if ( nBytes <= sizeof ( T ) ) {
            if ( nBytes == 0 ) {
                m_moveToNextBuf ();
            }
            else {
                T tmp;
                getNextElement ( elemSeq, tmp );
                m_push ( tmp );
                nElem--;
            }
        }
        else {
            size_t nElemThisSeg = m_vecCurrentNode.
                pushNetArray < T > ( elemSeq, nElem );
            assert ( nElem >= nElemThisSeg );
            nElem -= nElemThisSeg;
        }
    }
}

#if 0
template < class Elem >
inline void ByteSeqOut :: Iter ::
    allocBufPtr ( Ptr < Elem > & pBuf )
{
    m_ib.allocBufPtr ( pBuf );
}

template < class Elem >
inline void ByteSeqOut :: Iter ::
    releaseBufPtr ( Ptr < Elem > & pBuf )
{
    m_ib.releaseBufPtr ( pBuf );
}
#endif

} // end of namespace ca

#endif // ByteSeqOuth

