

/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include <stdexcept>

#include "db_access.h"
#include "dadbrIdent.h"
#include "daIntrospector.h"
#include "daCatalogSingle.h"
#include "daStringInFixedLenCharArray.h"

#define epicsExportSharedSymbols
#include "CatalogIterDBR.h"

#ifndef NELEMENTS
#   define NELEMENTS(array) (sizeof(array)/sizeof((array)[0]))
#endif

namespace ca {

using namespace da;

// this is a FrameSeqNet based on an IterIn. It is used
// to reveal network byte order DBR arrays. The parameter 
// T specifies the atomic primitive type of the DBR array.
template < class T >
class FrameSeqNetIterIn : 
    public FrameSeqNet {
public:
    FrameSeqNetIterIn ( IterIn &, size_t nElem );
    void rewind ();
    Frame < const Octet > next ( size_t nElemToBypass );
private:
    mutable IterInReader m_iterInReader;
    const size_t m_nElem;
    size_t m_nElemLeft;
    size_t m_nBytesInBuf;
    Octet m_segEdgeBuf[ sizeof ( T ) ];
};

template <>
class FrameSeqNetIterIn < dbr_string_t > : 
    public FrameSeq < const StringReader > {
public:
    FrameSeqNetIterIn ( IterIn &, size_t nElem );
    void rewind ();
    Frame < const StringReader > next ( size_t nElemToBypass );
private:
    mutable IterInReader m_iterInReader;
    const size_t m_nElem;
    size_t m_nElemLeft;
    size_t m_nBytesInBuf;
    dbr_string_t m_segEdgeBuf;
    StringInFixedLenCharArray :: 
        ScalarIntfConst < MAX_STRING_SIZE > m_scalarIntfConst;
};

class CatalogIterAlarmCond : public Catalog {
public:
    CatalogIterAlarmCond ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    class CatalogCondition : public Catalog {
    public:
        CatalogCondition ( IterIn & iterIn );
        void traverse ( Introspector & ) const;
        static size_t targetSize ();
    private:
        mutable IterInReader m_iterInReader;
    };
    CatalogCondition m_catalogCondition;
};

class CatalogIterAlarmAck : 
    public Catalog {
public:
    CatalogIterAlarmAck ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterDisplayLimits : public Catalog {
public:
    CatalogIterDisplayLimits ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterAlarmLimitsMajor : public Catalog {
public:
    CatalogIterAlarmLimitsMajor ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterAlarmLimitsMinor : public Catalog {
public:
    CatalogIterAlarmLimitsMinor ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterAlarmLimits : public Catalog {
public:
    CatalogIterAlarmLimits ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T, bool hasCtrl >
class CatalogIterLimits : public Catalog {
public:
    CatalogIterLimits ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterCtrlLimits : public Catalog {
public:
    CatalogIterCtrlLimits ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

class CatalogSts : 
    public Catalog {
public:
    CatalogSts ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    typedef CatalogSingle < dbr :: ID_condition, 
            CatalogIterAlarmCond > M_CatAlarmCond;
    M_CatAlarmCond m_catAlarm;
};

class CatalogStsAck : 
    public Catalog {
public:
    CatalogStsAck ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    class CatAlarm :
        public Catalog
    {
    public:
        CatAlarm ( IterIn & iterIn );
        void traverse ( Introspector & ) const;
    private:
        CatalogIterAlarmCond _catAlarmCond;
        CatalogIterAlarmAck _catAlarmAck;
    };
    CatAlarm m_catAlarm;
};

class CatalogTime : 
    public Catalog {
public:
    CatalogTime ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

class CatalogEnum : 
    public Catalog {
public:
    CatalogEnum ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    class M_CatalogStateSet : public Catalog {
    public:
        M_CatalogStateSet ( IterIn & iterIn );
        void traverse ( Introspector & ) const;
        static size_t targetSize ();
    private:
        mutable IterInReader m_iterInReader;
    };
    mutable IterInReader m_iterInReader;
};

template < class T, bool hasPrec = false, 
        bool hasCtrl = false, class PAD = void >
class CatalogGrCtrl : 
    public Catalog {
public:
    CatalogGrCtrl ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T, bool hasPrec = false, class PAD = void >
class CatalogCtrl : 
    public Catalog {
public:
    CatalogCtrl ( IterIn & iterIn );
    void traverse ( Introspector & ) const;
    static size_t targetSize ();
private:
    mutable IterInReader m_iterInReader;
};

template < class T >
class CatalogIterDBR :: M_CatalogAtomic : 
    public CatalogIterBase {
public:
    M_CatalogAtomic ( CatalogIterDBR & );
    void traverse ( Introspector & ) const;
private:
    // no data members here (see comment with
    // CatalogIterBase)
};

template < class F, class T, class PAD0, class PAD1 >
class CatalogIterDBR :: M_CatalogCompound : 
    public CatalogIterBase {
public:
    M_CatalogCompound ( CatalogIterDBR & );
    void traverse ( Introspector & ) const;
private:
    // no data members here (see comment with
    // CatalogIterBase)
};

epics :: Mutex CatalogIterDBR :: m_mutex;
epics :: Mutex CatalogIterDBR :: M_Handle :: 
        m_mutexNoop ( epics :: Mutex :: factoryNOOP );
tsFreeList < CatalogIterDBR > CatalogIterDBR :: m_freeList;
CatalogIterDBR :: M_Handle :: M_EnclLnk 
    CatalogIterDBR :: M_Handle :: m_enclLnk ( & CatalogIterDBR :: m_handle );

const CatalogIterDBR :: PCatFactory CatalogIterDBR :: m_pCatFactoryDBR [] =
{
    // DBR_STRING	returns a NULL terminated string
    // DBR_SHORT	returns an unsigned short
    // DBR_FLOAT	returns an IEEE floating point value
    // DBR_ENUM	    returns an unsigned short which is the enum item
    // DBR_CHAR	    returns an unsigned char
    // DBR_LONG	    returns an unsigned long
    // DBR_DOUBLE	returns a double precision floating point number
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_string_t > >, 
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_float_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_enum_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_char_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_long_t> >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_double_t > >,
    // DBR_STS_STRING   returns a string status structure (dbr_sts_string)
    // DBR_STS_SHORT    returns a short status structure (dbr_sts_short)
    // DBR_STS_FLOAT    returns a float status structure (dbr_sts_float)
    // DBR_STS_ENUM	    returns an enum status structure (dbr_sts_enum)
    // DBR_STS_CHAR	    returns a char status structure (dbr_sts_char)
    // DBR_STS_LONG	    returns a long status structure (dbr_sts_long)
    // DBR_STS_DOUBLE   returns a double status structure (dbr_sts_double)
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_string_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_float_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_enum_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_char_t, dbr_char_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_long_t> >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_double_t, dbr_long_t > >,
	// DBR_TIME_STRING	returns a string time structure (dbr_time_string)
	// DBR_TIME_SHORT	returns a short time structure (dbr_time_short)
	// DBR_TIME_FLOAT	returns a float time structure (dbr_time_float)
	// DBR_TIME_ENUM	returns an enum time structure (dbr_time_enum)
	// DBR_TIME_CHAR	returns a char time structure (dbr_time_char)
	// DBR_TIME_LONG	returns a long time structure (dbr_time_long)
	// DBR_TIME_DOUBLE	returns a double time structure (dbr_time_double)
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_string_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_short_t, dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_float_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_enum_t, dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_char_t, dbr_short_t, dbr_char_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_long_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogTime, dbr_double_t, dbr_long_t > >,
    // DBR_GR_STRING    returns a graphic string structure (dbr_sts_string)
    // DBR_GR_SHORT     returns a graphic short structure (dbr_gr_short)
    // DBR_GR_FLOAT     returns a graphic float structure (dbr_gr_float)
    // DBR_GR_ENUM      returns a graphic enum structure (dbr_gr_enum)
    // DBR_GR_CHAR      returns a graphic char structure (dbr_gr_char)
    // DBR_GR_LONG      returns a graphic long structure (dbr_gr_long)
    // DBR_GR_DOUBLE    returns a graphic double structure (dbr_gr_double)
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_string_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_short_t >, dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_float_t, true, 
                        false, dbr_short_t >, dbr_float_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogEnum, dbr_enum_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_char_t >, 
                            dbr_char_t, dbr_char_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_long_t >, dbr_long_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_double_t, true, 
                            false, dbr_short_t >, dbr_double_t > >,
    // DBR_CTRL_STRING  returns a control string structure (dbr_ctrl_int)
    // DBR_CTRL_SHORT   returns a control short structure (dbr_ctrl_short)
    // DBR_CTRL_FLOAT   returns a control float structure (dbr_ctrl_float)
    // DBR_CTRL_ENUM    returns a control enum structure (dbr_ctrl_enum)
    // DBR_CTRL_CHAR    returns a control char structure (dbr_ctrl_char)
    // DBR_CTRL_LONG    returns a control long structure (dbr_ctrl_long)
    // DBR_CTRL_DOUBLE  returns a control double structure (dbr_ctrl_double)
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogSts, dbr_string_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_short_t, 
                                false, true >, dbr_short_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_float_t, true, 
                            true, dbr_short_t >, dbr_float_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogEnum, dbr_enum_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_char_t, 
                        false, true >, dbr_char_t, dbr_char_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_long_t, 
                                false, true >, dbr_long_t > >,
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogGrCtrl < dbr_double_t, 
                        true, true, dbr_short_t >, dbr_double_t > >,
    // DBR_PUT_ACKT
    // DBR_PUT_ACKS
    // DBR_STSACK_STRING
    // DBR_CLASS_NAME
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_put_ackt_t > >, 
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_put_acks_t > >, 
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogCompound < CatalogStsAck, dbr_string_t > >, 
    & CatalogIterDBR :: m_catFactory < 
            M_CatalogAtomic < dbr_class_name_t > >
};

template < class T >
inline size_t voidIsZeroLenSizeOf ()
{
    return sizeof ( T );
}

template <>
inline size_t voidIsZeroLenSizeOf < void > ()
{
    return 0u;
}

template < class T >
inline FrameSeqNetIterIn < T > :: 
    FrameSeqNetIterIn ( IterIn & iterIn, size_t nElem ) :
    m_iterInReader ( iterIn, nElem * sizeof ( T ) ),
    m_nElem ( nElem ),
    m_nElemLeft ( nElem ),
    m_nBytesInBuf ( 0u )
{
}

template < class T >
void FrameSeqNetIterIn < T > :: 
    rewind ()
{
    m_iterInReader.rewind ();
    m_nElemLeft = m_nElem;
    m_nBytesInBuf = 0u;
}

// This could be packaged to call through a single common
// function passing the size of T as an argument, but that
// would run slower, and probably not substantially decrease 
// the code size.
template < class T >
Frame < const Octet > FrameSeqNetIterIn < T > :: 
    next ( size_t nElemToBypass )
{
    size_t nBytesToBypass = nElemToBypass * sizeof ( T );
    while ( nBytesToBypass ) {
        if ( nElemToBypass > m_nElemLeft ) {
            throw std :: runtime_error ( 
                "FrameSeqNetIterIn :: attempt to read past "
                "end of payload during bypass" );
        }
        assert ( m_nBytesInBuf <= nBytesToBypass );
        nBytesToBypass -= m_nBytesInBuf;
        m_nBytesInBuf = 0;
        m_iterInReader.moveForward ( nBytesToBypass );
        m_nElemLeft -= nElemToBypass;
    }
    if ( m_nElemLeft == 0 ) {
        throw std :: runtime_error ( 
            "FrameSeqNetIterIn :: attempt to read past "
            "end of payload" );
    }
    if ( m_nBytesInBuf ) {
        while ( m_nBytesInBuf < sizeof ( T ) ) {
            size_t nElemMax = sizeof ( T ) - m_nBytesInBuf;
            Vec < const epicsUInt8 > vec = m_iterInReader.read ( nElemMax );
            size_t vecLen = vec.nElem ();
            assert ( vecLen <= nElemMax );
            memcpy ( &m_segEdgeBuf[m_nBytesInBuf], vec.pFirst (), vecLen );
            m_nBytesInBuf += vecLen;
        }
        m_nBytesInBuf = 0u;
        m_nElemLeft--;
        return Frame < const Octet > ( m_segEdgeBuf, sizeof ( m_segEdgeBuf ) );
    }
    else {
        Vec < const epicsUInt8 > vec = 
            m_iterInReader.read ( m_nElemLeft * sizeof ( T ) );
        const Octet * const pFirst = vec.pFirst ();
        const size_t nBytesThisVec = vec.nElem ();
        const size_t nElemThisTime = nBytesThisVec / sizeof ( T );
        const size_t nBytesThisTime = nElemThisTime * sizeof ( T );
        const size_t trim = nBytesThisVec - nBytesThisTime;
        if ( trim ) {
            assert ( trim <= sizeof ( m_segEdgeBuf ) );
            memcpy ( m_segEdgeBuf, pFirst + nBytesThisTime, trim );
            vec.trimBack ( trim );
            m_nBytesInBuf = trim;
        }
        assert ( m_nElemLeft >= nElemThisTime );
        m_nElemLeft -= nElemThisTime;
        return Frame < const Octet > ( pFirst, nBytesThisTime );
    }
}

inline FrameSeqNetIterIn < dbr_string_t > :: 
    FrameSeqNetIterIn ( IterIn & iterIn, size_t nElem ) :
    m_iterInReader ( iterIn, nElem * sizeof ( dbr_string_t ) ),
    m_nElem ( nElem ),
    m_nElemLeft ( nElem ),
    m_nBytesInBuf ( 0u ),
    m_scalarIntfConst ( m_segEdgeBuf )
{
}

void FrameSeqNetIterIn < dbr_string_t > :: 
    rewind ()
{
    m_iterInReader.rewind ();
    m_nElemLeft = m_nElem;
    m_nBytesInBuf = 0u;
}

Frame < const StringReader > FrameSeqNetIterIn < dbr_string_t > :: 
    next ( size_t nElemToBypass )
{
    size_t nBytesToBypass = nElemToBypass * sizeof ( dbr_string_t );
    while ( nBytesToBypass ) {
        if ( nElemToBypass > m_nElemLeft ) {
            throw std :: runtime_error ( 
                "FrameSeqNetIterIn :: attempt to read past "
                "end of payload during bypass" );
        }
        assert ( m_nBytesInBuf <= nBytesToBypass );
        nBytesToBypass -= m_nBytesInBuf;
        m_nBytesInBuf = 0;
        m_iterInReader.moveForward ( nBytesToBypass );
        m_nElemLeft -= nElemToBypass;
    }
    if ( m_nElemLeft == 0 ) {
        throw std :: runtime_error ( 
            "FrameSeqNetIterIn :: attempt to read past "
            "end of payload" );
    }
    while ( m_nBytesInBuf < sizeof ( dbr_string_t ) ) {
        size_t nElemMax = sizeof ( dbr_string_t ) - m_nBytesInBuf;
        Vec < const epicsUInt8 > vec = m_iterInReader.read ( nElemMax );
        size_t vecLen = vec.nElem ();
        assert ( vecLen <= nElemMax );
        memcpy ( &m_segEdgeBuf[m_nBytesInBuf], vec.pFirst (), vecLen );
        m_nBytesInBuf += vecLen;
    }
    m_nBytesInBuf = 0u;
    m_nElemLeft--;
    return da :: Frame < const StringReader > ( m_scalarIntfConst );
}

//
// Alarm condition members
//
inline size_t CatalogIterAlarmCond :: targetSize ()
{
    return CatalogCondition :: targetSize ();
}

inline CatalogIterAlarmCond :: 
    CatalogIterAlarmCond ( IterIn & iterIn ) :
    m_catalogCondition ( iterIn )
{
}

void CatalogIterAlarmCond ::
    traverse ( Introspector & introspector ) const
{
    introspector.reveal ( da :: dbr :: pi_condition, m_catalogCondition );
}

inline size_t CatalogIterAlarmCond :: CatalogCondition :: targetSize ()
{
    return 2 * sizeof ( dbr_short_t );
}

inline CatalogIterAlarmCond :: CatalogCondition ::
    CatalogCondition ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

void CatalogIterAlarmCond :: CatalogCondition ::
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    introspector.reveal ( da :: dbr :: pi_status, iterIn.pop < dbr_short_t > () );
    introspector.reveal ( da :: dbr :: pi_severity, iterIn.pop < dbr_short_t > () );
}

//
// CatalogIterAlarmAck
//
inline size_t CatalogIterAlarmAck :: targetSize ()
{
    return 2u * sizeof ( dbr_short_t );
}
inline CatalogIterAlarmAck :: 
    CatalogIterAlarmAck ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}
void CatalogIterAlarmAck :: 
    traverse ( Introspector & i ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    i.reveal ( da :: dbr :: pi_pending, iterIn.pop < dbr_short_t > () );
    i.reveal ( da :: dbr :: pi_severity, iterIn.pop < dbr_short_t > () );
}

//
// Display limit members
//
template < class T >
inline size_t CatalogIterDisplayLimits < T > :: targetSize ()
{
    return 2 * sizeof ( T );
}

template < class T >
CatalogIterDisplayLimits < T > :: 
    CatalogIterDisplayLimits ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T >
void CatalogIterDisplayLimits < T > :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    introspector.reveal ( da :: dbr :: pi_upper, iterIn.pop < T > () );
    introspector.reveal ( da :: dbr :: pi_lower, iterIn.pop < T > () );
}

//
// Major alarm limit members
//
template < class T >
inline size_t CatalogIterAlarmLimitsMajor < T > :: 
    targetSize ()
{
    return 4 * sizeof ( T );
}

template < class T >
CatalogIterAlarmLimitsMajor < T > :: 
    CatalogIterAlarmLimitsMajor ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T >
void CatalogIterAlarmLimitsMajor < T > :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    introspector.reveal ( da :: dbr :: pi_upper, iterIn.pop < T > () );
    iterIn.pop < T > ();
    iterIn.pop < T > ();
    introspector.reveal ( da :: dbr :: pi_lower, iterIn.pop < T > () );
}

//
// Minor alarm limit members
//
template < class T >
inline size_t CatalogIterAlarmLimitsMinor < T > :: 
    targetSize ()
{
    return 2 * sizeof ( T );
}

template < class T >
CatalogIterAlarmLimitsMinor < T > :: 
    CatalogIterAlarmLimitsMinor ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T >
void CatalogIterAlarmLimitsMinor < T > :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    introspector.reveal ( da :: dbr :: pi_upper, iterIn.pop < T > () );
    introspector.reveal ( da :: dbr :: pi_lower, iterIn.pop < T > () );
}

//
// Alarm limit members
//
template < class T >
inline size_t CatalogIterAlarmLimits < T > :: 
    targetSize ()
{
    return 4 * sizeof ( T );
}

template < class T >
CatalogIterAlarmLimits < T > :: 
    CatalogIterAlarmLimits ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T >
void CatalogIterAlarmLimits < T > :: 
    traverse ( Introspector & introspector ) const
{
    {
        m_iterInReader.rewind ();
        IterIn iterIn ( m_iterInReader );
        CatalogIterAlarmLimitsMajor < T > catalogIterAlarmLimitsMajor ( iterIn );
        introspector.reveal ( da :: dbr :: pi_major, catalogIterAlarmLimitsMajor );
    }
    {
        m_iterInReader.rewind ();
        IterIn iterIn ( m_iterInReader );
        iterIn.pop < T > ();
        CatalogIterAlarmLimitsMinor < T > catalogIterAlarmLimitsMajor ( iterIn );
        introspector.reveal ( da :: dbr :: pi_minor, catalogIterAlarmLimitsMajor );
    }
}

//
// Gr limit members
//
template < class T, bool hasCtrl >
inline size_t CatalogIterLimits < T, hasCtrl > :: 
    targetSize ()
{
    return CatalogIterDisplayLimits < T > :: targetSize () + 
        CatalogIterAlarmLimits < T > :: targetSize () +
        ( hasCtrl ? CatalogIterCtrlLimits < T > :: targetSize () : 0 );
}

template < class T, bool hasCtrl >
CatalogIterLimits < T, hasCtrl > :: 
    CatalogIterLimits ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T, bool hasCtrl >
void CatalogIterLimits < T, hasCtrl > :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    {
        CatalogIterDisplayLimits < T > catalogIterDisplayLimits ( iterIn );
        introspector.reveal ( da :: dbr :: pi_display, catalogIterDisplayLimits );
    }
    {
        CatalogIterAlarmLimits < T > catalogIterAlarmLimits ( iterIn );
        introspector.reveal ( da :: dbr :: pi_alarm, catalogIterAlarmLimits );
    }
    if ( hasCtrl ) {
        CatalogIterCtrlLimits < T > catalogIterCtrlLimits ( iterIn );
        introspector.reveal ( da :: dbr :: pi_control, catalogIterCtrlLimits );
    }
}

//
// Control limit members
//
template < class T >
inline size_t CatalogIterCtrlLimits < T > :: 
    targetSize ()
{
    return 2 * sizeof ( T );
}

template < class T >
CatalogIterCtrlLimits < T > :: 
    CatalogIterCtrlLimits ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T >
void CatalogIterCtrlLimits < T > :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    introspector.reveal ( da :: dbr :: pi_upper, iterIn.pop < T > () );
    introspector.reveal ( da :: dbr :: pi_lower, iterIn.pop < T > () );
}

//
// CatalogIterBase
//
inline CatalogIterDBR :: CatalogIterBase :: 
    CatalogIterBase ( CatalogIterDBR & catalogIterDBR ) :
    m_catalogIterDBR ( catalogIterDBR )
{
}

//
// Atomic DBR
//
template < class T >
inline CatalogIterDBR :: M_CatalogAtomic < T > :: 
    M_CatalogAtomic ( CatalogIterDBR & catalogIterDBR ) :
    CatalogIterBase ( catalogIterDBR )
{
}

template < class T >
inline void popAndRevealValue ( 
    IterIn & iterIn, size_t nElem, Introspector & i )
{
    if ( nElem == 1u ) {
        i.reveal ( da :: dbr :: pi_value, iterIn.pop < T > () );
    }
    else {
        FrameSeqNetIterIn < T > iterFrameSeqNet ( iterIn, nElem );
        ArrayBoundsVector arrayBoundsVector ( nElem );
        i.revealNet < T > ( da :: dbr :: pi_value, arrayBoundsVector, 
                                iterFrameSeqNet );
    }
}

template < class T >
inline void popAndRevealValue ( 
    IterIn & iterIn, size_t nElem, Introspector & i, const Catalog & facets )
{
    if ( nElem == 1u ) {
        i.reveal ( da :: dbr :: pi_value, iterIn.pop < T > (), facets );
    }
    else {
        FrameSeqNetIterIn < T > iterFrameSeqNet ( iterIn, nElem );
        ArrayBoundsVector arrayBoundsVector ( nElem );
        i.revealNet < T > ( da :: dbr :: pi_value, arrayBoundsVector, 
                                iterFrameSeqNet, facets );
    }
}

template <>
void popAndRevealValue < dbr_string_t > ( 
    IterIn & iterIn, size_t nElem, Introspector & i )
{
    if ( nElem == 1 ) {
        using namespace StringInFixedLenCharArray;
        union {
            char buf[ MAX_STRING_SIZE ];
            epicsUInt8 ubuf[ MAX_STRING_SIZE ];
        };
        iterIn.popBytes ( ubuf, MAX_STRING_SIZE );
        ScalarIntfConst < MAX_STRING_SIZE > scalarIntfConst ( buf );
        i.reveal ( da :: dbr :: pi_value, scalarIntfConst );
    }
    else {
        FrameSeqNetIterIn < dbr_string_t > iterFrameSeqNet ( iterIn, nElem );
        ArrayBoundsVector arrayBoundsVector ( nElem );
        i.reveal ( da :: dbr :: pi_value, 
                            arrayBoundsVector, iterFrameSeqNet );
    }
}

template <>
void popAndRevealValue < dbr_string_t > ( 
    IterIn & iterIn, size_t nElem, Introspector & i, const Catalog & facets )
{
    if ( nElem == 1 ) {
        using namespace StringInFixedLenCharArray;
        union {
            char buf[ MAX_STRING_SIZE ];
            epicsUInt8 ubuf[ MAX_STRING_SIZE ];
        };
        iterIn.popBytes ( ubuf, MAX_STRING_SIZE );
        ScalarIntfConst < MAX_STRING_SIZE > scalarIntfConst ( buf );
        i.reveal ( da :: dbr :: pi_value, scalarIntfConst, facets );
    }
    else {
        FrameSeqNetIterIn < dbr_string_t > iterFrameSeqNet ( iterIn, nElem );
        ArrayBoundsVector arrayBoundsVector ( nElem );
        i.reveal ( da :: dbr :: pi_value, arrayBoundsVector, 
                            iterFrameSeqNet, facets );
    }
}

template < class T >
void CatalogIterDBR :: M_CatalogAtomic < T > :: 
    traverse ( Introspector & introspector ) const
{
    m_catalogIterDBR.m_iterInReader.rewind ();
    IterIn iterIn ( m_catalogIterDBR.m_iterInReader );
    popAndRevealValue < T > ( iterIn, m_catalogIterDBR.m_nElem, introspector );
}

//
// Compound DBR
//
template < class F, class T, class PAD0, class PAD1 >
inline CatalogIterDBR :: M_CatalogCompound < F, T, PAD0, PAD1 > :: 
    M_CatalogCompound ( CatalogIterDBR & catalogIterDBR ) :
    CatalogIterBase ( catalogIterDBR )
{
}
template < class F, class T, class PAD0, class PAD1 >
void CatalogIterDBR :: M_CatalogCompound < F, T, PAD0, PAD1 > :: 
    traverse ( Introspector & introspector ) const
{
    m_catalogIterDBR.m_iterInReader.rewind ();
    IterIn iterIn ( m_catalogIterDBR.m_iterInReader );
    F facets ( iterIn );
    iterIn.pop < PAD0 > ();
    iterIn.pop < PAD1 > ();
    popAndRevealValue < T > ( iterIn, m_catalogIterDBR.m_nElem, 
                                introspector, facets );
}

//
// STS DBR
//
inline CatalogSts :: 
    CatalogSts ( IterIn & iterIn ) :
    m_catAlarm ( iterIn )
{
}

void CatalogSts :: 
    traverse ( Introspector & introspector ) const
{
    introspector.reveal ( da :: dbr :: pi_alarm, m_catAlarm );
}

//
// CatalogStsAck
//
inline CatalogStsAck :: 
    CatalogStsAck ( IterIn & iterIn ) :
    m_catAlarm ( iterIn )
{
}
void CatalogStsAck :: 
    traverse ( Introspector & introspector ) const
{
    introspector.reveal ( da :: dbr :: pi_alarm, m_catAlarm );
}
inline CatalogStsAck :: CatAlarm ::
    CatAlarm ( IterIn & iterIn ) :
    _catAlarmCond ( iterIn ), _catAlarmAck ( iterIn )
{
}
void CatalogStsAck :: CatAlarm ::
    traverse ( Introspector & introspector ) const
{
    introspector.reveal ( da :: dbr :: pi_condition, _catAlarmCond );
    introspector.reveal ( da :: dbr :: pi_acknowledge, _catAlarmAck );
}

//
// Time DBR
//
inline CatalogTime :: 
    CatalogTime ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, 2 * sizeof ( dbr_short_t ) + 
                            sizeof ( epicsTimeStamp ) )
{
}

void CatalogTime :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    CatalogIterAlarmCond catalogIterAlarm ( iterIn );
    introspector.reveal ( da :: dbr :: pi_alarm, catalogIterAlarm );
    introspector.reveal ( da :: dbr :: pi_timeStamp, iterIn.pop < epicsTime > () );
}

//
// Enum DBR
//
size_t CatalogEnum :: targetSize ()
{
    return CatalogIterAlarmCond :: targetSize () + 
        M_CatalogStateSet :: targetSize ();
}

inline CatalogEnum :: 
    CatalogEnum ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

void CatalogEnum :: 
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    CatalogIterAlarmCond catalogIterAlarm ( iterIn );
    introspector.reveal ( da :: dbr :: pi_alarm, catalogIterAlarm );
    M_CatalogStateSet catalogStateSet ( iterIn );
    introspector.reveal ( da :: dbr :: pi_stateSet, catalogStateSet );
}

size_t CatalogEnum :: M_CatalogStateSet ::
    targetSize ()
{
    return sizeof ( dbr_short_t ) + 
        sizeof ( char ) * MAX_ENUM_STATES * MAX_ENUM_STRING_SIZE;
}

inline CatalogEnum :: M_CatalogStateSet ::
    M_CatalogStateSet ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

void CatalogEnum :: M_CatalogStateSet ::
    traverse ( Introspector & introspector ) const
{
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    using namespace StringInFixedLenCharArray;
    dbr_short_t nStr = iterIn.pop < dbr_short_t > ();
    if ( nStr < 0 ) {
        nStr = 0;
    }
    ArrayBoundsVector abv ( static_cast < size_t > ( nStr ) );
    char strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE];
    iterIn.popBytes ( reinterpret_cast < epicsUInt8 * > ( strs[0] ), sizeof ( strs ) );
    VectorIntfConst < MAX_ENUM_STRING_SIZE > 
        vic ( & strs[0], static_cast < size_t > ( nStr ) );
    introspector.reveal ( da :: dbr :: pi_labels, abv, vic );
}

//
// Graphic or Control DBR
//
template < class T, bool hasPrec, bool hasCtrl, class PAD >
inline size_t CatalogGrCtrl < T, hasPrec, hasCtrl, PAD > :: targetSize ()
{
    return CatalogIterAlarmCond :: targetSize () + 
        (hasPrec ? sizeof ( dbr_short_t ) : 0 ) +
        voidIsZeroLenSizeOf < PAD > () +
        sizeof ( char ) * MAX_UNITS_SIZE +
        CatalogIterLimits < T, hasCtrl > :: targetSize ();
}

template < class T, bool hasPrec, bool hasCtrl, class PAD >
inline CatalogGrCtrl < T, hasPrec, hasCtrl, PAD > :: 
    CatalogGrCtrl ( IterIn & iterIn ) :
    m_iterInReader ( iterIn, targetSize () )
{
}

template < class T, bool hasPrec, bool hasCtrl, class PAD >
void CatalogGrCtrl < T, hasPrec, hasCtrl, PAD > :: 
    traverse ( Introspector & introspector ) const
{
    using namespace StringInFixedLenCharArray;
    m_iterInReader.rewind ();
    IterIn iterIn ( m_iterInReader );
    CatalogIterAlarmCond catalogIterAlarm ( iterIn );
    introspector.reveal ( da :: dbr :: pi_alarm, catalogIterAlarm );
    if ( hasPrec ) {
        introspector.reveal ( da :: dbr :: pi_precision, 
            iterIn.pop < dbr_short_t > () );
    }
    iterIn.pop < PAD > (); // pop the pad0 member
    {
        union {
            char buf[ MAX_UNITS_SIZE ];
            epicsUInt8 ubuf[ MAX_UNITS_SIZE ];
        } overlay;
        iterIn.popBytes ( overlay.ubuf, MAX_UNITS_SIZE );
        ScalarIntfConst < MAX_UNITS_SIZE > scalarIntfConst ( overlay.buf );
        introspector.reveal ( da :: dbr :: pi_units, scalarIntfConst );
    }
    {
        CatalogIterLimits < T, hasCtrl > catalogIterGrLimits ( iterIn );
        introspector.reveal ( da :: dbr :: pi_limits, catalogIterGrLimits );
    }
}

Catalog & CatalogIterDBR :: m_catFactory ( unsigned type )
{
    if ( type >= NELEMENTS ( m_pCatFactoryDBR ) ) {
        throw std :: range_error ( "DBR type was out of range high side" );
    }
    return (this->*m_pCatFactoryDBR[type])();
}

Ptr < const Catalog > CatalogIterDBR :: factory ( unsigned dbrType, 
        IterIn & iterIn, size_t nElem, size_t payloadSize )
{
    epics :: Mutex :: Guard guard ( m_mutex );
    CatalogIterDBR * pCatalogIterDBR = 
        new ( m_freeList ) CatalogIterDBR ( dbrType, iterIn, nElem, payloadSize );
    return Ptr < const Catalog > ( pCatalogIterDBR->m_handle );
}

CatalogIterDBR :: M_Handle :: TargetMutexPair 
    CatalogIterDBR :: M_Handle :: m_targetMutexPair ()
{
    CatalogIterDBR & catalogIterDBR ( m_enclLnk.enclosure ( *this ) );
    // this is managed as a const Catalog reference and the 
    // server will not change the protocol input stream so 
    // its ok to allow access w/o locking
    return TargetMutexPair ( catalogIterDBR.m_catalog, m_mutexNoop );
}

HandleIntf < const Catalog > & CatalogIterDBR :: M_Handle :: m_clone ()
{
    m_refCtr.reference ();
    return *this;
}

void CatalogIterDBR :: M_Handle :: m_release ()
{
    if ( m_refCtr.unreference () ) {
        CatalogIterDBR & catalogIterDBR ( m_enclLnk.enclosure ( *this ) );
        epics :: Mutex :: Guard guard ( CatalogIterDBR :: m_mutex );
        CatalogIterDBR :: m_freeList.deleteProduct ( catalogIterDBR );
        // no code below this point due to above delete
    }
    // no code below this point due to above delete
}

} // end of namespace ca
