/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include "errlog.h"

#define epicsExportSharedSymbols
#include "caSvrEvQue.h"

namespace ca {

EvQue :: Event :: 
    Event ( const Guard & guard, EvQue & eventQueue ) :
    m_pCallback ( 0 ), m_pEventQueue ( & eventQueue )
{
    guard.assertIdenticalMutex ( eventQueue.m_interfaceMutex );
    eventQueue.m_reserve ( guard, 1u );
}

// explicit cleanup is faster because 
// we can reuse mutex ownership in more
// situations
void EvQue :: Event ::
    cleanup ( const Guard & guard )
{
    if ( m_pEventQueue ) {
        guard.assertIdenticalMutex ( 
            m_pEventQueue->m_interfaceMutex );
        if ( m_pCallback ) {
            throw std :: runtime_error (
                "EvQue :: Event :: cleanup (): "
                "calling cleanup while an event is pending"
                "is an interface violation" );
        }
        m_pEventQueue->m_release ( guard, 1u );
        m_pEventQueue = 0;
    }
}

EvQue :: Event :: 
    ~Event ()
{
    if ( m_pEventQueue ) {
        try {
            Guard guard ( 
                m_pEventQueue->m_interfaceMutex );
            this->cleanup ( guard );
        }
        catch ( std :: exception & except ) {
            errlogPrintf ( 
                "~Event (): Exception during implicit "
                "cleanup was \"%s\"\n", 
                except.what () );
        }
        catch ( ... ) {
            errlogPrintf ( 
                "~Event (): non-standard exception "
                "during implicit cleanup\n" );
        }
    }
}

void EvQue :: Event :: 
    show ( const Guard & guard, unsigned level ) const
{
    printf ( "EvQue :: Event\n" );
    if ( level > 0 && m_pCallback ) {
        // calling the show function in m_pCallback
        // here will result in infinite recusion if this
        // is called by the show function in an enclosing
        // class
        printf ( "EvQue :: Event: m_pCallback = %p\n",
            m_pCallback );
    }
}

void EvQue :: Event ::
    postSuccess ( const Guard & guard, Callback & callback, PCat & resp )
{
    if ( ! m_pEventQueue ) {
        throw std :: runtime_error (
            "EvQue :: Event :: postSuccess (): "
            "is being called after cleanup was called?" );
    }
    guard.assertIdenticalMutex ( 
        m_pEventQueue->m_interfaceMutex );
    if ( m_pCallback ) {
        throw std :: runtime_error (
            "EvQue :: Event :: postSuccess (): "
            "completion notify while callback is pending"
            "is an interface violation" );
    }
    m_pEventQueue->m_post (
        guard, m_successNotify, resp );
    m_pCallback = & callback;
}

void EvQue :: Event ::
    postFailure ( const Guard & guard, Callback & callback, 
            PCat & resp )
{
    if ( ! m_pEventQueue ) {
        throw std :: runtime_error (
            "EvQue :: Event :: postFailure (): "
            "is being called after cleanup was called?" );
    }
    guard.assertIdenticalMutex ( 
        m_pEventQueue->m_interfaceMutex );
    if ( m_pCallback ) {
        throw std :: runtime_error (
            "EvQue :: Event :: postFailure (): "
            "completion notify while callback is pending"
            "is an interface violation" );
    }
    m_pEventQueue->m_post ( guard, m_failNotify, resp );
    m_pCallback = & callback;
}

void EvQue :: Event :: 
    postFailure ( Guard & guard, 
            Callback & callback, const char * pContext )
{
    if ( ! m_pEventQueue ) {
        throw std :: runtime_error (
            "EvQue :: Event :: postFailure (pContext): "
            "is being called after cleanup was called?" );
    }
    PCat cp;
    FailureResponse * pFailureResponse =
        new ( m_pEventQueue->m_failRespFreeList )
            FailureResponse ( guard, *m_pEventQueue, pContext, cp );
    this->postFailure ( guard, callback, cp );
    cp.release ( guard );
}

#error odd
class EvQue :: Event :: SuccessNotify :: Link : 
    public EnclBinding :: Link < 
        Event, SuccessNotify, 
        & Event :: m_successNotify > 
{
};

void EvQue :: Event :: 
    m_callBackRelease ( const Guard & guard )
{
    Callback * pCallback = m_pCallback;
    assert ( pCallback );
    m_pCallback = 0;
    try {
        // dont touch this object after release()
        // function call because they may destroy this 
        // Event object as a side effect of their 
        // release() function being invoked
        pCallback->release ( guard );
        // read comment above before adding code here
    }
    catch ( std :: exception & except ) {
        errlogPrintf (  
            "Event :: m_callBackRelease: "
            "unexpected exception in "
            "Callback :: release was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf (  
            "Event :: m_callBackRelease: "
            "non-standard unexpected exception in "
            "Callback :: release\n" );
    }
    // read comment above before adding code here
}

void EvQue :: Event :: SuccessNotify ::
    action ( Guard & guard, PCat & pResp )
{
#error odd
    EnclBinding :: Ptr < Link > pEvent ( *this );
    assert ( pEvent->m_pCallback );
    pEvent->m_pCallback->success ( guard, pResp );
}

void EvQue :: Event :: SuccessNotify ::
    release ( const Guard & guard )
{
#error odd
    EnclBinding :: Ptr < Link > pEvent ( *this );
    pEvent->m_callBackRelease ( guard );
}

#error odd
class EvQue :: Event :: FailNotify :: Link : 
    public EnclBinding :: Link < 
        Event, FailNotify, 
        & Event :: m_failNotify > 
{
};

// the outbuf guard is expected here
void EvQue :: Event :: FailNotify ::
    action ( Guard & guard, PCat & pResp )
{
#error odd
    EnclBinding :: Ptr < Link > pEvent ( *this );
    assert ( pEvent->m_pCallback );
    pEvent->m_pCallback->fail ( guard, pResp );
}

// interface guard expected here
void EvQue :: Event :: FailNotify ::
    release ( const Guard & guard )
{
#error odd
    EnclBinding :: Ptr < Link > pEvent ( *this );
    pEvent->m_callBackRelease ( guard );
}

} // end of name space ca
