
#include "osiWireFormat.h"

#define epicsExportSharedSymbols
#include "IterIn.h"

namespace ca {

// this function is only appropriate when the
// iterator is operating on a finite length 
// stream (i.e. a datagram)
bool IterIn :: 
    isEndOfStream ()
{
    bool isEOS = ( m_vec.nElem () == 0u );
    if ( isEOS ) {
        m_vec = m_reader.read ( SIZE_MAX );
        isEOS = ( m_vec.nElem () == 0u );
    }
    return isEOS;
}

template <>
epicsUInt8 IterIn :: m_multiBufferPop < epicsUInt8 > ()
{
    epicsUInt8 tmp;
    if ( ! m_vec.popNative ( tmp ) ) {
        throw std :: length_error ( 
            "IterIn :: m_multiBufferPop () : "
            "end of stream detected" );
    }
    return tmp;
}

template <>
epicsUInt16 IterIn :: m_multiBufferPop < epicsUInt16 > ()
{
    // preserve big endian wire pop semantics
    // when crossing buffer boundaries
    epicsUInt8 b0 = this->pop < epicsUInt8 > ();
    epicsUInt8 b1 = this->pop < epicsUInt8 > ();
    return static_cast < epicsUInt16 > ( b0 ) << 8u | 
            static_cast < epicsUInt16 > ( b1 );
}

template <>
epicsUInt32 IterIn :: 
    m_multiBufferPop < epicsUInt32 > ()
{
    // preserve big endian wire pop semantics
    // when crossing buffer boundaries
    epicsUInt16 w0 = m_multiBufferPop < epicsUInt16 > ();
    epicsUInt16 w1 = m_multiBufferPop < epicsUInt16 > ();
    return static_cast < epicsUInt32 > ( w0 << 16u ) | 
            static_cast < epicsUInt32 > ( w1 );
}

template <>
epicsFloat64 IterIn :: 
    m_multiBufferPop < epicsFloat64 > ()
{
    union {
        epicsUInt64 tmpUInt64;
        epicsFloat64 tmpFloat64;
    };
    epicsUInt32 lw0 = m_multiBufferPop < epicsUInt32 > ();
    epicsUInt32 lw1 = m_multiBufferPop < epicsUInt32 > ();
    tmpUInt64 = ( static_cast < epicsUInt64 > ( lw0 ) << 32u ) | 
             static_cast < epicsUInt64 > ( lw1 );
    return tmpFloat64;
}

void IterIn :: 
    popBytes ( epicsUInt8 * pBuf, size_t bufSize )
{
    while ( bufSize ) {
        size_t nActual = m_vec.nElem ();
        if ( nActual == 0 ) {
            m_vec = m_reader.read ( SIZE_MAX );
            nActual = m_vec.nElem ();
            if ( nActual == 0 ) {
                throw std :: length_error ( 
                    "IterIn :: popBytes () : "
                    "end of stream detected" );
            }
        }
        if ( nActual > bufSize ) {
            nActual = bufSize;
        }
        memcpy ( pBuf, m_vec.pFirst (), nActual );
        m_vec.trimFront ( nActual );
        pBuf += nActual;
        bufSize -= nActual;
    }
}

// this pops a string of no more than maxStrSize 
// characters and appends it to dest 
void IterIn :: 
    popStringAppendTo ( StringEditor & dest, size_t maxStrSize )
{
    while ( maxStrSize ) {
        if ( m_vec.nElem () == 0u ) {
            m_vec = m_reader.read ( SIZE_MAX );
            if ( m_vec.nElem () == 0u ) {
                throw std :: length_error ( 
                    "IterIn :: popString () : "
                    "end of stream detected" );
            }
        }
        size_t nActual = 
            m_vec.popNetStringAppendTo ( dest, maxStrSize );
        if ( nActual == 0 ) {
            break;
        }
        if ( nActual > maxStrSize ) {
            throw std :: length_error ( 
                "IterIn :: popString () : "
                "Vec < uchar > :: popString removed more "
                "than requested?" );
        }
        maxStrSize -= nActual;
    }
}

Vec < const epicsUInt8 > IterIn :: 
    popVec ( size_t nBytesMax )
{
    Vec < const epicsUInt8 > tmp = 
            m_vec.popVec ( nBytesMax );
    if ( tmp.nElem () == 0 ) {
        m_vec = m_reader.read ( SIZE_MAX );
        tmp = m_vec.popVec ( nBytesMax );
        if ( tmp.nElem () == 0 ) {
            throw std :: length_error ( 
                "IterIn :: popVec () : "
                "end of stream detected" );
        }
    }
    return tmp;
}

void IterIn :: 
    moveForward ( size_t nBytes )
{
    while ( nBytes ) {
        if ( m_vec.nElem () == 0 ) {
            m_vec = m_reader.read ( SIZE_MAX );
	}
        size_t nActual = m_vec.trimFront ( nBytes );
        if ( nActual == 0u ) {
            throw std :: length_error ( 
                    "IterIn :: popVec () : "
                    "end of stream detected" );
        }
        assert ( nActual <= nBytes );
        nBytes -= nActual;
    }
}

} // end of namespace ca
