
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#ifndef IterInh
#define IterInh

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_IterInh
#   undef epicsExportSharedSymbols
#endif

#include "epicsTypes.h"
#include "epicsTime.h"

#ifdef epicsExportSharedSymbols_IterInh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caIntfIO.h"

#ifndef SIZE_MAX 
#   define SIZE_MAX ( static_cast < size_t > ( -1 ) )
#endif

class StringReader;
class StringEditor;

namespace ca {

struct Reader;

// The ByteVecPtrConst stored herein uses the same lock as 
// the Reader - which is typically a NOOP mutex.
class IterIn {
public:
    IterIn ( Reader & );
    bool isEndOfStream ();
    template < class T > T pop ();
    void popBytes ( epicsUInt8 * pBuf, size_t bufSize );  
    // this pops a string of no more than maxStrSize
    // characters and appends it to dest
    void popStringAppendTo ( StringEditor & dest, size_t maxStrSize );  
    Vec < const epicsUInt8 > popVec ( size_t nBytesMax );
    void moveForward ( size_t nBytes );
protected:
    // private data 
    Reader & m_reader;
    Vec < const epicsUInt8 > m_vec;
    // private functions
    template < class T > T m_multiBufferPop ();
    // IterIn does m_not_ have copy semantics due to its
    // being based on Reader, who's internal 
    // state isnt easily duplicated.
private:
    IterIn ( const IterIn & ); // disabled
    IterIn & operator = ( const IterIn & ); // disabled
};

inline IterIn ::
    IterIn ( Reader & reader ) :
    m_reader ( reader ), m_vec ( reader.read ( SIZE_MAX ) )
{
}

template <>
inline void IterIn :: pop < void > ()
{
}

template <>
inline epicsTime IterIn :: pop < epicsTime > ()
{
    epicsTimeStamp ets;
    ets.secPastEpoch = this->pop < epicsUInt32 > ();
    ets.nsec = this->pop < epicsUInt32 > ();
    return epicsTime ( ets );
}

template < class T >
inline T IterIn :: pop ()
{
    const size_t vSize = sizeof ( T );
    if ( m_vec.nElem () >= vSize ) {
        T tmp = WireGet < T > ( m_vec.pFirst () );
        m_vec.trimFront ( vSize );
        return tmp;
    }
    return m_multiBufferPop < T > ();
}

template <>
epicsUInt8 IterIn :: 
    m_multiBufferPop < epicsUInt8 > ();

template <>
epicsUInt16 IterIn :: 
    m_multiBufferPop < epicsUInt16 > ();

template <>
epicsUInt32 IterIn :: 
    m_multiBufferPop < epicsUInt32 > ();

template <>
inline epicsInt8 IterIn ::
    m_multiBufferPop < epicsInt8 > ()
{
    return static_cast < epicsInt8 > ( 
        m_multiBufferPop < epicsUInt8 > () ); 
}

template <>
inline epicsInt16 IterIn ::
    m_multiBufferPop < epicsInt16 > ()
{
    return static_cast < epicsInt16 > ( 
        m_multiBufferPop < epicsUInt16 > () ); 
}

template <>
inline epicsInt32 IterIn ::
    m_multiBufferPop < epicsInt32 > ()
{
    return static_cast < epicsInt32 > ( 
        m_multiBufferPop < epicsUInt32 > () ); 
}

template <>
inline epicsFloat32 IterIn ::
    m_multiBufferPop < epicsFloat32 > ()
{
    union {
        epicsUInt32 tmpUInt32;
        epicsFloat32 tmpFloat32;
    };
    tmpUInt32 = m_multiBufferPop < epicsUInt32 > ();
    return tmpFloat32;
}

#if 0
// THIS IS DIFFICULT TO IMPLEMENT BECAUSE IterIn
// DOES NOT HAVE REWIND SEMANTICS, BUT StringReader
// DOES
class StringFromIterIn :
    public StringReader {
public:
    StringFromIterIn ( IterIn &, size_t maxSize );
    ~StringFromIterIn ();
    size_t size () const;
    bool equals ( const StringReader & rhs ) const;
    bool contains ( size_t pos, const StringReader & rhs ) const;
    bool contains ( size_t pos, const char *str, size_t len ) const;
    char at ( size_t pos) const;
    size_t extract ( size_t pos, char *buf, size_t cap ) const;
private:
    // Its probably quite safe to limit the maximum
    // string size to no more than two partial ComBuf 
    // segments. That sets the string size limit at 
    // the size of a ComBuf - which is a huge string.
    //ByteVecPtrConst m_firstSeg;
    //ByteVecPtrConst m_secondSeg; 
    // we will scan the string for a nill terminator 
    // and set the m_strSize to be less than the ctor's 
    // maximum size as the situation dictates
    const size_t m_strSize;
};
#endif

} // end of namespace ca

#endif // ifndef IterInh
