
// Author: Jeff HIll

#include <algorithm>

#include "epicsStdio.h"
#include "errlog.h"
#include "ConstString.h"
#include "dadbr.h"
#include "FixedString.h"

#define epicsExportSharedSymbols
#include "casdef2.h"
#include "v3Proto.h"
#include "caHdrLargeArray.h"

#ifndef NELEMENTS
#   define NELEMENTS(array) (sizeof(array)/sizeof((array)[0]))
#endif

namespace ca {

inline size_t MsgMinderV3 :: 
    m_pushHdr ( const HdrLargeArray & hdr, 
            size_t payloadSize )
{
    //guard.assertIdenticalMutex ( m_pByteSeqOut->m_mutex );
    const size_t octWordCount = OCT_ROUND ( payloadSize );
    assert ( SIZE_MAX / 8 <= octWordCount );
    const size_t roundedPayloadSize = octWordCount * 8;
    if ( hdr.m_count >= 0xffff || 
            roundedPayloadSize >= 0xffff ) {
        if ( ! CA_V49( m_protoRev ) ) {
            throw std :: runtime_error (
                "RspThr :: m_pushVerThreeHeader: "
                "sending a size extended header, but "
                "this client doesnt converse in that "
                "protocol?" );
        }
        static const ca_uint16_t postfixSpecSize = 0xffff;
        m_iter.pushUInt16 ( hdr.m_cmmd ); 
        m_iter.pushUInt16 ( postfixSpecSize ); 
        m_iter.pushUInt16 ( hdr.m_dataType ); 
        m_iter.pushUInt16 ( postfixSpecSize );
        m_iter.pushUInt32 ( hdr.m_cid );
        m_iter.pushUInt32 ( hdr.m_available );
        m_iter.pushUInt32 ( roundedPayloadSize );
        m_iter.pushUInt32 ( hdr.m_count );
    }
    else {
        m_iter.pushUInt16 ( hdr.m_cmmd ); 
        m_iter.pushUInt16 ( 
            static_cast < ca_uint16_t > ( roundedPayloadSize ) ); 
        m_iter.pushUInt16 ( hdr.m_dataType ); 
        m_iter.pushUInt16 (  
            static_cast < ca_uint16_t > ( hdr.m_count ) );
        m_iter.pushUInt32 ( hdr.m_cid );
        m_iter.pushUInt32 ( hdr.m_available );
    }
    return roundedPayloadSize;
}

void MsgMinderV3 :: 
    pushResp ( const HdrLargeArray & hdr )
{
    const size_t payloadSize = 
        m_pushHdr ( hdr, hdr.m_postsize );
    if ( payloadSize ) {
        m_iter.pushZeros ( payloadSize );
    }
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_char_t & val )
{
    m_iter.pushInt8 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_short_t & val )
{
    m_iter.pushInt16 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_ushort_t & val )
{
    m_iter.pushUInt16 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_long_t & val )
{
    m_iter.pushInt32 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_ulong_t & val )
{
    m_iter.pushUInt32 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_float_t & val )
{
    m_iter.pushFloat32 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_double_t & val )
{
    m_iter.pushFloat64 ( val ); 
}

#if 0
inline void MsgMinderV3 ::
    m_pushVal ( const dbr_put_ackt_t & val )
{
    m_iter.pushUInt16 ( val ); 
}

inline void MsgMinderV3 ::
    m_pushVal ( const dbr_put_acks_t & val )
{
    m_iter.pushUInt16 ( val ); 
}
#endif

//
// Its probably a good long term strategy to push
// dbr types onto the wire as a two step process
// where we fetch into a dbr auto (stack temporary)
// first, and then subsequently push that dbr type 
// onto the wire because.
//
// o The da traversal order for the well defined
// hierarchy of commonly used properties cant be
// counted on to be the same as the order of the
// fields in the legacy dbr types.
//
// o We need to byte swap when pushing, and this
// must be in server specific code if DA's interface,
// and internal implementation, isnt going to balloon 
// significantly. Considering that we have to fetch 
// each field into a cache then a two step process 
// might not be that slow.
//
// o The dbr field order order and wire protocol order 
// are currently the same, but that could change.
//
template < unsigned typeCode >
void MsgMinderV3 :: 
    m_pushScalarPayload ( 
        const size_t payloadSize, 
        const Catalog & payloadSource )
{
    using namespace dbr;
    using namespace StringInFixedLenCharArray;
    typename TypeCodeTraits < typeCode > :: 
            DBR dbrCache;
    get ( payloadSource, pi_value, dbrCache );
    // sun c++ 5.9 seems to require that this const ref
    // be passed to push
    const typename TypeCodeTraits < typeCode > :: 
            DBR & dbrCacheConst = dbrCache;
    m_pushVal ( dbrCacheConst );
    assert ( sizeof ( dbrCache ) <= payloadSize );
    const size_t nPad = payloadSize - sizeof ( dbrCache );
    m_iter.pushZeros ( nPad );
}

template < unsigned typeCode >
void MsgMinderV3 :: 
    m_pushCmpdScalarPayload ( 
        const size_t payloadSize, 
        const Catalog & payloadSource )
{
    using namespace dbr;
    using namespace StringInFixedLenCharArray;
    typename TypeCodeTraits < typeCode > :: 
            DBR dbrCache;
    typename TypeCodeTraits < typeCode > :: 
            MutScalar mut ( dbrCache );
    get ( payloadSource, pi_value, mut );
    m_pushFacetsDBR ( dbrCache );
    // sun c++ 5.9 seems to require that 
    // this const ref be passed to push
    const typename TypeCodeTraits < typeCode > :: 
            DBR & dbrCacheConst = dbrCache;
    m_pushVal ( dbrCacheConst.value );
    assert ( sizeof ( dbrCache ) <= payloadSize );
    const size_t nPad = payloadSize - sizeof ( dbrCache );
    m_iter.pushZeros ( nPad );
}

template <>
void MsgMinderV3 :: 
    m_pushCmpdScalarPayload < DBR_GR_STRING > ( 
        const size_t payloadSize, 
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_GR_STRING hasnt "
        "been defined (scalar request)" );
}

template <>
void MsgMinderV3 :: 
    m_pushCmpdScalarPayload < DBR_CTRL_STRING > ( 
        const size_t payloadSize, 
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_CTRL_STRING hasnt "
        "been defined (scalar request)" );
}


#if 0
// This works for the following types
// dbr_char_t
// dbr_short_t
// dbr_ushort_t
// dbr_long_t
// dbr_ulong_t
// dbr_float_t
// dbr_double_t
template < class T >
size_t MsgMinderV3 :: 
    m_pushPayload_dbr_atomic ( const Catalog & src )
{
    T tmp;
    Clerk clerk ( src );
    clerk.get ( pi_PV, tmp );
    m_iter.push ( tmp );
    return sizeof ( tmp );
}

template < class T >
size_t MsgMinderV3 :: 
    m_pushPayload_dbr_sts ( const Catalog & src )
{
    T tmp;
    Clerk clerk ( src );
    clerk.get ( pi_PV, tmp );
    m_iter.push ( tmp );
    return sizeof ( tmp );
}
#endif

MsgMinderV3 :: PPushScalarPayload MsgMinderV3 :: 
        m_pPushScalarPayload[] = {
    & MsgMinderV3 :: m_pushScalarPayload < DBR_STRING >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_SHORT >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_FLOAT >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_ENUM >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_CHAR >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_LONG >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_STRING >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_SHORT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_ENUM >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_CHAR >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_LONG >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STS_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_STRING >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_SHORT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_ENUM >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_CHAR >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_LONG >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_TIME_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_STRING >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_SHORT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_ENUM >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_CHAR >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_LONG >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_GR_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_STRING >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_SHORT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_ENUM >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_CHAR >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_LONG >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_CTRL_DOUBLE >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_PUT_ACKT >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_PUT_ACKS >,
    & MsgMinderV3 :: m_pushCmpdScalarPayload < DBR_STSACK_STRING >,
    & MsgMinderV3 :: m_pushScalarPayload < DBR_CLASS_NAME >
};

#if 0
template < class T >
class ElemSeqIndexV3 : 
    public ElemSeqIndex {
public:
    ElemSeqIndexV3 ( size_t nElemTotal );
    void advance ( size_t nElem );
    Frame < T > & frame ();
private:
    static const size_t m_nBufElem = 512;
    T m_buf[m_nBufElem];
    Frame < T > m_frame;
    const size_t m_nElemTotal;
    size_t m_curPos;
    static size_t m_clampFrameLen ( size_t nElem );
};

template < class T >
inline size_t ElemSeqIndexV3 < T > :: 
    m_clampFrameLen ( size_t nElem )
{
    return std :: min ( nElem, m_nBufElem );
}

// no need to initialize m_buf to a known value
// here because we can count on the Data Access
// library to set every element or else throw 
// an exception
template < class T >
inline ElemSeqIndexV3 < T > :: 
    ElemSeqIndexV3 ( size_t nElemTotal ) :
    m_frame ( m_buf, m_clampFrameLen ( nElemTotal ) ),
    m_nElemTotal ( nElemTotal ),
    m_curPos ( curPos )
{
    // we dont init the buffer because we can
    // count on the DA library to set every 
    // element or else throw an exception
    // memset ( m_buf, sizeof ( m_buf ), '\0' );
}

template < class T >
void ElemSeqIndexV3 < T > :: 
    advance ( size_t nElem )
{
    if ( nElem > m_nElemTotal ) {
        throw std :: range_error ( 
            "ElemSeqIndexV3 :: advance (): "
            "advance request specified more "
            "elements than are in the sequence" );
    }
    if ( m_curPos > m_nElemTotal - nElem ) {
        throw std :: range_error ( 
            "ElemSeqIndexV3 :: advance (): "
            "attempt to position past the "
            "end of the element sequence" );
    }
    size_t nPad = 0;
    size_t nPush = nElem;
    if ( nElem > NELEMENTS ( m_buf ) ) {
        nPad = nElem - NELEMENTS ( m_buf );
        nPush = NELEMENTS ( m_buf );
    }
    for ( size_t i; i < nPush ; i++ ) {
        m_pushDBR ( guard, m_buf[i] );
    }
    m_iter.pushZeros ( 
        guard, nPad * sizeof ( T ) );
    m_curPos += nElem;
    size_t nFrameElem = 
        m_clampFrameLen ( nElemTotal - m_curPos  );
    m_frame = Frame < T > ( m_buf, nFrameElem );
}

template < class T >
inline Frame < T > & 
    ElemSeqIndexV3 < T > :: frame ()
{
    return m_frame;
}

#endif

// Note that T must be a concrete type.
template < class T >
class MsgMinderV3 :: MutatorVec : 
    public Mutator {
public:
    MutatorVec ( MsgMinderV3 & msgMinderV3, size_t nElem );
    ~MutatorVec ();
    void assignDefault ();
    void assignNexus ( const Catalog & facets );
    void assignScalarAtomic ( const Scalar & val );
    void assignScalarAtomicFaceted ( const Scalar & src, 
                    const Catalog & srcFacets );
    void assignScalarCatalog ( const Catalog & val );
    void assignArray ( const ArrayBounds & bounds, ElemSeq & );
    void assignArrayFaceted ( const ArrayBounds & bounds, 
                    ElemSeq &, const Catalog & srcFacets );
protected:
    MsgMinderV3 & m_msgMinderV3;
    size_t m_nElem;
};

// Note that E must be a concrete type.
template < class T, class E >
class MsgMinderV3 :: MutatorVecFaceted : 
    public MsgMinderV3 :: MutatorVec < E > {
public:
    MutatorVecFaceted ( MsgMinderV3 & msgMinderV3,
                                    size_t nElem );
    void assignScalarAtomic ( const Scalar & val );
    void assignScalarAtomicFaceted ( const Scalar & src, 
                    const Catalog & srcFacets );
    void assignArray ( const ArrayBounds & bounds, ElemSeq & );
    void assignArrayFaceted ( const ArrayBounds & bounds, 
                    ElemSeq &, const Catalog & srcFacets );
private:
    typename T :: DBR m_dbrCache;
    T m_mutFacets;
    using MutatorVec < E > :: m_msgMinderV3;
};

template < class T >
inline MsgMinderV3 :: MutatorVec < T > ::
    MutatorVec ( MsgMinderV3 & msgMinderV3, size_t nElem ) :
    m_msgMinderV3 ( msgMinderV3 ), m_nElem ( nElem )
{
}

template < class T >
inline MsgMinderV3 :: MutatorVec < T > ::
    ~MutatorVec ()
{
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignDefault ()
{
    throw UndefinedProperty ();
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignNexus ( const Catalog & /* srcFacets */ )
{
    throw UndefinedProperty ();
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignScalarAtomic ( const Scalar & src )
{
    if ( m_nElem == 1 ) {
        T tmp;
        get ( src, tmp );
	// sun c++ 5.9 seems to require that 
	// this const ref be passed to push
	// (when T is a fixed length string)
	const T & ctmp = tmp;
        m_msgMinderV3.m_pushVal ( ctmp );
    }
    else {
        throw ArrayBoundsIncompatible ();
    }
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignScalarAtomicFaceted ( const Scalar & src, 
        const Catalog & /* srcFacets */ )
{
    this->assignScalarAtomic ( src );
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignScalarCatalog ( const Catalog & src )
{
    throw IncompatibleTypes ( typeid ( T ), typeid ( src ) );
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignArray ( const ArrayBounds & bounds, 
                    ElemSeq & elemSeq )
{
    BoundsChecker bc ( bounds );
    bc.mustBeMatchingVector ( m_nElem );
    m_msgMinderV3.m_iter.pushArray < T > ( elemSeq, m_nElem );
}

template <>
void MsgMinderV3 :: MutatorVec < dbr_string_t > ::
    assignArray ( const ArrayBounds & bounds, 
                    ElemSeq & elemSeq )
{
    BoundsChecker bc ( bounds );
    bc.mustBeMatchingVector ( m_nElem );
    for ( size_t i = 0u; i < m_nElem; i++ ) {
        dbr_string_t tmp;
        getNextElement ( elemSeq, tmp );
	// sun c++ 5.9 seems to require that 
	// this const ref be passed to push
        const dbr_string_t & ctmp = tmp;
        m_msgMinderV3.m_pushOldString ( ctmp );
    }
}

template < class T >
void MsgMinderV3 :: MutatorVec < T > ::
    assignArrayFaceted ( 
            const ArrayBounds & bounds, 
            ElemSeq & elemSeq, 
            const Catalog & /* srcFacets */ )
{
    this->assignArray ( bounds, elemSeq );
}

template < class T, class E >
inline MsgMinderV3 :: MutatorVecFaceted < T, E > ::
    MutatorVecFaceted ( MsgMinderV3 & msgMinderV3,
                            size_t nElem ) :
    MsgMinderV3 :: 
        MutatorVec < E > ( msgMinderV3, nElem ),
    m_mutFacets ( m_dbrCache )
{
    assert ( typeid ( m_dbrCache.value ) == typeid ( E ) );
}

template < class T, class E >
void MsgMinderV3 :: MutatorVecFaceted < T, E > ::
    assignScalarAtomic ( const Scalar & src )
{
    throw UndefinedProperty ();
}

template < class T, class E >
void MsgMinderV3 :: MutatorVecFaceted < T, E > ::
    assignScalarAtomicFaceted ( const Scalar & src, 
        const Catalog & srcFacets )
{
    m_mutFacets.assignScalarCatalog ( srcFacets );
    m_msgMinderV3.m_pushFacetsDBR ( m_dbrCache );
    this->MutatorVec < E > :: 
        assignScalarAtomic ( src );
}

template < class T, class E >
void MsgMinderV3 :: MutatorVecFaceted < T, E > ::
    assignArray ( const ArrayBounds & bounds, 
                    ElemSeq & elemSeq )
{
    throw UndefinedProperty ();
}

template < class T, class E >
void MsgMinderV3 :: MutatorVecFaceted < T, E > ::
    assignArrayFaceted ( 
            const ArrayBounds & bounds, 
            ElemSeq & elemSeq, 
            const Catalog & srcFacets )
{
    m_mutFacets.assignScalarCatalog ( srcFacets );
    m_msgMinderV3.m_pushFacetsDBR ( m_dbrCache );
    this->MutatorVec < E > :: 
        assignArray ( bounds, elemSeq );
}

//
// the comment with m_pushScalarPayload applies
// to m_pushVectorPayload also except that we need
// to be particularly careful here because we are
// not assuming that the entire array needs to fit
// in the cache at once.
//
template < unsigned typeCode >
void MsgMinderV3 :: 
    m_pushCmpdVectorPayload ( 
        const size_t payloadSize, 
        const size_t nElem,
        const Catalog & payloadSource )
{
    using namespace dbr;
    if ( nElem == 0 ) {
        throw ArrayElementSubsetOutOfBounds ();
    }
    typedef typename 
        TypeCodeTraits < typeCode > :: ELEM ELEM;
    typedef typename 
        TypeCodeTraits < typeCode > :: MutSubord MutSubord;
    MutatorVecFaceted < MutSubord, ELEM >
        mutatorVecFaceted ( *this, nElem );
    Clerk clerk ( payloadSource );
    clerk.get ( pi_value, mutatorVecFaceted );
    size_t actualSize = dbr_size_n ( typeCode, nElem );
    assert ( actualSize <= payloadSize );
    const size_t nPad = payloadSize - actualSize;
    m_iter.pushZeros ( nPad );
    m_iter.commit ();
}

template <>
void MsgMinderV3 :: 
    m_pushCmpdVectorPayload < DBR_GR_STRING > ( 
        const size_t payloadSize, 
        const size_t nElem,
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_GR_STRING hasnt "
        "been defined (vector request)" );
}

template <>
void MsgMinderV3 :: 
    m_pushCmpdVectorPayload < DBR_CTRL_STRING > ( 
        const size_t payloadSize, 
        const size_t nElem,
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_CTRL_STRING hasnt "
        "been defined (vector request)" );
}

template < unsigned typeCode >
void MsgMinderV3 :: 
    m_pushVectorPayload ( 
        const size_t payloadSize, 
        const size_t nElem,
        const Catalog & payloadSource )
{
    using namespace dbr;
    if ( nElem == 0 ) {
        throw ArrayElementSubsetOutOfBounds ();
    }
    typedef typename 
        TypeCodeTraits < typeCode > :: DBR DBR;
    typedef typename 
        TypeCodeTraits < typeCode > :: ELEM ELEM;
    assert ( typeid ( DBR ) == typeid ( ELEM ) );
    MutatorVec < ELEM > mutatorVec ( *this, nElem );
    Clerk clerk ( payloadSource );
    clerk.get ( pi_value, mutatorVec );
    size_t actualSize = dbr_size_n ( typeCode, nElem );
    assert ( actualSize <= payloadSize );
    const size_t nPad = payloadSize - actualSize;
    m_iter.pushZeros ( nPad );
    m_iter.commit ();
}

template <>
void MsgMinderV3 :: 
    m_pushVectorPayload < DBR_GR_STRING > ( 
        const size_t payloadSize, 
        const size_t nElem,
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_GR_STRING hasnt been defined" );
}

template <>
void MsgMinderV3 :: 
    m_pushVectorPayload < DBR_CTRL_STRING > ( 
        const size_t payloadSize,
        const size_t nElem,
        const Catalog & payloadSource )
{
    throw std :: runtime_error ( 
        "a C type for DBR_CTRL_STRING hasnt been defined" );
}

MsgMinderV3 :: PPushVectorPayload MsgMinderV3 :: 
        m_pPushVectorPayload[] = {
    & MsgMinderV3 :: m_pushVectorPayload < DBR_STRING >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_SHORT >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_FLOAT >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_ENUM >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_CHAR >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_LONG >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_STRING >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_SHORT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_ENUM >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_CHAR >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_LONG >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STS_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_STRING >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_SHORT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_ENUM >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_CHAR >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_LONG >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_TIME_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_STRING >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_SHORT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_ENUM >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_CHAR >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_LONG >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_GR_DOUBLE >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_STRING >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_SHORT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_FLOAT >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_ENUM >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_CHAR >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_LONG >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_CTRL_DOUBLE >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_PUT_ACKT >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_PUT_ACKS >,
    & MsgMinderV3 :: m_pushCmpdVectorPayload < DBR_STSACK_STRING >,
    & MsgMinderV3 :: m_pushVectorPayload < DBR_CLASS_NAME >
};

void MsgMinderV3 :: 
    pushResp ( const HdrLargeArray & hdr, const Catalog & payload )
{
    const size_t payloadSize = 
        dbr_size_n ( hdr.m_dataType, hdr.m_count );
    const size_t roundedSize = 
        m_pushHdr ( hdr, payloadSize );
    if ( hdr.m_count == 1 ) {
        if ( hdr.m_dataType > 
            NELEMENTS ( m_pPushScalarPayload ) ) {
                throw std :: range_error (
                    "MsgMinderV3 :: pushResp: dbr "
                    "type out of range" );
        }
        ( this->*m_pPushScalarPayload[hdr.m_dataType] )
            ( roundedSize, payload );
    }
    else {
        if ( hdr.m_dataType > 
            NELEMENTS ( m_pPushVectorPayload ) ) {
                throw std :: range_error (
                    "MsgMinderV3 :: pushResp: dbr "
                    "type out of range" );
        }
        ( this->*m_pPushVectorPayload[hdr.m_dataType] )
            ( roundedSize, hdr.m_count, payload );
    }
}

void MsgMinderV3 :: 
    errorResponse ( 
        const HdrLargeArray & msg, 
        epicsUInt32 cid, const epicsUInt32 legacyStatus, 
        const char * const pFormat, ... )
{
	size_t stringSize = 0u;
    // allocate plenty of space for the message string
	char msgBuf[256]; 
	if ( pFormat ) {
	    va_list args;
		va_start ( args, pFormat );
		int status = epicsVsnprintf (
            msgBuf, sizeof ( msgBuf ), pFormat, args );
		if ( status < 0 ) {
			errlogPrintf (
				"MsgMinderV3 :: errorResponse "
                "formated IO error, format = \"%s\"\n", 
                pFormat );
		}
        else if ( static_cast < size_t > ( status ) 
                    >= sizeof ( msgBuf ) ) {
			errlogPrintf (
				"MsgMinderV3 :: errorResponse "
                "formated IO overflow avoided, "
                "format = \"%s\"\n", 
                pFormat );
            stringSize = sizeof ( msgBuf ) - 1;
            msgBuf [ stringSize ] = '\0';
        }
		else {
			stringSize = 
                static_cast < unsigned > ( status );
		}
	}
    unsigned hdrSize = sizeof ( caHdr );
    if ( msg.sizePostfixIsRequired () ) {
        if ( ! CA_V49( m_protoRev ) ) {
            throw std :: runtime_error (
                "MsgMinderV3 :: errorResponse: "
                "sending a size extended header, but"
                "this client doesnt converse in that "
                "protocol?" );
        }
        hdrSize += 2 * sizeof ( ca_uint32_t );
    }
    HdrLargeArray respHdr;
    respHdr.m_cmmd = CA_PROTO_ERROR;
    respHdr.m_postsize = hdrSize + stringSize + 1u;
    respHdr.m_dataType = 0;
    respHdr.m_count = 0;
    respHdr.m_cid = cid;
    respHdr.m_available = legacyStatus;
    m_pushHdr ( respHdr, respHdr.m_postsize );
    m_pushHdr ( msg, 0 );
    ConstString constString ( msgBuf, stringSize );
    m_iter.pushString ( constString );
    m_iter.commit ();
}

void MsgMinderV3 :: 
    errorResponse ( 
        const HdrLargeArray & msg, 
        ca_uint32_t cid, ca_uint32_t legacyStatus,
        const Catalog & diag )
{
    FixedString < 256 > failureContext ( "" );
    ConstString failureContextDflt ( "<Failure Context Not Supplied>" );
    Clerk clerk ( diag );
    clerk.get ( 
        Responder :: context, 
        failureContext, failureContextDflt );
    errorResponse ( msg, cid, legacyStatus, 
         failureContext.c_str () );
}

// we prefer for the implementation of m_pushOldString
// to optimize the length of a scaler string, and not 
// transmit on the wire the entire string as has been 
// the behavior in the past when the string is at the
// end of an atomic payload, but this improvement
// hasnt been installed, at this time. This is partly
// because in dbr structure embedded situations the 
// optimization would not be appropriate.
template < size_t N >
inline void MsgMinderV3 :: 
    m_pushOldString ( const char ( & pStrBuf ) [ N ] )
{
    size_t i;
    for ( i = 0; i < N - 1; i++ ) {
        epicsInt8 charVal = 
            static_cast < epicsInt8 > ( pStrBuf [ i ] );
        if ( charVal == '\0' ) {
            break;
        }
        m_iter.pushInt8 ( charVal );
    }
    m_iter.pushZeros ( N - i );
}

template < size_t N >
inline void MsgMinderV3 :: 
    m_pushVal ( const char ( & pStrBuf ) [ N ] )
{
    m_pushOldString ( pStrBuf );
}

template < size_t M, size_t N >
void MsgMinderV3 :: 
    m_pushOldStringArray ( 
        const char ( & pStrBuf ) [ M ][ N ], size_t nElem )
{
    // sunpro seems to use an integer type for M ?
    // so I was unable to use std :: min
    if ( nElem > M ) {
        nElem = M;
    }
    for ( size_t i = 0; i < nElem; i++ ) {
        this->m_pushOldString ( pStrBuf[i] );
    }
    m_iter.pushZeros ( ( M - nElem ) * N );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_string & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_stsack_string & dbr )
{
    m_iter.pushUInt16 ( dbr.status );
    m_iter.pushUInt16 ( dbr.severity );
    m_iter.pushUInt16 ( dbr.ackt );
    m_iter.pushUInt16 ( dbr.acks );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_int & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_short & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_float & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_enum & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_char & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt8 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_long & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_sts_double & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_string & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_short & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
    m_iter.pushUInt16 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_float & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_enum & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
    m_iter.pushUInt16 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_char & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
    m_iter.pushUInt16 ( 0 );
    m_iter.pushUInt8 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_long & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_time_double & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushUInt32 ( dbr.stamp.secPastEpoch );
    m_iter.pushUInt32 ( dbr.stamp.nsec );
    m_iter.pushUInt32 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_int & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt16 ( dbr.upper_disp_limit );
    m_iter.pushInt16 ( dbr.lower_disp_limit );
    m_iter.pushInt16 ( dbr.upper_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_warning_limit );
    m_iter.pushInt16 ( dbr.lower_warning_limit );
    m_iter.pushInt16 ( dbr.lower_alarm_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_short & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt16 ( dbr.upper_disp_limit );
    m_iter.pushInt16 ( dbr.lower_disp_limit );
    m_iter.pushInt16 ( dbr.upper_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_warning_limit );
    m_iter.pushInt16 ( dbr.lower_warning_limit );
    m_iter.pushInt16 ( dbr.lower_alarm_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_float & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.precision );
    m_iter.pushUInt16 ( 0 );
    m_pushOldString ( dbr.units );
    m_iter.pushFloat32 ( dbr.upper_disp_limit );
    m_iter.pushFloat32 ( dbr.lower_disp_limit );
    m_iter.pushFloat32 ( dbr.upper_alarm_limit );
    m_iter.pushFloat32 ( dbr.upper_warning_limit );
    m_iter.pushFloat32 ( dbr.lower_warning_limit );
    m_iter.pushFloat32 ( dbr.lower_alarm_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_enum & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.no_str );
    if ( dbr.no_str < 0 ) {
        throw std :: range_error ( 
            "dbr.no_str was out of range" );
    }
    m_pushOldStringArray ( dbr.strs, 
        static_cast < size_t > ( dbr.no_str ) );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_char & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt8 ( dbr.upper_disp_limit );
    m_iter.pushInt8 ( dbr.lower_disp_limit );
    m_iter.pushInt8 ( dbr.upper_alarm_limit );
    m_iter.pushInt8 ( dbr.upper_warning_limit );
    m_iter.pushInt8 ( dbr.lower_warning_limit );
    m_iter.pushInt8 ( dbr.lower_alarm_limit );
    m_iter.pushInt8 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_long & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt16 ( dbr.upper_disp_limit );
    m_iter.pushInt16 ( dbr.lower_disp_limit );
    m_iter.pushInt16 ( dbr.upper_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_warning_limit );
    m_iter.pushInt16 ( dbr.lower_warning_limit );
    m_iter.pushInt16 ( dbr.lower_alarm_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_gr_double & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.precision );
    m_iter.pushUInt16 ( 0 );
    m_pushOldString ( dbr.units );
    m_iter.pushFloat64 ( dbr.upper_disp_limit );
    m_iter.pushFloat64 ( dbr.lower_disp_limit );
    m_iter.pushFloat64 ( dbr.upper_alarm_limit );
    m_iter.pushFloat64 ( dbr.upper_warning_limit );
    m_iter.pushFloat64 ( dbr.lower_warning_limit );
    m_iter.pushFloat64 ( dbr.lower_alarm_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_int & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt16 ( dbr.upper_disp_limit );
    m_iter.pushInt16 ( dbr.lower_disp_limit );
    m_iter.pushInt16 ( dbr.upper_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_warning_limit );
    m_iter.pushInt16 ( dbr.lower_warning_limit );
    m_iter.pushInt16 ( dbr.lower_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_ctrl_limit );
    m_iter.pushInt16 ( dbr.lower_ctrl_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_short & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt16 ( dbr.upper_disp_limit );
    m_iter.pushInt16 ( dbr.lower_disp_limit );
    m_iter.pushInt16 ( dbr.upper_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_warning_limit );
    m_iter.pushInt16 ( dbr.lower_warning_limit );
    m_iter.pushInt16 ( dbr.lower_alarm_limit );
    m_iter.pushInt16 ( dbr.upper_ctrl_limit );
    m_iter.pushInt16 ( dbr.lower_ctrl_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_float & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.precision );
    m_iter.pushUInt16 ( 0 );
    m_pushOldString ( dbr.units );
    m_iter.pushFloat32 ( dbr.upper_disp_limit );
    m_iter.pushFloat32 ( dbr.lower_disp_limit );
    m_iter.pushFloat32 ( dbr.upper_alarm_limit );
    m_iter.pushFloat32 ( dbr.upper_warning_limit );
    m_iter.pushFloat32 ( dbr.lower_warning_limit );
    m_iter.pushFloat32 ( dbr.lower_alarm_limit );
    m_iter.pushFloat32 ( dbr.upper_ctrl_limit );
    m_iter.pushFloat32 ( dbr.lower_ctrl_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_enum & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.no_str );
    if ( dbr.no_str < 0 ) {
        throw std :: range_error ( 
            "dbr.no_str was out of range" );
    }
    m_pushOldStringArray ( dbr.strs, 
        static_cast < size_t > ( dbr.no_str ) );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_char & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt8 ( dbr.upper_disp_limit );
    m_iter.pushInt8 ( dbr.lower_disp_limit );
    m_iter.pushInt8 ( dbr.upper_alarm_limit );
    m_iter.pushInt8 ( dbr.upper_warning_limit );
    m_iter.pushInt8 ( dbr.lower_warning_limit );
    m_iter.pushInt8 ( dbr.lower_alarm_limit );
    m_iter.pushInt8 ( dbr.upper_ctrl_limit );
    m_iter.pushInt8 ( dbr.lower_ctrl_limit );
    m_iter.pushUInt8 ( 0 );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_long & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_pushOldString ( dbr.units );
    m_iter.pushInt32 ( dbr.upper_disp_limit );
    m_iter.pushInt32 ( dbr.lower_disp_limit );
    m_iter.pushInt32 ( dbr.upper_alarm_limit );
    m_iter.pushInt32 ( dbr.upper_warning_limit );
    m_iter.pushInt32 ( dbr.lower_warning_limit );
    m_iter.pushInt32 ( dbr.lower_alarm_limit );
    m_iter.pushInt32 ( dbr.upper_ctrl_limit );
    m_iter.pushInt32 ( dbr.lower_ctrl_limit );
}

void MsgMinderV3 :: 
    m_pushFacetsDBR ( const dbr_ctrl_double & dbr )
{
    m_iter.pushInt16 ( dbr.status );
    m_iter.pushInt16 ( dbr.severity );
    m_iter.pushInt16 ( dbr.precision );
    m_iter.pushUInt16 ( 0 );
    m_pushOldString ( dbr.units );
    m_iter.pushFloat64 ( dbr.upper_disp_limit );
    m_iter.pushFloat64 ( dbr.lower_disp_limit );
    m_iter.pushFloat64 ( dbr.upper_alarm_limit );
    m_iter.pushFloat64 ( dbr.upper_warning_limit );
    m_iter.pushFloat64 ( dbr.lower_warning_limit );
    m_iter.pushFloat64 ( dbr.lower_alarm_limit );
    m_iter.pushFloat64 ( dbr.upper_ctrl_limit );
    m_iter.pushFloat64 ( dbr.lower_ctrl_limit );
}

} // end of namespace ca
