/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *      ServerThr.cc,v 1.14 2009/12/04 23:28:07 jhill Exp
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include <string>
#include <stdexcept>
#include <algorithm>

#include "epicsVersion.h"
#include "errlog.h"

#define epicsExportSharedSymbols
#include "caIntfIO.h"
#include "ServerThr.h"
#include "caSvrCirMgr.h"

// include a version string for POSIX systems
static const char pVersionCAS[] = 
    "@(#) " EPICS_VERSION_STRING 
    ", CA Portable Server Library " 
    "2009/12/04 23:28:07";

namespace ca {
namespace Svr {

//epicsShareDef PCat catPtrEmpty;

PServer ServerThr :: 
    factory ( const PService & pService, 
    const PCat & pNameResChanDescr, 
    std :: auto_ptr < NewCircuitAgent > pNewCircuitAgent, 
    std :: auto_ptr < DatagramAgent > pDatagramAgent )
{
    ServerThr * pServerThr = new ServerThr ( pService, pNameResChanDescr, 
                                pNewCircuitAgent, pDatagramAgent );
    return PServer ( pServerThr->m_serverHandle );
}

ServerThr :: 
    ServerThr ( const PService & pService, const PCat & pNameResChanDescr,
        std :: auto_ptr < NewCircuitAgent > pNewCircuitAgent,
        std :: auto_ptr < DatagramAgent > pDatagramAgent ) :
    // the order here is significant
    m_thread ( 
        m_entryPt, 
        "CAS TCP Listen", 
        epicsThreadGetStackSize ( epicsThreadStackMedium ), 
        epicsThreadPriorityLow ), 
    m_pService ( pService ),
    m_pTimerQueue ( epicsTimerQueueActive :: allocate ( false ) ),
    m_pNewCircuitAgent ( pNewCircuitAgent ),
    m_debugLevel ( 0u ),
    //m_nEventsProcessed ( 0u ),
    //m_nEventsPosted ( 0u ),
    m_enabled ( true )
#if 0
    m_valueEvent ( registerEvent ( "value" ) ),
	m_logEvent ( registerEvent ( "log" ) ),
	m_alarmEvent ( registerEvent ( "alarm" ) )
#endif
{

    if ( m_pService.isNill () ) {
        throw std :: runtime_error ( 
            "Server :: Server (): attempting to construct "
            "CA server attached to a nill service" );
    }

    m_pDGMgr = ca :: Svr :: DG :: Mgr :: 
        factory ( pService, pNameResChanDescr, *m_pTimerQueue, 
            pDatagramAgent, Ptr < VecAllocator > ( m_vecAllocatorHandle ) );
    m_thread.start ();
}

ServerThr :: ~ServerThr ()
{
    m_pDGMgr.clear ();
#pragma message ( "more synchronization needed here" )
    m_enabled = false;
    m_pNewCircuitAgent->interruptBlockForNewClient ();
    m_thread.exitWait ();
	// delete all clients, the act of removing them
    // from the list causes them to be unreferenced
    // and therefore destroyed
    m_cirMgrList.clear ();
}

// When calling down the hierarchy with show we
// have a similar situation to callbacks where
// we must be careful to release the current lock
// so there are not unanticipated lock hierarchy 
// inversions. This approach makes the server
// design easier because we are allowed a more
// natural lock hierarchy.
void ServerThr :: show ( unsigned level ) const
{
    printf ( "Channel Access Server V%s\n",
        CA_VERSION_STRING ( CA_MINOR_PROTOCOL_REVISION ) );
    printf ( "\trevision %s\n", pVersionCAS );

    if ( level > 0 ) {
        // The approach below, where we transfer the
        // circuit list to a temporary, prevents
        // the unlocked list from being desturbed.
        // Since the contents of the list are 
        // reference counting smart pointers then 
        // everything works correctly. If a circuit
        // is unreferenced while the lock is applied
        // we keep a reference here until show completes,
        // and then the circuit is destroyed when the 
        // temporary list is destroyed.
        std :: list < Cir :: PMgr > tmpList ( m_cirMgrList );
        std :: list < Cir :: PMgr > :: 
            const_iterator iter = tmpList.begin ();
        {
            Mutex :: Unlocker unlocker ( m_mutex );
            while ( iter != tmpList.end () ) {
                (*iter)->show ( level );
                iter++;
            }
        }
        m_mutex.show ( level - 1 );
        m_vecAllocator.show ( level - 1 );
        m_thread.show ( level - 1 );
        {
            Mutex :: Unlocker unlocker ( m_mutex );
            // these objects are destroyed only 
            // by the servers dtor
            m_pDGMgr->show ( level - 1 );
            m_pService->show ( level - 1 );
            m_pTimerQueue->show ( level - 1 );
        }
        m_pNewCircuitAgent->show ( level - 1);
        m_refCtr.show ( level - 1 );
        printf ( "m_debugLevel = %u\n", m_debugLevel );
        printf ( "m_enabled = %u\n", m_enabled ? 'T' : 'F' );
    }
}

#if 0
void ServerThr :: destroyClient ( CirSvr & client )
{
    {
        Guard guard ( m_mutex );
	    m_cirMgrList.remove ( client );
    }
    delete & client;
}
#endif

////////////////////////////////////
// ServerThr :: M_EntryPt
////////////////////////////////////
ServerThr :: M_EntryPt :: M_EnclLnk
    ServerThr :: M_EntryPt :: m_enclLnk = 
            & ServerThr :: m_entryPt;
void ServerThr :: M_EntryPt ::
    run ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( serverThr.m_mutex );
    while ( serverThr.m_enabled ) {
        try {
            std :: auto_ptr < Circuit > pCircuit;
            {
                Guard :: Release release ( guard );
                pCircuit = serverThr.m_pNewCircuitAgent->blockForNewClient ();
            }
            if ( pCircuit.get () ) {
                Cir :: PMgr pMgr = 
                    Cir :: Mgr :: factory ( serverThr.m_pService, pCircuit, 
                        Ptr < VecAllocator > ( serverThr.m_vecAllocatorHandle ),
                        Ptr < Cir :: TerminateNotify > 
                            ( serverThr.m_terminateNotifyHandle ) );
	            serverThr.m_cirMgrList.push_back ( pMgr );
            }
        }
        catch ( std :: exception & excep ) {
            Guard :: Release release ( guard );
            errlogPrintf (
                "Server :: NewClientThread :: run: "
                "recovering from exception \"%s\"\n",
                excep.what () );
            epicsThreadSleep ( 60.0 );
        }
        catch ( ... ) {
            Guard :: Release release ( guard );
            errlogPrintf (
                "Server :: NewClientThread :: run: "
                "recovering from non-standard exception\n" );
            epicsThreadSleep ( 60.0 );
        }
    }
}
void ServerThr :: M_EntryPt ::
    show ( unsigned level ) const
{
    const ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( serverThr.m_mutex );
    serverThr.show ( level );
}

bool isInactive ( const Cir :: PMgr & pMgr )
{
    return ! pMgr->isActive ();
}

////////////////////////////////////
// M_CirTerminateNotify
////////////////////////////////////
ServerThr :: M_CirTerminateNotify :: M_EnclLnk
    ServerThr :: M_CirTerminateNotify :: m_enclLnk = 
            & ServerThr :: m_cirTerminateNotify;
void ServerThr :: M_CirTerminateNotify ::
    terminateNotify ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    // if we know at least one of them is inactive we can remove
    // them from the list an dthe unreference will cause them
    // to be destroyed
    std :: list < Cir :: PMgr > tmpList;
    remove_copy_if ( 
        serverThr.m_cirMgrList.begin (),
        serverThr.m_cirMgrList.end (),
        tmpList.begin (),
        isInactive );
    {
        // unreference (and destroy) all of them without 
        // holding the lock
        epics :: Mutex :: Unlocker unlocker ( serverThr.m_mutex );
        tmpList.clear ();
    }
}

////////////////////////////////////
// M_ServerHandle
////////////////////////////////////
ServerThr :: M_ServerHandle :: M_EnclLnk
    ServerThr :: M_ServerHandle :: m_enclLnk = 
            & ServerThr :: m_serverHandle;
ServerThr :: M_ServerHandle :: TargetMutexPair 
    ServerThr :: M_ServerHandle :: m_targetMutexPair ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( serverThr.m_server, serverThr.m_mutex );
}
HandleIntf < Server > & 
    ServerThr :: M_ServerHandle :: m_clone ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    serverThr.m_refCtr.reference ();
    return *this;
}
void ServerThr :: M_ServerHandle :: m_release ()
{
    ServerThr * pServerThr ( & m_enclLnk.enclosure ( *this ) );
    if ( pServerThr->m_refCtr.unreference () ) {
        delete pServerThr;
        // no code here considering above delete
    }
    // no code here considering above delete
}

////////////////////////////////////
// M_VecAllocatorHandle handle
////////////////////////////////////
ServerThr :: M_VecAllocatorHandle :: M_EnclLnk
    ServerThr :: M_VecAllocatorHandle :: m_enclLnk = 
            & ServerThr :: m_vecAllocatorHandle;
ServerThr :: M_VecAllocatorHandle :: TargetMutexPair 
    ServerThr :: M_VecAllocatorHandle :: m_targetMutexPair ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( serverThr.m_vecAllocator, serverThr.m_mutex );
}
HandleIntf < VecAllocator > & 
    ServerThr :: M_VecAllocatorHandle :: m_clone ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    serverThr.m_refCtr.reference ();
    return *this;
}
void ServerThr :: M_VecAllocatorHandle :: m_release ()
{
    ServerThr * pServerThr ( & m_enclLnk.enclosure ( *this ) );
    if ( pServerThr->m_refCtr.unreference () ) {
        delete pServerThr;
        // no code here considering above delete
    }
    // no code here considering above delete
}

////////////////////////////////////
// M_TerminateNotifyHandle handle
////////////////////////////////////
ServerThr :: M_TerminateNotifyHandle :: M_EnclLnk
    ServerThr :: M_TerminateNotifyHandle :: m_enclLnk = 
            & ServerThr :: m_terminateNotifyHandle;
ServerThr :: M_TerminateNotifyHandle :: TargetMutexPair 
    ServerThr :: M_TerminateNotifyHandle :: m_targetMutexPair ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( serverThr.m_cirTerminateNotify, serverThr.m_mutex );
}
HandleIntf < Cir :: TerminateNotify > & 
    ServerThr :: M_TerminateNotifyHandle :: m_clone ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    serverThr.m_refCtr.reference ();
    return *this;
}
void ServerThr :: M_TerminateNotifyHandle :: m_release ()
{
    ServerThr * pServerThr ( & m_enclLnk.enclosure ( *this ) );
    if ( pServerThr->m_refCtr.unreference () ) {
        delete pServerThr;
        // no code here considering above delete
    }
    // no code here considering above delete
}

void casVerifyFunc ( 
    const char * pFile, unsigned line, const char * pExp )
{
    fprintf ( stderr, 
        "the expression \"%s\" didnt evaluate "
        "to boolean true \n",
        pExp );
    fprintf ( stderr,
        "and therefore internal problems are "
        "suspected at line %u in \"%s\"\n",
        line, pFile );
    fprintf( stderr,
        "Please forward above text to "
        "johill@lanl.gov - thanks\n");
}

void serverToolDebugFunc (
    const char *pFile, unsigned line, const char *pComment )
{
    fprintf ( stderr,
        "Bad server tool response detected at "
        "line %u in \"%s\" because \"%s\"\n", 
        line, pFile, pComment );
}

#if 0
void Server :: m_sendBeacon ( ca_uint32_t beaconNo )
{
    Guard guard ( m_mutex );
    m_pIO->sendBeacon ( guard, beaconNo );
}

void Server :: destroyMonitor ( casMonitor &  mon )
{
    mon.~casMonitor ();
    m_casMonitorFreeList.release ( & mon );
}
#endif

#if 0
void Server :: updateEventsPostedCounter ( unsigned nNewPosts )
{
    Guard guard ( m_diagnosticCountersMutex );
    m_nEventsPosted += nNewPosts;
}

unsigned Server :: subscriptionEventsPosted () const
{
    Guard guard ( m_diagnosticCountersMutex );
    return m_nEventsPosted;
}

void Server :: incrEventsProcessedCounter ()
{
    Guard guard ( m_diagnosticCountersMutex );
    m_nEventsProcessed ++;
}

unsigned Server :: subscriptionEventsProcessed () const
{
    Guard guard ( m_diagnosticCountersMutex );
    return m_nEventsProcessed;
}
#endif

////////////////////////////////////
// M_Server
////////////////////////////////////
ServerThr :: M_Server :: M_EnclLnk
    ServerThr :: M_Server :: m_enclLnk = 
            & ServerThr :: m_server;
unsigned ServerThr :: M_Server :: 
    getDebugLevel () const
{
    const ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( serverThr.m_mutex );
    return serverThr.m_debugLevel;
}

void ServerThr :: M_Server :: 
    setDebugLevel ( unsigned debugLevel )
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( serverThr.m_mutex );
    serverThr.m_debugLevel = debugLevel;
}

void ServerThr :: M_Server :: generateBeaconAnomaly ()
{
    ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    serverThr.m_pDGMgr->generateBeaconAnomaly ();
}

void ServerThr :: M_Server :: show ( unsigned level ) const
{
    const ServerThr & serverThr ( m_enclLnk.enclosure ( *this ) );
    serverThr.show ( level );
}

#if 0
inline EventMask Server :: M_Server :: valueEventMask() const
{
    return this->m_valueEvent;
}

inline EventMask Server :: M_Server :: logEventMask() const
{
    return this->m_logEvent;
}

inline EventMask Server :: M_Server :: alarmEventMask() const
{
    return this->m_alarmEvent;
}
#endif

#if 0
casMonitor & Server :: casMonitorFactory ( 
    casChannelI & chan, caResId clientId, 
    const unsigned long count, const unsigned type, 
    const EventMask & mask, 
    casMonitorCallbackInterface & cb )
{
    return * new ( m_casMonitorFreeList ) casMonitor 
            ( clientId, chan, count, type, mask, cb );
}

void Server :: casMonitorDestroy ( casMonitor & cm )
{
    cm.~casMonitor ();
    m_casMonitorFreeList.release ( & cm );
}
#endif


#if 0 // old caServer stuff
#error defunct

#include "dbMapper.h"		// ait to dbr types 
#include "gddAppTable.h"	// EPICS application type table
#include "fdManager.h"

#define epicsExportSharedSymbols
#include "casMonitor.h"
#include "caServerI.h"

EventRegistry :: ~EventRegistry ()
{
    this->traverse ( & EventMaskEntry::destroy );
}

caServer :: caServer ()
{
    static bool init = false;
    
    if ( ! init ) {
        gddMakeMapDBR(gddApplicationTypeTable::app_table);
        init = true;
    }    

    this->pCAS = new caServerI ( *this );
}

caServer :: ~caServer()
{
	if (this->pCAS) {
		delete this->pCAS;
        this->pCAS = NULL;
	}
}

pvExistReturn caServer :: pvExistTest ( const casCtx & ctx, 
	const caNetAddr & /* clientAddress */, const char * pPVAliasName )
{
    return this->pvExistTest ( ctx, pPVAliasName );
}

pvExistReturn caServer :: pvExistTest ( const casCtx &, const char * )
{
	return pverDoesNotExistHere;
}

pvCreateReturn caServer :: createPV ( const casCtx &, const char * )
{
	return S_casApp_pvNotFound;
}

pvAttachReturn caServer :: pvAttach ( 
    const casCtx &ctx, const char *pAliasName )
{
	// remain backwards compatible (call deprecated routine)
	return this->createPV ( ctx, pAliasName );
}

EventMask caServer :: registerEvent (const char *pName) // X aCC 361
{
	if (this->pCAS) {
		return this->pCAS->registerEvent(pName);
	}
	else {
		EventMask emptyMask;
		printf("caServer:: no server internals attached\n");
		return emptyMask;
	}
}

void caServer :: show(unsigned level) const
{
	if (this->pCAS) {
		this->pCAS->show(level);
	}
	else {
		printf("caServer:: no server internals attached\n");
	}
}

void caServer :: setDebugLevel (unsigned level)
{
	if (pCAS) {
		this->pCAS->setDebugLevel(level);
	}
	else {
		printf("caServer:: no server internals attached\n");
	}
}

unsigned caServer :: getDebugLevel () const // X aCC 361
{
    if (pCAS) {
        return this->pCAS->getDebugLevel();
    }
    else {
        printf("caServer:: no server internals attached\n");
        return 0u;
    }
}

EventMask caServer :: valueEventMask () const // X aCC 361
{
    if (pCAS) {
        return this->pCAS->valueEventMask();
    }
    else {
        printf("caServer:: no server internals attached\n");
        return EventMask();
    }
}

EventMask caServer :: logEventMask () const // X aCC 361
{
    if (pCAS) {
        return this->pCAS->logEventMask();
    }
    else {
        printf("caServer:: no server internals attached\n");
        return EventMask();
    }
}

EventMask caServer :: alarmEventMask () const // X aCC 361
{
    if ( pCAS ) {
        return this->pCAS->alarmEventMask ();
    }
    else {
        printf ( "caServer:: no server internals attached\n" );
        return EventMask ();
    }
}

class epicsTimer & caServer :: createTimer ()
{
    return fileDescriptorManager.createTimer ();
}

unsigned caServer :: subscriptionEventsProcessed () const // X aCC 361
{
    if ( pCAS ) {
        return this->pCAS->subscriptionEventsProcessed ();
    }
    else {
        return 0u;
    }
}

unsigned caServer :: subscriptionEventsPosted () const // X aCC 361
{
    if ( pCAS ) {
        return this->pCAS->subscriptionEventsPosted ();
    }
    else {
        return 0u;
    }
}

#endif

} // end of namespace Svr
} // end of namespace ca

