
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

#ifndef ServerThrh
#define ServerThrh

#include <list>
#include <memory>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_ServerThrh
#   undef epicsExportSharedSymbols
#endif

// external headers included here
#include "tsFreeList.h"
#include "caProto.h"
#include "epicsTimer.h"
#include "epicsMutex.h"
#include "autoDestroyPtr.h"
#include "EnclLnk.h"

#ifdef epicsExportSharedSymbols_ServerThrh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caServer.h"
#include "caIntfIO.h"
#include "ComBuf.h"
#include "casdef2.h"
#include "caSvrDGMgr.h"
#include "caSvrCirMgr.h"

namespace ca {

using namespace da;
struct DatagramAgent;

namespace Svr {

struct IntfIO;
struct ServerIO;
struct NewCircuitAgent;

class ServerThr {
public:
    static PServer factory ( const PService &, const PCat & pNameResChanDescr, 
                            std :: auto_ptr < NewCircuitAgent >, 
                            std :: auto_ptr < DatagramAgent > );
    ServerThr ( const PService &, const PCat & pNameResChanDescr, 
                std :: auto_ptr < NewCircuitAgent >, 
                std :: auto_ptr < DatagramAgent > );
    ~ServerThr ();
    void show ( unsigned int level ) const;
private:
    class M_EntryPt : 
        public epicsThreadRunable {
    public:
        void run ();
        void show ( unsigned level ) const;
    private:
        typedef const epics :: EnclLnk < ServerThr, M_EntryPt > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_Server : 
        public Server {
    public:
	    unsigned getDebugLevel() const;
	    void setDebugLevel ( unsigned debugLevelIn );
        void generateBeaconAnomaly ();
    	void show ( unsigned level ) const;
    private:
        typedef const epics :: EnclLnk < ServerThr, M_Server > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_ServerHandle : 
        public HandleIntf < Server > {
    private:
		TargetMutexPair m_targetMutexPair ();
		HandleIntf < Server > & m_clone ();
		void m_release ();
        typedef const epics :: EnclLnk < ServerThr, M_ServerHandle > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_VecAllocator :
        public VecAllocator {
    public:
        Vec < epicsUInt8 > alloc ();
        size_t blockSize () const;
        //void destroy ( ComBuf & );
        void show ( unsigned level ) const;
    private:
        class M_Destructor :
            public ComBuf :: Destructor {
        public:
            void destroy ( ComBuf & );
        private:
            typedef const epics :: EnclLnk < M_VecAllocator, M_Destructor > M_EnclLnk;
            static M_EnclLnk m_enclLnk;
        };
        mutable epics :: Mutex m_mutex;
        M_Destructor m_destructor;
        tsFreeList < ComBuf, 0x10 > m_comBufFreeList;
        typedef const epics :: EnclLnk < ServerThr, M_VecAllocator > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_VecAllocatorHandle : 
        public HandleIntf < VecAllocator > {
    private:
		TargetMutexPair m_targetMutexPair ();
		HandleIntf < VecAllocator > & m_clone ();
		void m_release ();
        typedef const epics :: EnclLnk < ServerThr, M_VecAllocatorHandle > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_CirTerminateNotify : 
        public Cir :: TerminateNotify {
    public:
        void terminateNotify ();
    private:
        typedef const epics :: EnclLnk < ServerThr, M_CirTerminateNotify > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_TerminateNotifyHandle : 
        public HandleIntf < Cir :: TerminateNotify > {
    private:
		TargetMutexPair m_targetMutexPair ();
		HandleIntf < Cir :: TerminateNotify > & m_clone ();
		void m_release ();
        typedef const epics :: EnclLnk < ServerThr, M_TerminateNotifyHandle > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    // the order here is significant
    std :: list < Cir :: PMgr > m_cirMgrList;
    mutable epics :: Mutex m_mutex;
    M_ServerHandle m_serverHandle;
    M_VecAllocatorHandle m_vecAllocatorHandle;
    M_TerminateNotifyHandle m_terminateNotifyHandle;
    M_EntryPt m_entryPt;
    M_VecAllocator m_vecAllocator;
    M_CirTerminateNotify m_cirTerminateNotify;
    M_Server m_server;
    epicsThread m_thread;
    DG :: PMgr m_pDGMgr;
    PService m_pService;
    autoDestroyPtr < epicsTimerQueueActive, 
        & epicsTimerQueueActive :: release > m_pTimerQueue;
    std :: auto_ptr < NewCircuitAgent > m_pNewCircuitAgent;
    RefCtr m_refCtr;
	unsigned m_debugLevel;
    //unsigned m_nEventsProcessed; 
    //unsigned m_nEventsPosted; 
    bool m_enabled;
#if 0
    EventMask m_valueEvent; 
	EventMask m_logEvent; 
	EventMask m_alarmEvent; 
#endif
    //void m_sendBeacon ( ca_uint32_t beaconNo );
	//void m_newCircuitThread ( IntfIO & );
    ServerThr ( const ServerThr& ); // disabled
    ServerThr & operator = ( const ServerThr & ); // disabled
};

} // end of namespace Svr
} // end of namespace ca

#endif // ifndef ServerThrh
