
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
//
// caIntfIO.h,v 1.5 2009/12/04 23:28:07 jhill Exp
//

#ifndef caIntfIOh
#define caIntfIOh 

#include <memory>
#include <vector>

#include "caVec.h"
#include "caPtr.h"
#if 0
#include "caNetAddr.h"
#endif

namespace ca {

class NetAddr;

class CircuitTerminate :
    public std :: exception {
public:
    // CircuitTerminate ();
    ~CircuitTerminate () throw () {}
    const char * what () const throw ();
};

class CircuitDisconnect :
    public std :: exception {
public:
    CircuitDisconnect ( const char * );
    ~CircuitDisconnect () throw () {}
    const char * what () const throw ();
private:
    std :: string m_reason;
};

struct Reader { 
    virtual ~Reader () {}
    // The read function returns a reference counting 
    // pointer to the buffer containing the next segment
    // of the input stream.
    // o For circuit implementations, the read function 
    // is expected to block until at least one byte is 
    // available, or an orderly circuit shutdown condition
    // is detected (indicated by returning a nill pointer). 
    // o Datagram implementatiuons do not block, and 
    // instead return frame segments repeatedly until 
    // an end of frame is detected (indicated by returning 
    // a nill pointer). 
    //
    // The read function throws an exception if the 
    // circuit terminates for an unexpected reason. 
    //
    virtual Vec < const epicsUInt8 > read ( size_t nBytesMax ) = 0;
    // A rewind functionality is needed so that we can 
    // attach a Catalog to an input payload. There must be
    // a special implementation of Reader for storing the 
    // payload which has the ability to remeber of a list of 
    // Vec.
    // Input streams such as a sockets will of course not 
    // rewind, and will therefore throw an exception in 
    // their implementation of rewind.
    virtual void rewind () = 0;
};

struct Writer { 
    virtual ~Writer () {}
    // The write function is permitted to block until 
    // a sufficent number of bytes have been written. 
    // It throws an exception if there is a circuit 
    // disconnect or some other form of communication 
    // failure.
    //
    // When the server calls this function, the 
    // ByteVecPtr passed here is locked with a noop
    // mutex.
    //
    virtual void write ( const Vec  < const epicsUInt8 > & ) = 0;
};

struct CircuitOut { 
    virtual ~CircuitOut () {}
    virtual void write ( const Vec  < const epicsUInt8 > & ) = 0;
    virtual void shutdown () = 0; // also cancels a blocking write
    virtual void flush () = 0;
    virtual void show ( unsigned level ) const = 0;
};

struct CircuitIn { 
    virtual ~CircuitIn () {}
    // The read function returns the number of bytes 
    // that it copied to the destination buffer. The 
    // read function is expected to block until at least 
    // one byte is available. The read function must return 
    // any bytes that are currently available - even if 
    // only one byte of additional data is returned. It 
    // must not block until the number of bytes 
    // requested are available.
    // The read function throws an exception if the circuit 
    // terminates for any reason. It does not return zero to 
    // indicate a circuit disconnect.
    virtual Vec < const epicsUInt8 > read ( size_t nBytesMax ) = 0;
    virtual void shutdown () = 0; // also cancels a blocking read
    virtual void show ( unsigned level ) const = 0;
};

struct DatagramOut { 
    // To send a frame the following calls are required
    // beginFrame ()
    // writeFrameSeg ()
    // .
    // .
    // .
    // endFrame ()
    virtual ~DatagramOut () {}
    virtual void beginFrame ( const NetAddr & dest ) = 0;
    virtual void writeFrameSeg ( const Vec < const epicsUInt8 > & ) = 0;
    // endFrame is allowed to block
    virtual void endFrame () = 0;
    // implementation of endFrame must throw an exception 
    // when it is interrupted
    virtual void interruptEndFrame () = 0;
    virtual void show ( unsigned level ) const = 0;
    virtual void sendBeacon ( epicsUInt32 beaconNumber ) = 0;
};

struct DatagramIn {
    virtual ~DatagramIn () {}
    // o This function blocks until a datagram arrives.
    // o This function must throw an exception when it 
    // is interrupted (see interruptLoadFrame).
    virtual void loadFrame ( NetAddr & srcAddr, NetAddr & intfAddr ) = 0;
    // implementation of endFrame must throw an exception 
    // when it is interrupted
    virtual void interruptLoadFrame () = 0;
    // To obtain the entire frame, readFrameSeg is called 
    // repeatedly until zero bytes are returned. This
    // function does not block.
    virtual Vec < const epicsUInt8 > readFrameSeg ( size_t nBytesMax ) = 0;
#if 0
    // What address is to be used to create a virtual circuit
    // to the network interface that datagrams are being
    // received on.
    virtual NetAddr circuitAddress () const = 0;
    // What destination address is used when sending a 
    // broadcast on the network interface that datagrams 
    // are being received on.
    virtual NetAddr broadcastAddress () const = 0;
    // Whether the the DatagramAgent is known to 
    // receive only datagrams addressed to unicast 
    // addresses. If unknown, or uncertain, set unicastOnly 
    // to false.
#endif
    virtual bool unicastsOnly () const = 0;
    virtual void show ( unsigned level ) const = 0;
};

struct VecAllocator {
    virtual Vec < epicsUInt8 > alloc () = 0;
    virtual size_t blockSize () const = 0;
};

struct Circuit {
    virtual ~Circuit () {}
    virtual std :: auto_ptr < CircuitIn > 
        createInputStream ( const Ptr < VecAllocator > & ) = 0;
    virtual std :: auto_ptr < CircuitOut > createOutStream () = 0;
    virtual void show ( unsigned level ) const = 0;
};

struct DatagramAgent {
    virtual ~DatagramAgent () {}
    virtual size_t numberOfDatagramInLeft () const = 0;
    virtual std :: auto_ptr < DatagramIn > 
        createNextDatagramIn ( const Ptr < VecAllocator > & ) = 0;
    virtual std :: auto_ptr < DatagramOut > createDatagramOut () = 0;
    virtual void show ( unsigned level ) const = 0;
};

namespace Svr {

struct NewCircuitAgent {
    virtual ~NewCircuitAgent () {}
    virtual std :: auto_ptr < Circuit > blockForNewClient () = 0;
    virtual void interruptBlockForNewClient () = 0;
    virtual void show ( unsigned level ) = 0;
};

} // end of namespace Svr

} // end of namespace ca

#endif // caIntfIOh
