
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

//
//
//      Author: Jeffrey O. Hill
//              johill@lanl.gov
//

#include <string>
#include <stdexcept>

#include <cstdio>
#include <cstring>

#include "da.h"

#define epicsExportSharedSymbols
#define caNetAddrSock
#include "caNetAddr.h"

namespace ca {

using namespace da;
using namespace std;

const char pn_networkAddress[] = "networkAddress";
const char pn_internetV4[] = "internetV4";
const char pn_address[] = "address";
const char pn_port[] = "port";

ID_networkAddress pi_networkAddress;
ID_internetV4 pi_internetV4;
ID_address pi_address;
ID_port pi_port;

static class NetAddrSelfTest {
public:
    NetAddrSelfTest (); 
} NetAddrSelfTestDuringLoad;

//
// NetAddr :: stringConvert ()
//
void NetAddr :: 
    stringConvert ( char * pString, unsigned stringLength ) const
{
    if ( this->type == casnaInet ) {
        ipAddrToA ( &this->addr.ip, pString, stringLength );
        return;
    }
    if ( stringLength ) {
        strncpy ( pString, "<Undefined Address>", stringLength );
        pString[stringLength-1] = '\n';
    }
}

// 
// NetAddr :: clear()
//
void NetAddr :: 
    clear ()
{
    this->type = casnaUDF;
}

//
// NetAddr :: NetAddr()
//
NetAddr :: 
    NetAddr ()
{   
    this->clear();
}

bool NetAddr :: 
    isInet () const
{
    return this->type == casnaInet;
}

bool NetAddr :: 
    isValid () const
{
    return this->type != casnaUDF;
}

bool NetAddr :: 
    operator == ( const NetAddr & rhs ) const 
{
    if ( this->type != rhs.type ) {
        return false;
    }
    if ( this->type == casnaInet ) {
        bool addrEqual = ( 
            this->addr.ip.sin_addr.s_addr == 
            rhs.addr.ip.sin_addr.s_addr );
        bool portEqual = (
            this->addr.ip.sin_port == 
            rhs.addr.ip.sin_port );
        return addrEqual && portEqual;
    }
    else {
        return false;
    }
}

bool NetAddr :: 
    operator != ( const NetAddr & rhs ) const
{
    return ! this->operator == ( rhs );
}

//
// This is specified so that compilers will not use one of 
// the following assignment operators after converting to a 
// sockaddr_in or a sockaddr first.
//
// NetAddr NetAddr::operator =(const struct sockaddr_in&)
// NetAddr NetAddr::operator =(const struct sockaddr&)
//
NetAddr & NetAddr :: 
    operator = ( const NetAddr & naIn )
{
    this->addr = naIn.addr;
    this->type = naIn.type;
    return *this;
}   

void NetAddr :: 
    setSockIP ( unsigned long inaIn, unsigned short portIn )
{
    this->type = casnaInet;
    this->addr.ip.sin_family = AF_INET;
    this->addr.ip.sin_addr.s_addr = inaIn;
    this->addr.ip.sin_port = portIn;
}   

void NetAddr :: 
    setSockIP ( const struct sockaddr_in & sockIPIn )
{
    if ( sockIPIn.sin_family != AF_INET ) {
        throw logic_error ( 
                "NetAddr::setSockIP (): address wasnt IP" );
    }
    this->type = casnaInet;
    this->addr.ip = sockIPIn;
}   

void NetAddr :: 
    setSock ( const struct sockaddr & sock )
{
    if ( sock.sa_family != AF_INET ) {
        throw logic_error ( 
                "NetAddr::setSock (): address wasnt IP" );
    }
    this->type = casnaInet;
    const struct sockaddr_in *psip = 
        reinterpret_cast <const struct sockaddr_in*> ( & sock );
    this->addr.ip = *psip;
}   

NetAddr :: 
    NetAddr ( const struct sockaddr_in & sockIPIn )
{
    this->setSockIP ( sockIPIn );
}

NetAddr & NetAddr :: 
    operator = ( const struct sockaddr_in & sockIPIn )
{
    this->setSockIP ( sockIPIn );
    return *this;
}           

NetAddr & NetAddr :: 
    operator = ( const struct sockaddr & sockIn )
{
    this->setSock ( sockIn );
    return *this;
}       

struct sockaddr_in NetAddr :: 
    getSockIP () const
{
    if ( this->type != casnaInet ) {
        throw logic_error ( 
                "NetAddr::getSockIP (): address wasnt IP" );
    }
    return this->addr.ip;
}

struct sockaddr NetAddr :: 
    getSock () const
{
    if ( this->type != casnaInet ) {
        throw logic_error ( 
            "NetAddr::getSock (): address wasnt IP" );
    }
    union {
        struct sockaddr sa;
        struct sockaddr_in sain;
    } retVal;
    retVal.sain = this->addr.ip;
    return retVal.sa;
}

NetAddr :: 
    operator sockaddr_in () const
{
    return this->getSockIP ();
}

NetAddr :: 
    operator sockaddr () const
{
    return this->getSock ();
}

void NetAddr :: 
    show ( unsigned level ) const
{
    char string[64];
    this->stringConvert ( string, sizeof ( string ) );
    string [ sizeof ( string ) - 1 ] = '\0';
    printf ( "NetAddr: addr = %s\n", string );
}

void NetAddr :: 
    selfTest ()
{
    // the dummy field must be greater than or equal to the size of
    // each of the other entries in the union
    if ( sizeof ( this->addr ) != sizeof ( this->addr.pad ) ) {
        fprintf ( stderr, 
            "NetAddr::selfTest ():self test "
            "failed in %s at line %d\n", 
            __FILE__, __LINE__ );
        throw logic_error ( 
            "NetAddr::selfTest (): failed self test" );
    }
}

NetAddrSelfTest :: 
    NetAddrSelfTest ()
{
    NetAddr tmp;
    tmp.selfTest ();
}

class MutatorIPAddrV4 :
    public MutatorCatalog {
public:
    MutatorIPAddrV4 ( NetAddr & target, 
                    const NetAddr * pDflt = 0 );
    void assignDefault ();
    void assignScalarCatalog ( const Catalog & src );
private:
    NetAddr & m_target;
    const NetAddr * m_pDflt;
};

inline MutatorIPAddrV4 ::
    MutatorIPAddrV4 ( NetAddr & target,
            const NetAddr * pDflt ) :
    m_target ( target ), m_pDflt ( pDflt )
{
}

void MutatorIPAddrV4 ::
    assignDefault ()
{
    if ( m_pDflt == 0 ) {
        throw UndefinedProperty ();
    }
    if ( & m_target != m_pDflt ) {
        m_target = *m_pDflt;
    }
}

void MutatorIPAddrV4 ::
    assignScalarCatalog ( const Catalog & src )
{
    Clerk clerk ( src );
    // if they dont supply both a port and 
    // an address, but claim that its an
    // IP address then its fine to let the 
    // exception be caught (and handled 
    // as a programming error) at a higher 
    // level
    unsigned short port;
    unsigned long addr;
    clerk.get ( pi_address, addr );
    clerk.get ( pi_port, port );
    m_target.setSockIP ( addr, port );
}

NetAddr fetchNetworkAddress ( const Catalog & src,
                        const NetAddr * pDflt )
{
    // networkAddress / internetV4 / { port, address }
    typedef MutatorCatalogSingle < ID_internetV4, MutatorIPAddrV4 > MutNetAddr;
    NetAddr target;
    MutatorIPAddrV4 mutatorIPAddrV4 ( target, pDflt );
    MutNetAddr mutNetAddr ( mutatorIPAddrV4 );
    Clerk clerk ( src );
    clerk.get ( pi_networkAddress, mutNetAddr );
    return target;
}

InetAddrV4Catalog ::
    InetAddrV4Catalog ( unsigned address, unsigned port ) :
    m_address ( address ), m_port ( port )
{
}

void InetAddrV4Catalog :: 
    traverse ( Introspector & introspector ) const
{
    introspector.reveal ( pi_address, m_address );
    introspector.reveal ( pi_port, m_port );
}

void InetAddrV4Catalog :: 
    find ( const Ident & id, Viewer & viewer ) const
{
    if ( id == pi_address ) {
        viewer.reveal ( m_address );
    }
    else if ( id == pi_port ) {
        viewer.reveal ( m_port );
    }
}

} // end of namespace ca


