
#ifndef caPtrh
#define caPtrh

//
// Notes
// 1) The functions in type Handle are expected to be SMP 
// thread safe. See osiAtomic.h for some possibilities for
// building an efficent implementation.
// 2) Protections for multithreaded access to the Handle pointer 
// itself in Ptr < T > are expected to be provided by external 
// locking for the class that has member data of type Ptr < T >.
//

#ifndef assert // allow use of epicsAssert.h
#   include <cassert>
#endif
#include <cstdio>
#include <climits>
#include <exception>
#include <stdexcept>
#include <typeinfo>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caPtrh
#   undef epicsExportSharedSymbols
#endif

#include "errlog.h"
#include "epicsTypes.h"
#include "osiAtomic.h"
#include "epicsMutex.h"

#ifdef epicsExportSharedSymbols_caPtrh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caConstStrip.h"

namespace ca {

using std :: type_info;
using std :: exception;
using std :: type_info;

void ignoredExceptMsg ( const type_info & intfType, 
    const char * pFuncName );
void ignoredExceptMsg ( const type_info & intfType, 
    const char * pFuncName, exception & );

//
// Smart Pointer, with Guard Protected 
// Reference Counting 
//
template < class Target, class Mutex = epics :: Mutex > class HandleIntf;
template < class Target, class Handle = HandleIntf < Target > > class Ptr;

// clone copy semantics 
template < class Target, class Handle >
class Ptr {
public:
    typedef Handle Hdl;
    class Locker;
    Ptr (); // makes a nill pointer
    template < class T >
    Ptr ( const Ptr < T, Handle > & ); // clone src Ptr's Handle 
    Ptr ( Handle & ); // clone the handle passed in
    ~Ptr (); // functionally the same as calling clear
    void clear (); // Handle ownership released, become nill
    Ptr release (); // ownership transfer to returned Ptr, become nill
    bool isNill () const;
    bool isntNill () const;
    template < class T >
    Ptr & operator = ( const Ptr < T, Handle > & ); // clone src Ptr's Handle 
    Ptr & operator = ( Handle * ); // clone the handle passed in
    Locker operator -> () const; // exception thrown if nill
    //Locker operator * () const; // exception thrown if nill
    operator Locker () const; // exception thrown if nill
    Ptr clone () const;
    void replaceIfMatch ( const Ptr & pMatch, Ptr & pReplace );
private:
    Handle * m_pHandle;
};

// clone copy semantics 
template < class Target, class Handle >
class Ptr < Target, Handle > :: Locker {
public:
    Locker ( Handle & ); // become lock owner, clone handle passed in
    Locker ( const Locker & ); // lock owner add, clone src Locker's Handle
    Locker & operator = ( const Locker & ); // lock owner add, clone src Locker's Handle
    template < class P >
    bool identicalMutex ( const P & ptr ) const; // true if Ptr and locker use same Mutex
    template < class P >
    typename P :: Locker join ( const P & ptr ); // lcks must match, lck owner add, clone Ptr's Handle
    ~Locker (); // lock owner remove, releases Handle ownership
    Target * operator -> () const; 
    Target & operator * () const; 
    Ptr clone () const;
    template < class T >
    Locker & operator = ( const Ptr < T, Handle > & ); // retarget 
private:
    typedef typename Handle :: TargetMutexPair TargetMutexPair;
    // simplifies the ctor
    class GuardedTarget {
    public:
        GuardedTarget ( const TargetMutexPair & );
        GuardedTarget ( const GuardedTarget & );
        //Guard & guard () const; 
        Target * pTarget () const; 
		void retarget ( const TargetMutexPair & );
    private:
		typedef typename Handle :: Mutex :: Guard Guard;
        Guard m_guard; // needed for lock owner add {ctor, operator =, join}
        Target * m_pTarget; // needed for decent dereference performance
    };
    // yes, there are three pointers, but each one 
    // serves a unique purpose
    GuardedTarget m_guardedTarget;
    Handle * m_pHandle; // needed for release
    static void m_releaseExceptMsg ( Handle &, const char * pExceptName );
};

// If runtime polymorphic behavior is not required 
// then the overhead of a virtual interface might
// be unnecessary. In that situation, then a 
// specialization of this template will provide
// a simpler and more efficent implementation.
template < class T, class M >
class HandleIntf {
public:
	typedef M Mutex;
	typedef T Target;
    virtual ~HandleIntf () {}
    // transporting both together improves performance
    class TargetMutexPair {
    public:
        TargetMutexPair ( Target &, Mutex & );
        Target & target () const;
        Mutex & mutex () const;
    private:
        Target & m_target;
        Mutex & m_mutex;
    };
    TargetMutexPair targetMutexPair ();
    HandleIntf & clone ();
    void release () throw ();
private:
    virtual TargetMutexPair m_targetMutexPair () = 0;
    virtual HandleIntf & m_clone () = 0;
    virtual void m_release () = 0;
};

template < class Target, class Handle >
inline Ptr < Target, Handle > :: 
    Ptr () : m_pHandle ( 0 )
{
}

template < class Target, class Handle >
template < class T >
inline Ptr < Target, Handle > :: 
    Ptr ( const Ptr < T, Handle > & ptr ) :
    m_pHandle ( ptr.m_pHandle ? & ptr.m_pHandle->clone () : 0 )
{
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: 
    Ptr ( Handle & handle ) :
    m_pHandle ( & handle.clone () )
{
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: 
    ~Ptr ()
{
    Handle * pHandleTemp = m_pHandle;
    if ( pHandleTemp ) {
        pHandleTemp->release ();
    }
}

template < class Target, class Handle >
inline void Ptr < Target, Handle > :: 
    clear ()
{
    Handle * pHandleTemp = m_pHandle;
    if ( pHandleTemp ) {
        m_pHandle = 0;
        pHandleTemp->release ();
    }
}

template < class Target, class Handle >
inline Ptr < Target, Handle > 
    Ptr < Target, Handle > :: release ()
{ 
    Handle * pHandleTemp = m_pHandle;
    if ( pHandleTemp ) {
        m_pHandle = 0;
        return Ptr < Target, Handle > ( *pHandleTemp );
    }
    return Ptr < Target, Handle > ();
}

template < class Target, class Handle >
inline bool Ptr < Target, Handle > :: 
    isNill () const
{
    return ( m_pHandle == 0 );
}

template < class Target, class Handle >
inline bool Ptr < Target, Handle > :: 
    isntNill () const
{
    return ( m_pHandle != 0 );
}

template < class Target, class Handle >
template < class T >
inline Ptr < Target, Handle > & 
    Ptr < Target, Handle > :: 
        operator = ( const Ptr < T, Handle > & ptr )
{
    Handle * pHandleTemp = m_pHandle;
    m_pHandle = ptr.m_pHandle ? & ptr.m_pHandle->clone () : 0;
    if ( pHandleTemp ) {
        pHandleTemp->release ();
    }
    return *this;
}

template < class Target, class Handle >
inline Ptr < Target, Handle > & 
    Ptr < Target, Handle > :: 
        operator = ( Handle * pHandle )
{
    Handle * pHandleTemp = m_pHandle;
    m_pHandle = pHandle ? & pHandle->clone () : 0;
    if ( pHandleTemp ) {
        pHandleTemp->release ();
    }
}

template < class Target, class Handle >
inline void Ptr < Target, Handle > :: 
    replaceIfMatch ( const Ptr & pMatch, Ptr & pReplace )
{
    if ( m_pHandle == pMatch.m_pHandle ) {
        this->operator = ( pReplace );
    }
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: 
    operator Locker () const
{
    Handle * pHandleTemp = m_pHandle;
    if ( pHandleTemp == 0 ) {
        throw std :: runtime_error ( 
            "Ptr < T > :: operator -> "
            "nill pointer was dereferenced" );
    }
    return Locker ( pHandleTemp->clone () );
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: Locker :: GuardedTarget ::
    GuardedTarget ( const TargetMutexPair & targetMutexPair ) :
    m_guard ( targetMutexPair.mutex () ),
    m_pTarget ( & targetMutexPair.target () )
{
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: Locker :: GuardedTarget ::
    GuardedTarget ( const GuardedTarget & guardedTarget ) :
    m_guard ( guardedTarget.m_guard ),
    m_pTarget ( guardedTarget.pTarget () )
{
}

template < class Target, class Handle >
inline void
	Ptr < Target, Handle > :: Locker :: GuardedTarget ::
		retarget ( const TargetMutexPair & targetMutexPair )
{
	m_guard.retarget ( targetMutexPair.mutex () );
	m_pTarget = & targetMutexPair.target ();
}

#if 0
template < class Target, class Handle >
inline typename Ptr < Target, Handle > :: Locker :: Guard & 
    Ptr < Target, Handle > :: Locker :: GuardedTarget ::
        guard () const
{
    return m_guard;
}
#endif

template < class Target, class Handle >
inline Target * 
    Ptr < Target, Handle > :: Locker :: GuardedTarget ::
        pTarget () const
{
    return m_pTarget;
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: Locker ::
    Locker ( const Locker & locker ) :
    m_guardedTarget ( locker.m_guardedTarget ),
    m_pHandle ( & locker.m_pHandle->clone () )
{
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: Locker ::
    Locker ( Handle & handle ) :
    m_guardedTarget ( handle.targetMutexPair () ),
    m_pHandle ( & handle.clone () )
{
}

template < class Target, class Handle >
inline Ptr < Target, Handle > :: Locker ::
    ~Locker ()
{
    m_pHandle->release ();
}

template < class Target, class Handle >
inline typename Ptr < Target, Handle > :: Locker 
    Ptr < Target, Handle > :: operator -> () const
{
    // defer to operator Locker
    return *this;
}

#if 0
// Perhaps this poses a danger if this forwards to the 
// operator star in Locker. That would allow a user
// to initialize a raw reference and then dispose
// of the locker while keeping the raw reference
template < class Target, class Handle >
inline typename Ptr < Target, Handle > :: Locker 
    Ptr < Target, Handle > :: operator * () const
{
    // defer to operator Locker
    return *this;
}
#endif

template < class Target, class Handle >
template < class T >
typename Ptr < Target, Handle > :: Locker & 
    Ptr < Target, Handle > :: Locker ::
        operator = ( const Ptr < T, Handle > & ptr )
{
    Handle * pOldHandle = m_pHandle;
    Handle * pNewHandle = ptr.m_pHandle;
    if ( ! pNewHandle ) {
        throw std :: runtime_error ( 
            "Ptr < T > :: Locker :: operator = "
            "assignment using nill pointer" );
    }
    m_pHandle = & pNewHandle->clone ();
	m_guardedTarget.retarget ( pNewHandle->targetMutexPair () );
	pOldHandle->release ();
    return *this;
}

template < class Target, class Handle >
inline Target * Ptr < Target, Handle > :: Locker ::
    operator -> () const
{
    return m_guardedTarget.pTarget ();
}

template < class Target, class Handle >
inline Target & Ptr < Target, Handle > :: Locker ::
    operator * () const
{
    return *m_guardedTarget.pTarget ();
}


template < class Target, class Handle >
inline Ptr < Target, Handle > 
    Ptr < Target, Handle > :: Locker ::
		clone () const
{
    return Ptr < Target, Handle > ( & m_pHandle->clone () );
}

template < class T, class M >
inline HandleIntf  < T, M > :: TargetMutexPair :: 
    TargetMutexPair ( Target & target, Mutex & mutex ) :
    m_target ( target ), m_mutex ( mutex )
{
}

template < class T, class M >
inline T & 
    HandleIntf  < T, M > :: TargetMutexPair :: 
        target () const
{
    return m_target;
}

template < class T, class M >
inline M & 
    HandleIntf  < T, M > :: TargetMutexPair :: 
        mutex () const
{
    return m_mutex;
}

template < class T, class M >
inline typename HandleIntf  < T, M > :: TargetMutexPair 
    HandleIntf  < T, M > :: 
        targetMutexPair ()
{
    return m_targetMutexPair ();
}

template < class T, class M >
inline HandleIntf < T, M > & 
    HandleIntf  < T, M > :: 
        clone ()
{
    return m_clone ();
}

template < class T, class M >
inline void 
    HandleIntf  < T, M > :: 
        release () throw ()
{
    try {
        m_release ();
    }
    catch ( exception & except ) {
        ignoredExceptMsg ( typeid ( *this ),
            "m_release", except );
    }
    catch ( ... ) {
        ignoredExceptMsg ( typeid ( *this ),
            "m_release" );
    }
}

} // end of namespace ca

#endif // ifndef caPtrh
