
#ifndef caRefCtrh
#define caRefCtrh

#include <stdexcept>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caRefCtrh
#   undef epicsExportSharedSymbols
#endif

#include "osiAtomic.h"
#include "epicsAssert.h"

#ifdef epicsExportSharedSymbols_caRefCtrh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

namespace ca {

    // A thread safe reference counting class
    class RefCtr {
    public:
        RefCtr ();
        ~RefCtr ();
        void reference ();
        // returns true if this caused the ref count 
        // to go to zero
        bool unreference ();
        size_t count () const;
        void show ( unsigned level ) const;
    private:
        epics :: osi :: AtomicWord m_refCount;
    };

    inline RefCtr :: RefCtr () 
    {
        using epics :: osi :: atomicSet;
        epics :: osi :: atomicSet ( m_refCount, 0 );
    }

    inline RefCtr :: ~RefCtr ()
    {
        using epics :: osi :: atomicFetch;
        assert ( atomicFetch ( m_refCount ) == 0 );
    }
 
    inline void RefCtr :: reference ()
    {
        using epics :: osi :: AtomicWord;
        using epics :: osi :: atomicIncr;
        AtomicWord newVal = atomicIncr ( m_refCount );
        assert ( newVal >= 0 );
    }

    inline bool RefCtr :: unreference ()
    {   
        using epics :: osi :: AtomicWord;
        using epics :: osi :: atomicDecr;
        AtomicWord newVal = atomicDecr ( m_refCount );
        static const AtomicWord zeroVal = 0; 
        static const AtomicWord oneVal = 1; 
        // underflow indicates a programming error -
        // not easily corrected by catching an exception
        assert ( newVal != zeroVal - oneVal );
        return ( newVal == 0 );
    }

    inline size_t RefCtr :: count () const
    {
        using epics :: osi :: AtomicWord;
        using epics :: osi :: atomicFetch;
        AtomicWord result = atomicFetch ( m_refCount );
        assert ( result >= 0 );
        return static_cast < size_t > ( result );
    }

    inline void RefCtr :: show ( unsigned level ) const
    {
        using epics :: osi :: AtomicWord;
        using epics :: osi :: atomicFetch;
        using namespace std;
        AtomicWord atCurrent = atomicFetch ( m_refCount );;
        if ( atCurrent >= 0 ) {
            unsigned long ulCurrent = 
                static_cast < unsigned long > ( atCurrent );
            printf ( "RefCtr: m_refCount = %lu\n", ulCurrent );
        }
        else {
            printf ( "RefCtr: m_refCount = INVALID\n" );
        }
    }

} // end of namespace ca

#endif // ifndef caRefCtrh
