/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#error defunct

#include <string>
#include <stdarg.h>
#include <stdexcept>

#include "epicsVersion.h"
#include "errlog.h"

#include "addrList.h"

#define epicsExportSharedSymbols
#define caServerGlobal
#include "casIntfIO.h"
#include "caServer.h"
#include "caServerIO.h"
#include "beaconAnomalyGovernor.h"
#include "CirSvr.h"
#include "caHdrLargeArray.h"

// include a version string for POSIX systems
static const char pVersionCAS[] = 
    "@(#) " EPICS_VERSION_STRING 
    ", CA Portable Server Library " 
    "2009/11/24 01:34:22";

using namespace ca;
using namespace da;

Server :: Server () :
    m_clientFreeList ( m_mutex ),
    m_pBeaconAnomalyGov ( new beaconAnomalyGovernor ( *this ) ),
    m_pIO ( ServerIO :: factory () ),
    m_debugLevel ( 0u ),
    m_nEventsProcessed ( 0u ),
    m_nEventsPosted ( 0u ),
    m_valueEvent ( registerEvent ( "value" ) ),
	m_logEvent ( registerEvent ( "log" ) ),
	m_alarmEvent ( registerEvent ( "alarm" ) )
{
}

Server :: ~Server ()
{
    Guard guard ( m_mutex );
	// delete all clients
    while ( CirSvr * pClient = m_clientList.get() ) {
        pClient->destructor ( guard );
        m_clientFreeList.release ( guard, pClient );
	}
}

#if 0
void Server :: destroyClient ( CirSvr & client )
{
    {
        Guard guard ( m_mutex );
	    m_clientList.remove ( client );
    }
    delete & client;
}
#endif

extern Service :: Factory & serviceFactory;

#pragma message ("a thread needs to be spawned that calls this")
void Server :: m_newCircuitThread ( IntfIO & intfIO )
{
	while ( ! m_newCircuitThreadShutdown ) {
		Guard guard ( m_mutex );
		CirSvr * pClient = 
			new ( m_clientFreeList )
                CirSvr ( m_mutex, guard, *this, 
                    serviceFactory, fullDuplexByteStreamFactory );
		m_clientList.add ( *pClient );
	}
}

void casVerifyFunc ( const char * pFile, unsigned line, const char * pExp )
{
    fprintf(stderr, "the expression \"%s\" didnt evaluate to boolean true \n",
        pExp);
    fprintf(stderr,
        "and therefore internal problems are suspected at line %u in \"%s\"\n",
        line, pFile);
    fprintf(stderr,
        "Please forward above text to johill@lanl.gov - thanks\n");
}

void serverToolDebugFunc (const char *pFile, unsigned line, const char *pComment)
{
    fprintf ( stderr,
"Bad server tool response detected at line %u in \"%s\" because \"%s\"\n",
                line, pFile, pComment);
}

void Server :: m_sendBeacon ( ca_uint32_t beaconNo )
{
    Guard guard ( m_mutex );
    m_pIO->sendBeacon ( guard, beaconNo );
}

void Server :: generateBeaconAnomaly ()
{
    m_pBeaconAnomalyGov->start ();
}

#if 0
void Server :: destroyMonitor ( casMonitor &  mon )
{
    mon.~casMonitor ();
    m_casMonitorFreeList.release ( & mon );
}
#endif

void Server :: updateEventsPostedCounter ( unsigned nNewPosts )
{
    Guard guard ( m_diagnosticCountersMutex );
    m_nEventsPosted += nNewPosts;
}

unsigned Server :: subscriptionEventsPosted () const
{
    Guard guard ( m_diagnosticCountersMutex );
    return m_nEventsPosted;
}

void Server :: incrEventsProcessedCounter ()
{
    Guard guard ( m_diagnosticCountersMutex );
    m_nEventsProcessed ++;
}

unsigned Server :: subscriptionEventsProcessed () const
{
    Guard guard ( m_diagnosticCountersMutex );
    return m_nEventsProcessed;
}

void Server :: show ( unsigned level ) const
{
    int bytes_reserved;
    
    printf ( "Channel Access Server V%s\n",
        CA_VERSION_STRING ( CA_MINOR_PROTOCOL_REVISION ) );
    printf ( "\trevision %s\n", pVersionCAS );

    m_mutex.show ( level );
    
    {
        Guard guard ( m_mutex );
        tsDLIterConst < CirSvr > iterCl = m_clientList.firstIter ();
        while ( iterCl.valid () ) {
            iterCl->show ( guard, level );
            ++iterCl;
        }
    }
    
    bytes_reserved = 0u;
#if 0
    bytes_reserved += sizeof ( casClient ) *
        ellCount ( & this->freeClientQ );
    bytes_reserved += sizeof(casChannel) *
        ellCount( & this->freeChanQ );
    bytes_reserved += sizeof ( casEventBlock ) *
        ellCount ( & this->freeEventQ );
    bytes_reserved += sizeof ( casAsyncIIO ) *
        ellCount ( & this->freePendingIO );
#endif
    if ( level >= 1 ) {
        printf(
            "There are currently %d bytes on the server's free list\n",
            bytes_reserved );
#if 0
        printf (
            "%d client(s), %d channel(s), %d event(s) (monitors), and %d IO blocks\n",
            ellCount ( & this->freeClientQ ),
            ellCount ( & this->freeChanQ ),
            ellCount ( & this->freeEventQ ),
            ellCount ( & this->freePendingIO ) );
#endif
        printf( 
            "The server's integer resource id conversion table:\n");
    }
        
    return;
}

#if 0
casMonitor & Server :: casMonitorFactory ( 
    casChannelI & chan, caResId clientId, 
    const unsigned long count, const unsigned type, 
    const EventMask & mask, 
    casMonitorCallbackInterface & cb )
{
    return * new ( m_casMonitorFreeList ) casMonitor 
            ( clientId, chan, count, type, mask, cb );
}

void Server :: casMonitorDestroy ( casMonitor & cm )
{
    cm.~casMonitor ();
    m_casMonitorFreeList.release ( & cm );
}
#endif

//
//	Server::dumpMsg()
//
//	Debug aid - print the header part of a message.
//
//	dp arg allowed to be null
//
//
void Server :: dumpMsg ( 
    StringReader & hostName, 
    StringReader & userName,
    const HdrLargeArray & msg, 
    const char * pFormat, ... )
{
    va_list theArgs;
    va_start ( theArgs, pFormat );
    Server :: dumpMsg ( 
        hostName, userName, msg, pFormat, theArgs);
    va_end ( theArgs );
}

void Server :: dumpMsg ( 
    StringReader & hostName, 
    StringReader & userName,
    const HdrLargeArray & msg, 
    const char * pFormat, 
    va_list & theArgs )
{
    if ( pFormat ) {
        errlogPrintf ( "CAS: " );
        errlogVprintf ( pFormat, theArgs );
    }

    FixedString < 128 > hostNameStr = hostName;
    FixedString < 128 > userNameStr = userName;

    fprintf ( stderr, 
"CAS Request: %s on %s: cmd=%u cid=%u typ=%u cnt=%u psz=%u avail=%x\n",
        hostNameStr.c_str (),
        userNameStr.c_str (),
        msg.m_cmmd,
        msg.m_cid,
        msg.m_dataType,
        msg.m_count,
        msg.m_postsize,
        msg.m_available);

    //if ( msg.m_cmmd == CA_PROTO_WRITE && msg.m_dataType == DBR_STRING && dp ) {
    //    errlogPrintf("CAS: The string written: %s \n", (char *)dp);
    //}
}

void Server :: dumpMsg ( 
    const caNetAddr & addr,
    const HdrLargeArray & msg, 
    const char * pFormat, ... )
{
    va_list theArgs;
    va_start ( theArgs, pFormat );
    Server :: dumpMsg ( addr, msg, pFormat, theArgs);
    va_end ( theArgs );
}

void Server :: dumpMsg ( 
    const caNetAddr  & addr,
    const HdrLargeArray & msg, 
    const char * pFormat, 
    va_list & theArgs )
{
    if ( pFormat ) {
        errlogPrintf ( "CAS: " );
        errlogVprintf ( pFormat, theArgs );
    }

    char pHostName[64u];
    addr.stringConvert ( pHostName, sizeof ( pHostName ) );


    fprintf ( stderr, 
"CAS Request: %s: cmd=%u cid=%u typ=%u cnt=%u psz=%u avail=%x\n",
        pHostName,
        msg.m_cmmd,
        msg.m_cid,
        msg.m_dataType,
        msg.m_count,
        msg.m_postsize,
        msg.m_available);

    //if ( msg.m_cmmd == CA_PROTO_WRITE && msg.m_dataType == DBR_STRING && dp ) {
    //    errlogPrintf("CAS: The string written: %s \n", (char *)dp);
    //}
}


EventRegistry :: ~EventRegistry ()
{
    this->traverse ( & EventMaskEntry::destroy );
}

