/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include <memory>
#include <iostream>

#include "osiWireFormat.h"
#include "FixedString.h"
#include "errlog.h"
#include "tsFreeList.h"

#define epicsExportSharedSymbols
#include "caSvrCirMgr.h"
#include "caSvrCirMgrFreeLists.h"
#include "caSvrCirRcvThr.h"
#include "caSvrCirRspThr.h"
#include "caServer.h"
#include "caeventmask.h"
#include "caHdrLargeArray.h"

using namespace std;

class NonStandardExceptionSurrogate : 
    public exception {
    const char * what () const throw ();
};

const char * NonStandardExceptionSurrogate :: what () const throw ()
{
    return "non-standard exception";
}

static NonStandardExceptionSurrogate nonStandardExceptionSurrogate;

namespace ca {

std :: exception & nonstandardException = nonStandardExceptionSurrogate;

namespace Svr {

namespace Cir {

ReqNotify :: ~ReqNotify ()
{
}

static const caHdr nill_msg = { 0u, 0u, 0u, 0u, 0u, 0u };

//
// Mgr :: factory()
//
PMgr Mgr :: factory ( PService & pService, 
                     std :: auto_ptr < Circuit > pCircuit, 
                     const Ptr < VecAllocator > & pVecAllocator,
                     const Ptr < TerminateNotify > & pTerminateNotify )
{
    Mgr * pMgr = new Mgr ( pService, pCircuit, pVecAllocator, pTerminateNotify );
    return PMgr ( pMgr->m_handle );
}

//
// Mgr :: Mgr()
//
Mgr :: Mgr ( PService & pService, 
            std :: auto_ptr < Circuit > pCircuit, 
            const Ptr < VecAllocator > & pVecAllocator,
            const Ptr < TerminateNotify > & pTerminateNotify ) :
    m_noopMutex ( epics :: Mutex :: factoryNOOP ),
    m_eventQueue ( m_mutex, m_newEventSignal ),
    m_pRspThr ( RspThr :: factory ( Ptr < Mgr > ( m_handle ), 
            pCircuit->createOutStream (), pVecAllocator ) ),
    m_pRcvThr ( RcvThr :: factory ( pService, Ptr < Mgr > ( m_handle ), 
            m_pRspThr, pCircuit->createInputStream ( pVecAllocator ) ) ),
    m_pTerminateNotify ( pTerminateNotify ),
    m_pCircuit ( pCircuit ),
    m_pFreeLists ( new FreeLists () ),
    m_ioRequestsInProgress ( 0 ),
    m_isTerminated ( false )
{
    m_pRspThr->start ();
    m_pRcvThr->start ();
}

//
// Mgr :: ~Mgr ()
//
Mgr :: ~Mgr ()
{
    m_pRcvThr.clear (); // receive thread shutdown
    m_pRspThr.clear (); // response thread shutdown
    if ( m_danglingReqList.count () ) {
        errlogPrintf ( 
            "Smart pointer references to server Responder objects in the Mgr\n" );
        errlogPrintf ( 
            "persist in the Service when the Mgr destructor is running.\n" );
        errlogPrintf ( 
            "Since their free list and mutex is being destroyed a crash\n" );
        errlogPrintf ( 
            "will occur if they are handled by the Service in the future\n" );
        errlogPrintf (
            "These objects persist despite their corresponding Channel and Request\n" );
        errlogPrintf (
            "having been canceled and unreferenced.\n" );
        errlogPrintf ( 
            "Here is a dump of the offending objects\n" );
	    tsDLIter < ReqNotify > pReqNotify = m_danglingReqList.firstIter ();
	    while ( pReqNotify.valid () ) {
            pReqNotify->show ( 1000u );
            pReqNotify++;
	    }
    }
}

Mgr :: M_NewEventSignal :: M_EnclLnk
    Mgr :: M_NewEventSignal :: m_enclLnk =
        & Mgr :: m_newEventSignal;

void Mgr :: M_NewEventSignal :: signal ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
#pragma message ("potential for unneeded locking here")
#pragma message ("possible fix => epicsEvent which is reference counting like epics :: Mutex")
    mgr.m_pRspThr->eventSignal ();
}

void Mgr :: show ( unsigned level ) const
{
	printf ( "Mgr\n" );
	if ( level > 1u ) {
        m_pFreeLists->show ( level - 1 );
        // m_pFreeLists->m_badMsgLen.show ( level - 1 );
        m_mutex.show ( level - 1 );
        :: printf ( 
            "m_ioRequestsInProgress = %u ",
            m_ioRequestsInProgress );
	}
}

void Mgr :: FreeLists :: show ( unsigned level ) const
{
    printf ( "Mgr :: FreeLists\n" );
    if ( level > 0 ) {
        //m_reqFreeList.show ( level - 1 );
        //m_chanCreate.show ( level - 1 );
        m_chanRead.show ( level - 1 );
        m_chanWrite.show ( level - 1 );
        m_chanCircMon.show ( level - 1 );
        m_chanSubscribe.show ( level - 1 );
        //m_chanDelete.show ( level - 1 ),
        //m_hdrOnly.show ( level - 1 );
    }
}

void Mgr :: setPriority ( unsigned caPriority )
{
    // calculate new recv thread priority
    double tmp = caPriority - CA_PROTO_PRIORITY_MIN;
    tmp *= epicsThreadPriorityCAServerHigh - 
                epicsThreadPriorityCAServerLow;
    tmp /= CA_PROTO_PRIORITY_MAX - CA_PROTO_PRIORITY_MIN;
    tmp += epicsThreadPriorityCAServerLow;
    unsigned newPriority = static_cast < unsigned > ( tmp );
    m_pRspThr->setPriority ( newPriority );
    m_pRcvThr->setPriority ( newPriority );
}

void Mgr :: terminateNotify ()
{
    m_isTerminated = true;
    // most not hold the Mgr lock when
    // notifying backwards up the lock hierarchy
    epics :: Mutex :: Unlocker unlocker ( m_mutex );
    m_pTerminateNotify->terminateNotify ();
}

void Mgr :: postHdrOnlyResp ( const HdrLargeArray & hdr )
{
    PResp pResp = Chan :: HdrOnly :: factory ( *this, hdr );
    m_eventQueue.post ( pResp );
}

Mgr :: PChan Mgr ::
    createChannel ( caResId sid, caResId cid )
{
    return Chan :: factory ( *this, sid, cid );
}

#if 0
/*
 * Mgr :: m_noReadAccessEvent()
 *
 * substantial complication introduced here by the need for backwards
 * compatibility
 */
caStatus Server :: Mgr :: 
    m_noReadAccessEvent ( Guard & guard, casClientMon * pMon )
{
    guard.assertIdenticalMutex ( m_mutex );

	caHdr falseReply;
	unsigned size;
	caHdr * reply;
	int status;

	size = dbr_size_n ( pMon->getType(), pMon->getCount() );

	falseReply.m_cmmd = CA_PROTO_EVENT_ADD;
	falseReply.m_postsize = size;
	falseReply.m_dataType = pMon->getType();
	falseReply.m_count = pMon->getCount();
	falseReply.m_cid = pMon->getChannel().getCID();
	falseReply.m_available = pMon->getClientId();

	status = this->allocMsg ( size, &reply );
	if ( status ) {
		if( status == S_cas_hugeRequest ) {
#error we dont send clues to persons writing rogue clients
			return m_out.sendErr ( falseReply, ECA_TOLARGE, NULL );
		}
		return status;
	}
	else{
		/*
		 * New clients recv the status of the
		 * operation directly to the
		 * event/put/get callback.
		 *
		 * Fetched value is zerod in case they
		 * use it even when the status indicates 
		 * failure.
		 *
		 * The m_cid field in the protocol
		 * header is abused to carry the status
		 */
		*reply = falseReply;
		reply->m_postsize = size;
		reply->m_cid = ECA_NORDACCESS;
		memset((char *)(reply+1), 0, size);
		this->commitMsg ();
	}
	
	return S_cas_success;
}
#endif

const char * Mgr :: 
    BadProtocol :: what () const throw ()
{ 
    return "Invalid Protocol";
}

void Mgr :: 
    processEvents ()
{
    m_eventQueue.processEvents ();
}

///////////////////////////////////////////////////
// Mgr Handle
///////////////////////////////////////////////////

Mgr :: M_Handle :: M_EnclLnk
    Mgr :: M_Handle :: m_enclLnk =
        & Mgr :: m_handle;

Mgr :: M_Handle :: TargetMutexPair 
	Mgr :: M_Handle :: m_targetMutexPair ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
	return TargetMutexPair ( mgr, mgr.m_mutex );
}
HandleIntf < Mgr > & 
	Mgr :: M_Handle :: m_clone ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
	mgr.m_refCtr.reference ();
    return *this;
}
void Mgr :: M_Handle :: m_release ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
	if ( mgr.m_refCtr.unreference () ) {
		delete this;
		// no code past here considering above delete
	}
	// no code past here considering above delete
}

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca


//
//  Mgr::xSend()
//
#if 0
casStreamWriteIO::flushCondition Mgr::xSend ( char * pBufIn,
                                             size_t nBytesAvailableToSend,
                                             size_t nBytesNeedToBeSent,
                                             size_t & nActualBytes )
{
    casStreamWriteIO::flushCondition stat = casStreamWriteIO::flushDisconnect;
    size_t nActualBytesDelta;
    size_t totalBytes;

    assert ( nBytesAvailableToSend >= nBytesNeedToBeSent );
	
    totalBytes = 0u;
    while ( true ) {
        stat = this->osdSend ( &pBufIn[totalBytes],
                              nBytesAvailableToSend-totalBytes, nActualBytesDelta );
        if ( stat != casStreamWriteIO::flushProgress ) {
            if ( totalBytes > 0 ) {
                nActualBytes = totalBytes;
		        //
		        // !! this time fetch may be slowing things down !!
		        //
		        //this->lastSendTS = epicsTime::getCurrent();
                stat = casStreamWriteIO::flushProgress;
                break;
            }
            else {
                break;
            }
        }
		
        totalBytes += nActualBytesDelta;
		
        if ( totalBytes >= nBytesNeedToBeSent ) {
		    //
		    // !! this time fetch may be slowing things down !!
		    //
		    //this->lastSendTS = epicsTime::getCurrent();
            nActualBytes = totalBytes;
            stat = casStreamWriteIO::flushProgress;
            break;
        }
    }
	return stat;
}

//
// Mgr::xRecv()
//
casStreamReadIO::fillCondition Mgr::xRecv ( char * pBufIn, size_t nBytes,
                                 casStreamReadIO::fillParameter, size_t & nActualBytes )
{
	casStreamReadIO::fillCondition stat;
	
	stat = this->osdRecv ( pBufIn, nBytes, nActualBytes );
    //
    // this is used to set the time stamp for write GDD's
    //
    this->lastRecvTS = epicsTime::getCurrent ();
	return stat;
}
#endif

