

/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#ifndef caSvrCirMgrh
#define caSvrCirMgrh

#include <memory>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caSvrCirMgrh
#   undef epicsExportSharedSymbols
#endif

#include "caProto.h"
#include "tsDLList.h"
#include "EnclLnk.h"

#ifdef epicsExportSharedSymbols_caSvrCirMgrh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "casdef2.h"
#include "caNetAddr.h"
#include "caServer.h"
#include "caSvrEvQue.h"

namespace ca {

class IterIn;
class Server;
struct Circuit;
struct VecAllocator;
struct HdrLargeArray;

extern std :: exception & nonstandardException;

namespace Svr {

namespace Cir {

struct TerminateNotify {
    virtual void terminateNotify () = 0;
};

enum xBlockingStatus { xIsBlocking, xIsntBlocking };

struct StreamIO;
class casMonitor;
struct IntfIO;

// this needs to default to scalar string if the
// type and count are not available (this must
// not allow an exception to propagate). Nevertheless,
// log a warning message to the server's console if 
// there is a failure.
struct NativeTypeAndCountDBR {
    ca_uint16_t type;
    ca_uint32_t count;
};
extern NativeTypeAndCountDBR 
    extractNativeTypeAndCountDBR ( const Catalog & );
// this needs to log a message and default to csDisconn
// if the circuit state isnt available in the Catalog or 
// its out of range.
// if ( newState < 0 || newState > Channel :: csDisconn ) {
//     newState = Channel :: csDisconn;
//  }
extern void extractCircuitState ( const Catalog &, 
            Channel :: CircuitState & );

class RspThr;
class RcvThr;

struct ReqNotify : public tsDLNode < ReqNotify > {
    virtual ~ReqNotify () = 0;
    virtual void show ( unsigned level ) const = 0;
    virtual bool subscrDestroy ( caResId ) = 0;
    virtual void readLackingCallbackDestroy () = 0;
    virtual void enterFlowControlMode () = 0;
    virtual void exitFlowControlMode () = 0;
    virtual void channelDestroyNotify () = 0;
};

class Mgr;
typedef Ptr < Mgr > PMgr;

//
// Mgr 
//
class Mgr {
public:
    static PMgr factory ( PService &, std :: auto_ptr < Circuit >, 
        const Ptr < VecAllocator > &, const Ptr < TerminateNotify > & );
    Mgr ( PService &, std :: auto_ptr < Circuit >, 
        const Ptr < VecAllocator > &, const Ptr < TerminateNotify > & );
    ~Mgr ();  
    void show ( unsigned level ) const;
    class Chan;
    typedef Ptr < Chan > PChan;
    typedef Ptr < Mgr > PCirSvr;
    Mgr :: PChan createChannel ( caResId sid, caResId cid );
    void setPriority ( unsigned );
    void postHdrOnlyResp ( const HdrLargeArray & );
    void processEvents ();
    void terminateNotify ();
    bool isActive () const;
private:
    struct FreeLists;
    class M_NewEventSignal :
        public EvQue :: NewEventSignal {
    public:
        void signal ();
    private:
        typedef const epics :: EnclLnk < Mgr, 
                                    M_NewEventSignal > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_Handle : 
        public HandleIntf < Mgr > {
    private:
        TargetMutexPair m_targetMutexPair ();
        HandleIntf < Mgr > & m_clone ();
        void m_release ();
        typedef const epics :: EnclLnk < Mgr, 
                                    M_Handle > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };

    mutable Mutex m_mutex; // must be first
    mutable Mutex m_noopMutex; 
    M_NewEventSignal m_newEventSignal;
    // we print a warning message if the reference to
    // the response still exists during the Mgr dtor
    // and we know we have already done all of the following
    // o canceled the Request
    // o released the reference to the Request
    // o released the reference to the Channel
    tsDLList < ReqNotify > m_danglingReqList;
    NetAddr m_clientSideCircuitAddress;
    M_Handle m_handle;
    EvQue m_eventQueue;
    // The event thread must be shut down prior to destroying 
    // the RspThr object.
    Ptr < RspThr > m_pRspThr;
    Ptr < RcvThr > m_pRcvThr;
    Ptr < TerminateNotify > m_pTerminateNotify;
    std :: auto_ptr < Circuit > m_pCircuit;
    std :: auto_ptr < FreeLists > m_pFreeLists;
    size_t m_ioRequestsInProgress;
    RefCtr m_refCtr;
    bool m_isTerminated;

    // exceptions
    class BadProtocol;

    // static const unsigned m_maxSimlIO = 256;
    // caStatus createChannel ( const char * pName );
    // void m_sendErrWithEpicsStatus (  Guard &,
    //    const HdrLargeArray &, ca_uint32_t cid, 
    //    caStatus epicsStatus, caStatus clientStatus );

#if 0
    void read ( const gdd * & pDesc );
    void write ();

    void writeArrayData();
    void writeScalarData();
    void writeString();

    casStreamWriteIO::flushCondition xSend ( 
        char * pBuf, size_t nBytesAvailableToSend,
		size_t nBytesNeedToBeSent, size_t & nBytesSent );
    casStreamReadIO::fillCondition xRecv ( 
        char * pBuf, size_t nBytesToRecv,
		casStreamReadIO::fillParameter parm, size_t & nByesRecv );
	virtual casStreamWriteIO::flushCondition osdSend ( 
        const char *pBuf, size_t nBytesReq,
			size_t & nBytesActual ) = 0;
	virtual casStreamReadIO::fillCondition osdRecv ( 
        char *pBuf, size_t nBytesReq,
			size_t &nBytesActual ) = 0;
	virtual xBlockingStatus blockingState () const = 0;

    void casChannelDestroyFromInterfaceNotify ( casChannelI & chan, 
        bool immediatedSestroyNeeded );
#endif

#if 0
    void m_requestCreateFailCleanup ( 
        Guard &, Request &, Chan :: IO :: Notify &, 
        ca_uint32_t legacyStatus, const char * pReason );
#endif
    // disconnect old client that is unable to receive protocol
    // informing it about new situations
    class OldClientDisconnect : 
        public std :: exception {
    public:
        const char * what () const throw ();
    };
    Mgr ( const Mgr & ); // not implemented
    Mgr & operator = ( const Mgr & ); // not implemented
};

class Mgr :: BadProtocol : 
    public std :: exception {
    const char * what () const throw ();
};

inline bool Mgr :: isActive () const
{
    return ! m_isTerminated;
}

#if 0
class Mgr :: ChanCreateParms : 
    public Catalog {
public:
    ChanCreateParms ( 
        const char * pName, size_t nameLen );
    void traverse ( Viewer & ) const;
    void find ( const Ident & id, Viewer & ) const;
private:
    ConstString m_name;
};
#endif

#if 0
inline ca_uint16_t Mgr :: 
    protocolRevision ( Guard & guard ) const 
{
    guard.assertIdenticalMutex ( m_mutexId );
    return m_minorVersionNumber;
}
#endif

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca

#endif // caSvrCirMgrh

