/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#define epicsExportSharedSymbols
#include "caSvrCirMgrFreeLists.h"
#include "caSvrCirMgrChan.h"
#include "caSvrCirMgrChanRead.h"
#include "caSvrCirMgrChanWrite.h"
#include "caSvrCirMgrChanSubscribe.h"

#ifndef NELEMENTS
#   define NELEMENTS(array) (sizeof(array)/sizeof((array)[0]))
#endif

namespace ca {
namespace Svr {
namespace Cir {

const Mgr :: Chan :: PStateTrans 
Mgr :: Chan :: m_pStateChangeFunc [] = {
    & Chan :: m_becomeDisconnected,
    & Chan :: m_becomeUnresponsive,
    & Chan :: m_becomeResponsive
};

Ptr < Mgr :: Chan > 
    Mgr :: Chan :: factory ( Mgr & mgr, caResId sid, caResId cid )
{
    Chan * pChan = new ( mgr.m_pFreeLists->m_chan ) 
	    Chan ( mgr, sid, cid );
    return Ptr < Chan > ( pChan->m_mgrChanHandle );
}

Mgr :: Chan :: 
    Chan ( Mgr & mgr, caResId sid, caResId cid ) :
    m_respThrCirMon ( sid, cid ),
    m_mgr ( mgr ),
    m_state ( Channel :: csDisconn )
{
    assert ( 
        m_pStateChangeFunc[ Channel :: csUnresp ] == 
        & Chan :: m_becomeDisconnected );
    assert ( 
        m_pStateChangeFunc[ Channel :: csResp ] == 
        & Chan :: m_becomeUnresponsive );
    assert ( 
        m_pStateChangeFunc[ Channel :: csDisconn ] == 
        & Chan :: m_becomeResponsive );
}

Mgr :: Chan :: 
    ~Chan ()
{
    while ( ReqNotify * pReq = m_reqList.first () ) {
        pReq->channelDestroyNotify ();
        // since pReq could be destroyed 
        // by the serviceDestroyNotify function
        // it should not be accessed below here
    }
}

PResp Mgr :: Chan :: createReadResp ( const PReq & pReq, 
	const HdrLargeArray & hdr, caResId cid )
{
    return Read :: factory ( *this, pReq, hdr, cid );
}

PResp Mgr :: Chan :: createWriteResp ( const PReq & pReq, 
	const HdrLargeArray & hdr, caResId cid )
{
    return Write:: factory ( *this, pReq, hdr, cid );
}

PResp Mgr :: Chan :: createSubscResp ( const PReq & pReq, 
	const HdrLargeArray & hdr, caResId cid, const PCat & reqParms )
{
    return Subscribe :: factory ( *this, pReq, hdr, cid, reqParms );
}

Mgr :: Chan :: M_CirMonServiceResp :: M_EnclLnk 
    Mgr :: Chan :: M_CirMonServiceResp :: m_enclLnk = 
        & Mgr :: Chan :: m_cirMonServiceResp ;
  
// fetch the channels state, and execute a state transition
void Mgr :: Chan :: M_CirMonServiceResp :: 
    m_responseNotify ( const PCat & pCat )
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    unsigned newCircuitState = 
        Channel :: csDisconn;
    {
        // best to be conservative and avoid any chance 
        // of lock hierarchy reversal here
        epics :: Mutex :: Unlocker 
            unlocker ( chan.m_mgr.m_mutex );
        PCat :: Locker pCatLocked = pCat.operator -> ();
        Clerk clerk ( *pCatLocked );
        clerk.get ( Channel :: circuitState, newCircuitState );
    }
    if ( newCircuitState < NELEMENTS ( m_pStateChangeFunc ) ) {
        PStateTrans pMemFun = m_pStateChangeFunc[newCircuitState];
        (chan.*pMemFun)( pCat );
    }
    else {
        throw std :: range_error ( 
            "Channel :: circuitState "
            "supplied to circuit monitor m_responseNotify "
            "callback was out of range" );
    }
}

void Mgr :: Chan ::  
    show ( unsigned level ) const
{
     printf ( "ca :: Svr :: Mgr :: Chan\n" );
     if ( level > 0u ) {
         tsDLIterConst < ReqNotify > iter = m_reqList.firstIter ();
         while ( iter != m_reqList.invalidIterConst () ) {
             iter->show ( level - 1u );
         }
         printf ( "m_state = %d\n", m_state );
         m_refCtr.show ( level - 1u );
         // !!!! potential for show loops here via response
         // !!!! callback interface's show
         m_pCirMonReq->show ( level - 1u );
     }
}

void Mgr :: Chan :: M_CirMonServiceResp :: 
    m_show ( unsigned level ) const
{
    const Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.show ( level );
}

void Mgr :: Chan :: 
    m_becomeDisconnected ( const PCat & pCat )
{
    if ( m_state == Channel :: csDisconn ) {
        return;
    }
    m_mgr.m_eventQueue.post ( PResp ( m_cirMonRespThrHandle ), pCat );
    m_state = Channel :: csDisconn;
}

void Mgr :: Chan :: 
    m_becomeUnresponsive ( const PCat & pCat )
{
    if ( m_state == Channel :: csUnresp ) {
        return;
    }
    m_mgr.m_eventQueue.post ( PResp ( m_cirMonRespThrHandle ), pCat );
    m_state = Channel :: csUnresp;
}

void Mgr :: Chan :: 
    m_becomeResponsive ( const PCat & pCat )
{
    if ( m_state == Channel :: csResp ) {
        return;
    }
    m_mgr.m_eventQueue.post ( PResp ( m_cirMonRespThrHandle ), pCat );
    m_state = Channel :: csResp;
}

void Mgr :: Chan :: enterFlowControlMode ()
{
    tsDLIter < ReqNotify > iter = m_reqList.firstIter ();
    while ( iter != m_reqList.invalidIter () ) {
        iter->enterFlowControlMode ();
    }
}

void Mgr :: Chan :: exitFlowControlMode ()
{
    tsDLIter < ReqNotify > iter = m_reqList.firstIter ();
    while ( iter != m_reqList.invalidIter () ) {
        iter->exitFlowControlMode ();
    }
}

bool Mgr :: Chan :: unsubscribe ( caResId subscrId )
{
    bool success = false;
    tsDLIter < ReqNotify > iter = m_reqList.firstIter ();
    while ( iter != m_reqList.invalidIter () ) {
        success = iter->subscrDestroy ( subscrId );
        if ( success ) {
            break;
        }
    }
    return success;
}

void Mgr :: Chan :: readLackingCallbackCancel ()
{
    tsDLIter < ReqNotify > iter = m_reqList.firstIter ();
    while ( iter != m_reqList.invalidIter () ) {
        iter->readLackingCallbackDestroy ();
    }
}

//////////////////////////////////////////////////////
// M_MgrCHanHandle 
//////////////////////////////////////////////////////
Mgr :: Chan :: M_MgrChanHandle :: M_EnclLnk 
    Mgr :: Chan :: M_MgrChanHandle :: m_enclLnk = 
        & Chan :: m_mgrChanHandle;
Mgr :: Chan :: M_MgrChanHandle :: TargetMutexPair 
    Mgr :: Chan :: M_MgrChanHandle :: m_targetMutexPair ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( chan, chan.m_mgr.m_mutex );
}
HandleIntf < Mgr :: Chan > & 
    Mgr :: Chan :: M_MgrChanHandle :: m_clone ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_refCtr.reference ();
    return *this;
}
void Mgr :: Chan :: M_MgrChanHandle :: m_release ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_commonUnreference ();
    // no code below here due to potential destroy
}

//////////////////////////////////////////////////////
// M_CirMonRespThrHandle
//////////////////////////////////////////////////////
Mgr :: Chan :: M_CirMonRespThrHandle :: M_EnclLnk 
    Mgr :: Chan :: M_CirMonRespThrHandle :: m_enclLnk = 
        & Mgr :: Chan :: m_cirMonRespThrHandle ;
Mgr :: Chan :: M_CirMonRespThrHandle :: TargetMutexPair 
    Mgr :: Chan :: M_CirMonRespThrHandle :: m_targetMutexPair ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( chan.m_respThrCirMon, 
                        chan.m_mgr.m_noopMutex );
}
HandleIntf < Responder > & 
    Mgr :: Chan :: M_CirMonRespThrHandle :: m_clone ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_refCtr.reference ();
    return *this;
}
void Mgr :: Chan :: M_CirMonRespThrHandle :: m_release ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_commonUnreference ();
    // no code below here due to potential destroy
}

//////////////////////////////////////////////////////
// M_CirMonServRespHandle
//////////////////////////////////////////////////////
Mgr :: Chan :: M_CirMonServRespHandle :: M_EnclLnk 
    Mgr :: Chan :: M_CirMonServRespHandle :: m_enclLnk = 
        & Mgr :: Chan :: m_cirMonServRespHandle;
Mgr :: Chan :: M_CirMonServRespHandle :: TargetMutexPair 
    Mgr :: Chan :: M_CirMonServRespHandle :: m_targetMutexPair ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( chan.m_cirMonServiceResp, 
                        chan.m_mgr.m_mutex );
}
HandleIntf < Responder > & 
    Mgr :: Chan :: M_CirMonServRespHandle :: m_clone ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_refCtr.reference ();
    return *this;
}
void Mgr :: Chan :: M_CirMonServRespHandle :: m_release ()
{
    Chan & chan ( m_enclLnk.enclosure ( *this ) );
    chan.m_commonUnreference ();
    // no code below here due to potential destroy
}

void Mgr :: Chan :: m_commonUnreference ()
{
    if ( m_refCtr.unreference () ) {
        delete this;
        // no code below here due to above destroy
    }
    // no code below here due to above potential destroy
}

void Mgr :: Chan :: 
    postFailure ( const PResp & pResp, const char * pCtx )
{
    m_mgr.m_eventQueue.postFailure ( pResp, pCtx );
}

#if 0
class PVMetadata : 
    public Catalog {
public:
    PVMetadata ();
    void traverse ( Viewer & ) const;
    Catalog & operator = ( const Catalog & );
private:
    epicsType m_nativeType;
    static const size_t m_nbounds = 5;
    size_t m_bound;
};

PVMetadata :: PVMetadata () :
    m_nativeType ( epicsStringT ),
    m_bound ( 0 )
{
}

void PVMetadata :: 
    traverse ( Viewer & viewer ) const
{
    Property < unsigned > ntProp (
        pi_nativeType, m_nativeType );
    viewer.reveal ( ntProp );
    VectorProperty < size_t > bdProp (
        pi_bounds, m_bounds, m_nbounds );
    viewer.reveal ( bdProp );
}

Catalog & PVMetadata :: 
    operator = ( const Catalog & src )
{
    unsigned nt;
    Property < unsigned > ntProp (
        pi_nativeType, nt );
    src.extract ( ntProp );
    switch ( nt ) {
    case epicsInt8T;
    case epicsUInt8T;
    case epicsInt16T;
    case epicsUInt16T;
    case epicsEnum16T;
    case epicsInt32T;
    case epicsUInt32T;
    case epicsFloat32T;
    case epicsFloat64T;
    case epicsStringT;
    case epicsOldStringT;
        m_nativeType = static_cast < epicsType > ( nt );
        break;
    default:
        throw xxx ();
    }
    Property < size_t > bdProp (
        pi_bounds, m_bounds );
    src.extract ( bdProp );
}
 
//
// Property Hierarchy for observe method
// with response type of CircuitState
//
// o circuitState
// o interfaceName
// o processVariable (if interface name is processVariable)
//     o nativeType
//     o bounds (potentially an array)
//
void Mgr :: Chan :: 
    cirStateNotify ( const Catalog & resp )
{
    guard.assertIdenticalMutex ( m_mgr.m_mutex );
    CircuitState newState = csDisconn;
    unsigned cs;
    Property < unsigned > csProp (
        pi_circuitState, cs );
    resp.extract ( csProp );
    switch ( cs ) {
    case csDisconn:
    case csUnresp:
    case csResp:
        newState = static_cast < newState > ( csRaw );
        break;
    default:
        throw xxx ();
    }

    FixedLengthString < 64 > intfName;
    Property < StringEditor > intfNameProp (
        pi_interfaceName, intfName );
    resp.extract ( intfNameProp );
    static const char pPVName[] = "processVariable";
    static const size_t pvNameLen = sizeof ( pPVName );
    static const FixedLengthString < pvNameLen > 
        pvName ( pPVName, pvNameLen-1 );
    PVMetadata pvMetadata;
    if ( intfName.equals ( pvName ) ) {
        m_isPV = true;
        {
            Property < Catalog > pvProp (
                pi_processVariable, pvMetadata );
            resp.extract ( pvProp );
        }
#error user callbacks must not diretly place items in 
#error protocol stream for fear of lockup vulnerabilities
#error must post event instead
        m_mgr.m_createChanResponse ( 
            guard, *this, hdr, pvMetadata );
    }
    else {
#error this is a disconnect condition?
    }

    CircuitState oldState = m_state;
    m_state = newState;
#error needs to be an interface class
#error needs to specify the interface type?
    m_mgr.channelCircuitStateNotify ( 
        *this, pvMetadata, oldState, newState );
}

#error what does the native type mean?


    }
void Mgr :: Chan :: cirStateNotifyFailure ( 
    Guard & guard, const Abberation & )
{
    guard.assertIdenticalMutex ( m_mgr.m_mutex );

#error place channel in disconneted state
    CircuitState oldState = m_state;
    m_state = csDisconn;
    m_mgr.channelCircuitStateNotify ( 
        *this, pvMetadata, oldState, newState );
}
#endif

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca
