
/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include <sstream>

#include "errlog.h"
#include "db_access.h"

#define epicsExportSharedSymbols
#include "caServer.h"
#include "caSvrCirMgr.h"
#include "caSvrCirMgrChan.h"
#include "caSvrCirRcvThr.h"
#include "caSvrCirRcvThrChan.h"
#include "caSvrCirRspThr.h"
#include "caServerDefs.h"
#include "protoDiagnostics.h"
#include "caChannelDescription.h"
#include "caStringReaderIterIn.h"
#include "caChannelDescription.h"

#include "CatalogIterDBR.h"

namespace ca {
namespace Svr {
namespace Cir {

struct InvalidDataType : 
    public LegacyStatus < ECA_BADTYPE > {};

struct ArrayElementSubsetIsEmpty : 
    public LegacyStatus < ECA_BADCOUNT > 
{
    const char * what () const throw () { 
        return "Array element subset is empty";
    }
};

struct InvalidChannelId : 
    public LegacyStatus < ECA_BADCHID > {};

const RcvThr:: PMsgHandler  
    RcvThr :: m_msgHandlers[] =
{
    & RcvThr :: m_versionAction,
    & RcvThr :: m_subscribeAction,
    & RcvThr :: m_eventCancelAction,
    & RcvThr :: m_readAction,
    & RcvThr :: m_writeAction,
    & RcvThr :: m_uknownMessageAction, 
    & RcvThr :: m_uknownMessageAction, 
    & RcvThr :: m_uknownMessageAction, 
    & RcvThr :: m_eventsOffAction,
    & RcvThr :: m_eventsOnAction,
    & RcvThr :: m_readSyncAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_clearChannelAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_readAction,
    & RcvThr :: m_ignoreMsgAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_createChannelAction,
    & RcvThr :: m_writeAction,
    & RcvThr :: m_clientNameAction,
    & RcvThr :: m_hostNameAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_echoAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction
};

//
// RcvThr :: M_EntryPt :: M_EnclLnk :: m_enclLnk
//
const RcvThr :: M_EntryPt :: M_EnclLnk
    RcvThr :: M_EntryPt ::  
        m_enclLnk ( & RcvThr :: m_entryPt );

//
// RcvThr :: M_EntryPt :: run ()
//
void RcvThr :: M_EntryPt :: run ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    try {
        Guard guard ( rcvThr.m_mutex );
        IterIn iterIn ( rcvThr.m_circuitInReader );
        while ( ! rcvThr.m_threadShutdownCmd ) {
            rcvThr.m_processMsg ( iterIn );
        }
        return;
    }
    catch ( InvalidProtocol & except ) {
        rcvThr.m_msgStubProtocolErrorDiag ( except );
    }
    catch ( std :: exception & except ) {
        rcvThr.m_msgStubExceptDiag ( except );
    }
    catch ( ... ) {
        rcvThr.m_msgStubExceptDiag ( nonstandardException );
    }
    // most not hold the rcv thread lock when
    // notifying backwards up the lock hierarchy
    rcvThr.m_pMgr->terminateNotify ();
}

void RcvThr :: M_EntryPt :: 
    show ( unsigned level ) const
{
    printf ( "ca :: Svr :: Cir :: RcvThr :: EntryPt\n" );
    // we have to be careful what we show here because
    // of thread synchronization and deadlock issues
}

Ptr < RcvThr > RcvThr :: factory ( 
        const PService & pService, 
        const Ptr < Mgr > & pMgr, 
        const Ptr < RspThr > & pRspThr, 
        std :: auto_ptr < CircuitIn > circuitIn )
{
    RcvThr * pRcvThr = new RcvThr ( pService, pMgr, pRspThr, circuitIn );
    return Ptr < RcvThr > ( pRcvThr->m_handle );
}
 
RcvThr :: RcvThr ( const PService & pService, 
            const Ptr < Mgr > & pMgr, 
            const Ptr < RspThr > & pRspThr, 
            std :: auto_ptr < CircuitIn > circuitIn ) :
    m_circuitInReader ( circuitIn ),
    m_thread ( m_entryPt, "CAS-Cir-Rcv", 
        epicsThreadGetStackSize ( epicsThreadStackMedium ), 
        m_calcOffsetPriority ( epicsThreadPriorityCAServerLow ) ),
    m_pService ( pService ),
    m_pMgr ( pMgr ),
    m_pRspThr ( pRspThr ),
    m_minorVersionNumber ( 0u ),
    m_threadShutdownCmd ( false )
{
#pragma message ( "setup m_dbrTypePtrs here" )
}

void RcvThr :: start ()
{
    m_thread.start ();
}

RcvThr :: ~RcvThr ()
{
    while ( Chan * pChan = m_chanList.get() ) {
        chronIntIdRes < Chan > & res ( *pChan );
        m_chanTable.remove ( res );
        m_chanFreeList.deleteProduct ( *pChan );
    }
    m_circuitInReader.shutdown ();
    m_threadShutdownCmd = true;
    epics :: Mutex :: Unlocker unlocker ( m_mutex );
    m_thread.exitWait ();
}

unsigned RcvThr :: 
    m_calcOffsetPriority ( unsigned reqPriority )
{
    // calculate new receive thread priority
    // slightly higher than the response thread
    unsigned resPriority;
    epicsThreadBooleanStatus tbs = 
        epicsThreadHighestPriorityLevelBelow ( 
            reqPriority, & resPriority );
    if ( tbs != epicsThreadBooleanStatusSuccess ) {
        resPriority = reqPriority;
    }
    return resPriority;
}

void RcvThr :: 
    setPriority ( unsigned epicsPriority ) 
{
    m_thread.setPriority ( m_calcOffsetPriority ( epicsPriority ) );
}

void RcvThr :: 
    m_msgStubExceptDiag ( const std::exception & except ) const
{
	errlogDumpMsg ( 
        m_hostNameStr, 
        m_userNameStr, 
        m_curMsg, 
        "Exception: \"%s\" when processing CA circuit protocol "
        "- forcing disconnect",
        except.what () );
}

void RcvThr :: 
    m_msgStubProtocolErrorDiag ( const std::exception & except ) const
{
    errlogDumpMsg ( 
        m_hostNameStr, 
        m_userNameStr, 
        m_curMsg, 
        "CA Circuit Protocol Error: \"%s\" - forcing disconnect",
        except.what () );
}

void RcvThr :: 
    m_processMsg ( IterIn & iterIn )
{
    {
        m_curMsg.m_cmmd = iterIn.pop < ca_uint16_t > ();
        ca_uint16_t payloadSize = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_dataType = iterIn.pop < ca_uint16_t > ();
        ca_uint16_t elementCount = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_cid = iterIn.pop < ca_uint32_t > ();
        m_curMsg.m_available = iterIn.pop < ca_uint32_t > ();
        if ( payloadSize != 0xffff && elementCount != 0xffff ) {
            m_curMsg.m_postsize = payloadSize;
            m_curMsg.m_count = elementCount;
        }
        else {
            m_curMsg.m_postsize = iterIn.pop < ca_uint32_t > ();
            m_curMsg.m_count = iterIn.pop < ca_uint32_t > ();
        }
    }
    PMsgHandler pMsgHandler = & RcvThr :: m_uknownMessageAction;
    if ( m_curMsg.m_cmmd < NELEMENTS ( m_msgHandlers ) ) {
        pMsgHandler = m_msgHandlers[m_curMsg.m_cmmd];
    }
    {
        // a special reader that always removes exactly m_postsize 
        // bytes, no more and no less, from the input stream
        IterInReader iterInReader ( iterIn, m_curMsg.m_postsize );
        IterIn iterInPayload ( iterInReader );
        ( this->*pMsgHandler ) ( iterInPayload );
    }
}

//
// RcvThr :: show ()
//
void RcvThr :: 
    show ( unsigned level ) const
{
    printf ( "RcvThr for user \"%s\" and host \"%s\"\n",
        m_userNameStr.c_str (), m_hostNameStr.c_str () );
    if ( level > 0 ) {
	    m_lastRecvTS.show ( level - 1 );
        printf ( "m_chanTable\n" );
        m_chanTable.show ( level - 1 );
        m_thread.show ( level - 1 );
    }
}

//
// RcvThr :: m_uknownMessageAction()
//
void RcvThr :: 
    m_uknownMessageAction ( IterIn & )
{
    std :: stringstream diag;
    diag << "unexpected ca protocol request code = ";
    diag << m_curMsg.m_cmmd;
    diag <<  " in message";
    throw InvalidProtocol ( diag.str () );
}

//
// RcvThr :: m_ignoreMsgAction()
//
void RcvThr :: 
    m_ignoreMsgAction ( IterIn & )
{
}

//
// RcvThr :: m_versionAction()
//
#pragma message ( "it would be better to know the priority prior to starting the thread" )
#pragma message ( "so we dont get starved out by the threads of a lower priority client" )
#pragma message ( "we can get the version in a create channel req or here" )
#pragma message ( "if they send a non-zero version here then send a msg to the resp thread communicating the version, and lock out future version changes" )
void RcvThr :: m_versionAction ( IterIn & )
{
    // validate protocol
    if ( 
        m_curMsg.m_dataType > CA_PROTO_PRIORITY_MAX || 
        m_curMsg.m_dataType < CA_PROTO_PRIORITY_MIN ) {
            std :: string ();
        std :: stringstream diag;
        diag << "bad priority ";
        diag << m_curMsg.m_dataType;
        diag <<  " in ca protocol version msg";
        throw InvalidProtocol ( diag.str () );
    }
    m_pMgr->setPriority ( m_curMsg.m_dataType );
    m_setVersion ( m_curMsg.m_count );
}

//
// RcvThr :: m_versionAction()
//
void RcvThr :: m_setVersion ( ca_uint16_t clientVersion )
{
    if ( m_minorVersionNumber != 0u && m_minorVersionNumber != clientVersion ) {
        errlogPrintf (
            "attempt by the client to subsequently set "
            "it's version to a different value was ignored" );
    }
    else {
        m_minorVersionNumber = clientVersion;
    }
}

//
// RcvThr :: m_verifyRequest ()
//
RcvThr :: Chan & 
    RcvThr :: m_verifyRequest ()
{
	//
	// data type out of range ?
	//
	if ( m_curMsg.m_dataType > static_cast < unsigned > ( LAST_BUFFER_TYPE ) ) {
        throw InvalidDataType ();
	}
#pragma message ("dont allow this check to be redundant")
    bool invalidType = ( m_curMsg.m_dataType >= m_dbrTypePtrs.size () );
    if ( invalidType || m_dbrTypePtrs[ m_curMsg.m_dataType ].isNill () ) {
        std :: stringstream diag;
        diag << "specified dbr type ";
        diag << m_curMsg.m_dataType;
        diag <<  " isnt supported by this server";
        throw InvalidProtocol ( diag.str () );
    }


	//
	// no elements?
	//
	if ( m_curMsg.m_count == 0u ) {
        throw ArrayElementSubsetIsEmpty ();
	}

	//
	// channel exists for this resource id ?
	//
    chronIntId tmpId ( m_curMsg.m_cid );
	Chan * pChan = m_chanTable.lookup ( tmpId );
	if ( ! pChan ) {
        throw InvalidChannelId ();
	}
    return * pChan;
}

//
// RcvThr :: m_readAction()
//
void RcvThr :: m_readAction ( IterIn & )
{
    Chan & chan = m_verifyRequest ();
    chan.read ( m_curMsg );
}

//
// RcvThr :: m_writeAction()
//
void RcvThr :: m_writeAction ( IterIn & iterIn )
{	
    Chan & chan = m_verifyRequest ();
    PCat pPL = CatalogIterDBR :: factory ( m_curMsg.m_dataType, 
                    iterIn, m_curMsg.m_count, m_curMsg.m_postsize );
    chan.write ( m_curMsg, pPL );
}

//
// RcvThr :: m_subscribeAction()
//
void RcvThr :: m_subscribeAction ( IterIn & iterIn )
{
    Chan & chan = m_verifyRequest ();
    if ( m_curMsg.m_postsize < sizeof ( mon_info ) ) {
        throw InvalidProtocol ( 
            "subscription req msg payload size too small" );
    }
    mon_info monInfo;
    {
        monInfo.m_lval = iterIn.pop < epicsFloat32 > ();
        monInfo.m_hval = iterIn.pop < epicsFloat32 > ();
        monInfo.m_toval = iterIn.pop < epicsFloat32 > ();
        monInfo.m_mask = iterIn.pop < ca_uint16_t > ();
    }
#pragma message ("subscription parms will not be empty" )
    PCat pCatalog;
	chan.subscribe ( m_curMsg, pCatalog );
}

//
// RcvThr :: m_eventsOnAction()
//
void RcvThr :: m_eventsOnAction ( IterIn & )
{
    tsDLIter < Chan > iterChans = m_chanList.firstIter ();
    while ( iterChans.valid() ) {
        iterChans->enterFlowControlMode ();
        iterChans++;
    }
}

//
// RcvThr :: m_eventsOffAction()
//
void RcvThr :: m_eventsOffAction ( IterIn & )
{
    tsDLIter < Chan > iterChans = m_chanList.firstIter ();
    while ( iterChans.valid() ) {
        iterChans->exitFlowControlMode ();
        iterChans++;
    }
}

//
// RcvThr :: m_hostNameAction ()
//
void RcvThr :: m_hostNameAction ( IterIn & iterIn )
{
    if ( m_chanList.count () != 0 ) {
        throw InvalidProtocol ( 
            "Attempt to set user name after "
            "circuit initialization" );
    }
    // resize, m_and_ fill with zeros
    m_hostNameStr.resize ( m_curMsg.m_postsize + 1u );
    iterIn.popStringAppendTo ( m_hostNameStr, m_curMsg.m_postsize );
}

//
// RcvThr :: m_clientNameAction ()
//
void RcvThr :: m_clientNameAction ( IterIn & iterIn )
{
    if ( m_chanList.count () != 0 ) {
        throw InvalidProtocol ( 
            "Attempt to set user name after "
            "circuit initialization" );
    }
    m_hostNameStr.resize ( m_curMsg.m_postsize + 1u );
    iterIn.popStringAppendTo ( m_userNameStr, m_curMsg.m_postsize );
}

//
// RcvThr :: m_createChannelAction ()
//
void RcvThr :: m_createChannelAction ( IterIn & iterIn )
{
    // The available field is used (abused)
    // here to communicate the miner version number
    // starting with CA 4.1. The field was set to zero
    // prior to 4.1
    if ( m_curMsg.m_available > 0xffff ) {
        throw InvalidProtocol ( 
            "Channel create msg specifying invalid version" );
    }
    m_setVersion ( static_cast < ca_uint16_t > ( m_curMsg.m_available ) );

    StringReaderIterIn 
        stringReaderIterIn ( iterIn, m_curMsg.m_postsize );
    if ( stringReaderIterIn.size () == 0u ) {
        throw InvalidProtocol ( 
            "Channel create msg didnt specify "
            "the channel's name" );
    }

    // We shouldnt be receiving a connect message from 
    // an R3.11 client because we will not respond to their
    // search requests. If so, we log a message and rudely
    // disconnect the client.
    if ( ! CA_V44 ( m_minorVersionNumber ) ) {
        throw InvalidProtocol ( "R3.11 connect sequence from old client" );
    }
    try {
        // create channel description
        // o channel name
        // o !!!! channel priority (not passed in) !!!!
#pragma message ( "channel description not passed in" )
	PCat pChanDescr = ChannelDescription :: factory ( stringReaderIterIn );
        // create channel in service
        PChan pServiceChan = m_pService->createChannel ( pChanDescr );
	// create circuit monitor request
        Chan * pRcvThrChan = new ( m_chanFreeList ) 
                Chan ( *this, pServiceChan, m_curMsg.m_cid );
        m_chanTable.idAssignAdd ( *pRcvThrChan );
        m_chanList.add ( *pRcvThrChan );
        try {
	        // create channel in server
	        Mgr :: PChan pMgrChan = 
                    m_pMgr->createChannel ( pRcvThrChan->getSID (), m_curMsg.m_cid );
            // install server channel into rcv thr chan
	        pRcvThrChan->setServerChan ( pMgrChan );
#pragma message ("this wont do")
static PType m_pSubscrReqType;
static PType m_pSubscrResType;
            PReq pCirMonReq = pServiceChan->createRequest ( 
			    		Request :: subscribe, 
			    		m_pSubscrReqType, 
                        m_pSubscrResType );
	        PResp pResp = pMgrChan->installCircuitMonitor ( pCirMonReq );
            pCirMonReq->initiate ( PCat (), pResp );
        }
        catch ( ... ) {
            m_chanTable.remove ( *pRcvThrChan );
            m_chanList.remove ( *pRcvThrChan );
            m_chanFreeList.deleteProduct ( *pRcvThrChan );
            throw;
        }
    }
    catch ( const std :: exception & except ) {
        m_pRspThr->chanCreateFailResp ( m_curMsg.m_cid, except );
    }
    catch ( ... ) {
        m_pRspThr->chanCreateFailResp ( m_curMsg.m_cid, 
                                        nonstandardException );
    }
}

//
// RcvThr :: m_echoAction()
//
void RcvThr :: m_echoAction ( IterIn & )
{
    try {
        m_pMgr->postHdrOnlyResp ( m_curMsg );
    }
    catch ( exception & except ) {
        errlogPrintf ( 
            "Exception when creating echo "
            "request object was \"%s\".\n", 
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "Nonstandard exception when creating "
            "echo response object.\n" );
    }
}

//
// RcvThr :: m_clearChannelAction ()
//
void RcvThr :: m_clearChannelAction ( IterIn & )
{
	// Verify the channel
    chronIntId tmpId ( m_curMsg.m_cid );
	Chan * pChan = m_chanTable.remove ( tmpId );
	if ( pChan ) {
        m_chanList.remove ( *pChan );
        m_chanFreeList.deleteProduct ( *pChan );
	    // send delete confirmed message -
        // the constructor enqueues the response
        m_pMgr->postHdrOnlyResp ( m_curMsg );
    }
    else {
        throw InvalidProtocol (   
            "Invalid channel identifier in "
            "channel delete request" );
	}
}

//
// RcvThr :: m_eventCancelAction()
//
void RcvThr :: m_eventCancelAction ( IterIn & )
{
    chronIntId tmpId ( m_curMsg.m_cid );
    Chan * pChan = m_chanTable.lookup ( tmpId );
    if ( ! pChan ) {
        throw InvalidProtocol ( 
            "Invalid channel identifier in "
            "subscription delete request" );
    }
	bool success = pChan->unsubscribe ( m_curMsg.m_available );
    if ( success ) {
        // send delete confirmed message -
        // the constructor enqueues the response
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_EVENT_ADD;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = m_curMsg.m_dataType;
        respHdr.m_count = m_curMsg.m_count;
        respHdr.m_cid = m_curMsg.m_cid;
        respHdr.m_available = m_curMsg.m_available;
        m_pMgr->postHdrOnlyResp ( respHdr );
    }
    else {
        throw InvalidProtocol ( 
            "Invalid subscription identifier in "
            "subscription delete request" );
    }
}

void RcvThr :: m_readSyncAction ( IterIn & )
{
    m_readLackingCallbackCancel ();
    // send delete confirmed message -
    // the constructor enqueues the response
    m_pMgr->postHdrOnlyResp ( m_curMsg );
}

//
// RcvThr :: m_readLackingCallbackCancel ()
//
void RcvThr :: m_readLackingCallbackCancel ()
{
    tsDLIter < Chan > iterChans = m_chanList.firstIter ();
    while ( iterChans.valid() ) {
        tsDLIter < Chan > iterChansNext = iterChans;
        iterChansNext++;
        iterChans->readLackingCallbackCancel ();
        iterChans = iterChansNext;
    }
}

//
// M_CircuitInReader
//
RcvThr :: M_CircuitInReader :: M_EnclLnk
    RcvThr :: M_CircuitInReader ::   
        m_enclLnk ( & RcvThr :: m_circuitInReader );
Vec < const epicsUInt8 > RcvThr :: M_CircuitInReader ::
    read ( size_t nBytesMax )
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    Mutex :: Unlocker unlocker ( rcvThr.m_mutex );
    return m_pCircuitIn->read ( nBytesMax );
}
RcvThr :: M_CircuitInReader ::
    M_CircuitInReader ( std :: auto_ptr < CircuitIn > pCircuitIn ) :
    m_pCircuitIn ( pCircuitIn )
{
    if ( ! m_pCircuitIn.get () ) {
        throw std :: runtime_error ( 
               "RcvThr is being attached to a nill recv circuit" );
    }
}
void RcvThr :: M_CircuitInReader :: rewind ()
{
    throw std :: runtime_error ( 
           "M_CircuitInReader isnt rewindable" );
}
void RcvThr :: M_CircuitInReader :: shutdown ()
{
    m_pCircuitIn->shutdown ();
}

//
// M_Handle
//
RcvThr :: M_Handle :: M_EnclLnk
    RcvThr :: M_Handle ::   
        m_enclLnk ( & RcvThr :: m_handle );
RcvThr :: M_Handle :: TargetMutexPair 
    RcvThr :: M_Handle :: m_targetMutexPair ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( rcvThr, rcvThr.m_mutex );
}
HandleIntf < RcvThr > &
    RcvThr :: M_Handle :: m_clone ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    rcvThr.m_refCtr.reference ();
    return *this;
}
void RcvThr :: M_Handle :: m_release ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    if ( rcvThr.m_refCtr.unreference () ) {
        delete & rcvThr;
        // no code beyond here due to above delete
    }
    // no code beyond here due to above delete
}

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca

