
/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#define epicsExportSharedSymbols
#include "caSvrCirRspThr.h"
#include "caSvrCirRcvThrChan.h"
#include "FailureContext.h"

namespace ca {
namespace Svr {
namespace Cir {


RcvThr :: Chan :: 
    Chan ( 
        RcvThr & recvThrCtx, 
        PChan & pServiceChan, 
        caResId cid ) :
    m_pServiceChan ( pServiceChan ),
    m_recvThr ( recvThrCtx ),
    m_cid ( cid )
{
    if ( pServiceChan.isNill () ) {
        throw std :: invalid_argument ( 
            "creating a recv thread channel with a nill "
            "service channel is not allowed" );
    }
    if ( m_pServerChan.isNill () ) {
        throw std :: invalid_argument ( 
            "creating a recv thread channel with a nill "
            "server channel is not allowed" );
    }
}

void RcvThr :: Chan :: 
    show ( unsigned level ) const
{
    FixedString < 128 > str;
    m_pServiceChan->cannonicalName ( str );
    printf ( 
        "Chan: server id %8u name \"%s\"\n", 
        this->getSID (), str.c_str () );
    if ( level > 0 ) {
        m_pServiceChan->show ( level - 1 );
        m_pServerChan->show ( level - 1 );
    }
}

void RcvThr :: Chan ::
	m_exceptionResponse ( const PResp & pResp, 
	const std :: exception & except, const char * pCtx )
{
    // if pResp is nill we have no choice but to allow the
    // exception to disconnect the client
    if ( pResp.isNill () ) {
        throw;
    }
    std :: string ctx = "c++ exception, ";
    ctx += except.what ();
    ctx += ", ";
    ctx += pCtx;
    m_pServerChan->postFailure ( pResp, ctx.c_str () );
}

void RcvThr :: Chan :: 
    read ( const HdrLargeArray & reqMsg )
{
    // if we cant create the response object the resulting exception 
    // is currently allowed to cause a disconnect
    PResp pResp;
    try {
#pragma message ( "error recovery needs work here - must send correct response to client" )
        bool invalidType = ( reqMsg.m_dataType >= 
                m_recvThr.m_dbrTypePtrs.size () );
        if ( invalidType || 
            m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ].isNill () ) {
            throw std :: range_error ( 
                "specified dbr type isnt supported by this server" );
        }
        Ptr < const Type > voidReqType;
        // We are careful here not to cross the Mgr, RspThr, or
        // Service locks when creating or initializing the request so 
        // that there will not be a deadlock
        PReq pReq = m_pServiceChan->createRequest ( 
            Request :: read, voidReqType, 
            m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ] );
        pResp = m_pServerChan->createReadResp ( pReq, reqMsg, m_cid );
        pReq->initiate ( PCat (), pResp );
    }
    catch ( std :: exception & except ) {
        m_exceptionResponse ( pResp, except, 
            "RcvThr :: Chan :: read: "
            "exception when creating / initiating "
            "\"read\" request in service" );
    }
    catch ( ... ) {
        m_exceptionResponse ( pResp, nonstandardException, 
            "RcvThr :: Chan :: read: "
            "exception when creating / initiating "
            "\"read\" request in service" );
    }
}

void RcvThr :: Chan :: 
    write ( const HdrLargeArray & reqMsg, const PCat & pVal )
{
    // if we cant create the response object the resulting exception 
    // is currently allowed to cause a disconnect
    PResp pResp;
    try {
#pragma message ( "error recovery needs work here - must send correct response to client" )
        bool invalidType = ( reqMsg.m_dataType >= 
                m_recvThr.m_dbrTypePtrs.size () );
        if ( invalidType || 
            m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ].isNill () ) {
            throw std :: range_error ( 
                "specified dbr type isnt supported by this server" );
        }
        Ptr < const Type > voidRespType;
        // We are careful here not to cross the Mgr, RspThr, or
        // Service locks when creating or initializing the request so 
        // that there will not be a deadlock
        PReq pReq = m_pServiceChan->createRequest ( 
                    Request :: write, 
                    m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ], 
                    voidRespType );
        PResp pServiceResp = 
            m_pServerChan->createWriteResp ( pReq, reqMsg, m_cid );
        pReq->initiate ( pVal, pResp );
    }
    catch ( std :: exception & except ) {
        m_exceptionResponse ( pResp, except, 
            "RcvThr :: Chan :: write: "
            "exception when creating / initiating "
            "\"write\" request in service" );
    }
    catch ( ... ) {
        m_exceptionResponse ( pResp, nonstandardException, 
            "RcvThr :: Chan :: write: "
            "exception when creating / initiating "
            "\"write\" request in service" );
    }
}

void RcvThr :: Chan :: 
    subscribe ( const HdrLargeArray & reqMsg, const PCat & reqParms )
{
    // if we cant create the response object the resulting exception 
    // is currently allowed to cause a disconnect
    PResp pResp;
    try {
#pragma message ( "error recovery needs work here - must send correct response to client" )
        bool invalidType = ( reqMsg.m_dataType >= 
                m_recvThr.m_dbrTypePtrs.size () );
        if ( invalidType || 
            m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ].isNill () ) {
            throw std :: range_error ( 
                "specified dbr type isnt supported by this server" );
        }
#pragma message ("eventually the request modifier type needs to be valid")
        Ptr < const Type > voidReqType;
        // We are careful here not to cross the Mgr, RspThr, or
        // Service locks when creating or initializing the request so 
        // that there will not be a deadlock
        //
        // if we are unable to reserve sufficent event entries we 
        // have no choice but to allow the resulting exception to 
        // cause a disconnect 
        //
        PReq pReq = 
            m_pServiceChan->createRequest ( 
                Request :: subscribe, voidReqType, 
                m_recvThr.m_dbrTypePtrs[ reqMsg.m_dataType ] );
        PResp pResp = m_pServerChan->
            createSubscResp ( pReq, reqMsg, m_cid, reqParms );
        pReq->initiate ( reqParms, pResp );
#pragma message ( "is sending the first subscription update" )
#pragma message ( "a service responsibiliy?" )
#pragma message ( "if the library does it its more fool proof" )
        // 
#pragma message ( "catch exceptions occurring in the createSubscResp!!!!" )
#pragma message ( "if unable to obtain event queue space, the" )
#pragma message ( "server should send a error response as the first update" )
#pragma message ( "this should m_not_ cause a client disconnect" )
#pragma message ( "(avoid positive congestion feedback)" )
    }
    catch ( std :: exception & except ) {
        m_exceptionResponse ( pResp, except, 
            "RcvThr :: Chan :: subscribe: "
            "exception when creating / initiating "
            "\"subscribe\" request in service" );
    }
    catch ( ... ) {
        m_exceptionResponse ( pResp, nonstandardException, 
            "RcvThr :: Chan :: subscribe: "
            "exception when creating / initiating "
            "\"subscribe\" request in service" );
    }
}

bool RcvThr :: Chan ::
    unsubscribe ( ca_uint32_t clientIdIn )
{
    return m_pServerChan->unsubscribe ( clientIdIn );
}

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca
