
/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include <iostream>

#include "ConString.h"
#include "epicsStdio.h"
#include "errlog.h"
#include "epicsExit.h"

#define epicsExportSharedSymbols
#include "v3Proto.h"
#include "caIntfIO.h"
#include "caSvrCirMgr.h"
#include "caSvrCirRspThr.h"
#include "caAccessRights.h"
#include "caSvrCirRspThrRead.h"
#include "caSvrCirRspThrWrite.h"
#include "caSvrCirRspThrSubscribe.h"

using namespace std;

namespace ca {
namespace Svr {
namespace Cir {

static epicsThreadPrivateId rspThrContextId = 0;

extern "C" void m_rspThrExitHandlerCir (void *)
{
    if ( rspThrContextId ) {
        epicsThreadPrivateDelete ( rspThrContextId );
        rspThrContextId = 0;
    }
}

extern "C" void m_rspThrInitOnceCir ( void * )
{
    rspThrContextId = epicsThreadPrivateCreate ();
    if ( rspThrContextId ) {
        epicsAtExit ( m_rspThrExitHandlerCir, 0 );
    }
}

RspThr :: RspThr ( const Ptr < Mgr > & pMgr, 
        std :: auto_ptr < CircuitOut > pCircuitOut,
        const Ptr < VecAllocator > & pVecAllocator ) :
    m_thread ( m_entryPt, "CAS-Cir-Rsp", 
        epicsThreadGetStackSize ( epicsThreadStackMedium ), 
        epicsThreadPriorityCAServerLow ),
    m_byteSeqOut ( pVecAllocator ),
    m_pOut ( pCircuitOut ),
    m_pMgr ( pMgr ),
    m_minorVersionNumber ( 0 ),
    m_threadShutdownCmd ( false )
{
    static epicsThreadOnceId once = EPICS_THREAD_ONCE_INIT;
    epicsThreadOnce ( &once, m_rspThrInitOnceCir, 0 );
    if ( ! rspThrContextId ) {
        throw std :: runtime_error ( 
            "RspThr ctor, unable to allocate "
            "rspThrCtx :: rspThrContextId" );
    }
    if ( ! m_pOut.get () ) {
        throw std :: runtime_error ( 
            "RspThr ctor, m_pOut was nill?" );
    }
    m_sendVersion ();
}

RspThr :: ~RspThr ()
{
    m_threadShutdownCmd = true;
    m_event.signal ();
    m_thread.exitWait ();
}

//
// RspThr :: show ()
//
void RspThr :: show ( unsigned level ) const
{
    cout << "RspThr" << endl;
    if ( level > 0 ) {
        m_thread.show ( level - 1 );
        m_byteSeqOut.show ( level - 1 );
        printf ( "EvQue: m_event\n" );
        m_event.show ( level - 1 );
        if ( m_pOut.get () ) {
            m_pOut->show ( level - 1 );
        }
        cout << "last send time stamp" << endl;
        m_lastSendTS.show ( level -1 );
        cout << "m_minorVersionNumber = " 
            << m_minorVersionNumber << endl;
    }
}

class RspThr :: CircuitOutWriter : 
    public Writer {
public:
    CircuitOutWriter ( CircuitOut & );
private:
    void write ( const Vec  < const epicsUInt8 > & );
    CircuitOut & m_circuitOut;
};

void RspThr :: m_flush ()
{
    CircuitOutWriter circuitOutWriter ( *m_pOut );
    m_byteSeqOut.flush ( circuitOutWriter );
}

//
// RspThr :: m_accessRightsResponse ()
//
void RspThr :: m_accessRightsResponse ( caResId cid, const Catalog & resp )
{
    RspThr * pRspThr = reinterpret_cast < RspThr * > 
            ( epicsThreadPrivateGet ( rspThrContextId ) );
    assert ( pRspThr );

    // noop if this is an old client
	if ( CA_V41 ( pRspThr->m_minorVersionNumber ) ) {
        bool readAccess = false;
        bool writeAccess = false;
        AccessRights accessRights = extractAccessRights ( resp );
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_ACCESS_RIGHTS;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = 0;
        respHdr.m_count = 0;
        respHdr.m_cid = cid;
        respHdr.m_available = 0;
		if ( accessRights.readable () ) {
			respHdr.m_available |= CA_PROTO_ACCESS_RIGHT_READ;
		}
		if ( accessRights.writeable () ) {
			respHdr.m_available |= CA_PROTO_ACCESS_RIGHT_WRITE;
		}
        MsgMinderV3 msgMinderV3 ( pRspThr->m_byteSeqOut, 
                        pRspThr->m_minorVersionNumber );
        msgMinderV3.pushResp ( respHdr );
	}
}

// send minor protocol revision to the client
void RspThr :: m_sendVersion ()
{
    HdrLargeArray respHdr;
    respHdr.m_cmmd = CA_PROTO_VERSION;
    respHdr.m_postsize = 0;
    respHdr.m_dataType = 0;
    respHdr.m_count = CA_MINOR_PROTOCOL_REVISION;
    respHdr.m_cid = 0;
    respHdr.m_available = 0;
    MsgMinderV3 msgMinderV3 ( m_byteSeqOut, m_minorVersionNumber );
    msgMinderV3.pushResp ( respHdr );
    msgMinderV3.commit ();
}

void RspThr :: chanCreateFailResp ( caResId cid, 
                                    const std :: exception & except )
{
    errlogPrintf ( 
        "CAS: exception \"%s\" - \"%s\" when creating channel\n", 
        typeid (except).name (), except.what () );
    HdrLargeArray respHdr;
    respHdr.m_cmmd = CA_PROTO_CREATE_CH_FAIL;
    respHdr.m_postsize = 0;
    respHdr.m_dataType = 0;
    respHdr.m_count = 0;
    respHdr.m_cid = cid;
    respHdr.m_available = 0;
    MsgMinderV3 msgMinderV3 ( m_byteSeqOut, m_minorVersionNumber );
    msgMinderV3.pushResp ( respHdr );
    msgMinderV3.commit ();
}

void RspThr :: CircuitOutWriter ::
    write ( const Vec < const epicsUInt8 > & pVec )
{
    m_circuitOut.write ( pVec );
}

inline RspThr :: CircuitOutWriter :: 
    CircuitOutWriter ( CircuitOut & circuitOut ) :
    m_circuitOut ( circuitOut )
{
}

//
// M_EntryPt
//
RspThr :: M_EntryPt :: M_EnclLnk
    RspThr :: M_EntryPt ::   
        m_enclLnk ( & RspThr :: m_entryPt );
void RspThr :: M_EntryPt :: run ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    epicsThreadPrivateSet ( rspThrContextId, & rspThr );
    try {
        Guard guard ( rspThr.m_mutex );
        while ( ! rspThr.m_threadShutdownCmd ) {
            rspThr.m_pMgr->processEvents ();
            Guard :: Release release ( guard );
            rspThr.m_flush ();
            rspThr.m_event.wait ();
        }   
    }
    catch ( std :: exception & ) {
#pragma message ("missing code")
    }
    catch ( ... ) {
#pragma message ("missing code")
    }
}

void RspThr :: M_EntryPt :: show ( unsigned level ) const
{
    const RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( rspThr.m_mutex );
    rspThr.show ( level );
}

RspThr * RspThr :: pAttachedRspThr ()
{
    // we make the assumption that a RspThr was created
    // before any call to this function occurs.
    if ( rspThrContextId == 0) {
        throw std :: runtime_error ( 
            "response thread context objects are being "
            "used before the response thread has initialized?" );
    }
    RspThr * pRspThr = reinterpret_cast < RspThr * > ( 
        epicsThreadPrivateGet ( rspThrContextId ) );
    if ( pRspThr == 0 ) {
        throw std :: runtime_error ( 
            "response thread context objects are being "
            "used with some other thread?" );
    }
    return pRspThr;
}

inline void RspThr :: m_verifyCorrectThread () const
{
#   ifndef NDEBUG
        RspThr * pRspThrTP = reinterpret_cast < RspThr * > ( 
            epicsThreadPrivateGet ( rspThrContextId ) );
        assert ( pRspThrTP == this );
#   endif
}

//////////////////////////////////////////////
// M_Handle - handle for RspThr object
//////////////////////////////////////////////
RspThr :: M_Handle :: M_EnclLnk
    RspThr :: M_Handle ::   
        m_enclLnk ( & RspThr :: m_handle );
RspThr :: M_Handle :: TargetMutexPair  
	RspThr :: M_Handle :: m_targetMutexPair ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( rspThr, rspThr.m_mutex );
}
HandleIntf < RspThr > & 
    RspThr :: M_Handle :: m_clone ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    rspThr.m_refCtr.reference ();
    return rspThr.m_handle;
}
void RspThr :: M_Handle :: m_release ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    if ( rspThr.m_refCtr.unreference () ) {
        delete & rspThr;
        // no code below here considering above delete
    }
    // no code below here considering above delete
}

// RspThr factory
Ptr < RspThr > 
    RspThr :: factory ( const Ptr < Mgr > & pCirSvr, 
                std :: auto_ptr < CircuitOut > pCircuitOut,
                const Ptr < VecAllocator > & pVecAllocator )
{
    RspThr * pRspThr = new 
        RspThr ( pCirSvr, pCircuitOut, pVecAllocator );
    return Ptr < RspThr > ( pRspThr->m_handle );
}

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca

