
/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include "db_access.h"

#ifndef NELEMENTS
#   define NELEMENTS(A) (sizeof(A)/sizeof(A[0]))
#endif

#define epicsExportSharedSymbols
#include "caSvrCirRspThrCirMon.h"
#include "caHdrLargeArray.h"
#include "v3Proto.h"

namespace ca {
namespace Svr {
namespace Cir {

static ca_uint16_t dbrTypeTable[][4] = {
{DBR_FLOAT, DBR_FLOAT,  DBR_FLOAT,  DBR_DOUBLE}, // csFloat
{DBR_CHAR,  DBR_SHORT,  DBR_LONG,   DBR_DOUBLE}, // csInt
{DBR_ENUM,  DBR_ENUM,   DBR_ENUM,   DBR_ENUM},   // csEnum
{DBR_STRING,DBR_STRING, DBR_STRING, DBR_STRING}  // csString
};

const RspThr :: CircMon :: PStateNotifyFunc 
RspThr :: CircMon :: m_stateNotifyDispatch [] = {
    & CircMon :: m_disconnectNotify,
    & CircMon :: m_unresponsiveNotify,
    & CircMon :: m_responsiveNotify
};

class OldClientUnresponsiveDisconnect :
    public std :: exception {
    const char * what () const throw ();
};

const char * OldClientUnresponsiveDisconnect ::
    what () const throw () 
{
    return "OldClientUnresponsiveDisconnect";
}


RspThr :: CircMon :: 
    CircMon ( caResId sid, caResId cid ) :
    m_sid ( sid ),
    m_cid ( cid ),
    m_createChannelRespWasSent ( false )
{
    assert ( 
        m_stateNotifyDispatch[ Channel :: csUnresp ] == 
        & CircMon :: m_unresponsiveNotify );
    assert ( 
        m_stateNotifyDispatch[ Channel :: csResp ] == 
        & CircMon :: m_responsiveNotify );
    assert ( 
        m_stateNotifyDispatch[ Channel :: csDisconn ] == 
        & CircMon :: m_disconnectNotify );
}

#if 0
void RspThr :: CircMon :: 
    initiate ( PChan & pChan )
{
    try {
        PReq pReq = pChan->createRequest ( 
                            Service :: Channel :: subscribe, 
                            m_pSubscrReqType, m_pSubscrResType );
#pragma message ( "the request modifier parameters will" )
#pragma message ( "definately not be empty" )
        //PCat pReqModifiers;
        m_subscrReq.initiate ( guard, m_eventCallback, pReq 
                                /* pReqModifiers */ );
    }
    catch ( std :: exception & except ) {
        m_subscrReq.forceFailResponse ( 
            std :: string (            
                "RspThr :: Mgr initiate: "
                "exception when creating "
                "\"subscribe\" request in service was \"" ) +
            except.what () + 
            std :: string ( "\"" ) );
    }
    catch ( ... ) {
        m_subscrReq.forceFailResponse ( 
            "RspThr :: Mgr initiate: "
            "non-standard exception when creating "
            "\"subscribe\" request in service" );
    }
}
#endif

void RspThr :: CircMon :: 
    m_show ( unsigned level ) const
{
    printf ( 
        "RspThr :: CircMon\n" );
    if ( level > 0 ) {
        printf ( "m_cid = %u\n", m_cid );
        printf ( "m_sid = %u\n", m_sid );
    }
}

void RspThr :: CircMon :: 
    m_responseNotify ( const PCat & pCat) 
{
    unsigned newCircuitState = Channel :: csDisconn;
    // we are not concerned about lock hierarchy 
    // reversal here
    PCat :: Locker pCatLocked = pCat.operator -> ();
    Clerk clerk ( *pCatLocked );
    clerk.get ( Channel :: circuitState, newCircuitState );
    if ( newCircuitState < NELEMENTS ( m_stateNotifyDispatch ) ) {
        PStateNotifyFunc pMemFun = 
            m_stateNotifyDispatch[newCircuitState];
        (this->*pMemFun)( pCat );
    }
    else {
        m_disconnectNotify ( pCat );
    }
}

void RspThr :: CircMon ::
    m_unresponsiveNotify ( const PCat & pCat )
{
    RspThr * pRspThr = RspThr :: pAttachedRspThr ();
    if ( CA_V413 ( pRspThr->m_minorVersionNumber ) ) {
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_CH_CONN_STATE;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = 0;
        respHdr.m_count = 0;
        respHdr.m_cid = m_cid;
        respHdr.m_available = CA_CH_CONN_STATE_UNRESPONSIVE;
        MsgMinderV3 msgMinderV3 ( pRspThr->m_byteSeqOut,
                pRspThr->m_minorVersionNumber );
        msgMinderV3.pushResp ( respHdr );
        msgMinderV3.commit ();
    }
    else {
       // force disconnect
       throw OldClientUnresponsiveDisconnect ();
    }
}


//
// capped at what is a reasonable atomic 
// data size (64 bits, 8 bytes)
//
static unsigned calcSizeIdx ( unsigned n ) 
{
    unsigned idx = 3u;
    if ( n <= 1u ) {
        idx = 0u;
    }
    else if ( n <= 2u ) {
        idx = 1u;
    }
    else if (n <= 4u ) {
        idx = 2;
    }
    return idx;
}

void RspThr :: CircMon ::
    m_responsiveNotify ( const PCat & pCat )
{
#pragma message ("currently hardwired here for scalars")
    ca_uint16_t count = 1u;
    ca_uint16_t type = DBR_STRING;
    {
        PCat :: Locker pCatLocked = pCat.operator -> ();
        Clerk clerk ( *pCatLocked );
        unsigned nativeType = Channel :: csString;
        clerk.get ( Channel :: nativeType, nativeType, 
                        Channel :: csString );
        unsigned nativeTypeSize = 0u;
        clerk.get ( Channel :: nativeTypeSize, 
                        nativeTypeSize, 0u );
        unsigned sizeIdx = calcSizeIdx ( nativeTypeSize );
        type = dbrTypeTable[nativeTypeSize][sizeIdx];
    }
    RspThr * pRspThr = RspThr :: pAttachedRspThr ();
    if ( CA_V413 ( pRspThr->m_minorVersionNumber ) ) {
#pragma message ("missing code")
        //m_respThread.m_accessRightsResponse ( m_cid, resp );
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_CH_CONN_STATE;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = type;
        respHdr.m_count = count;
        respHdr.m_cid = m_cid;
        respHdr.m_available = CA_CH_CONN_STATE_RESPONSIVE;
        MsgMinderV3 msgMinderV3 ( pRspThr->m_byteSeqOut, 
                pRspThr->m_minorVersionNumber );
        msgMinderV3.pushResp ( respHdr );
        msgMinderV3.commit ();
    }
    else if ( ! m_createChannelRespWasSent ) {
        // A serious dissadvantage here is that the 
        // client's create channel request can 
        // languish for a really long time before 
        // it gets responded to. During that time 
        // period the client is unable to delete 
        // the channel data structure in the 
        // server that it has created.
#pragma message ("missing code")
        //pRspThr->m_accessRightsResponse ( m_cid, resp );
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_CREATE_CHAN;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = type;
        respHdr.m_count = count;
        respHdr.m_cid = m_cid;
        respHdr.m_available = m_sid;
        MsgMinderV3 msgMinderV3 ( pRspThr->m_byteSeqOut,
                    pRspThr->m_minorVersionNumber );
        msgMinderV3.pushResp ( respHdr );
        msgMinderV3.commit ();
        m_createChannelRespWasSent = true;
    }
}

void RspThr :: CircMon ::
    m_disconnectNotify ( const PCat & pCat )
{
    RspThr * pRspThr = RspThr :: pAttachedRspThr ();
    if ( CA_V413 ( pRspThr->m_minorVersionNumber ) ) {
        HdrLargeArray respHdr;
        respHdr.m_cmmd = CA_PROTO_CH_CONN_STATE;
        respHdr.m_postsize = 0;
        respHdr.m_dataType = 0;
        respHdr.m_count = 0;
        respHdr.m_cid = m_cid;
        respHdr.m_available = CA_CH_CONN_STATE_DISCONNECTED;
        MsgMinderV3 msgMinderV3 ( pRspThr->m_byteSeqOut,
                    pRspThr->m_minorVersionNumber );
        msgMinderV3.pushResp ( respHdr );
        msgMinderV3.commit ();
    }
    else {
       // force disconnect
       throw OldClientUnresponsiveDisconnect ();
    }
}

} // end of namespace Cir
} // end of namespace Svr
} // end of namespace ca
