/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include "caerr.h"
#include "errlog.h"
#include "osiWireFormat.h"

#define epicsExportSharedSymbols
#define caNetAddrSock
#include "caIntfIO.h"
#include "caNetAddr.h"
#include "caSvrDGMgr.h"
#include "caSvrDGMgrBeaconTimer.h"
#include "caSvrDGMgrBeaconAnomGov.h"
#include "caSvrDGMgrNameRes.h"
#include "caSvrDGRspThr.h"
#include "caSvrDGRcvThr.h"
#include "osiPoolStatus.h" // osi pool monitoring functions
#include "caServer.h"
#include "caChannelDescription.h"

namespace ca {
namespace Svr {
namespace DG {

PMgr Mgr :: factory ( const PService & pService, 
                const PCat & pNameResChanDescr,
                epicsTimerQueue & timerQueue, 
                std :: auto_ptr < DatagramAgent > pDatagramAgent, 
                const Ptr < VecAllocator > & pVecAllocator )
{
    Mgr * pMgr = new Mgr ( pService, pNameResChanDescr, 
                    timerQueue, pDatagramAgent, pVecAllocator );
    return PMgr ( pMgr->m_handle );
}

//
// Mgr :: Mgr()
//
Mgr :: Mgr ( 
        const PService & pService, 
        const PCat & pNameResChanDescr,
		epicsTimerQueue & timerQueue, 
		std :: auto_ptr < DatagramAgent > pDatagramAgent, 
		const Ptr < VecAllocator > & pVecAllocator ) :
    m_noopMutex ( epics :: Mutex :: factoryNOOP ),
    m_eventQueue ( m_mutex, m_newEventSignal ),
    m_pBeaconTimer ( new BeaconTimer ( *this, timerQueue ) ),
    m_pBeaconAnomGov ( new BeaconAnomGov ( *this, timerQueue ) )
{
    if ( ! pDatagramAgent.get () ) {
        throw std :: invalid_argument ( "ca :: Svr :: DG :: Mgr ctor: nill dg agent" );
    }
    if ( pVecAllocator.isNill () ) {
        throw std :: invalid_argument ( "ca :: Svr :: DG :: Mgr ctor: nill vec allocator" );
    }
    m_pRspThr = RspThr :: factory (  PMgr ( m_handle ), timerQueue, 
			                        pDatagramAgent->createDatagramOut (), 
                                    pVecAllocator );
    const size_t nDGIn = pDatagramAgent->numberOfDatagramInLeft ();
    m_recvThreadPtrVector.resize ( nDGIn );
    for ( size_t i = 0u; i < nDGIn; i++ ) {
        std :: auto_ptr < DatagramIn > pDatagramIn = 
            pDatagramAgent->createNextDatagramIn ( pVecAllocator );
        m_recvThreadPtrVector[i] =
			RcvThr :: factory ( pService, pNameResChanDescr,
                                PMgr ( m_handle ), pDatagramIn );
        if ( m_recvThreadPtrVector[i].isntNill () ) {
            m_recvThreadPtrVector[i]->start ();
        }
    }
    m_pBeaconTimer->start ();
}

//
// Mgr :: ~Mgr ()
//
Mgr :: ~Mgr ()
{
    // explicitly destroy recv threads first
	const size_t vecLen = m_recvThreadPtrVector.size ();
    for ( size_t i = 0u; i < vecLen; i++ ) {
        m_recvThreadPtrVector[i].clear ();
    }
    while ( ReqNotify * pReq = m_reqList.get () ) {
        pReq->serverDestroyNotify ();
        // since pReq could be destroyed 
        // by the serviceDestroyNotify function
        // it should not be accessed here
    }
}

//
// Mgr :: show()
//
void Mgr :: 
    show ( unsigned level ) const
{
    if ( level >= 1u ) {
        m_pRspThr->show ( level - 1 );
        m_mutex.show ( level - 1 );
	    const size_t vecLen = m_recvThreadPtrVector.size ();
        for ( size_t i = 0u; i < vecLen; i++ ) {
            m_recvThreadPtrVector[i]->show ( level - 1 );
        }
        m_pBeaconTimer->show ( level - 1 );
        m_pBeaconAnomGov->show ( level - 1 );
    }
}

void Mgr :: 
    generateBeaconAnomaly ()
{
    m_pBeaconAnomGov->start ();
}

Mgr :: InitiateParms 
    Mgr :: createNameResResp ( 
            const PReq & pReq, const DGParm & dgParms, caResId cid, 
            bool failRespRequired, const StringReader & name )
{
    return NameRes :: factory ( *this, pReq, dgParms,
                        cid, failRespRequired, name );
}

//////////////////////////////////////////////
// M_NewEventSignal
//////////////////////////////////////////////
Mgr :: M_NewEventSignal :: M_EnclLnk
    Mgr :: M_NewEventSignal ::   
        m_enclLnk ( & Mgr :: m_newEventSignal );
void Mgr :: M_NewEventSignal :: signal ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
#pragma message ("potential for unneeded locking here")
#pragma message ("possible fix => epicsEvent which is reference counting like epics :: Mutex")
    mgr.m_pRspThr->eventSignal ();
}

//////////////////////////////////////////////
// M_Handle
//////////////////////////////////////////////
Mgr :: M_Handle :: M_EnclLnk
    Mgr :: M_Handle ::   
        m_enclLnk ( & Mgr :: m_handle );
Mgr :: M_Handle :: TargetMutexPair 
    Mgr :: M_Handle :: m_targetMutexPair ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( mgr, mgr.m_mutex );
}
HandleIntf < Mgr > & Mgr :: M_Handle :: m_clone ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
    mgr.m_refCtr.reference ();
    return *this;
}
void Mgr :: M_Handle :: m_release ()
{
    Mgr & mgr ( m_enclLnk.enclosure ( *this ) );
    if ( mgr.m_refCtr.unreference () ) {
        delete & mgr;
        // no code after here considering above delete
    }
    // no code after here considering above delete
}

#if 0
//
// Mgr::xSend()
//
casStreamWriteIO::flushCondition Mgr::xSend ( char *pBufIn, // X aCC 361
        size_t nBytesAvailableToSend, size_t nBytesNeedToBeSent,
        size_t &nBytesSent ) 
{
    assert ( nBytesAvailableToSend >= nBytesNeedToBeSent );

    size_t totalBytes = 0;
    while ( totalBytes < nBytesNeedToBeSent ) {
        cadg *pHdr = reinterpret_cast < cadg * > ( & pBufIn[totalBytes] );

        assert ( totalBytes <= bufSizeT_MAX-pHdr->cadg_nBytes );
        assert ( totalBytes + pHdr->cadg_nBytes <= nBytesAvailableToSend );

        char * pDG = reinterpret_cast < char * > ( pHdr + 1 );
        unsigned sizeDG = pHdr->cadg_nBytes - sizeof ( *pHdr );

        if ( pHdr->cadg_addr.isValid() ) {
            casStreamWriteIO::flushCondition stat = 
                this->osdSend ( pDG, sizeDG, pHdr->cadg_addr );	
	        if ( stat != casStreamWriteIO::flushProgress ) {
                break;
            }
        }

        totalBytes += pHdr->cadg_nBytes;
    }

    if ( totalBytes ) {
		//
		// !! this time fetch may be slowing things down !!
		//
		//this->lastSendTS = epicsTime::getCurrent();
        nBytesSent = totalBytes;
        return casStreamWriteIO::flushProgress;
    }
    else {
        return casStreamWriteIO::flushNone;
    }
}

//
// Mgr::xRecv ()
//
casStreamReadIO::fillCondition Mgr::xRecv (
    char *pBufIn, size_t nBytesToRecv, // X aCC 361
        fillParameter parm, size_t &nByesRecv)
{
    const char *pAfter = pBufIn + nBytesToRecv;
    char *pCurBuf = pBufIn;
    size_t nDGBytesRecv;
    casStreamReadIO::fillCondition stat;
    cadg *pHdr;

    while (pAfter-pCurBuf >= static_cast<int>(MAX_UDP_RECV+sizeof(cadg))) {
        pHdr = reinterpret_cast < cadg * > ( pCurBuf );
	    stat = this->osdRecv ( reinterpret_cast < char * > ( pHdr + 1 ), 
            MAX_UDP_RECV, parm, nDGBytesRecv, pHdr->cadg_addr);
	    if (stat==casFillProgress) {
            pHdr->cadg_nBytes = nDGBytesRecv + sizeof(*pHdr);
            pCurBuf += pHdr->cadg_nBytes;
		    //
		    // !! this time fetch may be slowing things down !!
		    //
		    //this->lastRecvTS = epicsTime::getCurrent();
	    }
        else {
            break;
        }
    }

    nDGBytesRecv = pCurBuf - pBufIn;
    if (nDGBytesRecv) {
        nByesRecv = nDGBytesRecv;
        return casFillProgress;
    }
    else {
        return casFillNone;
    }
}

//
// Mgr :: asyncSearchResp()
//
//
// this results in many small UDP frames which unfortunately
// isnt particularly efficient
//
void Mgr :: m_searchResponse ( 
    Guard & guard, const caNetAddr & outAddr,
	const HdrLargeArray & msg, const pvExistReturn & retVal,
    ca_uint16_t protocolRevision, ca_uint32_t sequenceNumber )
{
    guard.assertIdenticalMutex ( m_mutex );

    if ( retVal.getStatus() != pverExistsHere ) {
        return;
    }

    void * pRaw;
    const OutBuf :: Ctx outctx = 
        m_out.pushCtx ( guard, sizeof(cadg), MAX_UDP_SEND, pRaw );

    cadg * pRespHdr = static_cast < cadg * > ( pRaw );

    // insert version header at the start of the reply message
    this->sendVersion ( guard );

    caHdr * pMsg = reinterpret_cast < caHdr * > ( pRespHdr + 1 );
    assert ( ntohs ( pMsg->m_cmmd ) == CA_PROTO_VERSION );
    if ( CA_V411 ( protocolRevision ) ) {
        pMsg->m_cid = htonl ( sequenceNumber );
        pMsg->m_dataType = htons ( sequenceNoIsValid );
    }

	m_searchResponse ( guard, msg, retVal );

    pRespHdr->cadg_nBytes = 
        outctx.pop ( guard, outctx ) + sizeof ( *pRespHdr );
    if ( pRespHdr->cadg_nBytes > sizeof ( *pRespHdr ) + sizeof (caHdr) ) {
        pRespHdr->cadg_addr = outAddr;
        m_out.commitRawMsg ( guard, pRespHdr->cadg_nBytes );
    }
}
#endif

#if 0
bool Mgr :: inBufFull () const
{
    Guard guard ( m_mutex );
    return m_in.full ( guard );
}

size_t Mgr :: inBufBytesAvailable () const
{
    Guard guard ( m_mutex );
    return m_in.bytesAvailable ( guard );
}

size_t Mgr :: outBufBytesPresent () const
{
    Guard guard ( m_mutex );
    return m_out.bytesPresent ( guard );
}

casStreamWriteIO :: flushCondition Mgr :: flush ()
{
    Guard guard ( m_mutex );
    return m_out.flush ( guard );
}

#endif

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca
