
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *      caSvrDGMgrBeaconAnomGov.cpp,v 1.5 2009/12/03 17:39:51 jhill Exp
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#define epicsExportSharedSymbols
#if 0
#include "caServer.h"
#endif
#include "caSvrDGMgrBeaconAnomGov.h"
#include "caSvrDGMgrBeaconTimer.h"

namespace ca {
namespace Svr {
namespace DG {

//
// the minimum period between beacon anomalies
// -------------------------------------------
//
// o Gateways imply a tradeoff towards less traffic while
// paying a price of a less tightly coupled environment.
// Therefore, clients should not expect instant reconnects 
// when a server is rebooted on the other side of a gateway. 
//
// o In a quiescent system servers should not need to be 
// rebooted and network connectivity should be constant.
//
// o Clients will keep trying to reach channels for around
// 15 min after the last beacon anomaly was seen.
//
// o The exponential backoff mechanism in the beacon timer
// ensures that clients will detect beacon anomalies 
// continuously for around 30 seconds after each beacon 
// anomaly is requested.
//
static const double 
    CAServerMinBeaconAnomalyPeriod = 60.0 * 5.0; // seconds

Mgr :: BeaconAnomGov :: 
    BeaconAnomGov ( Mgr & mgr, epicsTimerQueue & timerQueue ) :
    m_mgr ( mgr ), m_timer ( timerQueue.createTimer () ), m_anomalyPending ( false )
{
}

Mgr :: BeaconAnomGov :: 
    ~BeaconAnomGov ()
{
    m_timer.destroy ();
}

void Mgr :: BeaconAnomGov :: 
    start ()
{
    m_anomalyPending = true;
    epicsTimer :: expireInfo expInfo = m_timer.getExpireInfo ();
    if ( ! expInfo.active ) {
        m_mgr.m_pBeaconTimer->generateBeaconAnomaly ();
        m_timer.start ( m_timerNotify, CAServerMinBeaconAnomalyPeriod );
        m_anomalyPending = false;
    }
}

//
// M_TimerNotify
//
Mgr :: BeaconAnomGov :: M_TimerNotify :: M_EnclLnk
    Mgr :: BeaconAnomGov :: M_TimerNotify ::   
        m_enclLnk ( & Mgr :: BeaconAnomGov :: m_timerNotify );
epicsTimerNotify :: expireStatus 
    Mgr :: BeaconAnomGov :: M_TimerNotify ::
        expire ( const epicsTime & )
{
    BeaconAnomGov & beaconAnomGov ( m_enclLnk.enclosure ( *this ) );
    epics :: Mutex :: Guard guard ( beaconAnomGov.m_mgr.m_mutex );
    if ( beaconAnomGov.m_anomalyPending ) {
        beaconAnomGov.m_anomalyPending = false;
        beaconAnomGov.m_mgr.m_pBeaconTimer->generateBeaconAnomaly ();
    }
    return noRestart;
}
void Mgr :: BeaconAnomGov :: M_TimerNotify ::
    show ( unsigned level ) const
{
    const BeaconAnomGov & beaconAnomGov ( m_enclLnk.enclosure ( *this ) );
    epics :: Mutex :: Guard guard ( beaconAnomGov.m_mgr.m_mutex );
    beaconAnomGov.show ( level );
}

void Mgr :: BeaconAnomGov ::
    show ( unsigned level ) const
{
    printf ( "BeaconAnomGov: anomalyPending = %s\n",
        m_anomalyPending ? "T": "F" );
    if ( level ) {
        m_timer.show ( level - 1 );
    }
}

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca

