
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *      caSvrDGMgrBeaconTimer.cpp,v 1.9 2009/12/03 17:39:51 jhill Exp
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#if 0
#include "fdManager.h"
#endif

#include "envDefs.h"
#include "errlog.h"

#define epicsExportSharedSymbols
#if 0
#include "caServer.h"
#endif
#include "caSvrDGMgrBeaconTimer.h"

namespace ca {
namespace Svr {
namespace DG {

// the maximum beacon period if EPICS_CA_BEACON_PERIOD isnt available
static const double CAServerMaxBeaconPeriod = 15.0; // seconds

// the initial beacon period
static const double CAServerMinBeaconPeriod = 1.0e-3; // seconds

Mgr :: BeaconTimer :: 
    BeaconTimer ( Mgr & mgr, 
                epicsTimerQueue & timerQueue ) :
    m_beaconPeriod ( CAServerMinBeaconPeriod ), 
    m_maxBeaconInterval ( CAServerMaxBeaconPeriod ),
    m_timer ( timerQueue.createTimer() ), 
    m_mgr ( mgr )
{
    long weirdLongLegacyStatus;
    if ( envGetConfigParamPtr ( & EPICS_CAS_BEACON_PERIOD ) ) {
        weirdLongLegacyStatus = envGetDoubleConfigParam ( 
            & EPICS_CAS_BEACON_PERIOD, & m_maxBeaconInterval );
    }
    else {
        weirdLongLegacyStatus = envGetDoubleConfigParam ( 
            & EPICS_CA_BEACON_PERIOD, & m_maxBeaconInterval );
    }
	if ( weirdLongLegacyStatus ) {
		errlogPrintf (
			"fetch of EPICS env \"%s\" to double "
            "precision target failed\n", 
            EPICS_CAS_BEACON_PERIOD.name );
		errlogPrintf (
			"Setting \"%s\" = %f (the hard coded default)\n", 
            EPICS_CAS_BEACON_PERIOD.name, 
			CAServerMaxBeaconPeriod );
        m_maxBeaconInterval = CAServerMaxBeaconPeriod;
	}
    else if ( m_maxBeaconInterval < CAServerMinBeaconPeriod ) {
		errlogPrintf (
			"fetch of EPICS env \"%s\" to double "
            "precision target produced a below minimum result\n", 
            EPICS_CAS_BEACON_PERIOD.name );
		errlogPrintf (
			"Clamping \"%s\" = %f (the hard coded minimum)\n", 
            EPICS_CAS_BEACON_PERIOD.name, 
			CAServerMinBeaconPeriod );
        m_maxBeaconInterval = CAServerMinBeaconPeriod;
    }
}

void Mgr :: BeaconTimer ::
    start ()
{
    m_timer.start ( m_timerNotify, CAServerMinBeaconPeriod );
}

Mgr :: BeaconTimer ::
    ~BeaconTimer ()
{
    m_timer.destroy ();
}

void Mgr :: BeaconTimer ::
    generateBeaconAnomaly ()
{
    m_beaconPeriod = CAServerMinBeaconPeriod;
    m_timer.start ( m_timerNotify, CAServerMinBeaconPeriod );
}

void Mgr :: BeaconTimer :: 
    show ( unsigned level ) const
{
    printf ( "Beacon Timer\n" );
    if ( level > 0 ) {
        printf ("period = %f, max period = %f\n",
            m_beaconPeriod, m_maxBeaconInterval );
        if ( level > 1 ) {
            m_timer.show ( level - 2 );
        }
    }
}

///////////////////////////////////////////
// M_TimerNotify
///////////////////////////////////////////
Mgr :: BeaconTimer :: M_TimerNotify :: M_EnclLnk
    Mgr :: BeaconTimer :: M_TimerNotify ::   
        m_enclLnk ( & Mgr :: BeaconTimer :: m_timerNotify );
epicsTimerNotify :: expireStatus 
    Mgr :: BeaconTimer :: M_TimerNotify ::
        expire ( const epicsTime & /* currentTime */ )	
{
    BeaconTimer & beaconTimer ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( beaconTimer.m_mgr.m_mutex );
    beaconTimer.m_mgr.m_eventQueue.
        post ( PResp ( beaconTimer.m_respHandle ), PCat () );

    // double the period between beacons (but dont exceed max)
    if ( beaconTimer.m_beaconPeriod < 
            beaconTimer.m_maxBeaconInterval ) {
        beaconTimer.m_beaconPeriod += beaconTimer.m_beaconPeriod;
        if ( beaconTimer.m_beaconPeriod >= 
                beaconTimer.m_maxBeaconInterval ) {
	        beaconTimer.m_beaconPeriod = 
                beaconTimer.m_maxBeaconInterval;
        }
    }
    return expireStatus ( restart, beaconTimer.m_beaconPeriod );
}
void Mgr :: BeaconTimer :: M_TimerNotify ::
    show ( unsigned level ) const
{
    const BeaconTimer & beaconTimer ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( beaconTimer.m_mgr.m_mutex );
    beaconTimer.show ( level );
}

/////////////////////////////////////////////////////////
// M_RespHandle
/////////////////////////////////////////////////////////
Mgr :: BeaconTimer :: M_RespHandle :: M_EnclLnk
    Mgr :: BeaconTimer :: M_RespHandle ::   
        m_enclLnk ( & Mgr :: BeaconTimer :: m_respHandle );
Mgr :: BeaconTimer :: M_RespHandle :: TargetMutexPair 
    Mgr :: BeaconTimer :: M_RespHandle :: m_targetMutexPair ()
{
    BeaconTimer & beaconTimer ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( beaconTimer.m_rspThrBeacon, 
                            beaconTimer.m_mgr.m_noopMutex );
}
HandleIntf < Responder > & 
    Mgr :: BeaconTimer :: M_RespHandle :: m_clone ()
{
    return *this;
}
void Mgr :: BeaconTimer :: M_RespHandle :: m_release ()
{
    // lifespan is determined instead by Mgr
}

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca
