
/*
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include "caerr.h"
#include "osiWireFormat.h"
#include "errlog.h"

#define epicsExportSharedSymbols
#define caNetAddrSock
#include "caNetAddr.h"
#include "caSvrDGMgr.h"
#include "caSvrDGRcvThr.h"
#include "osiPoolStatus.h" // osi pool monitoring functions
#include "caServer.h"
#include "protoDiagnostics.h"
#include "IterIn.h"
#include "caChannelAddress.h"
#include "caChannelDescription.h"
#include "caStringReaderIterIn.h"

#ifndef NELEMENTS
#   define NELEMENTS(array) (sizeof(array)/sizeof((array)[0]))
#endif

namespace ca {
namespace Svr {
namespace DG {

RcvThr :: MsgHandler 
const RcvThr :: m_msgHandlers[] =
{
    & RcvThr :: m_versionAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,

    & RcvThr :: m_uknownMessageAction, 
    & RcvThr :: m_searchAction, 
    & RcvThr :: m_uknownMessageAction, 
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,

    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,

    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,

    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_echoAction,
    & RcvThr :: m_uknownMessageAction,

    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction,
    & RcvThr :: m_uknownMessageAction
};

Ptr < RcvThr > RcvThr :: 
    factory ( const PService & pService, 
    const PCat & pNameResChanDescr, 
    const PMgr & pMgr, InPtr & inPtr)
{
    RcvThr * pRcvThr = new RcvThr ( pService, 
                pNameResChanDescr, pMgr, inPtr );
    return Ptr < RcvThr > ( pRcvThr->m_handle );
}

RcvThr :: RcvThr ( const PService & pService, const PCat & pNameResChanDescr, 
            const PMgr & pMgr, InPtr & pIn ) :
    m_thread (
        m_threadIntf, "CAS-DG-Rcv", 
        epicsThreadGetStackSize ( epicsThreadStackBig ), 
        epicsThreadPriorityCAServerLow ),
    m_pIn ( pIn ),
    m_pChanNameRes ( m_pService->createChannel ( pNameResChanDescr ) ),
    m_pService ( pService ),
    m_pMgr ( pMgr ),
    m_pChannelAddressType ( m_pService->createType ( ChannelAddress () ) ),
    m_pChannelDescriptionType ( m_pService->createType ( ChannelDescription () ) ),
    m_threadShutdownCmd ( false )
{
    if ( m_pChannelAddressType.isNill () ) {
        throw std :: runtime_error ( 
            "RspThr ctor: Service didnt provide "
            "a channel address type\n" );
    }
    if ( m_pChannelDescriptionType.isNill () ) {
        throw std :: runtime_error ( 
            "RspThr ctor: Service didnt provide "
            "a channel description type\n" );
    }
	memset ( &m_curMsg, '\0', sizeof ( m_curMsg ) );
    m_thread.start ();
}

RcvThr :: ~RcvThr ()
{
    m_threadShutdownCmd = true;
    m_pIn->interruptLoadFrame ();
	epics :: Mutex :: Unlocker ulocker ( m_mutex );
    m_thread.exitWait ();
}

void RcvThr :: start ()
{
    m_thread.start ();
}

void RcvThr :: show ( unsigned level ) const
{
    printf ( "RcvThr\n" );
    if ( level > 0 ) {
		m_thread.show ( level - 1 );
        printf ( "RcvThr: m_lastRecvTS\n" );
	    m_lastRecvTS.show ( level - 1 );
        m_pIn->show ( level - 1);
        printf ( "RcvThr: m_threadShutdownCmd = %c\n",
            m_threadShutdownCmd ? 'T' : 'F' );
    }
}

//
// RcvThr
//
RcvThr :: M_ThreadIntf :: M_EnclLnk
    RcvThr :: M_ThreadIntf ::   
        m_enclLnk ( & RcvThr :: m_threadIntf );
void RcvThr :: M_ThreadIntf :: run ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( rcvThr.m_mutex );
    while ( ! rcvThr.m_threadShutdownCmd ) {
        NetAddr srcAddr;
        NetAddr intfAddr;
        // fetch the next datagram, expecting to block when 
        // input queue is empty
		{
			Guard :: Release release ( guard );
			rcvThr.m_pIn->loadFrame ( srcAddr, intfAddr );
		}
        DatagramInReader datagramInReader ( *rcvThr.m_pIn );
        IterIn iterIn ( datagramInReader );
        rcvThr.m_curDGParm = DGParm ( srcAddr, intfAddr );
        try {
            rcvThr.m_processDG ( iterIn );
        }
        catch ( exception & except ) {
            errlogPrintf (
                "C++ exception when processing "
                "datagram was \"%s\"\n",
                except.what () );
        }
        catch ( ... ) {
            errlogPrintf (
                "non-standard C++ exception when "
                "processing datagram\n" );
        }
    }
}

//
// RcvThr
//
RcvThr :: M_Handle :: M_EnclLnk
    RcvThr :: M_Handle ::   
        m_enclLnk ( & RcvThr :: m_handle );
RcvThr :: M_Handle :: TargetMutexPair 
    RcvThr :: M_Handle :: m_targetMutexPair ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( rcvThr, rcvThr.m_mutex );
}
HandleIntf < RcvThr > & 
    RcvThr :: M_Handle :: m_clone ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    rcvThr.m_refCtr.reference ();
    return *this;
}
void RcvThr :: M_Handle :: m_release ()
{
    RcvThr & rcvThr ( m_enclLnk.enclosure ( *this ) );
    if ( rcvThr.m_refCtr.unreference () ) {
        delete & rcvThr;
        // no code after here considering above delete
    }
    // no code after here considering above delete
}

#if 0
//
// Mgr::processMsg ()
// process any messages in the in buffer
//
void Mgr :: processMsg ( Guard & guard )
{
    guard.assertIdenticalMutex ( m_mutex );
    try {
	    while ( unsigned bytesLeft = m_in.bytesPresent ( guard ) ) {
            //
            // copy as raw bytes in order to avoid alignment problems
            //
            caHdr smallHdr;
            if ( bytesLeft < sizeof ( smallHdr ) ) {
                break;
            }

            char * rawMP = m_in.msgPtr ( guard );
            memcpy ( & smallHdr, rawMP, sizeof ( smallHdr ) );

            ca_uint32_t payloadSize = 
                AlignedWireRef < epicsUInt16 > ( smallHdr.m_postsize );
            ca_uint32_t nElem = 
                AlignedWireRef < epicsUInt16 > ( smallHdr.m_count );

            ca_uint32_t hdrSize;
            if ( payloadSize != 0xffff && nElem != 0xffff ) {
                hdrSize = sizeof ( smallHdr );
            }
            else {
                ca_uint32_t LWA[2];
                hdrSize = sizeof ( smallHdr ) + sizeof ( LWA );
                if ( bytesLeft < hdrSize ) {
                    break;
                }
                //
                // copy as raw bytes in order to avoid
                // alignment problems
                //
                memcpy ( LWA, rawMP + sizeof ( caHdr ), sizeof( LWA ) );
                payloadSize = AlignedWireRef < epicsUInt32 > ( LWA[0] );
                nElem = AlignedWireRef < epicsUInt32 > ( LWA[1] );
            }

            HdrLargeArray msgTmp;
            msgTmp.m_cmmd = 
                AlignedWireRef < epicsUInt16 > ( smallHdr.m_cmmd );
            msgTmp.m_postsize = 
                payloadSize;
            msgTmp.m_dataType = 
                AlignedWireRef < epicsUInt16 > ( smallHdr.m_dataType );
            msgTmp.m_count = 
                nElem;
            msgTmp.m_cid = 
                AlignedWireRef < epicsUInt32 > ( smallHdr.m_cid );
            msgTmp.m_available = 
                AlignedWireRef < epicsUInt32 > ( smallHdr.m_available );

            if ( payloadSize & 0x7 ) {
                Server :: dumpMsg ( msgTmp, 
                    "CAS: Datagram request wasn't 8 byte aligned" );
                m_in.removeMsg ( guard, bytesLeft );
                break;
            }

            unsigned msgSize = hdrSize + payloadSize;
            if ( bytesLeft < msgSize ) {
                if ( msgSize > m_in.bufferSize ( guard ) ) {
                    Server :: dumpMsg ( msgTmp, 
                        "client's request didnt fit within the "
                        "CA server's message buffer" );
                    m_in.removeMsg ( guard, bytesLeft );
                }
                break;
            }
            
            m_ctx.setMsg ( msgTmp, rawMP + hdrSize );

	        if ( this->getCAS().getDebugLevel() > 5u ) {
                char pHostName[64u];
                m_lastRecvAddr.stringConvert ( 
                    pHostName, sizeof ( pHostName ) );
		        Server :: dumpMsg ( 
                    pHostName, "?", msgTmp, 0 );
	        }

		    //
		    // Reset the context to the default
		    // (guarantees that previous message does not get mixed 
		    // up with the current message)
		    //
		    m_ctx.setChannel ( NULL );
		    m_ctx.setPV ( NULL );

		    //
		    // Call protocol stub
		    //
            Mgr :: MsgHandler pHandler;
		    if ( msgTmp.m_cmmd < NELEMENTS ( Mgr :: m_msgHandlers ) ) {
                pHandler = m_msgHandlers[msgTmp.m_cmmd];
		    }
            else {
                pHandler = & Mgr :: uknownMessageAction;
            }
		    ( this->*pHandler ) ( guard );

            m_in.removeMsg ( guard, msgSize );
	    }
    }
    catch ( std::exception & except ) {
		Server :: dumpMsg ( m_ctx.getMsg(), 
            "C++ exception \"%s\" in datagram server",
            except.what () );
    }
    catch ( ... ) {
		Server :: dumpMsg ( m_ctx.getMsg(),
            "nonstandard C++ exception in datagram server" );
    }
}
#endif

void RcvThr :: m_processDG ( IterIn & iterIn )
{
    while ( ! iterIn.isEndOfStream () ) {
        m_curMsg.m_cmmd = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_postsize = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_dataType = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_count = iterIn.pop < ca_uint16_t > ();
        m_curMsg.m_cid = iterIn.pop < ca_uint32_t > ();
        m_curMsg.m_available = iterIn.pop < ca_uint32_t > ();
        MsgHandler pHandler = & RcvThr :: m_uknownMessageAction;
	    if ( m_curMsg.m_cmmd < NELEMENTS ( m_msgHandlers ) ) {
            pHandler = m_msgHandlers [ m_curMsg.m_cmmd ];
	    }
        {
            // a special reader that always removes exactly m_postsize 
            // bytes, no more and no less, from the input stream
            IterInReader iterInReader ( iterIn, m_curMsg.m_postsize );
            IterIn iterInPayload ( iterInReader );
	        ( this->*pHandler ) ( iterInPayload );
        }
    }
}

void RcvThr :: m_badMessageAction ( const char * pFormat, va_list & vaList )
{
    errlogDumpMsg ( m_curDGParm.srcAddr (), m_curMsg, pFormat, vaList );
    throw InvalidProtocol ( "Unsupported request - or corrupt protocol" );
}

void RcvThr :: m_badMessageAction ( const char * pFormat, ... )
{
    va_list theArgs;
    va_start ( theArgs, pFormat );
    m_badMessageAction ( pFormat, theArgs );
    va_end ( theArgs );
}

void RcvThr :: m_uknownMessageAction ( IterIn & )
{
    m_badMessageAction (  
        "bad request code in datagram "
        "= %u\n", m_curMsg.m_cmmd );
}

void RcvThr :: m_versionAction ( IterIn & )
{
    ca_uint16_t protoRev = 
        static_cast < ca_uint16_t > ( 
            m_curMsg.m_count & 0xffff );
    ca_uint32_t seqNum = m_curMsg.m_cid;
    if ( ! CA_V411 ( protoRev ) ) {
        seqNum = 0;
    }
    m_curDGParm.setHdrParms ( seqNum, protoRev );
}

//
// echoAction()
//
void RcvThr :: m_echoAction ( IterIn & )
{
#if 0
    RcvThr :: Echo * pEchoResp = 
        new ( m_dgSvr.m_echoResReqFreeList )
            RcvThr :: Echo ( 
                guard, m_dgSvr.m_respCtx, 
                    m_curDGParm, m_curMsg, m_curPayload );
#endif
}

void RcvThr :: m_searchAction ( IterIn & iterIn )
{
	//
	// verify that we have sufficent memory for a PV and a
	// monitor prior to calling PV exist test so that when
	// the server runs out of memory we dont reply to
	// search requests, and therefore dont thrash through
	// caServer::pvExistTest() and casCreatePV::pvAttach()
	//
    if ( ! osiSufficentSpaceInPool ( 0 ) ) {
        return;
    }

    //
    // starting with V4.1 the count field is used (abused)
    // by the client to store the minor version number of 
    // the client.
    //
    // Old versions expect alloc of channel in response
    // to a search request. This is no longer supported.
    //
    if ( !CA_V44( m_curDGParm.protocolRevision () ) ) {
        //
        // old connect protocol was dropped when the
        // new API was added to the server (they must
        // now use clients at EPICS 3.12 or higher)
        //
        char pHostName[64u];
        m_curDGParm.hostName ( pHostName, sizeof ( pHostName ) );
        errlogPrintf ( 
            "client \"%s\" using EPICS R3.11 CA "
            "connect protocol was ignored\n", pHostName );
        return;
    }

    StringReaderIterIn 
        stringReaderIterIn ( iterIn, m_curMsg.m_postsize );
    if ( stringReaderIterIn.size () == 0u ) {
        throw InvalidProtocol ( 
			"Channel name resolve request msg didnt specify "
            "the channel's name" );
    }

    //
    // check the sanity of the message
    //
    if ( m_curMsg.m_postsize <= 1 ) {
        char pHostName[64u];
        m_curDGParm.hostName ( pHostName, sizeof ( pHostName ) );
        m_badMessageAction ( 
			"Name resolution request from \"%s\" didnt specify "
            "the channel's name\n", pHostName );
        return;
    }

    {
        // this organization avoids overlapping lock domains
        epics :: Mutex :: Unlocker unlocker ( m_mutex );
        PReq pReq = m_pChanNameRes->createRequest ( 
                        Request :: resolveName, 
                        m_pChannelDescriptionType,
                        m_pChannelAddressType );
#pragma message ( "failRespRequired is hardwaired" )
        Mgr :: InitiateParms initiateParms = 
            m_pMgr->createNameResResp ( pReq, m_curDGParm, 
                    m_curMsg.m_cid, false, stringReaderIterIn );
        pReq->initiate ( initiateParms.pReqParms, 
                        initiateParms.pResp );
    }
#if 0
    if ( pChanName[0] == '\0' ) {
        char pHostName[64u];
        m_lastRecvAddr.stringConvert ( pHostName, sizeof ( pHostName ) );
	    Server::dumpMsg ( pHostName, "?", msg, 
            "zero length PV name in UDP search request?\n" );
        return;
    }

    // check for an unterminated string before calling 
    // server tool by searching backwards through the 
    // string (some early versions of the client library 
    // might not be setting the pad bytes to nill)
    for ( unsigned i = msg.m_postsize-1; 
        pChanName[i] != '\0'; i-- ) {
        if ( i <= 1 ) {
            char pHostName[64u];
            m_lastRecvAddr.stringConvert ( 
                pHostName, sizeof ( pHostName ) );
	        Server::dumpMsg ( pHostName, "?", msg,  
                "unterminated PV name in UDP search request?\n" );
            return;
        }
    }
	if ( this->getCAS().getDebugLevel() > 6u ) {
		char pHostName[64u];
		this->hostName ( pHostName, sizeof ( pHostName ) );
		printf ( "\"%s\" is searching for \"%s\"\n", 
            pHostName, pChanName );
	}

#pragma message ("must ignore such requests if msg was sent to broadcast address?")
#pragma message ("appears not to be using the supplied channel name (incoming proto)?")
    bool failRespRequired = ( m_curMsg.m_dataType == DOREPLY );
    new ( m_dgSvr.m_nameResReqFreeList )
        RespCtx :: NameRes ( guard, m_curDGParm, m_curMsg.m_cid, 
                        failRespRequired, m_dgSvr.m_respCtx );
#endif
}

Vec < const epicsUInt8 > RcvThr :: DatagramInReader :: 
    read ( size_t nBytesMax )
{
    return m_datagramIn.readFrameSeg ( nBytesMax );
}

void RcvThr :: DatagramInReader :: 
    rewind ()
{
    throw std :: runtime_error ( 
        "DatagramInReader commanded to rewind, "
        "but this isnt supported" );
}

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca

