/*
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include "epicsExit.h"

#define epicsExportSharedSymbols
#include "caSvrDGMgr.h"
#include "caSvrDGRspThr.h"
#include "caSvrDGRspThrDGCoalescer.h"
#include "caSvrDGRspThrEcho.h"
#include "caSvrDGRspThrNameRes.h"
#include "caChannelAddress.h"
#include "caChannelDescription.h"

namespace ca {
namespace Svr {
namespace DG {

static epicsThreadPrivateId rspThrContextId = 0;

extern "C" void m_rspThrExitHandlerDG ( void * )
{
    if ( rspThrContextId ) {
        epicsThreadPrivateDelete ( rspThrContextId );
        rspThrContextId = 0;
    }
}

extern "C" void m_rspThrInitOnceDG ( void * )
{
    rspThrContextId = epicsThreadPrivateCreate ();
    if ( rspThrContextId ) {
        epicsAtExit ( m_rspThrExitHandlerDG, 0 );
    }
}

Ptr < RspThr > RspThr :: factory ( 
    const PMgr & pMgr, 
    epicsTimerQueue & timerQueue, 
    std :: auto_ptr < DatagramOut > pDatagramOut, 
    const Ptr < VecAllocator > & pVecAllocator )
{
    RspThr * pRspThr = new 
        RspThr ( pMgr, timerQueue, pDatagramOut, pVecAllocator );
    return Ptr < RspThr > ( pRspThr->m_handle );
}

RspThr :: RspThr (  const PMgr & pMgr, epicsTimerQueue & timerQueue,
                    std :: auto_ptr < DatagramOut > pDatagramOut, 
                    const Ptr < VecAllocator > & pVecAllocator ) :
    m_thread (
        m_entryPt, "CAS-DG-Rsp", 
        epicsThreadGetStackSize ( epicsThreadStackBig ), 
        epicsThreadPriorityCAServerLow ),
    m_pDGCoalescer ( new DGCoalescer ( *this, pVecAllocator ) ),
    m_pDatagramOut ( pDatagramOut ),
    m_pMgr ( pMgr ),
    m_threadShutdownCmd ( false )
{
    static epicsThreadOnceId once = EPICS_THREAD_ONCE_INIT;
    epicsThreadOnce ( &once, m_rspThrInitOnceDG, 0 );
    if ( ! rspThrContextId ) {
        throw std :: runtime_error ( 
            "RspThr ctor, unable to allocate "
            "rspThrCtx :: rspThrContextId" );
    }
    m_thread.start ();
}

RspThr :: ~RspThr ()
{
#pragma message ("thread shutdown code missing here")
}

//
// M_EntryPt
//
RspThr :: M_EntryPt :: M_EnclLnk
    RspThr :: M_EntryPt ::   
        m_enclLnk ( & RspThr :: m_entryPt );
void RspThr :: M_EntryPt :: run ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    epicsThreadPrivateSet ( rspThrContextId, & rspThr );
    Guard guard ( rspThr.m_mutex );
    try {
        while ( ! rspThr.m_threadShutdownCmd ) {
            rspThr.m_pMgr->processEvents ();
            {
                Guard :: Release release ( guard );
#pragma message ( "consider carefully if the lock should be applied when we call flush")
#pragma message ( "do we release inside of flush instead?")
                rspThr.m_pDGCoalescer->flush ();
                rspThr.m_event.wait ();
            }
        }   
    }
    catch ( std :: exception & ) {
#pragma message ("missing code")
    }
    catch ( ... ) {
#pragma message ("missing code")
    }
}
void RspThr :: M_EntryPt :: show ( unsigned level ) const
{
    const RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    Guard guard ( rspThr.m_mutex );
    rspThr.show ( level );
}

void RspThr :: show ( unsigned level ) const
{
    printf ( "RspThr ::\n" );
    if ( level > 0 ) {
        m_mutex.show ( level -1 );
        m_pDGCoalescer->show ( level -1 );
        printf ( "RspThr :: m_lastSendTS\n" );
        m_lastSendTS.show ( level - 1 );
        if ( m_pDatagramOut.get () ) {
            m_pDatagramOut->show ( level - 1 );
        }
    }
}

RspThr * RspThr :: pAttachedRspThr ()
{
    // we make the assumption that a RspThr was created
    // before any call to this function occurs.
    if ( rspThrContextId == 0) {
        throw std :: runtime_error ( 
            "response thread context objects are being "
            "used before the response thread has initialized?" );
    }
    RspThr * pRspThr = reinterpret_cast < RspThr * > ( 
        epicsThreadPrivateGet ( rspThrContextId ) );
    if ( pRspThr == 0 ) {
        throw std :: runtime_error ( 
            "response thread context objects are being "
            "used with some other thread?" );
    }
    return pRspThr;
}

/////////////////////////////////////////
// M_Handle - handle for RspThr object
////////////////////////////////////////
RspThr :: M_Handle :: M_EnclLnk
    RspThr :: M_Handle ::   
        m_enclLnk ( & RspThr :: m_handle );
RspThr :: M_Handle :: TargetMutexPair 
    RspThr :: M_Handle :: m_targetMutexPair ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( rspThr, rspThr.m_mutex );
}
HandleIntf < RspThr > & 
    RspThr :: M_Handle :: m_clone ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    rspThr.m_refCtr.reference ();
    return *this;
}
void RspThr :: M_Handle :: m_release ()
{
    RspThr & rspThr ( m_enclLnk.enclosure ( *this ) );
    if ( rspThr.m_refCtr.unreference () ) {
        delete & rspThr;
        // no code here due to above delete
    }
    // no code here due to above delete
}

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca

