
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

#define epicsExportSharedSymbols
#include "caSvrDGRspThrNameRes.h"
#include "caSvrDGRspThrDGCoalescer.h"
#include "FailureContext.h"
#include "osiSock.h"
#define NetaddrSock 
#include "caNetAddr.h"

namespace ca {
namespace Svr {
namespace DG {

RspThr :: NameRes :: 
    NameRes ( const DGParm & dgParm, caResId cid, 
            bool failRespRequired ) :
    m_dgParm ( dgParm ), m_cid ( cid ), 
        m_failRespRequired ( failRespRequired )
{
}

inline RspThr :: NameRes :: ResponseAgentSuccess ::
    ResponseAgentSuccess ( NameRes & nameRes, 
            const Catalog & respParms ) :
    m_nameRes ( nameRes ), m_respParm ( respParms )
{
}

void RspThr :: NameRes :: ResponseAgentSuccess ::
    assignDefault ()
{
    // ok, success indication wasnt found so we 
    // probe now for failure indication
    ResponseAgentFailure responseAgentFailure ( m_nameRes );
    get ( m_respParm, Responder :: failure, responseAgentFailure );
}

void RspThr :: NameRes :: ResponseAgentSuccess ::
    assignScalarCatalog ( const Catalog & successParms )
{
    //
    // cid field is abused to carry the IP 
    // address in CA_V48 or higher
    // (this allows a CA servers to serve
    // as a directory service)
    //
    // data type field is abused to carry the IP 
    // port number here CA_V44 or higher
    // (this allows multiple CA servers on one
    // host)
    //
    // this server does not respond to name
    // resolution requests prior to CA_V44
    // (this is screened when processing the 
    // request)
    //
    ca_uint32_t serverAddr = ntohl ( INADDR_BROADCAST );
    ca_uint16_t serverPort = CA_SERVER_PORT;
    try {
        if ( CA_V48( m_nameRes.m_dgParm.protocolRevision () ) ) {  
            const NetAddr intfAddrDflt = m_nameRes.m_dgParm.intfAddr ();
            NetAddr intfAddr = 
                fetchNetworkAddress ( successParms, & intfAddrDflt );
            if ( intfAddr.isInet () ) {
                sockaddr_in intfAddrIN = intfAddr.getSockIP ();
                if ( 
                    intfAddrIN.sin_addr.s_addr != INADDR_ANY && 
                    intfAddrIN.sin_addr.s_addr != INADDR_BROADCAST &&
                    intfAddrIN.sin_port != 0u ) {
                    serverAddr = htonl ( intfAddrIN.sin_addr.s_addr ); 
                    serverPort = htons ( intfAddrIN.sin_port );
                }
            }
        }
        caHdr respMsg;
        respMsg.m_cmmd = CA_PROTO_SEARCH;
        respMsg.m_postsize = sizeof ( ca_uint16_t );
        respMsg.m_dataType = serverPort;
        respMsg.m_count = 0;
        respMsg.m_cid = serverAddr;
        respMsg.m_available = m_nameRes.m_cid;
        RspThr * pRspThr = RspThr :: pAttachedRspThr ();
        pRspThr->m_pDGCoalescer->pushHeaderAndRev ( m_nameRes.m_dgParm, respMsg );
    }
    catch ( exception & except ) {
        std :: string ctx = typeid ( except ).name ();
        ctx += except.what ();
        FailureContext failureContext ( ctx.c_str () );
        m_nameRes.m_failResponse ( failureContext );
    }
    catch ( ... ) {
        FailureContext failureContext ( "nonstandard c++ exception" );
        m_nameRes.m_failResponse ( failureContext );
    }
}

inline RspThr :: NameRes :: ResponseAgentFailure ::
    ResponseAgentFailure ( NameRes & nameRes ) :
    m_nameRes ( nameRes )
{
}

void RspThr :: NameRes :: ResponseAgentFailure ::
    assignDefault ()
{
    FailureContext failureContext ( 
        "service name resolution response payload was "
        "success/failure ambiguous" );
    m_nameRes.m_failResponse ( failureContext );
}

void RspThr :: NameRes :: ResponseAgentFailure ::
    assignScalarCatalog ( const Catalog & failureParms )
{
    m_nameRes.m_failResponse ( failureParms );
}

void RspThr :: NameRes :: 
    m_responseNotify ( const PCat & pRespParms )
{
    // This tells the dtor that the m_responseNotify
    // from the Request has arrived so the request 
    // doesnt need to be canceled
    if ( pRespParms.isntNill () ) {
        PCat :: Locker pRespParmsLocked = pRespParms;
        ResponseAgentSuccess 
            responseAgentSuccess ( *this, *pRespParmsLocked ); 
        get ( *pRespParmsLocked, success, responseAgentSuccess );
    }
    else {
        FailureContext failureContext ( 
            "received service read response notification, "
            "but success/failure indication wasnt in "
            "response (empty) payload" );
        m_failResponse ( failureContext );
    }
}

void RspThr :: NameRes :: 
    m_failResponse ( const Catalog & )
{
    if ( m_failRespRequired ) {
        caHdr respMsg;
        respMsg.m_cmmd = CA_PROTO_NOT_FOUND;
        respMsg.m_postsize = 0;
        respMsg.m_dataType = 0;
        respMsg.m_count = 0;
        respMsg.m_cid = 0;
        respMsg.m_available = m_cid;
        RspThr * pRspThr = RspThr :: pAttachedRspThr ();
        pRspThr->m_pDGCoalescer->pushHeader ( m_dgParm, respMsg );
    }
}

void RspThr :: NameRes ::  
    m_show ( unsigned level ) const
{
    printf ( "RspThr :: NameRes\n" );
    if ( level > 0 ) {
        m_dgParm.show ( level - 1 );
        printf ( "m_cid = %u\n", m_cid );
        printf ( "m_failRespRequired = %c\n", 
            m_failRespRequired ? 'T' : 'F' );
    }
}

} // end of namespace DG
} // end of namespace Svr
} // end of namespace ca
