
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

// Author: Jeff HIll

#include "errlog.h"

#define epicsExportSharedSymbols
#include "caSvrEvQueSubscr.h"
#include "caSvrEvQueFailResp.h"

namespace ca {
namespace Svr {

EvQue :: EvQue ( Mutex & mutex, NewEventSignal & newEventSignal ) :
    m_mutex ( mutex ),
    m_noopMutex ( Mutex :: factoryNOOP ),
    m_newEventSignal ( newEventSignal ),
    m_nEventsAvailable ( 0 )
{
    //m_thread.start ();
}

EvQue :: ~EvQue ()
{
    while ( EntryBlock * pEntryBlock = m_entryBlockList.get () ) {
        m_entryBlockFreeList.deleteProduct ( *pEntryBlock );
    }
}

void EvQue :: processEvents ()
{
    while ( Subscr * pSubscr = m_listSubscrOVF.first () ) {
        pSubscr->postOVF ();
    }
    tsDLList < EntryBlock > pending;
    pending.add ( m_entryBlockList );
    tsDLList < EntryBlock > completed;
    {
        // not holding the lock when calling callbacks 
        // prevents unexpected deadlocks
        Mutex :: Unlocker unlocker ( m_mutex );
        while ( EntryBlock * pEntryBlock = pending.get () ) {
            pEntryBlock->action ();
            completed.add ( *pEntryBlock );
        }
    }
    m_destroyEntryBlocks ( completed );
}

void EvQue :: m_replRespIfMatch ( const PResp & pMatch, PResp & pReplace )
{
    tsDLIter < EntryBlock > pEntryBlock = 
        m_entryBlockList.firstIter ();
    while ( pEntryBlock.valid () ) {
        pEntryBlock->replRespIfMatch ( pMatch, pReplace );
        pEntryBlock++;
    }
}

void EvQue :: show ( unsigned level ) const
{
    printf ( "EvQue:\n" );
    if ( level > 0 ) {
        printf ( "EvQue: m_entryBlockList\n" );
        m_entryBlockList.show ( level - 1 );
        printf ( "EvQue: m_entryBlockFreeList\n" );
        m_entryBlockFreeList.show ( level - 1 );
        printf ( "EvQue: m_failRespFreeList\n" );
        m_failRespFreeList.show ( level - 1 );
        printf ( "EvQue: m_mutex\n" );
        m_mutex.show ( level - 1 );
        printf ( "EvQue: m_nEventsAvailable = %u\n",
            m_nEventsAvailable );
    }
}

void EvQue :: m_destroyEntryBlocks ( 
                    tsDLList < EntryBlock > & entryBlockList )
{
    while ( EntryBlock * pEntryBlock = entryBlockList.get () ) {
        m_entryBlockFreeList.deleteProduct ( *pEntryBlock );
    }
}

// Each entity that uses the event queue _MUST_
// reserve, in advance, the maximum numer of events 
// that it will have pending simultaneously. Posting 
// more simultaneous events than were reserved will 
// create potential for any event posting agent
// to randomly receive an overflow exception. 
// The reserve () function will throw an exception 
// in out-of-memory situations.
//
#pragma message ("decide about the issue described in comment below")
// If this fails we have no mechanism to report 
// the problem to the user via a response on the 
// event queue. We have no choice (by design) 
// other than to let a resulting exception 
// cause a client disconnect.
//
// An alternative to throwing an exception here 
// might be to use nothrow new and just block the 
// receive threadif we cant get the memory we need.
//
void EvQue :: reserve ( size_t nEvents )
{
    if ( nEvents <= m_nEventsAvailable ) {
        // the typical path going through here
        // is super efficent
        m_nEventsAvailable -= nEvents;
    }
    else {
        size_t nEventsAllocated = 0u;
        size_t nEventsNeeded = nEvents - m_nEventsAvailable;
        tsDLList < EntryBlock > entryBlockListTmp;
        try {
            while ( nEventsAllocated < nEventsNeeded ) {
                EntryBlock * pEntryBlock = 
                    new ( m_entryBlockFreeList ) 
                        EntryBlock ();
                entryBlockListTmp.add ( *pEntryBlock );
                nEventsAllocated += 
                    EntryBlock :: numberOfSlots ();
            }
        }
        catch ( ... ) {
            m_destroyEntryBlocks ( entryBlockListTmp );
            m_nEventsAvailable += nEventsAllocated;
            throw;
        }
        m_destroyEntryBlocks ( entryBlockListTmp );
        m_nEventsAvailable = 
            nEventsAllocated - nEventsNeeded;
    }
}

// release shouldnt throw an exception because it is
// called by destructors 
void EvQue :: release ( size_t nEvents )
{
    if ( m_nEventsAvailable <= UINT_MAX - nEvents ) {
        m_nEventsAvailable += nEvents;
    }
    else {
        m_nEventsAvailable = UINT_MAX;
        errlogPrintf ( 
            "EvQue :: m_release (): "
            "m_nEventsAvailable overflow detected\n" );
    }
}

void EvQue :: post ( const PResp & pResp, const PCat & pCat )
{
    EntryBlock * pLast = m_entryBlockList.last ();
    bool success = pLast->post ( pResp, pCat );
    if ( ! success ) {
        // this new should never throw an exception by design
        // because we call m_reserve ahead of time to guarantee 
        // adaquate blocks in reserve on the free list 
        EntryBlock * pEntryBlock = 
            new ( m_entryBlockFreeList )
                EntryBlock ();
        success = pEntryBlock->post ( pResp, pCat );
        assert ( success );
        m_entryBlockList.add ( *pEntryBlock );
    }
}

void EvQue :: post ( const PResp & pResp )
{
    this->post ( pResp, PCat () );
}

#if 0
void EvQue :: m_failResponse ( PResp & pResp, const char * pContext )
{
    FailResp * pFailResp = new ( m_failRespFreeList )
                                    FailResp ( *this, pContext );
    PCat pCat = pFailResp->createPtr ();
    pResp->responseNotify ( pCat );
}
#endif

void EvQue :: postFailure ( const PResp & pResp, 
                    const std :: string & failCtx )
{
    FailResp * pFailResp = new ( m_failRespFreeList ) 
                            FailResp ( *this, failCtx.c_str () );
    this->post ( pResp, pFailResp->createPtr () );
}

} // end of namespace Svr
} // end of name space ca

