
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

// Author: Jeff HIll

#ifndef caSvrEvQueh
#define caSvrEvQueh

#include <string>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caSvrEvQueh
#   undef epicsExportSharedSymbols
#endif

// includes for libraries external to ca server shared library
#include "tsDLList.h"
#include "tsFreeList.h"
#include "epicsMutex.h"

#ifdef epicsExportSharedSymbols_caSvrEvQueh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "casdef2.h"
#include "caRefCtr.h"

namespace ca {
namespace Svr {

typedef epics :: Mutex Mutex;
typedef Mutex :: Guard Guard;

class EvQue {
public:
    class NewEventSignal {
        virtual void signal () = 0;
    };
    class Event;
    class Subscr;
    struct Callback;
    class Reservation;
    EvQue ( Mutex &, NewEventSignal & );
    ~EvQue ();
    void post ( const PResp & );
    void post ( const PResp &, const PCat & );
    void postFailure ( const PResp &, const std :: string & failCtx );
    //void shutdown ();
    void show ( unsigned level ) const;
    void processEvents ();
    void reserve ( size_t nEvents );
    void release ( size_t nEvents );
private:
    struct Notify;
    //class Reservation; // this appears to be unused, currently
    class EntryBlock;
    class FailResp;
    tsDLList < EntryBlock > m_entryBlockList;
    tsFreeList < EntryBlock > m_entryBlockFreeList;
    typedef tsFreeList < FailResp, 16 > FailRespFreeList;
    FailRespFreeList m_failRespFreeList;
    tsDLList < Subscr > m_listSubscrOVF;
    Mutex m_mutex;
    Mutex m_noopMutex;
    size_t m_nEventsAvailable;
    NewEventSignal & m_newEventSignal;
    void m_failResponse ( PResp &, const char * pContext );
    void m_destroyEntryBlocks ( tsDLList < EntryBlock > & );
    void m_replRespIfMatch ( const PResp & pMatch, PResp & pReplace );
	EvQue ( const EvQue & ); // disabled
	EvQue & operator = ( const EvQue & ); // disabled
};

#if 0
struct EvQue :: Notify {
    virtual void action ( PCat & ) = 0;
    // a Guard for the interface mutex is passed
    virtual void release ( const Guard & ) = 0;
protected:
    virtual ~Notify () {} // see destroy mf
};

class EvQue :: Event 
{
public:
    Event ( const Guard &, EvQue & );
    void cleanup ( const Guard & );
    ~Event ();
    void postSuccess ( const Guard &, Callback &, PCat & );
    void postFailure ( const Guard &, Callback &, PCat & );
    void postFailure ( Guard &, Callback &, const char * pCtx );
    bool isPending ( const Guard & ) const;
    void show ( const Guard &, unsigned level ) const; 
private:
    class SuccessNotify :
        public Notify {
    public:
        void release ( const Guard & );
        void action ( Guard &, PCat & );
    };
    class FailNotify :
        public Notify {
    public:
        void release ( const Guard & );
        void action ( Guard &, PCat & );
    };
    SuccessNotify m_successNotify;
    FailNotify m_failNotify;
    Callback * m_pCallback;
    EvQue * m_pEventQueue;
    void m_callBackRelease ( const Guard & );
    Event ( const Event & ); // disabled
    Event & operator = ( const Event & ); // disabled
};

struct EvQue :: Callback {
    // a Guard for the notify mutex is passed
    virtual void success ( 
        Guard & cbGuard, PCat & parm ) = 0;
    // a Guard for the notify mutex is passed
    virtual void fail ( 
        Guard & cbGuard, PCat & parm ) = 0;
    // a Guard for the interface mutex is passed
    // (the user deallocates resoures in this func)
    virtual void release ( 
        const Guard & intfGuard ) = 0;
protected:
    virtual ~Callback () {} // see release mf
};

template < size_t N >
class EvQue :: Reservation {
public:
    Reservation ( EvQue & );
    ~Reservation ();
    size _t size () const;
private:
    EvQue & m_eventQueue;
    Reservation ( const Reservation & ); // disabled
    Reservation & operator = ( const Reservation & ); // disabled
};

#endif

class EvQue :: EntryBlock : 
    public tsDLNode < EntryBlock >
{
public:
    EntryBlock ();
    ~EntryBlock ();
    bool post ( const PResp &, const PCat & );
    void action ();
    //void release ();
    void show ( unsigned level ) const;
    static size_t numberOfSlots ();
    void replRespIfMatch ( const PResp & pMatch, PResp & pReplace );
private:
	struct Entry {
        Entry ( PResp pResponder, PCat pPayload );
		void show ( unsigned level ) const;
		PResp m_pResponder;
		PCat m_pPayload;
	}; 
    static const size_t m_nEntries = 0x100;
    // entries are stored in a raw buf and they are allocated
    // with placement new in order to avoid high ctor overhead
    // for a dedicated array
    char m_entries [ sizeof ( Entry ) * m_nEntries ];
    size_t m_nextEntry;
    EntryBlock ( const EntryBlock & ); // disabled
    EntryBlock & operator = ( const EntryBlock & ); // disabled
}; 


////////////////////////////////////////////////////////////
// Reservation inline functions
////////////////////////////////////////////////////////////
#if 0
template < size_t N >
inline EvQue :: Reservation < N > :: 
    Reservation ( EvQue & eventQueue ) :
    m_eventQueue ( eventQueue )
{
    m_eventQueue.m_reserve ( N );
}

template < size_t N >
inline EvQue :: Reservation < N > :: 
    ~Reservation () 
{
    m_eventQueue.m_release ( N );
}

template < size_t N >
inline size_t EvQue :: Reservation < N > :: 
    size () const
{
    return N;
}
#endif

////////////////////////////////////////////////////////////
// EntryBlock inline functions
////////////////////////////////////////////////////////////

inline EvQue :: EntryBlock ::
    EntryBlock () :
    m_nextEntry ( 0 )
{
}

inline size_t EvQue :: EntryBlock :: 
    numberOfSlots ()
{
    return m_nEntries;
}

////////////////////////////////////////////////////////////
// Entry inline functions
////////////////////////////////////////////////////////////

inline EvQue :: EntryBlock :: Entry :: 
    Entry ( PResp pResponder, PCat pPayload ) :
    m_pResponder ( pResponder ), m_pPayload ( pPayload )
{
}

#if 0
inline void EvQue :: Entry ::
    action ()
{
    if ( m_pResponder.isntNill () ) {
        m_pResponder->responseNotify ( m_pPayload );
        // we prefer to call deallocators now w/o holding the 
        // event queue's lock (deadlock avoidance)
        m_pPayload.clear (); 
        m_pResponder.clear ();
    }
}

inline void EvQue :: Entry ::
    release ()
{
    m_pPayload.clear (); 
    m_pResponder.clear ();
}

void EvQue :: Entry :: 
    replRespIfMatch ( const PResp & pMatch, PResp & pReplace )
{
    m_pResponder.replaceIfMatch ( pMatch, pReplace );
}
#endif

////////////////////////////////////////////////////////////
// Event inline functions
////////////////////////////////////////////////////////////
#if 0
inline bool EvQue :: Event :: 
    isPending ( const Guard & ) const
{
    return m_pCallback != 0;
}
#endif

} // end of namesapce Svr
} // end of namespace ca

#endif // ifndef caSvrEvQueh

