
/*
 *      caSvrEvQueSubscr.cpp,v 1.7 2009/12/03 17:39:51 jhill Exp
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include "errlog.h"

#define epicsExportSharedSymbols
#include "caSvrEvQueSubscr.h"

namespace ca {
namespace Svr {

EvQue :: Subscr :: 
    Subscr ( EvQue & eventQueue, const PResp pResp, const PCat & /* reqParms */ ) :
    m_pResp ( pResp ),
    m_eventQueue ( eventQueue ), 
    m_flowControlIsActive ( false )
{
#pragma message ("if this fails we could try one entry and keep a dynamic maximum")
    m_eventQueue.reserve ( m_maxEventQueueEntires );
}

EvQue :: Subscr :: ~Subscr ()
{
    m_eventQueue.release ( m_maxEventQueueEntires );
    // verify that we are not still on the event queue's
    // subscription overflow list
    assert ( m_pPayloadOVF.isNill () );
}

void EvQue :: Subscr :: 
    show ( unsigned level ) const
{
    printf ( "EvQue :: Subscr\n" );
    if ( level > 0 ) {
        printf ( "EvQue :: Subscr: m_refCtr\n" );
        m_refCtr.show ( level - 1u );
        if ( m_pResp.isntNill () ) {
            printf ( "EvQue :: Subscr: m_pResp\n" );
            m_pResp->show ( level - 1 );
        }
        printf ( "m_flowControlIsActive = %c\n", 
            m_flowControlIsActive ? 'T' : 'F' );
    }
}

void EvQue :: Subscr :: 
    cancel ()
{
    // purge references directly to this object from the event queue
    m_eventQueue.m_replRespIfMatch ( PResp ( m_handleResp ), m_pResp );
    m_pResp.clear ();
}

void EvQue :: Subscr :: 
    responseNotify ( const PCat & pPayload )
{
    if ( m_pResp.isNill () ) {
        return;
    }
    const size_t nLeft = m_remainingEntries ();
    if ( m_pPayloadOVF.isNill () ) {
        // leave space in the reservation for one overflow 
        // event (which is posted by the event queue 
        // processing thread)
        if ( m_flowControlIsActive || nLeft <= 1u ) {
            m_pPayloadOVF = pPayload;
            m_eventQueue.m_listSubscrOVF.add ( *this );
        }
        else {
            m_eventQueue.post ( PResp ( m_handleResp ), pPayload );
        }
    }
    else {
        // leave space in the reservation for one overflow 
        // event (which is posted by the event queue 
        // processing thread)
        if ( m_flowControlIsActive || nLeft <= 1u ) {
            m_pPayloadOVF = pPayload;
        }
        // leave space in the reservation for one overflow 
        // event (which is posted by the event queue 
        // processing thread)
        else if ( nLeft > 2u ) {
            m_eventQueue.post ( PResp ( m_handleResp ), m_pPayloadOVF );
            m_eventQueue.post ( PResp ( m_handleResp ), pPayload );
            m_pPayloadOVF.clear ();
            m_eventQueue.m_listSubscrOVF.remove ( *this );
        }
        // else there are two left
        else {
            m_eventQueue.post ( PResp ( m_handleResp ), pPayload );
            m_pPayloadOVF.clear ();
            m_eventQueue.m_listSubscrOVF.remove ( *this );
        }
    }
}

void EvQue :: Subscr :: 
    postOVF ()
{
    assert ( m_pPayloadOVF.isntNill () );
    m_eventQueue.post ( PResp ( m_handleResp ), m_pPayloadOVF );
    m_eventQueue.m_listSubscrOVF.remove ( *this );
    m_pPayloadOVF.clear ();
}

//
// M_Responder
//
EvQue :: Subscr :: M_Responder :: M_EnclLnk
    EvQue :: Subscr :: M_Responder ::  
        m_enclLnk ( & Subscr :: m_responder );
void EvQue :: Subscr :: M_Responder ::
    m_responseNotify ( const PCat & pPayload )
{
    // this pass through Responder allows the Subscr
    // to track event queue references efficently using
    // its m_refCtr member
    Subscr & subscr ( m_enclLnk.enclosure ( *this ) );
    assert ( subscr.m_pResp.isntNill () );
    subscr.m_pResp->responseNotify ( pPayload );
}
void EvQue :: Subscr :: M_Responder ::
    m_show ( unsigned level ) const
{
    const Subscr & subscr ( m_enclLnk.enclosure ( *this ) );
    subscr.show ( level );
    if ( level > 0 ) {
        subscr.m_pResp->show ( level - 1u );
    }
}

/////////////////////////////////////////////////////////
// M_HandleResp 
/////////////////////////////////////////////////////////
EvQue :: Subscr :: M_HandleResp :: M_EnclLnk
    EvQue :: Subscr :: M_HandleResp ::  
        m_enclLnk ( & Subscr :: m_handleResp );
// we dont impose a locking overhead here because 
// this is just a pass through for the m_pResp field
EvQue :: Subscr :: M_HandleResp :: TargetMutexPair 
    EvQue :: Subscr :: M_HandleResp ::
        m_targetMutexPair ()
{
    Subscr & subscr ( m_enclLnk.enclosure ( *this ) );
    return TargetMutexPair ( subscr.m_responder, subscr.m_eventQueue.m_noopMutex );
}
HandleIntf < Responder > & 
    EvQue :: Subscr :: M_HandleResp :: 
        m_clone ()
{
    Subscr & subscr ( m_enclLnk.enclosure ( *this ) );
    subscr.m_refCtr.reference ();
    return *this;
}
void EvQue :: Subscr :: M_HandleResp :: 
    m_release ()
{
    Subscr & subscr ( m_enclLnk.enclosure ( *this ) );
    subscr.m_refCtr.unreference ();
}

} // end of namepace Svr
} // end of namespace ca

