
/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

// Author: Jeff HIll

#ifndef caSvrEvQueSubscrh
#define caSvrEvQueSubscrh

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caSvrEvQueSubscrh
#   undef epicsExportSharedSymbols
#endif

#include "EnclLnk.h"

#ifdef epicsExportSharedSymbols_caSvrEvQueSubscrh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caSvrEvQue.h"

namespace ca {
namespace Svr {

class EvQue :: Subscr : 
    public tsDLNode < Subscr >
{
public:
    typedef ca :: Responder Responder;
    Subscr ( EvQue &, const PResp pResp, const PCat & reqParms );
    ~Subscr (); 
    void enterFlowControlMode ();
    void exitFlowControlMode ();
    void show ( unsigned level ) const; 
    void responseNotify ( const PCat & );
    void postOVF ();
    void cancel ();
private:
    class M_Responder :         
        public Responder {
    private:
        void m_responseNotify ( const PCat & );
        void m_show ( unsigned ) const;
        typedef const epics :: EnclLnk < Subscr, 
                                M_Responder > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    class M_HandleResp : 
        public HandleIntf < Responder > {
    private:
		TargetMutexPair m_targetMutexPair ();
		HandleIntf < Responder > & m_clone ();
		void m_release ();
        typedef const epics :: EnclLnk < Subscr, 
                                M_HandleResp > M_EnclLnk;
        static M_EnclLnk m_enclLnk;
    };
    M_Responder m_responder;
    M_HandleResp m_handleResp;
    // This is typically a reference to the RspThr :: Subscribe
    // which is posted to the event queue.
    PResp m_pResp;
    PCat m_pPayloadOVF;
    EvQue & m_eventQueue;
    // this ends up being a count of the number of event
    // references to this object on the event queue
    RefCtr m_refCtr;
    bool m_flowControlIsActive;
    static const size_t m_maxEventQueueEntires = 16u;
    size_t m_remainingEntries () const;
    Subscr ( const Subscr & ); // not impl
    Subscr & operator = ( const Subscr & ); // not impl

    //
    // strategy for dealing with subscription update production
    // rate exceeds subscription update consumption rate
    // ------------------------------------------------------
    // If an update is posted and the maximum number of pending 
    // event entries for a subscription is exceeded we set the 
    // overflow payload ptr in the subscription, dont post the 
    // event, and we add the subscription to the event queue's
    // overflow subscription list.
    //
    // If an update for this subscription is posted when 
    // the overflow payload ptr is set we take one of two 
    // possible actions.
    // 1) If the maximum number of pending event entries will not
    // be exceeded we post two events. First the overflow payload
    // is posted. Second, the currently posted payload is posted. 
    // We remove the subscription from the event queue's overflow
    // subscription list. Othersise, if only one slot is available
    // we discard the overflow payload, and post only the current 
    // payload.
    // 2) Otherwise we just replace the overflow pointer thereby 
    // releasing its previous contents.
    //
    // When the event queue is processed the following steps are 
    // performed.
    // o For any subscriptions in overflow subscription list we
    // post their overflow payload to the event queue.
    // o We remove all events from the queue
    // o process all events that have been removed
    //
    // If at any time memory cant be allocated when posting an 
    // event to the queue then the overflow payload in the
    // subscription is set and the subscription is added to the 
    // event queue's overflow syuibscription list.
    // 
};

inline size_t EvQue :: Subscr ::
    m_remainingEntries () const
{
    const size_t currentCount = m_refCtr.count ();
    if ( currentCount < m_maxEventQueueEntires ) {
        return m_maxEventQueueEntires - currentCount;
    }
    else {
        return 0u;
    }
}

inline void EvQue :: Subscr :: 
    enterFlowControlMode ()
{
    m_flowControlIsActive = true;
}

inline void EvQue :: Subscr :: 
    exitFlowControlMode ()
{
    m_flowControlIsActive = false;
}

} // end of namesapce Svr
} // end of namespace ca

#endif // ifndef caSvrEvQueSubscrh

