
#ifndef caVech
#define caVech

// so ... I could change this to not allow access except
// through a guarded pointer object.
//
// there are some advantages
// o No need to pass (and verify) a guard in every interface 
// o objects that are shared about are likely to be shared by
// many threads and they should have interlocked access.
//
// there are some disadvantages
// o There isnt any way that the internal function can 
// release the Guard. If the Guard is explicitly passed
// then the function can temporarily release the Guard
// although admitedly this original design can cause trouble 
// if they create a Guard :: Release and dont destroy it before 
// returning from the Guard protected function. PERHAPS
// THE BETTER ALTERNATIVE IS TO PROVIDE AN ALTERNATIVE 
// INTERFACE THAT ISNT PROTECTED BY A LOCK.
// o there isnt any way to explicitly verify a guard but
// perhaps this is nolonger necessary.
//
// ISSUES WITH PTR INTERFACE
// o the ptr can be nill (which requires frequent testing 
// for nillness, but this isnt easy to fix.
//

#ifndef assert // allow use of epicsAssert.h
#   include <cassert>
#endif
#include <cstdio>
#include <climits>
#include <exception>
#include <stdexcept>
#include <typeinfo>

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_caVech
#   undef epicsExportSharedSymbols
#endif

#include "epicsTypes.h"
#include "osiWireFormat.h"
#include "BaseString.h"
#include "daElemSeq.h"

#ifdef epicsExportSharedSymbols_caVech
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caConstStrip.h"

namespace ca {

using namespace da;

using std :: exception;

template < typename > struct VecBinding;
template < class T > class ElementBlock;

template < class T >
class Vec {
public:
    typedef typename ConstStrip < T > :: Prod ConstStripT;
    typedef typename VecBinding < ConstStripT > :: Handle Handle;
    Vec (); 
    Vec ( const Vec & ); 
    Vec ( const Vec &, size_t nElem, size_t offset = 0u ); 
    Vec ( Handle &, T * pFirst, size_t nElem ); 
    ~Vec (); 
    void clear ();
    Vec & operator = ( const Vec & ); 
    operator Vec < const T > ();
    T * pFirst () const;
    size_t nElem () const;
    bool popNative ( ConstStripT & );
    Vec < T > popVec ( size_t nElemMax );
    size_t trimFront ( size_t nElemMax );
    size_t trimBack ( size_t nElemMax );
    size_t pushZeros ( size_t nElemMax );
    void show ( unsigned level ) const;
    // Notes concerning popNet, pushNet
    // 1) implemented only for T = epicsUint8
    // 2) both presume that Vec needs to be in wire byte order
    size_t popNetStringAppendTo ( StringEditor &, size_t maxChar );
    template < class D >
    bool pushNet ( const D & ); 
    template < class D >
    size_t pushNetArray ( ElemSeq &, size_t nElemMax ); 
    size_t pushNetString ( const StringReader &, size_t pos );
private:
    Handle * m_pHandle;
    T * m_pFirst;
    size_t m_nElem;
    void m_sanityCheck () const;
};

struct HandleVec {
    virtual ~HandleVec () {}
    virtual HandleVec & clone () = 0;
    virtual void release () = 0;
};

// indirect access to the Handle type through 
// the Binding template allows any Handle type 
// Xxxx to be specifed directly (without needing to
// make a specialization that derives from Xxxx).
template < class T >
struct VecBinding {
    typedef HandleVec Handle;
};

template < class T >
inline Vec < T > :: 
    Vec () : m_pHandle ( 0 ), m_pFirst ( 0 ), m_nElem ( 0 )
{
}

template < class T >
inline void Vec < T > :: 
    m_sanityCheck () const
{
#   ifndef NDEBUG
        if ( m_pHandle ) {
            if ( (m_pFirst == 0) == (m_nElem == 0) ) {
                throw std :: 
                    invalid_argument ( 
                        "Vec ctor or assign: "
                        "m_pFirst m_nElem inconsistency" );
            }
        }
        else if ( m_pFirst || m_nElem ) {
            throw std :: 
                invalid_argument ( 
                    "Vec ctor or assign: "
                    "m_pFirst m_nElem inconsistent with nill handle" );
        }
#   endif
}

template < class T >
Vec < T > :: 
    Vec ( const Vec & vec ) :
    m_pHandle ( vec.m_pHandle ? & vec.m_pHandle->clone () : 0 ), 
    m_pFirst ( vec.m_pFirst ), 
    m_nElem ( vec.m_nElem )
{
    m_sanityCheck ();
}

template < class T >
Vec < T > :: 
    Vec ( const Vec & vec, size_t nElem, size_t offset ) :
    m_pHandle ( 0 ), m_pFirst ( 0 ), m_nElem ( 0 )
{
    if ( offset > m_nElem ) {
        throw std :: 
            range_error ( "Vec ctor: offset + nElem exceeding vec size" );
    }
    if ( nElem ) {
        const size_t nLeft = m_nElem - offset;
        if ( nElem > nLeft ) {
            nElem = nLeft;
        }
        if ( nElem ) {
            m_pHandle = & vec.m_pHandle->clone (); 
            m_pFirst = m_pFirst + offset; 
            m_nElem = nElem;
            m_sanityCheck ();
        }
    }
}

template < class T >
Vec < T > :: 
    Vec ( Handle & handle, T * pFirst, size_t nElem ) :
    m_pHandle ( & handle.clone () ), m_pFirst ( pFirst ), 
        m_nElem ( nElem )
{
    if ( m_pFirst == 0 || m_nElem == 0u ) {
        throw std :: 
            range_error ( "Vec ctor: empty vec bind to handle?" );
    }
}

template < class T >
inline Vec < T > :: 
    ~Vec ()
{
    if ( m_pHandle ) {
        m_pHandle->release ();
    }
}

template < class T >
void Vec < T > :: clear ()
{
    if ( m_pHandle ) {
        m_pHandle->release ();
        m_pHandle = 0u;
        m_pFirst = 0u;
        m_nElem = 0u;
    }
    else {
        assert ( m_pFirst == 0 );
        assert ( m_nElem == 0 );
    }
}

template < class T >
Vec < T > & Vec < T > :: 
    operator = ( const Vec & vec )
{
    if ( m_pHandle != vec.m_pHandle ) {
        if ( m_pHandle ) {
            m_pHandle->release ();
        }
        m_pHandle = vec.m_pHandle;
        if ( vec.m_pHandle ) {
            m_pHandle = & vec.m_pHandle->clone ();
        }
    }
    m_pFirst = vec.m_pFirst; 
    m_nElem = vec.m_nElem;
    m_sanityCheck ();
    return *this;
}

template < class T >
void Vec < T > :: 
    show ( unsigned level ) const
{
    printf ( "Vec < %s >\n", typeid ( T ).name () );
    if ( level > 0u ) {
        printf ( "m_pFirst = %p, m_nElem = %lu\n", m_pFirst, 
            static_cast < unsigned long > ( m_nElem ) );
    }
}

template < class T >
inline bool Vec < T > :: 
    popNative ( ConstStripT & dest )
{
    if ( m_nElem ) {
        dest = *m_pFirst++;
        m_nElem--;
        return true;
    }
    return false;
}

template <>
template < class D >
inline bool Vec < epicsUInt8 > :: 
    pushNet ( const D & src )
{
    const size_t vSize = sizeof ( D );
    if ( m_nElem >= vSize ) {
        WireSet ( src, m_pFirst );
        m_pFirst += vSize;
        m_nElem -= vSize;
        return true;
    }
    return false;
}

template <>
template < class D >
size_t Vec < epicsUInt8 > :: 
    pushNetArray ( ElemSeq & elemSeq, size_t nElemMax )
{
    size_t nElemCur = m_nElem / sizeof ( D );
    if ( nElemCur ) {
        nElemCur = std :: min ( nElemMax, nElemCur );
        const size_t nBytes = nElemCur * sizeof ( D );
        elemSeq.getNextNetFrame < D > ( m_pFirst, nBytes );
        m_nElem -= nElemCur;
        m_pFirst += nElemCur;
    }
    return nElemCur;
}

template < class T >
size_t Vec < T > :: 
    pushZeros ( size_t nElemMax )
{
    if ( nElemMax > m_nElem ) {
        nElemMax = m_nElem;
    }
    memset ( m_pFirst, '\0', nElemMax );
    m_nElem -= nElemMax;
    m_pFirst += nElemMax;
    return nElemMax;
}


template <>
inline size_t Vec < epicsUInt8 > :: 
    pushNetString ( const StringReader & str, size_t pos )
{
    size_t n = str.extract ( pos, 
        reinterpret_cast < char * > ( m_pFirst ), m_nElem );
    m_pFirst += n;
    assert ( m_nElem >= n );
    m_nElem -= n;
    return n;
}

template < class T >
inline Vec < T > Vec < T > :: 
    popVec ( size_t nElemMax )
{
    if ( nElemMax > m_nElem ) {
        nElemMax = m_nElem;
    }
    Vec tmp ( *this, nElemMax );
    m_pFirst += tmp.nElem ();
    m_nElem -= tmp.nElem ();
    return tmp;
}

template < class T >
inline size_t Vec < T > ::
    trimFront ( size_t nElem )
{
    if ( nElem > m_nElem ) {
        nElem = m_nElem;
    }
    m_pFirst += nElem;
    m_nElem -= nElem;
    return nElem;
}

template < class T >
inline size_t Vec < T > :: 
    trimBack ( size_t nElem )
{
    if ( nElem > m_nElem ) {
        nElem = m_nElem;
    }
    m_nElem -= nElem;
    return nElem;
}

template < class T >
inline T * Vec < T > :: 
    pFirst () const
{
    return m_pFirst;
}

template < class T >
inline size_t Vec < T > :: 
    nElem () const
{
    return m_nElem;
}

template < class T >
inline Vec < T > :: operator Vec < const T > ()
{
    if ( m_pHandle ) {
        return Vec < const T > ( *m_pHandle, m_pFirst, m_nElem );
    }
    else {
        return Vec < const T > ();
    }
}

#if 0
template <>
template < class V >
inline bool VecGuard < const epicsUInt8 > :: 
    pop ( V & val )
{
    const size_t vSize = sizeof ( V );
    bool success = m_hdl.nElems () >= vSize + m_pos;
    if ( success ) {
        WireGet ( m_hdl.pTarget () + m_pos, val );
        m_pos += vSize;
    }
    return success;
}
#endif

#if 0
template < class T >
class ElementBlock {
public:
    ElementBlock ( T & target );
    // empty element blocks are not allowed
    ElementBlock ( T * pFirstElem, size_t nElem );
    T & first () const;
    T * pFirst () const;
    size_t nElem () const;
    T & operator [] ( size_t ) const;
    template < class V >
    bool pop ( V & val );
    void advance ( size_t ); // remove elems from start
    void trim ( size_t ); // remove elems from end
private:
    T * m_pFirst;
    size_t m_nElem;
};

template < class T >
inline ElementBlock < T > :: 
    ElementBlock ( T & target ) :
    m_pFirst ( & target ),
    m_nElem ( 1u )
{
}

template < class T >
inline ElementBlock < T > :: 
    ElementBlock ( T * pFirstElem, size_t nElem ) :
    m_pFirstElem ( pFirstElem ), m_nElem ( nElem )
{
    if ( m_nElem > 0 && m_pFirst == 0 ) {
        throw std :: invalid_argument ( 
            "ElementBlock ctor: the elem "
            "ptr was nill when the elem count wast" );
    }
}

template < class T >
inline T & ElementBlock < T > :: 
    first () const
{
    // ptr nillness is checked in the ctor
    return *m_pFirst;
}

template < class T >
inline T * ElementBlock < T > :: 
    pFirst () const
{
    // ptr nillness is checked in the ctor
    return m_pFirst;
}

template < class T >
inline T & ElementBlock < T > :: 
    operator [] ( size_t idx ) const
{
    assert ( idx < m_nElem );
    return m_pFirst[idx];
}

template < class T >
inline size_t ElementBlock < T > :: 
    nElem () const
{
    return m_nElem;
}

template <>
template < class V >
inline bool ElementBlock < const epicsUInt8 > :: 
    pop ( V & dest )
{
    const size_t vSize = sizeof ( V );
    bool success = ( m_nElem >= vSize );
    if ( success ) {
        WireGet ( m_pFirst, dest );
        m_pFirst += vSize;
        m_nElem -= vSize;
    }
    return success;
}
#endif

#if 0
template <>
inline size_t VecGuard < const epicsUInt8 > :: 
    popString ( size_t maxElems, StringEditor & se )
{
#error still needs work
    bool success = ( m_nElem > 0u );
    if ( success ) {
        assert ( sizeof ( char ) == sizeof ( *m_pFirstElem ) );
        se.append ( reinterpret_cast < char > ( m_pFirstElem ), m_nElem );
        m_pFirstElem += m_nElem;
        m_nElem = 0u;
    }
    return success;
}

template <>
template < class V >
bool VecGuard < epicsUInt8 > :: 
    push ( const V & val )
{
}

// returns the number of characters pushed
template <>
size_t VecGuard < epicsUInt8 > :: 
    pushString ( size_t pos, const StringReader & )
{
}

template <>
size_t VecGuard < epicsUInt8 > :: 
    pushZeros ( size_t len )
{
}
#endif

#if 0
// Visual c++ 2008 seems to require this
// specialization while both g++ and comeau dont?
// This is related to the compiler's choosing
// what type const T will be, when T is 
// "char [N]". Will it be a const array, or an 
// array of const elements. The standard says
// "const char [N]". Presumable, vc++ decides to
// strip const from the array applying it instead
// to the element leading to a conclusion that 
// Ptr < T >, and not Ptr < const T >, is a base 
// class for Ptr < T >. That conclusion could 
// lead to vc++'s decision not to compile the 
// code, but perhaps that behavior is incorrect.
template < size_t N >
class Ptr < const char [ N ] > {
public:
    typedef const char FixedStrConst [ N ];
    Ptr ( FixedStrConst & );
    const FixedStrConst * operator -> () const;
    const FixedStrConst & operator * () const;
protected:
    Guard m_guard;
    FixedStrConst & m_target;
    Ptr ( const Ptr & ); // disabled
    Ptr & operator = ( const Ptr & ); // disabled
};
#endif

} // end of namespace ca

#endif // ifndef caVech
