/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
//
//  casdefII.h,v 1.3 2009/11/24 01:34:27 jhill Exp
//
//      Author: Jeffrey O. Hill
//              johill@lanl.gov
//              (505) 665 1831
//      Date:   1-95
//

#error defunct 

#ifndef includecasdefh
#define includecasdefh

#ifdef epicsExportSharedSymbols
#   define epicsExportSharedSymbols_casdefh
#   undef epicsExportSharedSymbols
#endif

#include "daData.h"   // EPICS data descriptors 

#ifdef epicsExportSharedSymbols_casdefh
#   define epicsExportSharedSymbols
#   include "shareLib.h"
#endif

#include "caNetAddr.h"
#include "casEventMask.h"   // EPICS event select class 

//
// pv exist test return
//
// If the server tool does not wish to start another simultaneous
// asynchronous IO operation or if there is not enough memory
// to do so return pverDoesNotExistHere (and the client will
// retry the request later).
//
enum pvExistReturnEnum { pverExistsHere, pverDoesNotExistHere, 
    pverAsyncCompletion };

class casCtx;
class casChannel;

class epicsShareClass pvExistReturn { // X aCC 361
public:
    // most server tools will use this
    pvExistReturn ( pvExistReturnEnum s = pverDoesNotExistHere );
    // directory service server tools will use this (see caNetAddr.h)
    pvExistReturn ( const caNetAddr & );
    ~pvExistReturn ();
    const pvExistReturn & operator = ( pvExistReturnEnum rhs );
    const pvExistReturn & operator = ( const caNetAddr & rhs );
    pvExistReturnEnum getStatus () const;
    caNetAddr getAddr () const;
    bool addrIsValid () const;
private:
    caNetAddr address;
    pvExistReturnEnum status;
};

//
// createChannelReturn
//
class epicsShareClass createChannelReturn {
public:
    createChannelReturn (); // failed
    createChannelReturn ( casPV & pv );
    const createChannelReturn & operator = ( caStatus rhs );
    const createChannelReturn & operator = ( casPV * pPVIn );
    caStatus getStatus() const ;
    casPV * getPV() const;
private:
    casChannel * pPV;
};

//
// caServer - Channel Access Server API Class
//
class caServer {
    friend class casPVI;
public:
    epicsShareFunc caServer ();
    epicsShareFunc virtual ~caServer() = 0;

    //
    // pvExistTest()
    //
    // This function is called by the server library when it needs to 
    // determine if a named PV exists (or could be created) in the 
    // server tool.
    //
    // The request is allowed to complete asynchronously
    // (see Asynchronous IO Classes below).
    //
    // The server tool is encouraged to accept multiple PV name
    // aliases for the same PV here. 
    //
    // example return from this procedure:
    // return pverExistsHere;   // server has PV
    // return pverDoesNotExistHere; // server does know of this PV
    // return pverAsynchCompletion; // deferred result 
    //
    // Return pverDoesNotExistHere if too many simultaneous
    // asynchronous IO operations are pending against the server. 
    // The client library will retry the request at some time
    // in the future.
    //
    epicsShareFunc virtual pvExistReturn pvExistTest ( const casCtx & ctx, 
        const caNetAddr & clientAddress, const char * pAliasName );

    //
    // pvAttach() 
    //
    // This function is called m_every_ time that a client attaches to the PV.
    // The name supplied here will be either a canonical PV name or an alias 
    // PV name.
    //  
    // The request is allowed to complete asynchronously
    // (see Asynchronous IO Classes below).
    //
    // IMPORTANT: 
    // It is a responsibility of the server tool to detect attempts by 
    // the server library to attach to an existing PV. If the PV does not 
    // exist then the server tool should create it. Otherwise, the server 
    // tool typically will return a pointer to the preexisting PV. 
    //
    // The server tool is encouraged to accept multiple PV name aliases 
    // for the same PV here. 
    //
    // In most situations the server tool should avoid PV duplication 
    // by returning a pointer to an existing PV if the PV alias name
    // matches a preexisting PV's name or any of its aliases.
    //
    // example return from this procedure:
    // return pPV;  // success (pass by pointer)
    // return PV;   // success (pass by ref)
    // return S_casApp_pvNotFound; // no PV by that name here
    // return S_casApp_noMemory; // no resource to create pv
    // return S_casApp_asyncCompletion; // deferred completion
    //
    // Return S_casApp_postponeAsyncIO if too many simultaneous
    // asynchronous IO operations are pending against the server. 
    // The server library will retry the request whenever an
    // asynchronous IO operation (create or exist) completes
    // against the server.
    //
    // In certain specialized rare situations the server tool may choose
    // to create client private process variables that are not shared between
    // clients. In this situation there might be several process variables
    // with the same name. One for each client. For example, a server tool
    // might be written that provides access to archival storage of the
    // historic values of a set of process variables. Each client would 
    // specify its date of interest by writing into a client private process 
    // variable. Next the client would determine the current value of its 
    // subset of named process variables at its privately specified date by 
    // attaching to additional client private process variables.
    //
    epicsShareFunc virtual createChannelReturn createChannel ( 
        const casCtx & ctx, const char * pAliasName,
        const char * const pUserName, const caNetAddr & hostAddress,
        const char * const pHostName );

    //
    // obtain an event mask for a named event type
    // to be used with casPV::postEvent()
    //
    epicsShareFunc EventMask registerEvent ( const char *pName );

    //
    // common event masks 
    // (what is currently used by the CA clients)
    //
    epicsShareFunc EventMask valueEventMask () const; // DBE_VALUE 
    epicsShareFunc EventMask logEventMask () const;  // DBE_LOG 
    epicsShareFunc EventMask alarmEventMask () const; // DBE_ALARM 
    epicsShareFunc EventMask accessRightsEventMask () const; 

    epicsShareFunc void setDebugLevel ( unsigned level );
    epicsShareFunc unsigned getDebugLevel () const;

    //
    // dump internal state of server to standard out
    //
    epicsShareFunc virtual void show ( unsigned level ) const;

    //
    // server Abberation counters (allowed to roll over)
    //
    // subscriptionEventsPosted()
    //      - number of events posted by server tool to the event queue
    // subscriptionEventsProcessed() 
    //      - number of events removed by server library from the event queue
    //
    epicsShareFunc unsigned subscriptionEventsPosted () const;
    epicsShareFunc unsigned subscriptionEventsProcessed () const;

    epicsShareFunc void generateBeaconAnomaly ();

    // caStatus enableClients ();
    // caStatus disableClients ();

private:
    class caServerI * pCAS;
};

//
// casChannel 
//
class casChannel {
public:
    epicsShareFunc casChannel ( const casCtx & ctx );
    epicsShareFunc virtual ~casChannel ();
    epicsShareFunc virtual epicsType bestExternalType () const;
    epicsShareFunc virtual unsigned maxDimension () const; // return zero if scalar
    epicsShareFunc virtual aitIndex maxBound ( unsigned dimension ) const;
    epicsShareFunc virtual void getName ( const char *pBuf, unsigned bufLength ) const;
    epicsShareFunc virtual bool readAccess () const;
    epicsShareFunc virtual bool writeAccess () const;
    epicsShareFunc virtual bool confirmationRequested () const;
    epicsShareFunc virtual void read ( const casCtx &, daData & );
    epicsShareFunc virtual void write ( const casCtx &, const daData & );
    epicsShareFunc virtual void show ( unsigned level ) const;

    epicsShareFunc void postEvent ( const EventMask &, const daData & );
    epicsShareFunc caServer & getCAS () const;

private:
    class casChannelI * pChanI;

    casChannel ( const casChannel & );
    casChannel & operator = ( const casChannel & );
};

//
// Asynchronous IO Classes
//
// The following virtual functions allow for asynchronous completion:
//
//  Virtual Function          Asynchronous IO Class
//  -----------------         ---------------------
//  caServer::pvExistTest()   casAsyncPVExist
//  caServer::createChannel() casAsyncCreateChannel
//  casPV::read()             casAsyncRead
//  casPV::write()            casAsyncWrite
//
// To initiate asynchronous completion create a corresponding
// asynchronous IO object from within one of the virtual 
// functions shown in the table above. Use the member function 
// "postIOCompletion()" to inform the server library that the 
// requested operation has completed.
//

//
// casAsyncRead 
// 
class casAsyncRead {
public:
    epicsShareFunc casAsyncRead ( const casCtx & ctx );
    epicsShareFunc virtual ~casAsyncRead (); 
    epicsShareFunc void postCompletion ( const daData & );
    epicsShareFunc void postCompletionFailure ( 
        const char * pExplanation );
    epicsShareFunc caServer * getCAS () const;
private:
    class casAsyncReadI * pAsyncReadI;

    casAsyncRead ( const casAsyncRead & );
    casAsyncRead & operator = ( const casAsyncRead & );
    friend class casAsyncReadI;
};

//
// casAsyncWrite 
//
class casAsyncWrite {
public:
    epicsShareFunc casAsyncWrite ( const casCtx & ctx );
    epicsShareFunc virtual ~casAsyncWrite (); 
    epicsShareFunc void postIOCompletion ();
    epicsShareFunc void postIOCompletionFailure ( 
        const char * pExplanation );
    epicsShareFunc caServer * getCAS () const;
private:
    class casAsyncWriteIOI * pAsyncWriteI;

    casAsyncWriteIO ( const casAsyncWrite & );
    casAsyncWriteIO & operator = ( const casAsyncWrite & );
    friend class casAsyncWriteI;
};

//
// casAsyncPVExist
//
class casAsyncPVExist {
public:
    epicsShareFunc casAsyncPVExist ( const casCtx & ctx );
    epicsShareFunc virtual ~casAsyncPVExist (); 
    epicsShareFunc caStatus postIOCompletion ( const pvExistReturn & retValIn );
    epicsShareFunc caServer * getCAS () const;
private:
    class casAsyncPVExistIOI * pAsyncPVExistI;
    casAsyncPVExistIO ( const casAsyncPVExist & );
    casAsyncPVExistIO & operator = ( const casAsyncPVExist & );
    friend class casAsyncPVExistI;
};

//
// casAsyncCreateChannel
//
class casAsyncCreateChannel {
public:
    epicsShareFunc casAsyncCreateChannel ( const casCtx & ctx );
    epicsShareFunc virtual ~casAsyncCreateChannel (); 
    epicsShareFunc caStatus postIOCompletion ( const createChannelReturn & retValIn );
    epicsShareFunc caServer * getCAS () const;
private:
    class casAsyncPVAttachI * pAsyncPVAttachI;
    casAsyncPVAttachIO ( const casAsyncPVAttach & );
    casAsyncPVAttachIO & operator = ( const casAsyncPVAttach & );
    friend class casAsyncPVAttachI;
};

// TODO:
// .03  document new event types for limits change etc
// .04  certain things like native type cant be changed during
//  pv id's life time or we will be required to have locking
//  (doc this)
// .08  Need a mechanism by which an error detail string can be returned
//  to the server from a server app (in addition to the normal
//  error constant)
// .12  Should the server have an interface so that all PV names
//  can be obtained (even ones created after init)? This
//  would be used to implement update of directory services and 
//  wild cards? Problems here with knowing PV name structure and 
//  maximum permutations of name components.
// .16  go through this file and make sure that we are consistent about
//  the use of const - use a pointer only in spots where NULL is
//  allowed?
// NOTES:
// .01  When this code is used in a multi threaded situation we
//  must be certain that the derived class's virtual function 
//  are not called between derived class destruction and base
//  class destruction (or prevent problems if they are).
//  Possible solutions
//  1) in the derived classes destructor set a variable which
//  inhibits use of the derived classes virtual function.
//  Each virtual function must check this inhibit bit and return 
//  an error if it is set
//  2) call a method on the base which prevents further use
//  of it by the server from the derived class destructor.
//  3) some form of locking (similar to (1) above)
//

#endif // ifdef includecasdefh (this must be the last line in this file) 

