/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE is distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

//
// Author: Jeff Hill
//

#include <string>
#include <stdexcept>

#include "errlog.h"
#include "osiSock.h"
#include "osiAtomic.h"
#include "epicsThread.h"

#define epicsExportSharedSymbols
#include "SockCir.h"
#include "caPtr.h"

namespace ca {

// SockCir :: CirIO ()
SockCir :: SockCir ( SOCKET sock ) :
    m_sock ( sock ), 
    m_circuitInHasExisted ( false ),
    m_circuitOutHasExisted ( false )
{
    using epics :: osi :: atomicSet;
    atomicSet ( m_sockCloseHasCompleted, false );
	/*
	 * see TCP(4P) this seems to make unsollicited single events much
	 * faster. I take care of queue up as load increases.
	 */
	static const int yes = true;
	int status = setsockopt ( m_sock, IPPROTO_TCP, TCP_NODELAY,
                    reinterpret_cast < const char * > ( & yes ), sizeof ( yes ) );
	if ( status < 0 ) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf (
			"CAS: %s TCP_NODELAY option set failed %s\n",
			__FILE__, sockErrBuf );
        throw std::runtime_error ( 
            "setsockopt() TCP_NODELAY failed" );
	}

	/*
	 * turn on KEEPALIVE so if the client crashes
	 * this task will find out and exit
	 */
	status = setsockopt ( m_sock, SOL_SOCKET, SO_KEEPALIVE,
					reinterpret_cast < const char * > ( & yes ), sizeof ( yes ) );
	if ( status < 0 ) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf (
			"CAS: %s SO_KEEPALIVE option set failed %s\n",
			__FILE__, sockErrBuf );
        throw std::runtime_error ( 
            "setsockopt() SO_KEEPALIVE failed" );
	}

	/*
	 * some concern that vxWorks will run out of mBuf's
	 * if this change is made
	 *
	 * joh 11-10-98
	 */
#if 0

	/*
	 * set TCP buffer sizes to be synergistic
	 * with CA internal buffering
	 */
	static const int i = MAX_MSG_SIZE;
	status = setsockopt ( m_sock, SOL_SOCKET, SO_SNDBUF,
					(const char *)&i, sizeof(i));
	if ( status < 0 ) {
		errlogPrintf("CAS: SO_SNDBUF set failed\n");
        throw std::runtime_error ( 
            "setsockopt() SO_SNDBUF failed" );
	}
	status = setsockopt(
					sock,
					SOL_SOCKET,
					SO_RCVBUF,
					(const char *)&i,
					sizeof(i));
	if(status < 0){
		errlogPrintf("CAS: SO_RCVBUF set failed\n");
        throw std::runtime_error ( 
            "setsockopt() SO_RCVBUF failed" );
	}
#endif

}

// SockCir :: ~CirIO ()
SockCir :: ~SockCir ()
{
    using epics :: osi :: atomicTestAndSet;
    if ( atomicTestAndSet ( m_sockCloseHasCompleted ) ) {
        epicsSocketDestroy ( m_sock );
    }
}

// SockCir :: createInputStream ()
std :: auto_ptr < CircuitIn > 
    SockCir :: createInputStream ( const Ptr < VecAllocator > & pVecAllocator )
{
    if ( m_circuitInHasExisted ) {
        throw std :: runtime_error ( 
            "SockCir :: createInputStream: called more than once" );
    }
    m_circuitOutHasExisted = true;
    M_CircuitIn * pIn = new M_CircuitIn ( *this, pVecAllocator );
    return std :: auto_ptr < CircuitIn > ( pIn );
}

// SockCir :: createOutStream ()
std :: auto_ptr < CircuitOut > 
    SockCir :: createOutStream ()
{
    if ( m_circuitOutHasExisted ) {
        throw std :: runtime_error ( 
            "SockCir :: createOutStream: called more than once" );
    }
    m_circuitOutHasExisted = true;
    M_CircuitOut * pOut = new M_CircuitOut ( *this );
    return std :: auto_ptr < CircuitOut > ( pOut );
}

// SockCir :: show ()
void SockCir :: show ( unsigned level ) const
{
	printf ( "SockCir\n" );
	if ( level > 1u ) {
        osiSockAddr sa;
        osiSocklen_t size = sizeof ( sa );
        getsockname ( m_sock, & sa.sa, & size ); 
        char buf[64];
        ipAddrToA ( &sa.ia, buf, sizeof ( buf ) );
		printf ( "client addr = \"%s\"\n", buf );
        using epics :: osi :: atomicFetch;
		printf ( "m_sockCloseHasCompleted = %c\n", 
            atomicFetch ( m_sockCloseHasCompleted ) ? 'T' : 'F' );
	}
}

SockCir :: M_CircuitOut ::
    M_CircuitOut ( SockCir & sockCir ) :
    m_sockCir ( sockCir )
{
}

void SockCir :: M_CircuitOut ::
    write ( const Vec  < const epicsUInt8 > & vecIn )
{
    if ( m_sockCir.m_sock == INVALID_SOCKET ) {
        throw CircuitTerminate ();
    }
    Vec  < const epicsUInt8 > vec ( vecIn );
    while ( vec.nElem () ) {
        int status = send ( m_sockCir.m_sock, 
            reinterpret_cast < const char * > 
                ( vec.pFirst () ), vec.nElem (), 0 );
        if ( status == 0 ) {
            throw CircuitTerminate ();
        }
        else if ( status < 0 ) {
            int anerrno = SOCKERRNO;
            if ( anerrno == SOCK_EINTR ) {
                return;
            }
            if ( anerrno == SOCK_ENOBUFS ) {
                errlogPrintf ( 
                    "cas: system low on network buffers - "
                    "hybernating for 1 second\n" );
                epicsThreadSleep ( 1.0 );
                return;
            }
            m_sockCir.m_throwDisconnect ();
        }
        size_t nElemSent = static_cast < size_t > ( status ); 
        assert ( nElemSent <= vec.nElem () ); 
        vec.trimFront ( nElemSent );
    }
}

void SockCir :: M_CircuitOut ::
    flush ()
{
}

void SockCir :: M_CircuitOut :: show ( unsigned level ) const
{
    m_sockCir.show ( level );
}

void SockCir:: M_CircuitOut :: shutdown ()
{
    using epics :: osi :: atomicTestAndSet;
    epicsSocketSystemCallInterruptMechanismQueryInfo mech =
        epicsSocketSystemCallInterruptMechanismQuery ();
    switch ( mech ) {
    case esscimqi_socketBothShutdownRequired:
        :: shutdown ( m_sockCir.m_sock, SHUT_WR );
        break;
    default:
        errlogPrintf (
            "CAS: epicsSocketSystemCallInterruptMechanismQuery "
            "returned an unexpected value\n" );
    case esscimqi_socketCloseRequired:
        if ( atomicTestAndSet ( m_sockCir.m_sockCloseHasCompleted ) ) {
            epicsSocketDestroy ( m_sockCir.m_sock );
        }
    }
}

void SockCir :: m_throwDisconnect ()
{
    char sockErrBuf[64];
    epicsSocketConvertErrnoToString ( 
            sockErrBuf, sizeof ( sockErrBuf ) );
    osiSockAddr sa;
    osiSocklen_t size = sizeof ( sa );
    getsockname ( m_sock, & sa.sa, & size ); 
    char buf[64];
    ipAddrToA ( &sa.ia, buf, sizeof ( buf ) );
    std :: string reason;
    reason += "TCP socket send to \"";
    reason += buf;
    reason += "\" failed because \"";
    reason += sockErrBuf;
    reason += "\"";
    errlogPrintf ( "CAS: %s\n", reason.c_str () );
    throw CircuitDisconnect ( reason.c_str () );
}

SockCir :: M_CircuitIn ::
    M_CircuitIn ( SockCir & sockCir, 
        const Ptr < VecAllocator > & pVecAllocator ) :
    m_sockCir ( sockCir ), m_pVecAllocator ( pVecAllocator )
{
    if ( pVecAllocator.isNill () ) {
        throw std :: invalid_argument ( 
            "M_CircuitIn :: M_CircuitIn: pVecAllocator arg was nill" );
    }
}

Vec < const epicsUInt8 > SockCir :: M_CircuitIn ::
    read ( size_t nBytesMax )
{
    if ( m_sockCir.m_sock == INVALID_SOCKET ) {
        throw CircuitTerminate ();
    }
    Vec < epicsUInt8 > vec = m_pVecAllocator->alloc ();
    int nchars = recv ( m_sockCir.m_sock, reinterpret_cast < char * > 
                            ( vec.pFirst () ), vec.nElem (), 0 );
    if ( nchars == 0 ) {
        throw CircuitTerminate ();
    }
    else if ( nchars < 0 ) {
        int myerrno = SOCKERRNO;
        if ( myerrno == SOCK_EINTR ) {
            return Vec < const epicsUInt8 > ();
        }
        if ( myerrno == SOCK_ENOBUFS ) {
            errlogPrintf ( 
                "CAS: system low on network buffers - "
                "hybernating for 1 second\n" );
            epicsThreadSleep ( 1.0 );
            return Vec < const epicsUInt8 > ();
        }
        m_sockCir.m_throwDisconnect ();
    }
    size_t nBytesActual = static_cast < size_t > ( nchars );
    assert ( nBytesActual <= vec.nElem () );
    vec.trimBack ( nBytesActual );
    return vec;
}

void SockCir :: M_CircuitIn :: shutdown ()
{
    using epics :: osi :: atomicTestAndSet;
    epicsSocketSystemCallInterruptMechanismQueryInfo mech =
        epicsSocketSystemCallInterruptMechanismQuery ();
    switch ( mech ) {
    case esscimqi_socketBothShutdownRequired:
        :: shutdown ( m_sockCir.m_sock, SHUT_RD );
        break;
    default:
        errlogPrintf (
            "CAS: epicsSocketSystemCallInterruptMechanismQuery "
            "returned an unexpected value\n" );
    case esscimqi_socketCloseRequired:
        if ( atomicTestAndSet ( m_sockCir.m_sockCloseHasCompleted ) ) {
            epicsSocketDestroy ( m_sockCir.m_sock );
        }
    }
}

void SockCir :: M_CircuitIn :: show ( unsigned level ) const
{
    m_sockCir.show ( level );
}

#if 0
// SockCir:::optimumBufferSize()
bufSizeT SockCir :: optimumBufferSize () 
{

	int n;
	int size;
	int status;

	/* fetch the TCP send buffer size */
	n = sizeof ( size) ;
	status = getsockopt ( this->sock, SOL_SOCKET,
			SO_SNDBUF, (char *) & size, & n );
	if ( status < 0 || n != sizeof ( size ) ) {
		size = 0x400;
	}

	if (size<=0) {
		size = 0x400;
	}
printf("the tcp buf size is %d\n", size);
	return (bufSizeT) size;
}
#endif

} // end of namespace ca

