/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
//
//
// Author Jeff Hill
//

#include <string>
#include <stdexcept>

#include <stdio.h>

#include "errlog.h"
#define epicsAssertAuthor "Jeff Hill johill@lanl.gov"
#include "epicsAssert.h"

#define epicsExportSharedSymbols
#include "casdef.h"
#include "caNetAddr.h"
#include "sockIntfIO.h"
#include "casStreamIO.h"
#include "casStreamOS.h"

//
// 5 appears to be a TCP/IP built in maximum
//
const unsigned caServerConnectPendQueueSize = 5u;

//
// sockIntfIO::sockIntfIO()
//
sockIntfIO :: sockIntfIO ( const caNetAddr & addrIn ) : 
	sock ( INVALID_SOCKET ),
    addr ( addrIn.getSockIP() )
{
	int status;
	osiSocklen_t addrSize;
    bool portChange;

	if ( ! osiSockAttach () ) {
        throw std::runtime_error ( 
            "server was unable to initialize the socket library" );
	}

	/*
	 * Setup the server socket
	 */
	this->sock = epicsSocketCreate ( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if (this->sock==INVALID_SOCKET) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		printf ( "No socket error was %s\n", sockErrBuf );
        throw std::runtime_error ( 
            "server was unable to create a socket" );
	}

    epicsSocketEnableAddressReuseDuringTimeWaitState ( this->sock );

	status = bind ( this->sock,
                      reinterpret_cast <sockaddr *> (&this->addr),
                      sizeof(this->addr) );
	if (status<0) {
		if (SOCKERRNO == SOCK_EADDRINUSE) {
			//
			// enable assignment of a default port
			// (so the getsockname() call below will
			// work correctly)
			//
			this->addr.sin_port = ntohs (0);
			status = bind(
                    this->sock,
                    reinterpret_cast <sockaddr *> (&this->addr),
                    sizeof(this->addr) );
		}
		if (status<0) {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
			char buf[64];
			ipAddrToA (&this->addr, buf, sizeof(buf));
			errPrintf ( S_cas_bindFail,
				__FILE__, __LINE__,
				"- bind TCP IP addr=%s failed because %s",
				buf, sockErrBuf );
            epicsSocketDestroy (this->sock);
            throw std::runtime_error ( 
                "server was unable to bind to a network interface" );
		}
        portChange = true;
	}
    else {
        portChange = false;
    }

	addrSize = ( osiSocklen_t ) sizeof (this->addr);
	status = getsockname ( 
                this->sock, 
                reinterpret_cast <sockaddr *> ( &this->addr ), 
                &addrSize );
	if (status) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf ( "CAS: getsockname() error %s\n", 
			sockErrBuf );
        epicsSocketDestroy (this->sock);
        throw std::runtime_error ( 
            "getsockname() failed" );
	}

	//
	// be sure of this now so that we can fetch the IP 
	// address and port number later
	//
    assert (this->addr.sin_family == AF_INET);

    if ( portChange ) {
        errlogPrintf ( "cas warning: Configured TCP port was unavailable.\n");
        errlogPrintf ( "cas warning: Using dynamically assigned TCP port %hu,\n", 
            ntohs (this->addr.sin_port) );
        errlogPrintf ( "cas warning: but now two or more servers share the same UDP port.\n");
        errlogPrintf ( "cas warning: Depending on your IP kernel this server may not be\n" );
        errlogPrintf ( "cas warning: reachable with UDP unicast (a host's IP in EPICS_CA_ADDR_LIST)\n" );
    }

    status = listen(this->sock, caServerConnectPendQueueSize);
    if ( status < 0 ) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf ( "CAS: listen() error %s\n", sockErrBuf );
        epicsSocketDestroy (this->sock);
        throw std::runtime_error ( 
            "listen() failed" );
    }
}

//
// sockIntfIO::~sockIntfIO()
//
sockIntfIO :: ~sockIntfIO()
{
	if (this->sock != INVALID_SOCKET) {
		epicsSocketDestroy (this->sock);
	}

	osiSockRelease ();
}

//
// newStreamIO :: byteStreamFactory ()
//
casStreamOS & sockIntfIO :: byteStreamFactory () 
{
    static bool oneMsgFlag = false;

    struct sockaddr	newAddr;
    osiSocklen_t length = ( osiSocklen_t ) sizeof ( newAddr );
    SOCKET newSock = epicsSocketAccept ( this->sock, & newAddr, & length );
    if ( newSock == INVALID_SOCKET ) {
        int errnoCpy = SOCKERRNO;
        if ( ! oneMsgFlag ) {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
            errlogPrintf ( "CAS: %s accept error \"%s\"\n",
                __FILE__, sockErrBuf );
            oneMsgFlag = true;
        }
        throw runtime_error ( "unable to create socket" );
    }
    else if ( sizeof ( newAddr ) > (size_t) length ) {
        epicsSocketDestroy ( newSock );
        errlogPrintf ( "CAS: accept returned bad address len?\n" );
        throw runtime_error ( "accept returned bad address len" );
    }
    oneMsgFlag = false;
    ioArgsToNewStreamIO args;
    args.addr = newAddr;
    args.sock = newSock;
    casStreamOS	* pOS = new ( std::nothrow ) 
                    casStreamOS ( args );
    if ( ! pOS ) {
        errMessage ( S_cas_noMemory, 
            "unable to create data structures for a new client" );
        epicsSocketDestroy ( newSock );
        throw std :: badAlloc ();
    }
    return pOS;
}

//
// sockIntfIO::setNonBlocking()
//
void sockIntfIO :: setNonBlocking()
{
        int status;
        osiSockIoctl_t yes = true;
 
        status = socket_ioctl(this->sock, FIONBIO, &yes); // X aCC 392
        if ( status < 0 ) {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
            errlogPrintf (
                "%s:CAS: server non blocking IO set fail because \"%s\"\n",
                            __FILE__, sockErrBuf );
        }
}
 
//
// sockIntfIO::getFD()
//
int sockIntfIO :: getFD() const
{
        return this->sock;
}

//
// sockIntfIO::show()
//
void sockIntfIO :: show(unsigned level) const
{
	if (level>2u) {
		printf(" sockIntfIO::sock = %d\n", this->sock);
	}
}

//
// sockIntfIO::serverAddress ()
// (avoid problems with GNU inliner)
//
caNetAddr sockIntfIO :: serverAddress () const
{
    return caNetAddr (this->addr);
}

