/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
//
//

#include <string>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "sockStreamIO.h"

// sockStreamIO::sockStreamIO()
sockStreamIO :: sockStreamIO ( caServerI & cas, clientBufMemoryManager & bufMgr,
                          const ioArgsToNewStreamIO & args ) :
	CirSvr ( cas, bufMgr ), sock ( args.sock ), addr (  args.addr), 
        blockingFlag ( xIsBlocking ), sockHasBeenShutdown ( false )
 {
	assert ( sock >= 0 );
	int yes = true;
	int	status;

	/*
	 * see TCP(4P) this seems to make unsollicited single events much
	 * faster. I take care of queue up as load increases.
	 */
	status = setsockopt ( this->sock, IPPROTO_TCP, TCP_NODELAY,
							( char * ) & yes, sizeof ( yes ) );
	if ( status < 0 ) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf (
			"CAS: %s TCP_NODELAY option set failed %s\n",
			__FILE__, sockErrBuf );
        throw std::runtime_error ( 
            "setsockopt() TCP_NODELAY failed" );
	}

	/*
	 * turn on KEEPALIVE so if the client crashes
	 * this task will find out and exit
	 */
	status = setsockopt ( sock, SOL_SOCKET, SO_KEEPALIVE,
					(char *) & yes, sizeof ( yes ) );
	if (status<0) {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf (
			"CAS: %s SO_KEEPALIVE option set failed %s\n",
			__FILE__, sockErrBuf );
        throw std::runtime_error ( 
            "setsockopt() SO_KEEPALIVE failed" );
	}

	/*
	 * some concern that vxWorks will run out of mBuf's
	 * if this change is made
	 *
	 * joh 11-10-98
	 */
#if 0
	int i;

	/*
	 * set TCP buffer sizes to be synergistic
	 * with CA internal buffering
	 */
	i = MAX_MSG_SIZE;
	status = setsockopt(
					sock,
					SOL_SOCKET,
					SO_SNDBUF,
					(char *)&i,
					sizeof(i));
	if ( status < 0 ) {
		errlogPrintf("CAS: SO_SNDBUF set failed\n");
        throw std::runtime_error ( 
            "setsockopt() SO_SNDBUF failed" );
	}
	i = MAX_MSG_SIZE;
	status = setsockopt(
					sock,
					SOL_SOCKET,
					SO_RCVBUF,
					(char *)&i,
					sizeof(i));
	if(status < 0){
		errlogPrintf("CAS: SO_RCVBUF set failed\n");
        throw std::runtime_error ( 
            "setsockopt() SO_RCVBUF failed" );
	}
#endif

}

// sockStreamIO::~sockStreamIO()
sockStreamIO :: ~sockStreamIO()
{
	epicsSocketDestroy ( this->sock );
}

// sockStreamIO::osdSend()
casStreamWriteIO :: flushCondition sockStreamIO :: osdSend ( 
    const char *pInBuf, size_t nBytesReq, 
    size_t &nBytesActual )
{
    int	status;
    
    if ( nBytesReq == 0 ) {
        nBytesActual = 0;
        return casStreamWriteIO::flushNone;
    }
    
    status = send (this->sock, (char *) pInBuf, nBytesReq, 0);
    if (status == 0) {
        return casStreamWriteIO::flushDisconnect;
    }
    else if (status<0) {
        int anerrno = SOCKERRNO;
        
        if ( anerrno == SOCK_EINTR || 
                anerrno == SOCK_EWOULDBLOCK ) {
            return casStreamWriteIO::flushNone;
        }

        if ( anerrno == SOCK_ENOBUFS ) {
            errlogPrintf ( 
                "cas: system low on network buffers - hybernating for 1 second\n" );
            epicsThreadSleep ( 1.0 );
            return casStreamWriteIO::flushNone;
        }

        if (  
            anerrno != SOCK_ECONNABORTED &&
            anerrno != SOCK_ECONNRESET &&
            anerrno != SOCK_EPIPE &&
            anerrno != SOCK_ETIMEDOUT ) {

            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
 			char buf[64];
            ipAddrToA (&this->addr, buf, sizeof(buf));
			errlogPrintf (
"CAS: TCP socket send to \"%s\" failed because \"%s\"\n",
				buf, sockErrBuf );
        }
        return casStreamWriteIO::flushDisconnect;
    }
    nBytesActual = (size_t) status;
    return casStreamWriteIO::flushProgress;
}

// sockStreamIO::osdRecv()
casStreamReadIO :: fillCondition
sockStreamIO :: osdRecv ( char * pInBuf, size_t nBytes, // X aCC 361
                      size_t & nBytesActual )
{
    int nchars;
    
    nchars = recv (this->sock, pInBuf, nBytes, 0);
    if ( nchars == 0 ) {
        return casFillDisconnect;
    }
    else if ( nchars < 0 ) {
        int myerrno = SOCKERRNO;
        char buf[64];

        if ( myerrno == SOCK_EWOULDBLOCK ||
                myerrno == SOCK_EINTR ) {
            return casFillNone;
        }

        if ( myerrno == SOCK_ENOBUFS ) {
            errlogPrintf ( 
                "CAS: system low on network buffers - hybernating for 1 second\n" );
            epicsThreadSleep ( 1.0 );
            return casFillNone;
        }

        if (
            myerrno != SOCK_ECONNABORTED &&
            myerrno != SOCK_ECONNRESET &&
            myerrno != SOCK_EPIPE &&
            myerrno != SOCK_ETIMEDOUT ) {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
            ipAddrToA (&this->addr, buf, sizeof(buf));
            errlogPrintf(
		"CAS: client %s disconnected because \"%s\"\n",
                buf, sockErrBuf );
        }
        return casFillDisconnect;
    }
    else {
    	nBytesActual = (size_t) nchars;
        return casFillProgress;
    }
}

// sockStreamIO::forceDisconnect()
void sockStreamIO :: forceDisconnect ()
{
    // !!!! other OS specific wakeup will be required here when we 
    // !!!! switch to a threaded implementation
	if ( ! this->sockHasBeenShutdown ) {
        int status = ::shutdown ( this->sock, SHUT_RDWR );
        if ( status == 0 ) {
            this->sockHasBeenShutdown = true;
        }
        else {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
            errlogPrintf ("CAC TCP socket shutdown error was %s\n", 
                sockErrBuf );
        }
	}
}

// sockStreamIO :: show()
void sockStreamIO :: osdShow (unsigned level) const
{
	printf ( "sockStreamIO at %p\n", 
        static_cast <const void *> ( this ) );
	if (level>1u) {
		char buf[64];
		ipAddrToA(&this->addr, buf, sizeof(buf));
		printf (
			"client=%s, port=%x\n",
			buf, ntohs(this->addr.sin_port));
	}
}

// sockStreamIO::xSsetNonBlocking()
void sockStreamIO :: xSetNonBlocking()
{
	int status;
	osiSockIoctl_t yes = true;

	status = socket_ioctl(this->sock, FIONBIO, &yes); // X aCC 392
	if (status>=0) {
		this->blockingFlag = xIsntBlocking;
	}
	else {
        char sockErrBuf[64];
        epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
		errlogPrintf ( "%s:CAS: TCP non blocking IO set fail because \"%s\"\n", 
				__FILE__, sockErrBuf );
        throw std::runtime_error ( 
            "ioctl() FIONBIO failed" );
	}
}

// sockStreamIO::blockingState()
xBlockingStatus sockStreamIO :: blockingState() const
{
	return this->blockingFlag;
}

// sockStreamIO::incomingBytesPresent()
size_t sockStreamIO :: incomingBytesPresent () const // X aCC 361
{
    int status;
    osiSockIoctl_t nchars = 0;
    
    status = socket_ioctl ( this->sock, FIONREAD, &nchars ); // X aCC 392
    if ( status < 0 ) {
        int localError = SOCKERRNO;
        if (
            localError != SOCK_ECONNABORTED &&
            localError != SOCK_ECONNRESET &&
            localError != SOCK_EPIPE &&
            localError != SOCK_ETIMEDOUT ) 
        {
            char sockErrBuf[64];
            epicsSocketConvertErrnoToString ( sockErrBuf, sizeof ( sockErrBuf ) );
            char buf[64];
            ipAddrToA ( &this->addr, buf, sizeof(buf) );
            errlogPrintf ("CAS: FIONREAD for %s failed because \"%s\"\n",
                buf, sockErrBuf );
        }
        return 0u;
    }
    else if ( nchars < 0 ) {
        return 0u;
    }
    else {
        return ( size_t ) nchars;
    }
}

// sockStreamIO::hostName()
void sockStreamIO :: hostName ( char * pInBuf, unsigned bufSizeIn ) const
{
	ipAddrToA ( & this->addr, pInBuf, bufSizeIn );
}

// sockStreamIO:::optimumBufferSize()
size_t sockStreamIO :: optimumBufferSize () 
{

#if 0
	int n;
	int size;
	int status;

	/* fetch the TCP send buffer size */
	n = sizeof ( size) ;
	status = getsockopt ( this->sock, SOL_SOCKET,
			SO_SNDBUF, (char *) & size, & n );
	if ( status < 0 || n != sizeof ( size ) ) {
		size = 0x400;
	}

	if (size<=0) {
		size = 0x400;
	}
printf("the tcp buf size is %d\n", size);
	return (size_t) size;
#else
// this needs to be MAX_TCP (until we fix the array problem)
	return (size_t) MAX_TCP; // X aCC 392
#endif
}

// sockStreamIO::getFD()
int sockStreamIO :: getFD() const
{
	return this->sock;
}

const caNetAddr sockStreamIO :: getAddr() const
{
	return caNetAddr ( this->addr );
}
