/*************************************************************************\
* Copyright (c) 2008 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to the Software License Agreement
* in the file LICENSE that is included with this distribution.
\*************************************************************************/
/*	Author: Andrew Johnson
 *
 * FIXME: Jobs still to be done:
 *  Finish defining and implement numeric conversion functions
 *  Convert exceptions to use Jeff's approach
 *  Design, implement and test C wrappers
 *  Create the ChunkedString implementation
 */

/* BaseString.h */

#ifndef INC_BaseString_H
#define INC_BaseString_H

#include <stddef.h>	// size_t
#include <stdarg.h>	// va_list

#ifdef __cplusplus
#include <string.h>	// strlen()

#include "shareLib.h"

// **** Base Classes ****

class epicsShareClass StringReader {
protected:
    virtual ~StringReader() {}
public:
    // String length
    virtual size_t size() const = 0;
    
    // Comparison functions
    virtual bool equals(const StringReader & rhs) const = 0;
    virtual bool contains(size_t pos, const StringReader & rhs) const = 0;
    virtual bool contains(size_t pos, const char *buf, size_t len) const = 0;
    
    // Indexed byte access
    virtual char at(size_t pos) const = 0;
    
    // Copy out to buffer, no terminator
    virtual size_t extract(size_t pos, char *buf, size_t cap) const = 0;
    
    // Overloaded operator, not virtual
    const char operator[] (size_t pos) const;
};

class epicsShareClass StringEditor :
    public StringReader {
public:
    // Length adjustment
    virtual void resize(size_t len, char fill = '\0') = 0;
    
    // Buffer storage management
    virtual size_t capacity() const = 0;
    virtual void reserve(size_t cap) = 0;
    
    // Assignment
    virtual void assign(const StringReader & src, size_t off = 0) = 0;
    virtual void assign(const char *buf, size_t len) = 0;
    
    // Append
    virtual void append(char chr) = 0;
    virtual void append(const StringReader & src, size_t off = 0) = 0;
    virtual void append(const char *buf, size_t len) = 0;
    
    // Overloaded operators, not virtual
    StringEditor& operator = (const StringReader &rhs);
    StringEditor& operator = (const char *rhs);
    StringEditor& operator += (char chr);
    StringEditor& operator += (const StringReader &rhs);
    StringEditor& operator += (const char *rhs);
};

// **** Notes for implementors ****

// StringReader::equals() should be implemented by checking that the two
// string sizes match and then calling rhs.contains(pos, str, len) as
// many times as necessary (just once if you have a contiguous buffer)
// to present your whole string buffer.  You may stop prematurely if the
// contains() call returns false, but otherwise you should continue with
// increasing values for pos each time.  This approach allows specialized
// StringReader implementations direct access to your string buffer to do
// a fast hash calculation say.  An implementation that doesn't work this
// way should not expect full interoperability with other specialised
// StringReader classes.


// **** Inline member functions ****

// These inline overloaded operators are deliberately not virtual,
// as they immediately redirect to another function that is.

inline const char StringReader::operator[] (size_t pos) const {
    return at(pos);
}

inline StringEditor& StringEditor::operator = (const StringReader &rhs) {
    assign(rhs);
    return *this;
}

inline StringEditor& StringEditor::operator = (const char *str) {
    assign(str, strlen(str));
    return *this;
}

inline StringEditor& StringEditor::operator += (char chr) {
    append(chr);
    return *this;
}

inline StringEditor& StringEditor::operator += (const StringReader &rhs) {
    append(rhs);
    return *this;
}

inline StringEditor& StringEditor::operator += (const char *str) {
    append(str, strlen(str));
    return *this;
}


// **** Non-member non-friend functions ****

// Equality comparison
inline bool operator == (const StringReader & lhs, const StringReader & rhs) {
    return lhs.equals(rhs);
}
inline bool operator != (const StringReader & lhs, const StringReader & rhs) {
    return ! lhs.equals(rhs);
}


// Numeric conversions: read from string
// FIXME: Error reporting: return status, or throw?
void convert(long & val, const StringReader & str,
	     size_t *ppos = NULL, int base = 0);
void convert(unsigned long & val, const StringReader & str,
	     size_t *ppos = NULL, int base = 0);
void convert(double & val, const StringReader & str,
	     size_t *ppos = NULL);


// Numeric conversions: append to string
void convert(StringEditor & str, long value, int base = 10);
void convert(StringEditor & str, unsigned long value, int base = 10);
void convert(StringEditor & str, double value); // FIXME: prec, ... etc.

void bsvprintf(StringEditor & str, const char *fmt, va_list args);
void bsprintf(StringEditor & str, const char *fmt, ...);

#endif	/* __cplusplus */

/**** C Wrappers ****/

/* TBD ... */


/*****************************************************************************

Helper Implementations
----------------------

ConString.h:
    ConStringReader : Implements most string operations where the underlying
	string buffer is contiguous.  Concrete implementations may derive from
	this class to fully implement the StringReader interface.
    ConStringEditor : Like ConStringReader, but for strings that must be
	modifyable through the string interface.

SegString.h:
    SegStringReader : Implements most string operations where the underlying
	string buffer is segmented.  Concrete implementations may derive from
	this class to fully implement the StringReader interface.
    SegStringEditor : Like SegStringReader, but for strings that must be
	modifyable through the string interface.


Concrete Implementations
------------------------

ConstString : Concrete implementation of StringReader for wrapping string
    literals and other const char* buffers.  Can also be wrapped around a
    regular char array to provide a StringReader interface for it, although
    if the string's length changes a new wrapper object must be created.

FixedString : Template implementation of StringEditor, using a fixed buffer
    which is stored in the class, whose size is set by a template parameter.

VariableString : Implementation of StringEditor, using new[] and delete[] to
    provide a contiguous, variable length string. The buffer increases in size
    by multiples of 2 as necessary, but does not normally decrease.

StandardString : Implementation of StringEditor using a std::string buffer.

ChunkedString : Implementation of StringEditor, using fixed size buffer
    segments that are allocated off and returned to a freelist as necessary.
    This is the type I want to use in the IOC database for string fields.

*/

#endif	/* INC_BaseString_H */
