/*************************************************************************\
* Copyright (c) 2008 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to the Software License Agreement
* in the file LICENSE that is included with this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// ConString.cpp

#include <string.h>
#include <stdexcept>
#include <algorithm>

#define epicsExportSharedSymbols
#include "ConString.h"

// **** ConStringReader ****

// required for windows dll export
ConStringReader :: ConStringReader () {}
ConStringReader :: ~ConStringReader () {}

// Comparison functions
bool ConStringReader::equals(const StringReader & rhs) const {
    if (size() != rhs.size()) return false;
    return rhs.contains(0, getBuffer(), size());
}
bool ConStringReader::contains(size_t pos, const StringReader & rhs) const {
    size_t len = size();
    if (len >= pos) {
	len -= pos;
	size_t rhsLen = rhs.size();
	if (len >= rhsLen)
	    return rhs.contains(0, getBuffer() + pos, rhsLen);
    }
    return false;
}
bool ConStringReader::contains(size_t pos, const char *str, size_t len) const {
    if (!str || len + pos > size()) return false;
    return ! memcmp(getBuffer() + pos, str, len);
}

// Individual byte access
char ConStringReader::at(size_t pos) const {
    if (pos >= size()) throw std::out_of_range("FIXME");
    return getBuffer()[pos];
}

// Copy out to buffer
size_t ConStringReader::extract(size_t pos, char *buf, size_t cap) const {
    size_t len = size();
    if (!buf || len <= pos) return 0;
    len -= pos;
    if (cap < len) len = cap;
    memcpy(buf, getBuffer() + pos, len);
    return len;
}


// **** ConStringEditor ****

// required for windows dll export
ConStringEditor :: ConStringEditor () {}
ConStringEditor :: ~ConStringEditor () {}

// Comparison functions
bool ConStringEditor::equals(const StringReader & rhs) const {
    if (size() != rhs.size()) return false;
    return rhs.contains(0, getBuffer(), size());
}
bool ConStringEditor::contains(size_t pos, const StringReader & rhs) const {
    size_t len = size();
    if (len >= pos) {
	len -= pos;
	size_t rhsLen = rhs.size();
	if (len >= rhsLen)
	    return rhs.contains(0, getBuffer() + pos, rhsLen);
    }
    return false;
}
bool ConStringEditor::contains(size_t pos, const char *str, size_t len) const {
    if (!str || len + pos > size()) return false;
    return ! memcmp(getBuffer() + pos, str, len);
}

// Individual byte access
char ConStringEditor::at(size_t pos) const {
    if (pos >= size()) throw std::out_of_range("FIXME");
    return getBuffer()[pos];
}

// Copy out to buffer
size_t ConStringEditor::extract(size_t pos, char *buf, size_t cap) const {
    size_t len = size();
    if (!buf || len <= pos) return 0;
    len -= pos;
    if (cap < len) len = cap;
    memcpy(buf, getBuffer() + pos, len);
    return len;
}

// Length adjustment
void ConStringEditor::resize(size_t len, char fill) {
    if (!getBuffer() || len > size()) {
	reserve(len);
	memset(getBuffer() + size(), fill, len - size());
    }
    setSize(len);
}

// Assignment
void ConStringEditor::assign(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len > off) {
	len -= off;
	reserve(len);
	src.extract(off, getBuffer(), len);
    } else {
	len = 0;
    }
    setSize(len);
}
void ConStringEditor::assign(const char *str, size_t len) {
    if (len) {
    	if (str) {
	    reserve(len);
	    memcpy(getBuffer(), str, len);
	} else {
	    len = 0;
    	}
    }
    setSize(len);
}

// Append
void ConStringEditor::append(char chr) {
    size_t pos = size();
    size_t nsz = pos + 1;
    reserve(nsz);
    getBuffer()[pos] = chr;
    setSize(nsz);
}
void ConStringEditor::append(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len > off) {
	len -= off;
	size_t pos = size();
	size_t nsz = pos + len;
	reserve(nsz);
	src.extract(off, getBuffer() + pos, len);
	setSize(nsz);
    }
}
void ConStringEditor::append(const char *str, size_t len) {
    if (str && len) {
	size_t pos = size();
	size_t nsz = pos + len;
	reserve(nsz);
	memcpy(getBuffer() + pos, str, len);
	setSize(nsz);
    }
}
