/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// FixedString.h

#ifndef INC_FixedString_H
#define INC_FixedString_H

#include <stdexcept>
#include <cassert>

#include "ConString.h"

// This implementation is a template for fixed size strings. There is a
// base class that provides most of the code, the template adds storage

template <int CAP>
class FixedString :
    public ConStringEditor {
private:
    // Data
    char str_[CAP+1];	// character buffer
    size_t len_;	// Length, excluding the \0 byte

    // Private methods for base class use
    virtual const char * getBuffer() const;
    virtual char * getBuffer();
    virtual void setSize(size_t len);

public:
    // Constructors
    FixedString();
    FixedString(const FixedString<CAP> &rhs);
    FixedString(const StringReader &rhs);
    FixedString(const char *str, size_t len = 0);

    // String length
    virtual size_t size(void) const;

    // Access for C compatibility
    const char *c_str() const;
    char *c_str();

    // Buffer storage management
    virtual size_t capacity() const;
    virtual void reserve(size_t cap);

    // Assignment
    FixedString<CAP>& operator = (const FixedString<CAP> &rhs);
    FixedString<CAP>& operator = (const StringReader &rhs);
};


// Inline member functions

// Private methods for base class use
template <int CAP>
inline const char * FixedString<CAP>::getBuffer() const {
    return str_;
}
template <int CAP>
inline char * FixedString<CAP>::getBuffer() {
    return str_;
}
template <int CAP>
inline void FixedString<CAP>::setSize(size_t len) {
    assert(CAP >= len);
    len_ = len;
    str_[len] = '\0';	// Maintain string terminator
}

// Constructors
template <int CAP>
inline FixedString<CAP>::FixedString() {
    setSize(0);
}
template <int CAP>
inline FixedString<CAP>::FixedString(const FixedString<CAP> &rhs) :
    len_(rhs.len_) {
    memcpy(str_, rhs.str_, len_ + 1);	// Copies \0 byte as well
}
template <int CAP>
inline FixedString<CAP>::FixedString(const StringReader &rhs) {
    assign(rhs);
}
template <int CAP>
inline FixedString<CAP>::FixedString(const char *str, size_t len) {
    if (!str) throw std::invalid_argument("FIXME");
    if (!len) len = strlen(str);
    if (len > CAP) len = CAP;	// Truncate string, assume expert user
    memcpy(str_, str, len);
    setSize(len);
}

// Assignment
template <int CAP>
inline FixedString<CAP>& FixedString<CAP>::operator = (const FixedString<CAP> &rhs) {
    len_ = rhs.len_;
    memcpy(str_, rhs.str_, len_ + 1);
    return *this;
}
template <int CAP>
inline FixedString<CAP>& FixedString<CAP>::operator = (const StringReader &rhs) {
    assign(rhs);
    return *this;
}

// String length
template <int CAP>
inline size_t FixedString<CAP>::size(void) const {
    return len_;
}

// Buffer storage management
template <int CAP>
inline size_t FixedString<CAP>::capacity() const {
    return CAP;
}
template <int CAP>
void FixedString<CAP>::reserve(size_t req) {
    if (req > CAP) throw std::length_error("FIXME");
}

// Access for C compatibility
template <int CAP>
inline const char * FixedString<CAP>::c_str() const {
    return str_;
}
template <int CAP>
inline char * FixedString<CAP>::c_str() {
    return str_;
}

#endif
