/*************************************************************************\
* Copyright (c) 2008 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to the Software License Agreement
* in the file LICENSE that is included with this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// SegString.cpp

#include <assert.h>
#include <string.h>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "SegString.h"

// **** SegStringReader ****

// Comparison functions
bool SegStringReader::equals(const StringReader & rhs) const {
    size_t len = size();
    if (len != rhs.size()) return false;
    const char *seg;
    size_t dim, pos = 0;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (! rhs.contains(pos, seg, dim)) return false;
	pos += dim;
	len -= dim;
    }
    return true;
}
bool SegStringReader::contains(size_t pos, const StringReader & str) const {
    size_t len = size();
    if (len < pos) return false;
    len -= pos;
    if (len < str.size()) return false;
    const char *seg;
    size_t dim, off = 0;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (! str.contains(off, seg, dim)) return false;
	off += dim;
	pos += dim;
	len -= dim;
    }
    return true;
}
bool SegStringReader::contains(size_t pos, const char *buf, size_t len) const {
    if (!buf || len + pos > size()) return false;
    const char *seg;
    size_t dim;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (memcmp(seg, buf, dim)) return false;
	buf += dim;
	pos += dim;
	len -= dim;
    }
    return true;
}

// Individual byte access
char SegStringReader::at(size_t pos) const {
    if (pos >= size())
	throw std::out_of_range("FIXME");
    const char *seg;
    size_t dim = 0;
    getSegment(seg, dim, pos);
    assert(dim > 0);
    return *seg;
}

// Copy out to buffer
size_t SegStringReader::extract(size_t pos, char *buf, size_t cap) const {
    size_t len = size();
    if (pos >= len) return 0;
    len -= pos;
    if (cap > len) cap = len;
    len = 0;
    size_t dim;
    const char *seg;
    while (cap && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (cap < dim) dim = cap;
	memcpy(buf, seg, dim);
	pos += dim;
	len += dim;
	buf += dim;
	cap -= dim;
    }
    return len;
}


// **** SegStringEditor ****

// Comparison functions
bool SegStringEditor::equals(const StringReader &rhs) const {
    size_t len = size();
    if (len != rhs.size()) return false;
    const char *seg;
    size_t dim, pos = 0;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (! rhs.contains(pos, seg, dim)) return false;
	pos += dim;
	len -= dim;
    }
    return true;
}
bool SegStringEditor::contains(size_t pos, const StringReader &str) const {
    size_t len = str.size();
    if (pos + len > size()) return false;
    const char *seg;
    size_t dim, off = 0;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (! str.contains(off, seg, dim)) return false;
	off += dim;
	pos += dim;
	len -= dim;
    }
    return true;
}
bool SegStringEditor::contains(size_t pos, const char *buf, size_t len) const {
    if (!buf || len + pos > size()) return false;
    const char *seg;
    size_t dim;
    while (len && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (len < dim) dim = len;
	if (memcmp(seg, buf, dim)) return false;
	buf += dim;
	pos += dim;
	len -= dim;
    }
    return true;
}

// Individual byte access
char SegStringEditor::at(size_t pos) const {
    if (pos >= size())
	throw std::out_of_range("FIXME");
    const char *seg;
    size_t dim = 0;
    getSegment(seg, dim, pos);
    assert(dim > 0);
    return *seg;
}

// Copy out to buffer
size_t SegStringEditor::extract(size_t pos, char *buf, size_t cap) const {
    size_t len = size();
    if (pos >= len) return 0;
    len -= pos;
    if (cap > len) cap = len;
    len = 0;
    size_t dim;
    const char *seg;
    while (cap && getSegment(seg, dim, pos)) {
	assert(dim > 0);
	if (cap < dim) dim = cap;
	memcpy(buf, seg, dim);
	pos += dim;
	len += dim;
	buf += dim;
	cap -= dim;
    }
    return len;
}

// Length adjustment
void SegStringEditor::resize(size_t len, char fill) {
    size_t pos = size();
    if (len > pos) {
	reserve(len);
	len -= pos;
	char *seg;
	size_t dim;
	while (len && getSegment(seg, dim, pos)) {
	    assert(dim > 0);
	    if (len < dim) dim = len;
	    memset(seg, fill, dim);
	    pos += dim;
	    len -= dim;
	}
	len = pos;
    }
    setSize(len);
}

// Assignment
void SegStringEditor::assign(const StringReader & src, size_t off) {
    size_t pos = 0;
    size_t len = src.size();
    if (len > off) {
	len -= off;
	reserve(len);
	char *seg;
	size_t dim;
	while (len && getSegment(seg, dim, pos)) {
	    assert(dim > 0);
	    if (len < dim) dim = len;
	    src.extract(off, seg, dim);
	    off += dim;
	    pos += dim;
	    len -= dim;
	}
    }
    setSize(pos);
}

void SegStringEditor::assign(const char *buf, size_t len) {
    size_t pos = 0;
    if (buf) {
	reserve(len);
	char *seg;
	size_t dim;
	while (len && getSegment(seg, dim, pos)) {
	    assert(dim > 0);
	    if (len < dim) dim = len;
	    memcpy(seg, buf, dim);
	    buf += dim;
	    pos += dim;
	    len -= dim;
	}
    }
    setSize(pos);
}

// Append
void SegStringEditor::append(char chr) {
    size_t pos = size();
    size_t nsz = pos + 1;
    reserve(nsz);
    char *seg;
    size_t dim;
    if (getSegment(seg, dim, pos)) {
    	assert(dim > 0);
    	*seg = chr;
    }
    setSize(nsz);
}
void SegStringEditor::append(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len > off) {
	len -= off;
	size_t pos = size();
	reserve(pos + len);
	char *seg;
	size_t dim;
	while (len && getSegment(seg, dim, pos)) {
	    assert(dim > 0);
	    if (len < dim) dim = len;
	    src.extract(off, seg, dim);
	    off += dim;
	    pos += dim;
	    len -= dim;
	}
	setSize(pos);
    }
}
void SegStringEditor::append(const char *buf, size_t len) {
    if (buf && len) {
	size_t pos = size();
	reserve(pos + len);
	char *seg;
	size_t dim;
	while (len && getSegment(seg, dim, pos)) {
	    assert(dim > 0);
	    if (len < dim) dim = len;
	    memcpy(seg, buf, dim);
	    buf += dim;
	    pos += dim;
	    len -= dim;
	}
	setSize(pos);
    }
}
