/*************************************************************************\
* Copyright (c) 2009 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to the Software License Agreement
* in the file LICENSE that is included with this distribution.
\*************************************************************************/
//      Author: Andrew Johnson

// SegTree.cpp

#define epicsExportSharedSymbols
#include "SegTree.h"

// A SegLeaf<N,M> is a SegTree<N,M> that holds M bytes of buffer data.

template <int N, int M>
class SegLeaf :
    public SegTree<N, M> {
private:
    // Data
    char data_[M];

    // Constructor for factory use
    SegLeaf() {}

    // Virtual methods
    void recycle();
    size_t capacity() const;
    size_t limit() const;
    size_t grow(size_t capacity);
    void getSegment(const char *&buf, size_t &len, size_t idx) const;
    void getSegment(char *&buf, size_t &len, size_t idx);

    friend SegTree<N, M>* SegTree<N, M>::Factory(size_t limit);
};


// A SegTwig<N,M> is a SegTree<N,M> that points to up to N child SegTrees.
// SegTwig has the following invariant characteristics:
//      1. node_[0] always points to an existing child SegTree.
//      2. Child branches are all the same length.

template <int N, int M>
class SegTwig :
    public SegTree<N, M> {
private:
    // Data
    size_t cap_;
    size_t seg_;
    SegTree<N, M> *node_[N];

    // Constructors for factory use
    SegTwig(size_t limit);
    SegTwig(SegTree<N, M> *node);

    // Virtual methods
    void recycle();
    size_t capacity() const;
    size_t limit() const;
    size_t grow(size_t capacity);
    void getSegment(const char *&buf, size_t &len, size_t idx) const;
    void getSegment(char *&buf, size_t &len, size_t idx);

    friend SegTree<N, M>* SegTree<N, M>::Factory(size_t limit);
    friend void SegTree<N, M>::Factory(SegTree<N, M>* &tree, size_t cap);
};


// SegTree methods

// Protected Factory
template <int N, int M>
inline SegTree<N, M>* SegTree<N, M>::Factory(size_t limit) {
    if (limit > M)
        return new SegTwig<N, M>(limit);                // FIXME: new
    else
        return new SegLeaf<N, M>;                       // FIXME: new
}

// Public Factory
template <int N, int M>
void SegTree<N, M>::Factory(SegTree<N, M>* &tree, size_t cap) {
    if (!tree) {
        tree = SegTree<N, M>::Factory(cap);             // FIXME: new
    } else {
        while (tree->limit() < cap) {
            tree = new SegTwig<N, M>(tree);             // FIXME: new
        }
    }
    tree->grow(cap);
}


// SegLeaf methods

// Storage management
template <int N, int M>
void SegLeaf<N, M>::recycle() {
    delete this;                                        // FIXME: delete
}

template <int N, int M>
size_t SegLeaf<N, M>::capacity() const {
    return M;
}

template <int N, int M>
size_t SegLeaf<N, M>::limit() const {
    return M;
}

template <int N, int M>
size_t SegLeaf<N, M>::grow(size_t target) {
    return 0;   // Can't grow...
}

// Data Accessors
template <int N, int M>
void SegLeaf<N, M>::getSegment(char *&buf, size_t &len, size_t idx) {
    buf = &data_[idx];
    len = M - idx;
}

template <int N, int M>
void SegLeaf<N, M>::getSegment(const char *&buf, size_t &len, size_t idx) const {
    buf = &data_[idx];
    len = M - idx;
}


// SegTwig methods

// Constructors
template <int N, int M>
SegTwig<N, M>::SegTwig(size_t limit) {
    SegTree<N, M> *node = SegTree<N, M>::Factory(limit / N);
    cap_ = node->capacity();
    seg_ = node->limit();
    node_[0] = node;
}

template <int N, int M>
SegTwig<N, M>::SegTwig(SegTree<N, M> *node) :
    cap_(node->capacity()), seg_(node->limit()) {
    node_[0] = node;
}

// Recycle
template <int N, int M>
void SegTwig<N, M>::recycle() {
    for (int i = 0; i < N && node_[i]; i++) {
        node_[i]->recycle();
        node_[i] = NULL;
    }
    delete this;                                        // FIXME: delete
}

// Storage management
template <int N, int M>
size_t SegTwig<N, M>::capacity() const {
    return cap_;
}

template <int N, int M>
size_t SegTwig<N, M>::limit() const {
    return N * seg_;
}

template <int N, int M>
size_t SegTwig<N, M>::grow(size_t target) {
    size_t lim = limit();
    if (target > lim)
        target = lim;
    size_t inc = 0;
    int i = cap_ / seg_;
    while (target > cap_ + inc) {
        if (!node_[i]) {
            node_[i] = SegTree<N, M>::Factory(seg_);
            inc += M;
        }
        inc += node_[i]->grow(target - i * seg_);
        i++;
    }
    cap_ += inc;
    return inc;
}

// Data Accessors
template <int N, int M>
void SegTwig<N, M>::getSegment(char *&buf, size_t &len, size_t idx) {
    int i = idx / seg_;
    node_[i]->getSegment(buf, len, idx - i * seg_);
}

template <int N, int M>
void SegTwig<N, M>::getSegment(const char *&buf, size_t &len, size_t idx) const {
    int i = idx / seg_;
    node_[i]->getSegment(buf, len, idx - i * seg_);
}


// Instantiations needed by ChunkedString
template class SegTree<16, 64>;
template class SegLeaf<16, 64>;
template class SegTwig<16, 64>;

