/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// StandardString.cpp

#include <assert.h>
#include <string.h>
#include <algorithm>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "StandardString.h"

// This implementation wraps a std::string and implements all of the
// StringEditor functionality.  It provides c_str() for interfacing to
// standard C library routines, std_str() for access to the underlying
// string object, and a swap() function.

// Constructor
StandardString::StandardString(const char *str, size_t len) :
    str_() {
    if (!str) throw std::invalid_argument("FIXME");
    if (!len) len = strlen(str);
    str_.assign(str, len);
}

// Comparison
bool StandardString::equals(const StringReader & rhs) const {
    if (size() != rhs.size()) return false;
    return rhs.contains(0, str_.c_str(), size());
}
bool StandardString::contains(size_t pos, const StringReader & rhs) const {
    size_t len = size();
    if (len >= pos) {
	len -= pos;
	size_t rhsLen = rhs.size();
	if (len >= rhsLen)
	    return rhs.contains(0, str_.c_str() + pos, rhsLen);
    }
    return false;
}
bool StandardString::contains(size_t pos, const char *str, size_t len) const {
    if (!str || len + pos > size()) return false;
#ifdef vxWorks
    return ! str_.substr(pos, len).compare(str, len);
#else
    return ! str_.compare(pos, len, str, len);
#endif
}

// Individual byte access
char StandardString::at(size_t pos) const {
    return str_.at(pos);
}

// Copy out to buffer
size_t StandardString::extract(size_t pos, char *buf, size_t cap) const {
    return str_.copy(buf, cap, pos);
}

// Length adjustment
void StandardString::resize(size_t len, char fill) {
    str_.resize(len, fill);
}

// Buffer storage management
void StandardString::reserve(size_t req) {
    str_.reserve(req);
}

// Assignment
void StandardString::assign(const StringReader & src, size_t off) {
    StandardString tgt;     // Allow for self-assignment...
    tgt.append(src, off);
    str_.swap(tgt.str_);
}

void StandardString::assign(const char *str, size_t len) {
    if (str && len) {
	str_.assign(str, len);
    } else {
#ifdef vxWorks
	str_.erase();
#else
	str_.clear();
#endif
    }
}

// Append
void StandardString::append(char chr) {
    str_.append(1, chr);
}

void StandardString::append(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len <= off) return;
    reserve(str_.size() + len - off);
    char buf[80];
    while (off < len) {
	size_t seg = src.extract(off, buf, sizeof(buf));
	str_.append(buf, seg);
	off += seg;
    }
}

void StandardString::append(const char *str, size_t len) {
    if (str && len) {
	str_.append(str, len);
    }
}
