/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// StandardString.h

#ifndef INC_StandardString_H
#define INC_StandardString_H

#include <string>
#include "BaseString.h"

// This implementation wraps a std::string

class epicsShareClass StandardString :
    public StringEditor {
private:
    std::string str_;

public:
    // Constructors
    StandardString();
    StandardString(const StandardString &rhs);
    StandardString(const StringReader &rhs);
    StandardString(const std::string &rhs);
    StandardString(const char *str, size_t len = 0);

    // Destructor
    virtual ~StandardString();

// StringReader virtual functions
    // String length
    virtual size_t size(void) const;

    // Comparison functions
    virtual bool equals(const StringReader & rhs) const;
    virtual bool contains(size_t pos, const StringReader & rhs) const;
    virtual bool contains(size_t pos, const char *str, size_t len) const;

    // Individual byte access
    virtual char at(size_t pos) const;

    // Copy out to buffer
    virtual size_t extract(size_t pos, char *buf, size_t cap) const;

// StringEditor virtual functions
    // Length adjustment
    virtual void resize(size_t len, char fill = '\0');

    // Buffer storage management
    virtual size_t capacity() const;
    virtual void reserve(size_t cap);

    // Assignment
    virtual void assign(const StringReader & src, size_t off = 0);
    virtual void assign(const char *str, size_t len);

    // Append
    virtual void append(char chr);
    virtual void append(const StringReader & src, size_t off = 0);
    virtual void append(const char *str, size_t len);

// Other functions
    // Access for C compatibility
    const char *c_str() const;

    // Access for C++ compatibility
    const std::string& std_str() const;
    std::string& std_str();

    // Swap storage
    void swap(StandardString & src);

    // Overloaded operators
    StandardString& operator = (const StandardString &rhs);
    StandardString& operator = (const StringReader &rhs);
    StandardString& operator = (const char *rhs);
    StandardString& operator += (char chr);
    StandardString& operator += (const StandardString &rhs);
    StandardString& operator += (const StringReader &rhs);
    StandardString& operator += (const char *rhs);
};


// Inline member functions

// Constructors
inline StandardString::StandardString() :
    str_() {}

inline StandardString::StandardString(const StringReader &rhs) :
    str_() {
    assign(rhs);
}

inline StandardString::StandardString(const StandardString &rhs) :
    str_(rhs.str_) {}

inline StandardString::StandardString(const std::string &rhs) :
    str_(rhs) {}

// Destructor
inline StandardString::~StandardString() {}

// String length
inline size_t StandardString::size() const {
    return str_.size();
}

// Buffer storage management
inline size_t StandardString::capacity() const {
    return str_.capacity();
}

// Access for C compatibility
inline const char * StandardString::c_str() const {
    return str_.c_str();
}

// Access for C++ compatibility
inline const std::string& StandardString::std_str() const {
    return str_;
}

inline std::string& StandardString::std_str() {
    return str_;
}

// Swap strings
inline void StandardString::swap(StandardString & src) {
    str_.swap(src.str_);
}

// Overloaded operators
inline StandardString& StandardString::operator = (const StandardString &rhs) {
    str_ = rhs.str_;
    return *this;
}

inline StandardString& StandardString::operator = (const StringReader &rhs) {
    assign(rhs);
    return *this;
}

inline StandardString& StandardString::operator = (const char *rhs) {
    str_.assign(rhs);
    return *this;
}

inline StandardString& StandardString::operator += (char chr) {
    str_.append(1, chr);
    return *this;
}


inline StandardString& StandardString::operator += (const StandardString &rhs) {
    str_ += rhs.str_;
    return *this;
}

inline StandardString& StandardString::operator += (const StringReader &rhs) {
    append(rhs);
    return *this;
}

inline StandardString& StandardString::operator += (const char *rhs) {
    str_.append(rhs);
    return *this;
}


#endif
