/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// VariableString.cpp

#include <assert.h>
#include <string.h>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "VariableString.h"

// This implementation uses new/delete for buffer management, and increases
// the buffer size by doubling only.  It maintains a \0 byte after the last
// character in the buffer which is not included in the reported length or
// string capacity, hence all the cap*2+1 math in reserve(), and it provides
// a c_str() function for interfacing to standard C library routines.

// Constructor
VariableString::VariableString(const char *str, size_t len) :
    str_(NULL), len_(0), cap_(0) {
    if (str) {
	if (!len) len = strlen(str);
	reserve(len);
	memcpy(str_, str, len);
	setSize(len);
    }
}

// Destructor
VariableString::~VariableString() {
    if (str_) {
	delete [] str_;
	str_ = NULL;
    }
}

// Comparison
bool VariableString::equals(const StringReader & rhs) const {
    if (size() != rhs.size()) return false;
    return rhs.contains(0, str_, size());
}
bool VariableString::contains(size_t pos, const StringReader & rhs) const {
    size_t len = size();
    if (len >= pos) {
	len -= pos;
	size_t rhsLen = rhs.size();
	if (len >= rhsLen)
	    return rhs.contains(0, str_ + pos, rhsLen);
    }
    return false;
}
bool VariableString::contains(size_t pos, const char *str, size_t len) const {
    if (!str || len + pos > size()) return false;
    return ! memcmp(str_ + pos, str, len);
}

// Individual byte access
char VariableString::at(size_t pos) const {
    if (pos >= size()) throw std::out_of_range("FIXME");
    return str_[pos];
}

// Copy out to buffer
size_t VariableString::extract(size_t pos, char *buf, size_t cap) const {
    size_t len = size();
    if (!buf || len <= pos) return 0;
    len -= pos;
    if (cap < len) len = cap;
    memcpy(buf, str_ + pos, len);
    return len;
}

// Length adjustment
void VariableString::resize(size_t len, char fill) {
    if (!str_ || len > size()) {
	reserve(len);
	memset(str_ + size(), fill, len - size());
    }
    setSize(len);
}

// Buffer storage management
void VariableString::reserve(size_t req) {
    if (req <= cap_) return;
    size_t newcap = cap_ ? cap_ : 15u;	// Minimum capacity
    while (newcap < req)
	newcap = newcap * 2 + 1;	// Grow exponentially
    char *str = new char [newcap + 1];  // Add space for '\0'
    if (str_ && len_ > 0) {
	memcpy(str, str_, len_+1);
	delete [] str_;
    }
    str_ = str;
    cap_ = newcap;
}

// Assignment
void VariableString::assign(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len > off) {
	len -= off;
	reserve(len);
	len = src.extract(off, str_, len);
    } else {
	len = 0;
    }
    setSize(len);
}
void VariableString::assign(const char *str, size_t len) {
    if (len) {
    	if (str) {
	    reserve(len);
	    memcpy(str_, str, len);
	} else {
	    len = 0;
    	}
    }
    setSize(len);
}

// Append
void VariableString::append(char chr) {
    size_t len = size() + 1;
    reserve(len);
    str_[size()] = chr;
    setSize(len);
}
void VariableString::append(const StringReader & src, size_t off) {
    size_t len = src.size();
    if (len <= off) return;
    len = len - off + size();
    reserve(len);
    src.extract(off, str_ + size(), src.size());
    setSize(len);
}
void VariableString::append(const char *str, size_t len) {
    if (str && len) {
	reserve(size() + len);
	memcpy(str_ + size(), str, len);
	setSize(len_ + len);
    }
}

// Swap storage
void VariableString::swap(VariableString & src) {
    char * str = str_;
    size_t len = len_;
    size_t cap = cap_;
    str_ = src.str_;
    len_ = src.len_;
    cap_ = src.cap_;
    src.str_ = str;
    src.len_ = len;
    src.cap_ = cap;
}
