/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// VariableString.h

#ifndef INC_VariableString_H
#define INC_VariableString_H

#include "BaseString.h"

// This implementation uses new/delete for buffer management, and increases
// the buffer size in powers of 2 only.

class epicsShareClass VariableString :
    public StringEditor {
private:
    // Data
    char *str_;     // character buffer, may be NULL
    size_t len_;    // Length, excluding the \0 byte
    size_t cap_;    // str_ has 1 extra byte allocated

    // Private methods for local class use
    virtual void setSize(size_t len);

public:
    // Constructors
    VariableString(const VariableString &rhs);
    VariableString(const StringReader &rhs);
    VariableString(const char *str = NULL, size_t len = 0);

    // Destructor
    virtual ~VariableString();

// StringReader virtual functions
    // String length
    virtual size_t size(void) const;

    // Comparison functions
    virtual bool equals(const StringReader & rhs) const;
    virtual bool contains(size_t pos, const StringReader & rhs) const;
    virtual bool contains(size_t pos, const char *str, size_t len) const;

    // Individual byte access
    virtual char at(size_t pos) const;

    // Copy out to buffer
    virtual size_t extract(size_t pos, char *buf, size_t cap) const;

// StringEditor virtual functions
    // Length adjustment
    virtual void resize(size_t len, char fill = '\0');

    // Buffer storage management
    virtual size_t capacity() const;
    virtual void reserve(size_t cap);

    // Assignment
    virtual void assign(const StringReader & src, size_t off = 0);
    virtual void assign(const char *str, size_t len);

    // Append
    virtual void append(char chr);
    virtual void append(const StringReader & src, size_t off = 0);
    virtual void append(const char *str, size_t len);

// Other functions
    // Access for C compatibility
    const char *c_str() const;

    // Swap storage
    void swap(VariableString & src);

    // Overloaded operators
    VariableString& operator = (const VariableString &rhs);
    VariableString& operator = (const StringReader &rhs);
};


// Inline member functions

// Private methods
inline void VariableString::setSize(size_t len) {
    len_ = len;
    str_[len] = '\0';	// Maintain string terminator
}

// Constructors
inline VariableString::VariableString(const StringReader &rhs) :
    str_(NULL), len_(0), cap_(0) {
    assign(rhs);
}

inline VariableString::VariableString(const VariableString &rhs) :
    str_(NULL), len_(0), cap_(0) {
    assign(rhs);
}

// String length
inline size_t VariableString::size() const {
    return len_;
}

// Buffer storage management
inline size_t VariableString::capacity() const {
    return cap_;
}

// Assignment
inline VariableString& VariableString::operator = (const VariableString &rhs) {
    assign(rhs);
    return *this;
}

inline VariableString& VariableString::operator = (const StringReader &rhs) {
    assign(rhs);
    return *this;
}

// Access for C compatibility
inline const char * VariableString::c_str() const {
    return str_;
}

#endif
