/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*  
 *
 *                              
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *                                  
 *  Copyright, The Regents of the University of California.
 *                                  
 *           
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 *	505 665 1831
 */

#ifndef autoDestroyPtrh
#define autoDestroyPtrh

//
// autoDestroyPtr's sole purpose is to make certain that we
// clean up properly if an exception is thrown while in the
// middle of constructing a complex object. The class does
// not provide transfer of ownership features so its
// behavior and implementation are less complex compared
// to auto_ptr.
//
// o more flexible than auto_ptr because, when the target is 
// destroyed, it calls a member function instead of delete
// o it works only when the target is a class (it does not 
// work with primitive atomic or array data)
// o unlike auto_ptr ownership does not transfer during 
// assignment, and in fact the code will not compile if
// assignment between two autoDestroyPtr objects is attempted.
// For many situations that behavior avoids bugs.
// o autoDestroyPtr is more efficent compared to auto_ptr 
// because the target is a reference.  
//
template < class T, void ( T :: * PM ) ( void ) 
                        = & T :: destroy >
class autoDestroyPtr {
public:
    autoDestroyPtr ( T & );
    ~autoDestroyPtr ();
    T & operator * () const;
    T * operator -> () const;
private:
    T & _target;
    // not implemented
    autoDestroyPtr < T, PM > & 
        operator = ( 
            const autoDestroyPtr < T, PM > & );
    autoDestroyPtr < T, PM > ( 
            const autoDestroyPtr < T, PM > & );
};

template < class T, void ( T :: * PM ) ( void ) >
inline autoDestroyPtr < T, PM > :: 
    autoDestroyPtr ( T & target ) :
        _target ( target ) 
{
}

template < class T, void ( T :: * PM ) ( void ) >
inline autoDestroyPtr < T, PM > :: 
    ~autoDestroyPtr ()
{
    ( _target.* PM )();
}

template < class T, void ( T :: * PM ) ( void ) >
inline T & autoDestroyPtr < T, PM > :: 
    operator * () const
{
    return _target;
}

template < class T, void ( T :: * PM ) ( void ) >
inline T * autoDestroyPtr < T, PM > :: 
    operator -> () const
{
    return & _target;
}

#endif // #ifdef autoDestroyPtrh
