/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

#ifndef tsFreeList_h
#define tsFreeList_h

/*
 *
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *
 *  Copyright, 1986, The Regents of the University of California.
 *
 *  Author: Jeff Hill
 */

//
// To use this class:
//
// tsFreeList < T > myFreeList ();
// T * p = new ( myFreeList ) T ( ... );
// if ( polymorphicDestroy ) {
//   // see invokeDestructor < T > ( *p)
//   // and also FreeListAutoPtr < FreeList < T > >
//   myFreeList.deleteProduct ( p );
// } else {
//   p->~T ();
//   :: operator delete ( p, myFreeList );
// }
//
// NOTES:
//
// 1) A Mutex interface reference to a NOOP 
// Mutex implementation may be passed to the 
// free list constructor if mutual exclusion 
// isnt required, and its overhead needs to 
// be minimized
//

//
// TODO: 
// 1) Maybe this should allow free list chaining so that a free 
// list (in a different lock domain) is used to provide any new
// Chunk, and the list of chunks is returned to this free list 
// when the free list is destroyed.
//

#include <new>
#include <memory>
#include <iostream>
#include <stdexcept>
#include <climits>
#include <cassert>
#include <typeinfo>

template < typename T, size_t N = 256 > struct FreeListAttr;

template < typename T, typename ATTR = FreeListAttr < T > > 
        class FreeList;

// allow for class specific overides
template < typename T >
inline void invokeDestructor ( T & target );

// its important to make the arguments here selective
// (they are based on FREELIST rather than T)
template < typename T, typename ATTR >
void * operator new ( size_t, FreeList < T, ATTR > & );

// its important to make the arguments here selective
// (they are based on FREELIST rather than T)
template < typename T, typename ATTR >
void * operator new ( size_t, const std :: nothrow_t &,
                        FreeList < T, ATTR > & );

// this is called by the compiler when handling 
// exception thrown in T's constructor, but it 
// will also be explicitly called by the user when 
// reclaiming storage for the object
// (arguments here need to be symetric with new)
template < typename T, typename ATTR >
void operator delete ( void *, FreeList < T, ATTR > & );

// this is called by the compiler when handling 
// exception thrown in T's constructor, but it 
// will also be explicitly called by the user when 
// reclaiming storage for the object
// (arguments here need to be symetric with new)
template < typename T, typename ATTR >
void operator delete ( void *, const std :: nothrow_t &,
                        FreeList < T, ATTR > & );

// Implementation Notes
// --------------------
// A very interesting idea here might be to implement free 
// list chaining. An intf for a Chunk allocator in a different 
// lock domain might be optionally passed here? The idea
// would be for chunks to be allocated from an allocator
// in another lock domain. If that fails then fall back to
// creation of chunks with new. When the free list is destroyed
// any chunks allocated from the foreign lock domain would
// be released through a deallocator in the foreign lock
// domain. So perhaps all that is needed is to pass a ref
// to an abstract chunk allocator/deallocator interface when 
// creating the free list.

template < typename T, typename ATTR >
class FreeList {
public:
    typedef T Product;
    typedef ATTR Attr;
    FreeList ();
    // this calls T's destructor using the
    // invokeDestructor < T > function template, 
    // and then reclaims the storage for T
    void deleteProduct ( T & );
    void show ( unsigned level ) const;
private:
    // tornado 2.2 does not support out of
    // template class union declaration
    union Item {
        char pad [ sizeof ( T ) ];
        union Item * _pNext;
    };
    class Chunk;
    typedef FreeList < T, ATTR > This;
    std :: auto_ptr < Chunk > _pFirstChunk;
    Item * _pFreeList;
    size_t _nItemsAllocated;
    void _prealloc ();
    template < bool noThrow >
    void * _allocate ( size_t size );
    void _release ( void * p );
    friend void * operator new < T, ATTR > ( size_t, This & );
    friend void * operator new < T, ATTR > ( size_t, 
                        const std :: nothrow_t &, This & );
    friend void operator delete < T, ATTR > ( void *, This & );
    friend void operator delete < T, ATTR > ( void *, 
                            const std :: nothrow_t &, This & );
    FreeList ( const This & ); //disabled
    This & operator = ( const This & ); // disabled
};

// Notes:
// 1) If _p isnt nill when the destructor runs the mutex 
// might be taken recursively. This can be avoided in
// non-exception path code by calling release, but cant
// always be avoided in exception path code.
// 2) This class assumes that the free list and the free list
// product are in the same lock domain. Sometimnes that
// is possible, but not always.
// 3) This class assumes that the free list product has a
// public explicit destructor as follows that takes a Guard 
// reference as its sole argument, and posibly also a private 
// or protected implicit destructor.
// void destructor ( Guard & );
template < class T >
class FreeListAutoPtr {
public:
    typedef typename T :: Product Target;
    FreeListAutoPtr ( T &, Target * pTarget );
    ~FreeListAutoPtr ();
    void assign ( FreeListAutoPtr < T > & );
    Target & operator * () const;
    Target * operator -> () const;
    Target * get () const;
    Target * release ();
private:
    T & _freeList;
    Target * _p;
    FreeListAutoPtr < T > & operator = ( 
            const FreeListAutoPtr < T > & ); // disabled
    FreeListAutoPtr < T > ( 
            const FreeListAutoPtr < T > & ); // disabled
};

template < typename T, size_t N >
struct FreeListAttr {
    static const size_t nElemPerChunk = N;
    static const size_t nElemPrealloc = 0;
    typedef FreeListAttr < T, N > This;
};

// backwards compatibility
template < class T, size_t N = 256 > 
class tsFreeList : 
    public FreeList < T, FreeListAttr < T, N > > {
public:
    typedef FreeListAttr < T, N > Attr;
    typedef FreeList < T, Attr > Base; 
    tsFreeList ();
};

// fake C99 compiler compliance as necessary
#ifndef SIZE_MAX 
#   define SIZE_MAX ( static_cast < size_t > ( -1 ) )
#endif

template < typename T >
inline void invokeDestructor ( T & target )
{
    target.~T ();
}

// its important to make the arguments here selective
// (they are based on FREELIST rather than T)
template < typename T, typename ATTR >
void * operator new ( size_t size, FreeList < T, ATTR > & freeList )
{
    return freeList. template _allocate < false > ( size );
}

// its important to make the arguments here selective
// (they are based on FREELIST rather than T)
template < typename T, typename ATTR >
void * operator new ( size_t size, const std :: nothrow_t &,
                        FreeList < T, ATTR > & freeList )
{
    return freeList.template _allocate < true > ( size );
}

// arguments here need to be symetric with new
template < typename T, typename ATTR >
void operator delete ( void * p, FreeList < T, ATTR > & freeList )
{
    return freeList._release ( p );
}

// arguments here need to be symetric with new
template < typename T, typename ATTR >
void operator delete ( void * p, const std :: nothrow_t &,
                        FreeList < T, ATTR > & freeList )
{
    return freeList._release ( p );
}

template < typename T, typename ATTR >
class FreeList < T, ATTR > :: Chunk {
public:
    Chunk ( Chunk * pNext, Item * & pFirst );
    void show ( unsigned ) const;
    static size_t nItems ();
private:
#   ifdef EPICS_FREELIST_DEBUG
        static const size_t _nItems = 1;
#   else
        static const size_t _nItems = 
                        ATTR :: nElemPerChunk;
#   endif
    Item _items [ _nItems ];
    std :: auto_ptr < Chunk > _pNext;
};

template < typename T, typename ATTR >
FreeList < T, ATTR > :: Chunk :: 
    Chunk ( Chunk * pNext, Item * & pFirst ) :
    _pNext ( pNext )
{
    for ( size_t i=0; i < _nItems - 1; i++ ) {
        _items[i]._pNext = & _items[i+1];
    }
    _items[ _nItems - 1 ]._pNext = 0;
    pFirst = _items;
}

template < typename T, typename ATTR >
inline size_t FreeList < T, ATTR > :: Chunk :: 
    nItems ()
{
    return _nItems;
}

template < typename T, typename ATTR >
void FreeList < T, ATTR > :: Chunk :: 
    show ( unsigned level ) const
{
    // The value of N will be revealed by the type 
    // name - I presume
    std :: cout << "%s" << typeid ( *this ) . name () << '\n';
    if ( level > 0 ) {
        std :: cout << "_pNext = " << _pNext.get () << '\n';
        _pNext->show ( level );
    }
}

template < typename T, class ATTR >
inline FreeList < T, ATTR > :: 
    FreeList () : 
    _pFreeList ( 0 ),
    _nItemsAllocated ( 0 )
{
    _prealloc ();
}

template < typename T, class ATTR >
inline void FreeList < T, ATTR > ::
    _prealloc ()
{
    size_t nElemAlloc = 0u;
    while ( nElemAlloc < ATTR :: nElemPrealloc ) {
        std :: auto_ptr < Chunk > 
            pChunkTmp ( 
                new Chunk ( 
                    _pFirstChunk.get (), _pFreeList ) );
        _pFirstChunk.release ();
        _pFirstChunk = pChunkTmp;
        nElemAlloc += Chunk :: nItems ();
    }
}

template < typename T, class ATTR >
void FreeList < T, ATTR > :: 
    show ( unsigned level ) const
{
    std :: cout << "%s" << typeid ( *this ) . name () << '\n';
    if ( level > 0 ) {
        std :: cout << 
            "_pChunkList = " << _pFirstChunk.get() << '\n';
        if ( _pFirstChunk.get() && level > 1 ) {
            _pFirstChunk->show ( level - 2 );
        }
        std :: cout << 
            "_pFreeList = " << _pFreeList << '\n';
        std :: cout << 
            "_nItemsAllocated = " << _nItemsAllocated << '\n';
    }
}

template < typename T, class ATTR >
template < bool noThrow >
void * FreeList < T, ATTR > :: 
    _allocate ( size_t size ) 
{
    if ( size != sizeof ( T ) ) {
        // throw exception (or assert fail) here if they have 
        // "new ( tsFreeList < Apple > ) Orange (...)"
        // This is creation of Oranges with the Apples
        // free list - a programmer error. 
        assert ( ! noThrow );
        throw std :: runtime_error (
            "FreeList < ATTR > :: allocate () "
            "size requested does not match FreeList "
            "product size" );
    }
    if ( ! _pFreeList ) {
        std :: auto_ptr < Chunk > pAPChunk ( 
            new ( std :: nothrow )
                Chunk ( _pFirstChunk.get (), _pFreeList ) );
        if ( pAPChunk.get () == 0 ) {
            if ( noThrow ) {
                return 0;
            }
            else {
                throw std :: bad_alloc ();
            }
        }
        _pFirstChunk.release ();
        _pFirstChunk = pAPChunk;
    }
    Item * pItem = _pFreeList;
    _pFreeList = pItem->_pNext;
    void * p = reinterpret_cast < void * > ( pItem );
#   ifdef EPICS_FREELIST_DEBUG
        // erase the past 
        memset ( p, 0xaa, sizeof ( Item ) );
#   endif
    return p;
}

template < typename T, typename ATTR >
inline void FreeList < T, ATTR > :: 
    deleteProduct ( T & product )
{
    invokeDestructor ( product );
    _release ( & product );
}

template < typename T, class ATTR >
void FreeList < T, ATTR > :: 
    _release ( void * p ) 
{
    if ( p ) {
#       ifdef EPICS_FREELIST_DEBUG
            // erase the past
            memset ( p, 0x55, sizeof ( Item ) );
#       endif
        Item * pItem = reinterpret_cast < Item * > ( p );
        pItem->_pNext = _pFreeList;
        _pFreeList = pItem;
    }
}

template < class T >
inline FreeListAutoPtr < T > :: 
FreeListAutoPtr ( T & freeList, typename T :: Product * p ) :
    _freeList ( freeList ), _p ( p )
{
}

template < class T >
inline void FreeListAutoPtr < T > :: 
    assign ( FreeListAutoPtr < T > & ap ) 
{
    if ( _p != ap._p ) {
        if ( _p ) {
            _freeList.deleteProduct ( *_p );
        }
        _p = ap._p;
    }
}

template < class T >
inline FreeListAutoPtr < T > :: ~FreeListAutoPtr ()
{
    if ( _p ) {
        _freeList.deleteProduct ( *_p );
    }
}

template < class T >
inline typename T :: Product & FreeListAutoPtr < T > :: 
    operator * () const
{
    return * _p;
}

template < class T >
inline typename T :: Product * FreeListAutoPtr < T > :: 
    operator -> () const
{
    return _p;
}

template < class T >
inline typename T :: Product * FreeListAutoPtr < T > :: 
    get () const
{
    return _p;
}

template < class T >
inline typename T :: Product * FreeListAutoPtr < T > :: 
    release ()
{
    Target * pTmp ( _p );
    return _p = 0, pTmp;
}

template < class T, size_t N > 
inline tsFreeList < T, N > :: 
    tsFreeList () :
    Base ()
{
}

#endif // tsFreeList_h
