/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*  
 *
 *                              
 *                    L O S  A L A M O S
 *              Los Alamos National Laboratory
 *               Los Alamos, New Mexico 87545
 *                                  
 *  Copyright, 1986, The Regents of the University of California.
 *                                  
 *           
 *	Author Jeffrey O. Hill
 *	johill@lanl.gov
 */

#include <string>
#include <stdexcept>
#include <cstdio>
#include <climits>

#define epicsExportSharedSymbols
#include "ipAddrToAsciiAsynchronous.h"
#include "epicsThread.h"
#include "epicsMutex.h"
#include "epicsEvent.h"
#include "epicsExit.h"
#include "tsDLList.h"
#include "tsFreeList.h"
#include "errlog.h"

typedef epics :: Mutex :: Guard Guard;

extern "C" {
static void ipAddrToAsciiEngineGlobalMutexConstruct ( void * );
static void ipAddrToAsciiEngineShutdownRequest ( void * );
}

// - this class executes the synchronous DNS query
// - it creates one thread
class ipAddrToAsciiEnginePrivate : 
    private ipAddrToAsciiEngine, 
    private epicsThreadRunable {
public:
    ipAddrToAsciiEnginePrivate ();
    ~ipAddrToAsciiEnginePrivate ();
private:

    // - this class implements the asynchronous DNS query
    // - it completes early with the host name in dotted IP 
    //   address form if the ipAddrToAsciiEngine is destroyed 
    //   before IO completion or if there are too many items 
    //   already in the engine's queue.
    class Transaction : 
        private ipAddrToAsciiTransaction,
        public tsDLNode < Transaction > {
    public:
        Transaction ( 
            class ipAddrToAsciiEnginePrivate & engineIn );
        void cleanup ( Guard & guard );
        ~Transaction ();
        osiSockAddr address () const;
    private:
        osiSockAddr _addr;
        ipAddrToAsciiEnginePrivate & _engine;
        ipAddrToAsciiCallBack * _pCB;
        bool _inProgress;
        void ipAddrToAscii ( 
            const osiSockAddr &, ipAddrToAsciiCallBack & );
        void show ( unsigned level ) const; 
        void destroy (); 
        friend class ipAddrToAsciiEnginePrivate;
        Transaction & operator = ( const Transaction & );
        Transaction ( const Transaction & );
    };

    char nameTmp [1024];
    tsFreeList < Transaction, 0x80 > _transactionFreeList;
    tsDLList < Transaction > labor;
    mutable epics :: Mutex mutex;
    epicsEvent laborEvent;
    epicsEvent destructorBlockEvent;
    epicsThread thread;
    Transaction * pCurrent;
    unsigned cancelPendingCount;
    bool exitFlag;
    bool callbackInProgress;
    static epics :: Mutex * pGlobalMutex;
    static ipAddrToAsciiEnginePrivate * pEngine;
    static unsigned numberOfReferences;
    Transaction & createTransaction ();
    void show ( unsigned level ) const; 
    void destroy ();
    static bool shutdownRequest;
    void run ();
	ipAddrToAsciiEnginePrivate ( const ipAddrToAsciiEngine & );
	ipAddrToAsciiEnginePrivate & operator = ( const ipAddrToAsciiEngine & );
    friend class ipAddrToAsciiEngine;
    friend class ipAddrToAsciiTransactionPrivate;
    friend void ipAddrToAsciiEngineShutdownRequest ( void * );
    friend void ipAddrToAsciiEngineGlobalMutexConstruct ( void * );
#   if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
		friend class Transaction;
#   endif
};

#ifdef _MSC_VER
#   pragma warning ( push )
#   pragma warning ( disable:4660 )
#endif

template class tsFreeList 
    < ipAddrToAsciiEnginePrivate :: Transaction, 0x80 >;

#ifdef _MSC_VER
#   pragma warning ( pop )
#endif

epics :: Mutex * ipAddrToAsciiEnginePrivate :: pGlobalMutex = 0;
ipAddrToAsciiEnginePrivate * 
    ipAddrToAsciiEnginePrivate :: pEngine = 0;
unsigned ipAddrToAsciiEnginePrivate :: numberOfReferences = 0u;
bool ipAddrToAsciiEnginePrivate :: shutdownRequest = false;
static epicsThreadOnceId ipAddrToAsciiEngineGlobalMutexOnceFlag = 0;

// the users are not required to supply a show routine
// for there transaction callback class
void ipAddrToAsciiCallBack :: show ( unsigned /* level */ ) const {}

static void ipAddrToAsciiEngineShutdownRequest ( void * )
{
    bool deleteGlobalMutexCondDetected = false;
    {
        Guard guard ( 
            * ipAddrToAsciiEnginePrivate :: pGlobalMutex );
        ipAddrToAsciiEnginePrivate :: shutdownRequest = true;
        deleteGlobalMutexCondDetected = 
            ( ipAddrToAsciiEnginePrivate :: numberOfReferences == 0 );
    }
    if ( deleteGlobalMutexCondDetected ) {
        delete ipAddrToAsciiEnginePrivate :: pGlobalMutex;
        ipAddrToAsciiEnginePrivate :: pEngine = 0;
    }
}

static void ipAddrToAsciiEngineGlobalMutexConstruct ( void * )
{
    ipAddrToAsciiEnginePrivate :: pGlobalMutex = 
                                    new epics :: Mutex ();
    epicsAtExit ( ipAddrToAsciiEngineShutdownRequest, 0 );
}

// for now its probably sufficent to allocate one 
// DNS transaction thread for all codes sharing
// the same process that need DNS services but we 
// leave our options open for the future
ipAddrToAsciiEngine & ipAddrToAsciiEngine :: create ()
{
    epicsThreadOnce (
        & ipAddrToAsciiEngineGlobalMutexOnceFlag,
        ipAddrToAsciiEngineGlobalMutexConstruct, 0 );
    // since we must not own lock when checking this flag
    // this diagnostic has imperfect detection, but never
    // incorrect detection
    if ( ipAddrToAsciiEnginePrivate :: shutdownRequest ) {
        throw std :: runtime_error ( 
            "ipAddrToAsciiEngine::allocate (): "
            "attempts to create an "
            "ipAddrToAsciiEngine while the exit "
            "handlers are running are rejected");
    }
    Guard guard ( * ipAddrToAsciiEnginePrivate::pGlobalMutex );
    if ( ! ipAddrToAsciiEnginePrivate :: pEngine ) {
        ipAddrToAsciiEnginePrivate::pEngine = 
            new ipAddrToAsciiEnginePrivate ();
    }
    ipAddrToAsciiEnginePrivate::numberOfReferences++;
    return * ipAddrToAsciiEnginePrivate :: pEngine;
}

ipAddrToAsciiEnginePrivate :: ipAddrToAsciiEnginePrivate () :
    _transactionFreeList (),
    thread ( *this, "ipToAsciiProxy",
        epicsThreadGetStackSize(epicsThreadStackBig),
        epicsThreadPriorityLow ),
    pCurrent ( 0 ), cancelPendingCount ( 0u ), exitFlag ( false ),  
    callbackInProgress ( false )
{
    this->thread.start (); // start the thread
}

ipAddrToAsciiEnginePrivate :: ~ipAddrToAsciiEnginePrivate ()
{
    {
        Guard guard ( this->mutex );
        this->exitFlag = true;
    }
    this->laborEvent.signal ();
    this->thread.exitWait ();
}

// for now its probably sufficent to allocate one 
// DNS transaction thread for all codes sharing
// the same process that need DNS services but we 
// leave our options open for the future
void ipAddrToAsciiEnginePrivate :: destroy ()
{
    bool deleteGlobalMutexCondDetected = false;
    epicsThreadOnce (
        & ipAddrToAsciiEngineGlobalMutexOnceFlag,
        ipAddrToAsciiEngineGlobalMutexConstruct, 0 );
    {
        Guard guard ( 
            * ipAddrToAsciiEnginePrivate::pGlobalMutex );
        assert ( 
            ipAddrToAsciiEnginePrivate :: numberOfReferences > 0u );
        ipAddrToAsciiEnginePrivate::numberOfReferences--;
        if ( ipAddrToAsciiEnginePrivate::numberOfReferences == 0u ) {
            deleteGlobalMutexCondDetected = 
                ipAddrToAsciiEnginePrivate :: shutdownRequest;
        }
    }
    if ( deleteGlobalMutexCondDetected ) {
        delete ipAddrToAsciiEnginePrivate :: pGlobalMutex;
        ipAddrToAsciiEnginePrivate :: pEngine = 0;
    }
}

void ipAddrToAsciiEnginePrivate :: show ( unsigned level ) const
{
    Guard guard ( this->mutex );
    printf ( 
        "ipAddrToAsciiEngine at %p with %u "
        "requests pending\n", 
        static_cast <const void *> (this), this->labor.count () );
    if ( level > 0u ) {
        tsDLIterConst < Transaction > pItem 
            = this->labor.firstIter ();
        while ( pItem.valid () ) {
            pItem->show ( level - 1u );
            pItem++;
        }
    }
    if ( level > 1u ) {
        printf ( "mutex:\n" );
        this->mutex.show ( level - 2u );
        printf ( "laborEvent:\n" );
        this->laborEvent.show ( level - 2u );
        printf ( "exitFlag  boolean = %u\n", this->exitFlag );
        printf ( "exit event:\n" );
    }
}

ipAddrToAsciiEnginePrivate :: Transaction & 
    ipAddrToAsciiEnginePrivate :: createTransaction ()
{
    Guard guard ( mutex );
    Transaction * pT = new ( _transactionFreeList ) Transaction ( *this );
    return * pT;
}

void ipAddrToAsciiEnginePrivate :: run ()
{
    Guard guard ( this->mutex );
    while ( ! this->exitFlag ) {
        {
            Guard :: Release unguard ( guard );
            this->laborEvent.wait ();
        }
        while ( true ) {
            Transaction * pItem = this->labor.get ();
            if ( ! pItem ) {
                break;
            }
            osiSockAddr addr = pItem->_addr;
            this->pCurrent = pItem;
    
            if ( this->exitFlag )
            {
                sockAddrToDottedIP ( & addr.sa, this->nameTmp, 
                    sizeof ( this->nameTmp ) );
            }
            else {
                Guard :: Release unguard ( guard );
                // depending on DNS configuration, this 
                // could take a very long timeso we release 
                // the lock
                sockAddrToA ( &addr.sa, this->nameTmp, 
                    sizeof ( this->nameTmp ) );
            }

            // the ipAddrToAsciiTransactionPrivate destructor 
            // is allowed to set pCurrent to nill and avoid 
            // blocking on a slow DNS operation
            if ( ! this->pCurrent ) {
                continue;
            }

            this->callbackInProgress = true;

            {
                Guard :: Release unguard ( guard );
                // dont call callback with lock applied
                this->pCurrent->_pCB->
                    transactionComplete ( this->nameTmp );
            }

            this->callbackInProgress = false;

            if ( this->pCurrent ) {
                this->pCurrent->_inProgress = false;
                this->pCurrent = 0;
            }
            if ( this->cancelPendingCount  ) {
                this->destructorBlockEvent.signal ();
            }
        }
    }
}

ipAddrToAsciiEnginePrivate :: Transaction :: Transaction 
    ( ipAddrToAsciiEnginePrivate & engineIn ) :
    _engine ( engineIn ), _pCB ( 0 ), 
    _inProgress ( false )
{
    memset ( & _addr, '\0', sizeof ( _addr ) );
    _addr.sa.sa_family = AF_UNSPEC;
}

ipAddrToAsciiEnginePrivate :: Transaction :: 
    ~Transaction ()
{
    if ( _inProgress ) {
        Guard guard ( _engine.mutex );
        this->cleanup ( guard );
    }
}

void ipAddrToAsciiEnginePrivate :: Transaction :: 
    cleanup ( Guard & guard )
{
    while ( _inProgress ) {
        if ( _engine.pCurrent == this && 
                _engine.callbackInProgress && 
                ! _engine.thread.isCurrentThread() ) {
            assert ( _engine.cancelPendingCount 
                < UINT_MAX );
            _engine.cancelPendingCount++;
            {
                Guard :: Release unguard ( guard );
                _engine.destructorBlockEvent.wait ();
            }
            assert ( _engine.cancelPendingCount > 0u );
            _engine.cancelPendingCount--;
            if ( ! _inProgress ) {
                if ( _engine.cancelPendingCount ) {
                    _engine.destructorBlockEvent.signal ();
                }
                break;
            }
        }
        else {
            if ( _engine.pCurrent == this ) {
                _engine.pCurrent = 0;
            }
            else {
                _engine.labor.remove ( *this );
            }
            _inProgress = false;
        }
    }
}

// we use destroy() and not combo of vitual destructor with
// class specific operator delete because we like using
// one guard protect both destructor and also operator delete
void ipAddrToAsciiEnginePrivate :: Transaction :: destroy ()
{
    Guard guard ( _engine.mutex );
    this->cleanup ( guard );
    _engine._transactionFreeList.deleteProduct ( *this );
}

void ipAddrToAsciiEnginePrivate :: Transaction :: ipAddrToAscii ( 
    const osiSockAddr & addrIn, ipAddrToAsciiCallBack & cbIn )
{
    bool success = false;
    {
        Guard guard ( _engine.mutex );
        // put some reasonable limit on queue expansion
        if ( !_inProgress && _engine.labor.count () < 16u ) {
            _addr = addrIn;
            _pCB = & cbIn;
            _inProgress = true;
            _engine.labor.add ( *this );
            success = true;
        }
    }

    if ( success ) {
        _engine.laborEvent.signal ();
    }
    else {
        char autoNameTmp[256];
        sockAddrToDottedIP ( & addrIn.sa, autoNameTmp, 
            sizeof ( autoNameTmp ) );
        cbIn.transactionComplete ( autoNameTmp );
    }
}

osiSockAddr ipAddrToAsciiEnginePrivate :: Transaction :: 
    address () const
{
    return _addr;
}

void ipAddrToAsciiEnginePrivate :: Transaction :: 
    show ( unsigned level ) const
{
    Guard guard ( _engine.mutex );
    char ipAddr [64];
    sockAddrToDottedIP ( &_addr.sa, 
        ipAddr, sizeof ( ipAddr ) );
    printf ( 
        "ipAddrToAsciiTransactionPrivate for "
        "address %s\n", ipAddr );
    if ( level > 0u ) {
        printf ( "\tengine %p\n",
            static_cast <void *> ( & _engine ) );
        _pCB->show ( level - 1u );
    }
}
