/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/* epicsMutex.c */
/*	Author: Marty Kraimer and Jeff Hill	Date: 03APR01	*/

/*
 * NOTES:
 * 1) VERIFY_CORRECT_OWNER feature is turned off by defualt because
 * it slows down the system, but might prove to be quite useful 
 * during regression tests.
 * 2) SHOW_LAST_OWNER feature is is currently turned off because 
 * it slows down the system at run time, and because its not 
 * currently safe to convert a thread id to a thread name because
 * the thread may have exited and thereby rendering the thread 
 * id invalid.
 */

#include <new>
#include <cstddef>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "epicsMutex.h"
#include "epicsStdioRedirect.h"
#include "epicsThread.h"
#include "errlog.h"
#include "epicsThread.h"
#include "tsDLList.h"
#include "tsFreeList.h"
#include "osiAtomic.h"

#undef max

namespace epics {

using namespace osi;

static struct MutexNOOP : 
    public Mutex :: Intf {
    void lock () {}
    void lockRecursive () {}
    void unlock () {}
    void mustUnlock () {}
    bool tryLock () { return true; }
    void show ( unsigned level ) const {}
    void reference () {}
    void unreference () {}
} mutexNOOP;

static struct SemaphoreFactoryNOOP : 
    public Mutex :: Factory {
public:
    Mutex :: Intf & create ( const SrcCodeLocation & );
private:
} semaphoreFactoryNOOP;
    
Mutex :: Factory & 
    Mutex :: factoryNOOP = semaphoreFactoryNOOP;

static SrcCodeLocation 
    _locationUnspecified  ( "Src File Name UKN", 0 );
SrcCodeLocation & Mutex :: 
    locationUnspecified = _locationUnspecified;

static void epicsMutexGlobalOnceFunc ( void * );

//
// Implementation Notes
// --------------------
//
// Note 1
// ======
//
// If the user calls "delete pMutex" and there is a 
// virtual Mutex :: ~Intf then a proper class 
// specific delete in ConcreteMutex will run. However, 
// the ~Intf destructor executes prior to
// when such an operator delete is called. Therefore, 
// we have only two viable designs considering that
// a ConcreteMutex must never remain on the list in
// ConcreteMutexFactory after it is destroyed (that
// would place other threads in the system accessing 
// ConcreteMutexFactory's member functions at
// unacceptable risk).
//
// 1) Mutex registration occurs in the ConcreteMutex
// constructor and unregistration occurs in the 
// ConcreteMutex destructor. With this design it 
// isnt possible to use the same Guard to protect
// both operations. Both operations means in this
// design either both free list allocation and 
// ConcreteMutexFactory registration, or both 
// free list deallocation and ConcreteMutex :: Factory
// unregistration.
//
// 2) Disable direct invocation of "delete pMutex" by 
// making Mutex :: Intf :: ~Intf a protected
// NOOP member function. Include also a public release
// member function that is used to destroy any mutex
// implementing the interface. With this design it 
// _is_ possible to use the same Guard to protect
// both operations. Both operations means in this
// design either both free list allocation and 
// ConcreteMutexFactory registration, or both 
// free list deallocation and ConcreteMutexFactory
// unregistration. Since Intf :: ~Intf ()
// is protected (not invokable by users then it is
// also not specified to be virtual).
//
// The design in option two was chosen because of its 
// clearly visible perfomance benefits.
//

class ConcreteMutex : 
    public tsDLNode < ConcreteMutex >,
    public Mutex :: Intf {
public:
    ConcreteMutex ( const SrcCodeLocation & );
    ~ConcreteMutex (); 
    void lock (); 
    void lockRecursive (); 
    void unlock ();
    void mustUnlock ();
    bool tryLock (); 
    void show ( unsigned level ) const;
    void reference ();
    void unreference ();
private:
    SrcCodeLocation _srcCodeLocation;
    epicsMutexOSD * _pMutexOSD;
#   ifdef VERIFY_CORRECT_OWNER
    epicsThreadId _lastOwner;
#   endif
    AtomicWord _referenceCount; 
    size_t _lockAcquireCount;
};

class ConcreteMutexFactory {
public:
    ConcreteMutexFactory ();
    ~ConcreteMutexFactory ();
    Mutex :: Intf & create ( const SrcCodeLocation & );
    void destroy ( ConcreteMutex & );
    void show ( bool onlyLocked, unsigned level ) const;
private:
    struct MutexFirst : 
        public ConcreteMutex {
    public:
        MutexFirst ();
        void destroy ();
    };
    class MutexFirstFactory :
        public Mutex :: Factory {
    public:
        MutexFirstFactory ( ConcreteMutexFactory & );
        Mutex :: Intf & create ( const SrcCodeLocation & );
    private:
        ConcreteMutexFactory & _privateFactory;
    };
    MutexFirstFactory _mutexFirstFactory;
    MutexFirst _mutexFirst;
    mutable Mutex _mutex;
    tsDLList < ConcreteMutex > _mutexList;
    tsFreeList < ConcreteMutex > _mutexParmFreeList;
#   if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
	friend struct MutexFirst;
	friend class MutexFirstFactory;
#   endif
};

// a simple facade class that guarantees that the
// mutex system is initialized on-demand even if
// file scope object constructors in this file
// dont run first
static struct ConcreteMutexFactoryFacade : 
    public Mutex :: Factory {
public:
    Mutex :: Intf & create ( const SrcCodeLocation & );
    void destroy ( ConcreteMutex & );
    void show ( bool onlyLocked, unsigned level ) const;
private:
    static epicsThreadOnceId _once;
    static ConcreteMutexFactory * _pPrivateFactory;
    friend void epicsMutexGlobalOnceFunc ( void * );
} concreteMutexFactoryFacade;

epicsThreadOnceId ConcreteMutexFactoryFacade :: 
        _once = EPICS_THREAD_ONCE_INIT;

ConcreteMutexFactory * 
    ConcreteMutexFactoryFacade :: _pPrivateFactory = 0;

Mutex :: Factory & 
    Mutex :: factoryConcrete = concreteMutexFactoryFacade;

Mutex :: 
    Mutex ( const Guard & guard ) :
    _pIntf ( 0 )
{
    if ( ! guard._pMutex ) {
        throw std :: runtime_error ( 
            "Mutex constructor: specified Guard "
            "that is currently released" );
    }
    _pIntf = guard._pMutex->_pIntf;
    _pIntf->reference ();
}

// packaged in a static standalone function to avoid code bloat
void Mutex :: 
    _throwLockDomainMismatch ()
{
    throw std :: runtime_error (
        "Mutex domain missmatch detected" );
}

void Mutex :: Guard :: 
    retarget ( Mutex & mutex )
{
    if ( ! _pMutex ) {
        throw std :: runtime_error ( 
            "Guard retarget: Guard "
            "is currently released?" );
    }
    if ( _pMutex->_pIntf != mutex._pIntf ) {
        _pMutex->_pIntf->unlock ();
        _pMutex = & mutex;
        _pMutex->_pIntf->lock ();
    }
}

Mutex :: Guard :: Release :: 
    Release ( Mutex :: Guard & guard ) :
    Mutex :: Unlocker ( *guard._pMutex ),
    _guard ( guard )
{
    guard._pMutex = 0;
}

Mutex :: Guard :: Release :: 
    ~Release ()
{
    _guard._pMutex = & _mutex;
}

void Mutex :: Guard :: 
    _throwReleaseInProgress ()
{
    throw std :: runtime_error (
        "Guard's mutex id is being verified "
        "when a Release is active" );
}

Mutex :: Unlocker ::
    Unlocker ( Mutex & mutex ) :
    _mutex ( mutex )
{
    _mutex.mustUnlock ();
}

Mutex :: Unlocker ::
    ~Unlocker ()
{
    _mutex.lock ();
}

ConcreteMutex :: 
    ConcreteMutex ( const SrcCodeLocation & srcCodeLoc ) :
    _srcCodeLocation ( srcCodeLoc ),
    _pMutexOSD ( epicsMutexOsdCreate() ),
#   ifdef VERIFY_CORRECT_OWNER
    _lastOwner ( 0 ),
#   endif
    _lockAcquireCount ( 0u )
{
    atomicSet ( _referenceCount, 0 );
    if ( ! _pMutexOSD ) {
        throw std :: runtime_error ( 
            "ConcreteMutex: unable to allocate OS specific mutex" );
    }
}

ConcreteMutex :: ~ConcreteMutex ()
{
    epicsMutexOsdDestroy ( _pMutexOSD );
    //
    // unfortunately its not easy to add a
    // reliable assert test for 
    // "_lockAcquireCount == 0u" here 
    // because it will be flaky if the thread 
    // that unlocked last isnt the thread that 
    // decrements the reference count to zero
    //
    // this is because the _referenceCount is
    // protected by osiAtomicIncr/Decr while
    // _lockAcquireCount is protected by the
    // lock
    //
    // nevertheless, many OS will detect this 
    // type of problem
    //
    // assert ( _lockAcquireCount == 0 );

    //
    // testing the reference count should be 
    // ok because we presumably just finished 
    // executing the atomicDecr ( _referenceCount );
    assert ( atomicFetch ( _referenceCount ) == 0 );
}

void ConcreteMutex :: show ( unsigned level ) const
{
    printf ( 
        "Mutex %p created in source file %s at "
        "line number %d, lock cnt %u\n",
        reinterpret_cast < const void * > ( this ), 
        _srcCodeLocation.pFileName (), 
        _srcCodeLocation.lineNumber (),
        _lockAcquireCount );
    if ( level > 0 ) {
        epicsMutexOsdShow ( _pMutexOSD, level - 1 );
#       if defined ( VERIFY_CORRECT_OWNER ) && defined ( SHOW_LAST_OWNER )
            char threadName [255];
            if ( _lastOwner ) {
#               error currently, its not safe to fetch name for stale thread 
                epicsThreadGetName ( _lastOwner,
                    threadName, sizeof ( threadName ) );
            }
            else {
                strcpy ( threadName, "<not used>" );
            }
        
            printf ( "last owner \"%s\"\n", threadName, );
#       endif
        AtomicWord refCnt = atomicFetch ( _referenceCount );
        if ( refCnt >= 0 ) {
            printf ( "ref cnt = %lu\n", 
                static_cast < unsigned long > ( refCnt ) );
        }
        else {
            printf ( "ref cnt = INVALID\n" );
        }
    }
}

void ConcreteMutex :: 
    lock ()
{
    epicsMutexLockStatus status = 
        epicsMutexOsdLock ( _pMutexOSD );
    if ( status != epicsMutexLockOK ) {
        throw std :: runtime_error ( 
            "invalid or corrupt mutex "
            "identifier passed to lock request" );
    }
#   ifdef VERIFY_CORRECT_OWNER
        _lastOwner = epicsThreadGetCurrent ();
#   endif
    assert ( _lockAcquireCount < SIZE_MAX );
    _lockAcquireCount++;
}

// this causes the recursive lock count to increment
// but doesnt take the lock
void ConcreteMutex :: 
    lockRecursive ()
{
    // Guard is verified at a higher level
    assert ( _lockAcquireCount < SIZE_MAX );
    assert ( _lockAcquireCount > 0u );
    _lockAcquireCount++;
}

void ConcreteMutex :: 
    mustUnlock () 
{
    if ( _lockAcquireCount != 1 ) {
        throw std :: runtime_error ( 
            "MutexManager :: mustUnlock (): "
            "lock count wasnt one" );
    }
    this->unlock ();
}

// inline so that this integrates into C callable 
// function with minimized overhead
void ConcreteMutex :: 
    unlock ()
{
#   ifdef VERIFY_CORRECT_OWNER
        assert ( _lastOwner == epicsThreadGetIdSelf() );
#   endif
    assert ( _lockAcquireCount > 0u );
    if ( --_lockAcquireCount == 0 ) {
        epicsMutexOsdUnlock ( _pMutexOSD );
    }
}

// inline so that this integrates into C callable 
// function with minimized overhead
inline bool ConcreteMutex :: 
    tryLock ()
{
    epicsMutexLockStatus status = 
        epicsMutexOsdTryLock ( _pMutexOSD );
    if ( status != epicsMutexLockOK ) {
        if ( status != epicsMutexLockTimeout ) {
            throw std :: runtime_error ( 
                "invalid or corrupt mutex identifier "
                "passed to tryLock()" );
        }
        return false;
    }
     assert ( _lockAcquireCount < SIZE_MAX );
    _lockAcquireCount++;
#   ifdef VERIFY_CORRECT_OWNER
        _lastOwner = epicsThreadGetIdSelf();
#   endif
    return true;
}

void ConcreteMutex :: 
    reference ()
{
    using namespace std;
    AtomicWord newVal = atomicIncr ( _referenceCount );
    assert ( newVal > 0 );
}

void ConcreteMutex :: 
    unreference ()
{
    AtomicWord newVal = atomicDecr ( _referenceCount );
    static const AtomicWord zeroVal = 0; 
    static const AtomicWord oneVal = 1; 
    assert ( newVal != zeroVal - oneVal );
    if ( newVal == 0 ) {
        concreteMutexFactoryFacade.destroy ( *this );
        // dont touch this object after this point in func
        // considering above destroy
    }
    // dont touch this  object after this point in func
    // considering above destroy
}

Mutex :: Intf & 
    SemaphoreFactoryNOOP :: 
        create ( const SrcCodeLocation & )
{
    return mutexNOOP;
}

ConcreteMutexFactory :: 
    ConcreteMutexFactory () : 
    _mutexFirstFactory ( *this ), 
    _mutex ( _mutexFirstFactory ),
    _mutexParmFreeList ()
{
}

ConcreteMutexFactory :: ~ConcreteMutexFactory () 
{
    assert ( _mutexList.count () ==  0 );
}

/*
 * The file name string passed here must exist, and 
 * remain unmodified, for the entire lifespan of 
 * of the mutex.
 */
Mutex :: Intf & ConcreteMutexFactory :: 
    create ( const SrcCodeLocation & srcCodeLoc )
{
    epics :: Mutex :: Guard guard ( _mutex );
    ConcreteMutex & mutex = * 
        new ( _mutexParmFreeList ) ConcreteMutex ( srcCodeLoc );
    _mutexList.add ( mutex );
    return mutex;
}

void ConcreteMutexFactory :: 
    destroy ( ConcreteMutex & mutex )
{
    epics :: Mutex :: Guard guard ( _mutex );
    _mutexList.remove ( mutex );
    _mutexParmFreeList.deleteProduct ( mutex );
}

void ConcreteMutexFactory :: 
    show ( bool onlyLocked, unsigned level ) const
{
    epics :: Mutex :: Guard guard ( _mutex );
    printf ( "ConcreteMutexFactory with %u mutexes instantiated\n",
        _mutexList.count () );
    if ( level > 0 ) {
        tsDLIterConst < ConcreteMutex > 
            iter ( _mutexList.firstIter () );
        while ( iter.valid () ) {
            if ( onlyLocked ) {
                // we are willing to bypass the const rules in 
                // order to sense the lock's state, but we are
                // also careful to return the lock to its original
                // unlocked condition
                ConcreteMutex & current ( 
                    const_cast < ConcreteMutex & > ( *iter ) );
                if ( current.tryLock () ) {
                    current.unlock ();
                }
                else {
                    iter->show ( level - 1 );
                }
            }
            else {
                iter->show ( level - 1 );
            }
            iter++;
        }
        if ( level > 1 ) {
            _mutexParmFreeList.show ( level - 2 );
        }
    }
}

inline ConcreteMutexFactory :: MutexFirstFactory ::  
    MutexFirstFactory ( ConcreteMutexFactory & privateFactory ) :
    _privateFactory ( privateFactory )
{
}

Mutex :: Intf & 
    ConcreteMutexFactory :: MutexFirstFactory :: 
        create ( const SrcCodeLocation & )
{
    return _privateFactory._mutexFirst;
}

inline ConcreteMutexFactory :: MutexFirst :: 
    MutexFirst () :
    ConcreteMutex ( srcCodeLocation () )
{
}

void ConcreteMutexFactory :: MutexFirst :: 
    destroy () 
{
    // intentional NOOP
}

static void epicsMutexGlobalOnceFunc ( void * )
{
    assert ( ConcreteMutexFactoryFacade :: _pPrivateFactory == 0 );
    ConcreteMutexFactoryFacade :: _pPrivateFactory = 
        new ( std :: nothrow ) ConcreteMutexFactory ();
}

Mutex :: Intf & 
    ConcreteMutexFactoryFacade ::
        create ( const SrcCodeLocation & srcCodeLocation )
{
    epicsThreadOnce ( & _once, epicsMutexGlobalOnceFunc, 0 );
    if ( ! _pPrivateFactory ) {
        _once = EPICS_THREAD_ONCE_INIT;
        throw std :: runtime_error ( 
            "unable to initialize the concrete mutex factory" );
    }
    return _pPrivateFactory->create ( srcCodeLocation );
}

void ConcreteMutexFactoryFacade ::
    destroy ( ConcreteMutex & mutex )
{
    assert ( _pPrivateFactory );
    _pPrivateFactory->destroy ( mutex );
}

void ConcreteMutexFactoryFacade :: 
    show ( bool onlyLocked, unsigned level ) const
{
    epicsThreadOnce ( & _once, epicsMutexGlobalOnceFunc, 0 );
    if ( _pPrivateFactory ) {
        _pPrivateFactory->show ( onlyLocked, level );
    }
}

namespace legacy {

void epicsMutex :: lock () 
{   
    _pIntf->lock (); 
}

void epicsMutex :: unlock () 
{ 
    _pIntf->unlock (); 
}

bool epicsMutex :: tryLock () { 
    return _pIntf->tryLock (); 
}

} // end of namespace legacy

} // end of namespace epics

epicsMutexId epicsShareAPI epicsMutexOsiCreate (
    const char * pFileName, int lineno )
{
    epicsMutexId pMutex = 0;
    try {
        pMutex = & epics :: concreteMutexFactoryFacade.
                create ( epics :: SrcCodeLocation ( pFileName, lineno ) );
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while creating a"
            "mutex was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while creating"
            "a mutex\n" );
    }
    return pMutex;
}

epicsMutexId epicsShareAPI epicsMutexOsiMustCreate (
    const char * pFileName, int lineno )
{
    epicsMutexId pmutexNode = 
        epicsMutexOsiCreate ( pFileName, lineno );
    assert ( pmutexNode );
	pmutexNode->reference ();
    return pmutexNode;
}

void epicsShareAPI epicsMutexDestroy ( epicsMutexId pmutexNode )
{
    try {
        pmutexNode->unreference ();
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex destroy function was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex destroy function\n" );
    }
}

void epicsShareAPI epicsMutexUnlock ( epicsMutexId pmutexNode )
{
    // There will be some small amount of additional
    // overhead here for the try / catch block but 
    // this will be insignificant compared to the overhead 
    // in the OS for unlocking the mutex.
    // There are also some diagnostic benefits when a 
    // C++ compiler maps hardware exceptions into C++ 
    // exceptions.
    try {
        pmutexNode->unlock ();
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex show function was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex show function\n" );
    }
}

epicsMutexLockStatus epicsShareAPI epicsMutexLock (
    epicsMutexId pmutexNode )
{
    // There will be some small amount of additional
    // overhead here for the try / catch block but 
    // this will probably be insignificant compared
    // to the overhead in the OS for locking the mutex.
    // There are also some diagnostic benefits when a 
    // C++ compiler maps hardware exceptions into C++ 
    // exceptions.
    epicsMutexLockStatus status = epicsMutexLockOK;
    try {
        pmutexNode->lock ();
    }
    catch ( const std :: exception & except ) {
        status = epicsMutexLockError;
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex lock function was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        status = epicsMutexLockError;
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex lock function\n" );
    }
    return status;
}

epicsMutexLockStatus epicsShareAPI epicsMutexTryLock (
    epicsMutexId pmutexNode )
{
    // There will be some small amount of additional
    // overhead here for the try / catch block, and for 
    // converting between the bool C++ status and the
    // epicsMutexLockStatus status, but this will 
    // be insignificant compared to the overhead in the 
    // OS for locking the mutex.
    // There are also some diagnostic benefits when a 
    // C++ compiler maps hardware exceptions into C++ 
    // exceptions.
    epicsMutexLockStatus status = epicsMutexLockOK;
    try {
        bool success = pmutexNode->tryLock ();
        if ( ! success ) {
            status = epicsMutexLockTimeout;
        }
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex tryLock function was \"%s\"\n",
            except.what () );
        status = epicsMutexLockError;
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex tryLock function\n" );
        status = epicsMutexLockError;
    }    
    return status;
}

void epicsShareAPI epicsMutexShow (
    epicsMutexId pmutexNode, unsigned  int level )
{
    try {
        pmutexNode->show ( level );
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex show function was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex show function\n" );
    }
}

void epicsShareAPI epicsMutexShowAll ( 
    int onlyLocked, unsigned  int level )
{
    try {
        epics :: concreteMutexFactoryFacade.
            show ( onlyLocked != 0, level );
    }
    catch ( const std :: exception & except ) {
        errlogPrintf ( 
            "C++ exception while calling "
            "mutex factory show function was \"%s\"\n",
            except.what () );
    }
    catch ( ... ) {
        errlogPrintf ( 
            "non-standard C++ exception while calling "
            "mutex factory show function\n" );
    }
}




