/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef epicsMutexh
#define epicsMutexh

#include <stddef.h> /* define size_t */

#include "epicsAssert.h"

#include "shareLib.h"

#ifdef __cplusplus


#ifndef NDEBUG
extern "C" {
typedef struct epicsThreadOSD * epicsThreadId;
epicsShareFunc epicsThreadId epicsShareAPI epicsThreadGetIdSelf ( void );
} /* end of extern "C" */
#endif

namespace epics {

class SrcCodeLocation;
class epicsShareClass MutexManager;

class epicsShareClass Mutex {
public:
    struct Intf;
    class Guard;
    class Unlocker;
    struct Factory;
   
    static Factory & factoryNOOP;
    static Factory & factoryConcrete;
    static SrcCodeLocation & locationUnspecified;

    Mutex ( Factory & = factoryConcrete, 
        const SrcCodeLocation & = locationUnspecified );
    explicit Mutex ( Mutex & );
    explicit Mutex ( const Guard & );
    ~Mutex ();
    void assertIdentical ( const Mutex & ) const;
    bool isIdentical ( const Mutex & ) const;
    void lock (); 
    void lockRecursive (); /* incr recursive lock cnt */
    void unlock (); 
    void mustUnlock (); /* recursive lock count must decr to zero */
    bool tryLock (); /* T returned => own lock */
    void show ( unsigned level ) const;
protected:
    /*
     * If the user calls scheduleMutexCleanup(),
     * a member function of Guard, then this 
     * member can become nill when the Guard 
     * destructor runs. Nevertheless, we dont
     * bracket access to this member with a 
     * lock / unlock pair as one might intially
     * conclude would be required on SMP systems,
     * to enforce cache coherency, because it is
     * assumed that scheduleMutexCleanup() is
     * called only after an application has
     * guaranteed that there is only one thread 
     * accessing the mutex, during a destruct 
     * sequence, until the mutex is destroyed.
     */
    Intf * _pIntf;
private:
    Mutex & operator = ( const Mutex & ); /* disabled */
    static void _throwLockDomainMismatch ();
    // deal with archaic vxWorks compiler
#if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
        friend class Guard;
#   endif
};

/*
 * Automatically acquires and releases the mutex.
 * The Guard constructor acquires the mutex and the
 * Guard destructor releases the mutex.
 * Useful for protecting class interfaces, managing
 * lock granularity, and writing exception safe
 * code. A Guard object's constructor and destructor
 * must always be run by the same thread. A Guard
 * is typically instantiated as an automatic (stack)
 * variable. Constructing one guard from another
 * does not incur locking overhead but does
 * (recursively) lock the lock.
 */
class epicsShareClass Mutex :: Guard {
public:
    explicit Guard ( Mutex & );
    explicit Guard ( const Guard & );
    ~Guard ();
    bool isIdenticalMutex ( const Mutex & ) const;
    void assertIdenticalMutex ( const Mutex & ) const;
    class Release;
    void retarget ( Mutex & );
private:
    Mutex * _pMutex;
    Guard & operator = ( const Guard & );
    static void _throwReleaseInProgress ();
    friend class Mutex;
    class ReleaseInProgressSetPIntf;
#   if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
        friend class Release;
#   endif
};

class epicsShareClass Mutex :: Unlocker {
public:
    explicit Unlocker ( Mutex & );
    ~Unlocker ();
protected:
    Mutex & _mutex;
    Unlocker ( Unlocker & );
    Unlocker & operator = ( const Unlocker & );
};

/*
 * Automatically releases and reacquires the mutex.
 * The Release constructor releases the mutex and the
 * Release destructor obtains the mutex.
 * The Release constructor must be run by the same
 * thread that created the Guard, and the Release
 * object's destructor must always be run by the
 * same thread as its constructor. A Release
 * is typically instantiated as an automatic (stack)
 * variable.
 *
 * However, the user should be very careful to not
 * pass a Guard which currently has a Release
 * referencing it.
 *
 * A lock that is recursively locked must not be 
 * released.
 */
class epicsShareClass Mutex :: Guard :: Release :
    public Mutex :: Unlocker {
public:
    explicit Release ( Guard & );
    ~Release ();
private:
    Guard & _guard;
    Release ( const Release & );
    Release & operator = ( const Release & );
};
    
struct Mutex :: Intf {
public:
    virtual void lock () = 0; 
    virtual void lockRecursive () = 0; /* incr recursive lock cnt */
    virtual void unlock () = 0; 
    virtual void mustUnlock () = 0; /* recursive lock count must decr to zero */
    virtual bool tryLock () = 0; /* T returned => own lock */
    virtual void show ( unsigned level ) const = 0;
    virtual void reference () = 0;
    virtual void unreference () = 0;
protected:
    /*
     * destroy with delete has been intentionally disabled
     * ---------------------------------------------------
     * Call Intf :: release () to destroy 
     * what was created by the semaphore factory. 
     * See also the implementation notes in 
     * epicsMutex.cpp
     */
    virtual ~Intf () {}
};

/*
 * a semaphore is always created by the factory with its 
 * reference count set to one
 */
struct epicsShareClass Mutex :: Factory {
    virtual ~Factory () {}
    virtual Intf & create ( const SrcCodeLocation & ) = 0;
};

#define srcCodeLocation() SrcCodeLocation ( __FILE__, __LINE__ )

/*
 * The file name string passed here must exist, and 
 * remain unmodified, for the entire lifespan of 
 * of the program.
 */
class epicsShareClass SrcCodeLocation {
public:
    SrcCodeLocation ( const char * pFileName, int lineno );
    const char * pFileName () const;
    int lineNumber () const;
private:
    const char * _pFileName;
    int _lineNumber;
};

namespace legacy {

/* legacy epicsMutex interface */
struct epicsShareClass epicsMutex : 
        private Mutex {
    void lock ();
    void unlock ();
    bool tryLock ();
};

} /* end of namespace epics :: legacy */

} /* end of namespace epics */

#endif /*__cplusplus*/

#ifdef __cplusplus
extern "C" {
    typedef struct epics :: Mutex :: Intf * epicsMutexId;
#else
    typedef struct EpicsMutexIntf * epicsMutexId;
#endif /*__cplusplus*/

typedef enum {
    epicsMutexLockOK,epicsMutexLockTimeout,epicsMutexLockError
} epicsMutexLockStatus;

#define epicsMutexCreate() epicsMutexOsiCreate(__FILE__,__LINE__)

epicsShareFunc epicsMutexId epicsShareAPI epicsMutexOsiCreate(
    const char *pFileName,int lineno);
#define epicsMutexMustCreate() epicsMutexOsiMustCreate(__FILE__,__LINE__)
epicsShareFunc epicsMutexId epicsShareAPI epicsMutexOsiMustCreate(
    const char *pFileName,int lineno);
epicsShareFunc void epicsShareAPI epicsMutexDestroy(epicsMutexId id);
epicsShareFunc void epicsShareAPI epicsMutexUnlock(epicsMutexId id);
epicsShareFunc epicsMutexLockStatus epicsShareAPI epicsMutexLock(
    epicsMutexId id);
#define epicsMutexMustLock(ID) {                        \
    epicsMutexLockStatus status = epicsMutexLock(ID);   \
    assert(status == epicsMutexLockOK);                 \
}
epicsShareFunc epicsMutexLockStatus epicsShareAPI epicsMutexTryLock(
    epicsMutexId id);
epicsShareFunc void epicsShareAPI epicsMutexShow(
    epicsMutexId id,unsigned  int level);
epicsShareFunc void epicsShareAPI epicsMutexShowAll(
    int onlyLocked,unsigned  int level);

/* 
 * NOTES (for library users):
 *  o pFileName passed to epicsMutexOsiMustCreate _must_ both exist, and
 *      remain constant, for the full lifetime of the mutex
 * NOTES (for OSD stub developers):
 *  o epicsMutex MUST implement recursive locking
 *  o epicsMutex should implement priority inheritance and deletion safe
 */

/* 
 * The following is the interface to the OS dependent 
 * implementation and should NOT be called directly by 
 * user code
 */
struct epicsMutexOSD * epicsMutexOsdCreate(void);
void epicsMutexOsdDestroy(struct epicsMutexOSD *);
void epicsMutexOsdUnlock(struct epicsMutexOSD *);
epicsMutexLockStatus epicsMutexOsdLock(struct epicsMutexOSD *);
epicsMutexLockStatus epicsMutexOsdTryLock(struct epicsMutexOSD *);
void epicsMutexOsdShow(struct epicsMutexOSD *,unsigned  int level);

#ifdef __cplusplus
}
#endif

#include "osdMutex.h"

#ifdef __cplusplus

namespace epics { 

inline Mutex :: 
    Mutex ( Factory & semFactory, 
        const SrcCodeLocation & srcCodeLoc ) : 
    _pIntf ( & semFactory.create ( srcCodeLoc ) )
{
    _pIntf->reference ();
}

inline Mutex :: 
    Mutex ( Mutex & mutex ) :
    _pIntf ( mutex._pIntf )
{
    _pIntf->reference ();
}

inline Mutex ::
    ~Mutex ()
{
    _pIntf->unreference ();
}

inline void Mutex :: 
    show ( unsigned level ) const
{
    _pIntf->show ( level );
}

inline bool Mutex :: 
    isIdentical ( const Mutex & mutex ) const
{
    return _pIntf == mutex._pIntf;
}

inline void Mutex :: 
    assertIdentical ( const Mutex & mutex ) const
{
#   ifndef NDEBUG
        if ( ! this->isIdentical ( mutex ) ) {
            Mutex :: _throwLockDomainMismatch ();
        }
#   endif
}

inline void Mutex :: 
    lock ()
{
    _pIntf->lock ();
}

inline void Mutex :: 
    lockRecursive ()
{
    _pIntf->lockRecursive ();
}

inline void Mutex :: 
    unlock ()
{
    _pIntf->unlock ();
}

inline void Mutex :: 
    mustUnlock ()
{
    _pIntf->mustUnlock ();
}

inline bool Mutex :: 
    tryLock ()
{
   return _pIntf->tryLock ();
}

inline Mutex :: Guard :: 
    Guard ( Mutex & mutex ) :
    _pMutex ( & mutex )
{
    mutex.lock ();
}   

inline Mutex :: Guard :: 
    Guard ( const Guard & guard ) :
    _pMutex ( guard._pMutex )
{
    if ( _pMutex == 0 ) {
        _throwReleaseInProgress ();
    }
    // this is a more efficent type of lock where the
    // lock reference count is increased, but the 
    // mutex lock function in the OS isnt called
    _pMutex->lockRecursive ();
}   

inline Mutex :: Guard :: 
    ~Guard ()
{
    // destroying a Guard while a Release is in 
    // progress is an exceptionally bad idea, but
    // throwing from within the destructor just 
    // makes the situation worse
    assert ( _pMutex );
    _pMutex->unlock ();
}

inline bool Mutex :: Guard :: 
    isIdenticalMutex ( const Mutex & mutex ) const
{
    if ( _pMutex == 0 ) {
        _throwReleaseInProgress ();
    }
    return ( _pMutex->_pIntf == mutex._pIntf );
}

inline void Mutex :: Guard :: 
    assertIdenticalMutex ( const Mutex & mutex ) const
{
#   ifndef NDEBUG
        if ( ! this->isIdenticalMutex ( mutex ) ) {
            Mutex :: _throwLockDomainMismatch ();
        }
#   endif
}

inline SrcCodeLocation :: 
    SrcCodeLocation ( const char * pFileName, int lineno ) :
    _pFileName ( pFileName ? pFileName : 
        "<nill ptr> supplied for File Name" ),
    _lineNumber ( lineno )
{
}

inline const char * SrcCodeLocation :: pFileName () const
{
    return _pFileName;
}

inline int SrcCodeLocation :: lineNumber () const
{
    return _lineNumber;
}

} /* end of namespace epics */

#endif /* __cplusplus */

#endif /* epicsMutexh */
