/*************************************************************************\
* Copyright (c) 2009 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to the Software License Agreement
* in the file LICENSE that is included with this distribution.
\*************************************************************************/
// Testing BaseString

#include <assert.h>
#include <stdio.h>

#include "BaseString.h"
#include "ConstString.h"
#include "FixedString.h"
#include "VariableString.h"
#include "StandardString.h"
#include "ChunkedString.h"

// Testing infrastructure

static int tests = 0;
static int passes = 0;
static int exceptions = 0;
static bool except = false;

// Macros to wrap, catch and report any exceptions

// Test where success is a true value
#define OK(test, desc) {	\
    bool res;			\
    try { res = test; } 	\
    catch (...) {		\
	except = true;		\
	res = false;		\
	printf("# Exception evaluating '%s'\n", #test);\
    }				\
    ok(res, desc); }

// Test where success requires an exception
#define EXCEPT(expr, desc) {	\
    bool res = false;		\
    try { expr; }		\
    catch (...) {		\
	res = true;		\
    }				\
    ok(res, desc); }

// Non-test expressions that might throw
#define TRAP(expr)		\
    try {			\
	expr;			\
	printf("#        %s\n", #expr);\
    }				\
    catch (...) {		\
	except = true;		\
	printf("# Exception executing '%s'\n", #expr);\
    }
#define QUIET(expr)		\
    try {			\
	expr;			\
    }				\
    catch (...) {		\
	except = true;		\
	printf("# Exception executing '%s'\n", #expr);\
    }

// Non-test expression
#define DO(expr)		\
    expr;			\
    printf("#        %s\n", #expr);

void ok(bool res, const char *desc) {
    if (res) {
	++tests; ++passes;
	printf("ok %3d - %s\n", tests, desc);
    } else {
	++tests;
	printf("not ok %d - %s\n", tests, desc);
    }
    if (except) {
	++exceptions;
	except = false;
	if (res) {
	    printf ("# test %d passed despite exception!\n", tests);
	}
    }
}

int results() {
    printf("\n# Passed %d of %d tests with %d unexpected exceptions.\n",
	   passes, tests, exceptions);
    int failures = tests - passes;
    if (failures)
    	printf("# %d test%s failed.\n", failures, failures>1 ? "s" : "");
    puts("");
    return (!failures && !exceptions) ? 0 : 1;
}

// Tests

void testReader(const char *className, const StringReader &rd, const char *buf) {
    printf("\n# Testing %s's StringReader\n", className);
    const size_t buflen = strlen(buf);
    OK(rd.size() == buflen, "rd.size()");
    bool identical = true;
    for (size_t i=0; i<buflen; i++) {
	QUIET(identical &= (rd[i] == buf[i]));
    }
    OK(identical, "rd[n]");
    EXCEPT(rd[buflen], "past-end exception");
    OK(rd.contains(0, buf, buflen), "rd.contains(buf)");
    OK(!rd.contains(0, "Not buf", 7), "!rd.contains(Not buf)");
    OK(rd.equals(rd), "rdequals(rd)");
    OK(rd == rd, "rd == rd");
    OK(!(rd != rd), "!(rd != rd)");
    OK(rd.contains(0, rd), "rd.contains(0, rd)");
    OK(!rd.contains(1, rd), "!rd.contains(1, rd)");
    char *tmp = new char[buflen + 1];
    tmp[buflen] = '*';
    OK(rd.extract(0, tmp, buflen) == buflen, "rd.extract() length");
    OK(!memcmp(tmp, buf, buflen), "rd.extract() contents");
    OK(tmp[buflen] == '*', "rd.extract() beyond");
    delete tmp;
}

void testEditor(const char *className, StringEditor &ed) {
    printf("\n# Testing %s's StringEditor\n", className);
    const size_t buflen = ed.size();
    assert(buflen > 0);
    char *tmp = new char[buflen];
    TRAP(ed.extract(0, tmp, buflen));
    TRAP(ed.resize(buflen-1));
    OK(ed.size() == buflen-1, "ed.resize(smaller)");
    OK(!ed.contains(0, tmp, buflen), "!ed.contains(buf)")
    TRAP(ed.append(tmp[buflen-1]));
    OK(ed[buflen-1] == tmp[buflen-1], "ed.append(char)");
    TRAP(ed.resize(buflen-1));
    TRAP(ed.append(&tmp[buflen-1], 1));
    OK(ed[buflen-1] == tmp[buflen-1], "ed.append(buf)");
    TRAP(ed.resize(buflen-1));
    TRAP(ed.resize(buflen, tmp[buflen-1]));
    OK(ed.size() == buflen, "ed.resize(larger) size");
    OK(ed.contains(0, tmp, buflen), "ed.resize(larger) content");
    TRAP(ed = "Greetings");
    OK(ed.contains(0, "Greetings", 9), "ed = str");
    OK(ed.contains(4, "ting", 4), "ed.contains(sub-str)");
    OK(!ed.contains(7, "gs", 3), "!ed.contains(spill)");
    TRAP(ed += ',');
    OK(ed[9] == ',', "ed += char");
    TRAP(ed += " friend");
    OK(ed.contains(10, " friend", 7), "ed += str");
    TRAP(ed.assign(tmp, buflen));
    OK(ed.contains(0, tmp, buflen), "ed.assign(buf)");
    TRAP(ed.assign(ed));
    OK(ed.contains(0, tmp, buflen), "ed.assign(ed)");
    TRAP(ed.reserve(1));
    OK(ed.capacity() >= ed.size(), "ed.reserve(smaller)");
    TRAP(ed.reserve(buflen*2));
    OK(ed.capacity() >= buflen*2, "ed.reserve(larger)");
    TRAP(ed.append(ed));
    OK(ed.size() == buflen*2, "ed.append(ed) size");
    OK(ed.contains(0, tmp, buflen) && ed.contains(buflen, tmp, buflen),
       "ed.append(ed) content");
    TRAP(ed.resize(ed.capacity()));
    OK(ed.size() == ed.capacity(), "ed.resize(capacity)");
    ed.resize(buflen);
    delete tmp;
}

void testPair(const char *classes, const StringReader &rd, StringEditor &ed) {
    printf("\n# Testing %s\n", classes);
    size_t rlen = rd.size();
    assert(rlen > 0);

    // Assign, then check comparisons
    TRAP(ed = rd);
    OK(ed.equals(rd), "ed.equals(rd)");
    OK(rd.equals(ed), "rd.equals(ed)");
    OK(ed == rd, "ed == rd");
    OK(rd == ed, "rd == ed");
    OK(!(ed != rd), "!(ed != rd)");
    OK(!(rd != ed), "!(rd != ed)");
    OK(ed.contains(0, rd), "ed.contains(0, rd)");
    OK(rd.contains(0, ed), "rd.contains(0, ed)");

    // Remove last char of ed, check comparisons again
    TRAP(ed.resize(rlen - 1));
    OK(!ed.equals(rd), "!ed.equals(rd)");
    OK(!rd.equals(ed), "!rd.equals(ed)");
    OK(ed != rd, "ed != rd");
    OK(rd != ed, "rd != ed");
    OK(!(ed == rd), "!(ed == rd)");
    OK(!(rd == ed), "!(rd == ed)");
    OK(!ed.contains(0, rd), "!ed.contains(0, rd)");
    OK(rd.contains(0, ed), "rd.contains(0, ed)");

    // Replace last char but with \0, check again
    TRAP(ed.resize(rlen));
    OK(!ed.equals(rd), "!ed.equals(rd)");
    OK(!rd.equals(ed), "!rd.equals(ed)");
    OK(ed != rd, "ed != rd");
    OK(rd != ed, "rd != ed");
    OK(!(ed == rd), "!(ed == rd)");
    OK(!(rd == ed), "!(rd == ed)");
    OK(!ed.contains(0, rd), "!ed.contains(0, rd)");
    OK(!rd.contains(0, ed), "!rd.contains(0, ed)");

    TRAP(ed.assign(rd));
    OK(ed == rd, "ed == rd");
    TRAP(ed.assign(rd, 5));
    OK(ed.size() == rlen-5, "ed.size() == rlen-5");
    TRAP(ed.append(rd));
    OK(ed.size() == rlen*2 - 5, "ed.size() == rlen*2 - 5");
    OK(ed != rd, "ed != rd");

    TRAP(ed.resize(1));
    TRAP(ed += rd);
    OK(ed.size() == rlen + 1, "ed += rd");
}

#define FS_SIZE 80

int main() {
    puts("\n# Testing copy and assignment operators");
    const char *greeting = "Hello, World!";

    // ConstString constructors store the pointer
    DO(ConstString cs1 = greeting);
    OK(cs1.c_str() == greeting, "ConstString(const char *)");
    DO(ConstString cs2 = cs1);
    OK(cs2.c_str() == greeting, "ConstString(ConstString&)");

    // VariableString constructors & assignment should not copy the pointer
    DO(VariableString vs1 = greeting);
    OK(vs1.c_str() != greeting, "VariableString(const char *)");
    DO(VariableString vs2 = vs1);
    OK(vs2.c_str() != vs1.c_str(), "VariableString(VariableString&)");
    DO(vs2.resize(0));
    DO(vs1.swap(vs2));
    OK(vs1.size() == 0, "VariableString.swap(VariableString)");
    DO(vs1 = vs2);
    OK(vs2.c_str() != vs1.c_str(), "VariableString = VariableString");

    // FixedString constructors & assignment should not copy the pointer
    DO(FixedString<FS_SIZE> fs1 = greeting);
    OK(fs1.c_str() != greeting, "FixedString(const char *)");
    DO(FixedString<FS_SIZE> fs2 = fs1);
    OK(fs1.c_str() != fs2.c_str(), "FixedString(FixedString&)");
    DO(FixedString<FS_SIZE> fs3);
    OK(fs3.size() == 0, "FixedString()");
    DO(fs3 = fs1);
    OK(fs3.c_str() != fs1.c_str(), "FixedString = FixedString");

    // StandardString constructor should not copy the pointer
    DO(StandardString ss1 = greeting);
    OK(ss1.c_str() != greeting, "StandardString(const char *)");

    // std::string may use reference counting internally, so we don't
    // care whether its copy constructor copies the pointer
    DO(StandardString ss2 = ss1);
    OK(!strcmp(ss1.c_str(), ss2.c_str()), "StandardString(StandardString&)");
    DO(ss2.resize(0));
    DO(ss1.swap(ss2));
    OK(ss1.size() == 0, "StandardString.swap(StandardString)");
    DO(ss1 = ss2);
    OK(ss1.std_str() == ss2.std_str(), "StandardString = StandardString");

    // ChunkedString doesn't support c_str()
    DO(ChunkedString ks1 = greeting);
    DO(ChunkedString ks2 = cs1);
    OK(ks1 == cs1, "ChunkedString = ConstString");

    testReader("ConstString", cs1, greeting);
    testReader("VariableString", vs1, greeting);
    testEditor("VariableString", vs1);
    testReader("FixedString", fs1, greeting);
    testEditor("FixedString", fs1);
    testReader("StandardString", ss1, greeting);
    testEditor("StandardString", ss1);
    testReader("ChunkedString", ks1, greeting);
    testEditor("ChunkedString", ks1);
    testPair("rd=ConstString, ed=VariableString", cs1, vs1);
    testPair("rd=ConstString, ed=FixedString", cs1, fs1);
    testPair("rd=ConstString, ed=StandardString", cs1, ss1);
    testPair("rd=ConstString, ed=ChunkedString", cs1, ks1);
    testPair("rd=VariableString, ed=VariableString", vs1, vs2);
    testPair("rd=VariableString, ed=FixedString", vs1, fs2);
    testPair("rd=VariableString, ed=StandardString", vs1, ss2);
    testPair("rd=VariableString, ed=ChunkedString", vs1, ks2);
    testPair("rd=FixedString, ed=VariableString", fs1, vs1);
    testPair("rd=FixedString, ed=FixedString", fs1, fs2);
    testPair("rd=FixedString, ed=StandardString", fs1, ss2);
    testPair("rd=FixedString, ed=ChunkedString", fs1, ks2);
    testPair("rd=StandardString, ed=VariableString", ss1, vs1);
    testPair("rd=StandardString, ed=FixedString", ss1, fs2);
    testPair("rd=StandardString, ed=StandardString", ss1, ss2);
    testPair("rd=StandardString, ed=ChunkedString", ss1, ks2);
    testPair("rd=ChunkedString, ed=VariableString", ks1, vs1);
    testPair("rd=ChunkedString, ed=FixedString", ks1, fs2);
    testPair("rd=ChunkedString, ed=StandardString", ks1, ss2);
    testPair("rd=ChunkedString, ed=ChunkedString", ks1, ks2);

    return results();
}
