/*************************************************************************\
* Copyright (c) 2008 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* EPICS BASE Versions 3.13.7 and higher are distributed subject to the
* Software License Agreement in the file LICENSE that is included with
* this distribution.
\*************************************************************************/
//	Author: Andrew Johnson

// TestSegString.h

#ifndef INC_TestSegString_H
#define INC_TestSegString_H

#include <stdexcept>
#include <cassert>

#include "SegString.h"

// This test implementation of a segmented string uses a fixed buffer,
// with the segments being allocated in reverse order.
// All character operations are done by the SegStringEditor routines,
// since that's what we're trying to test.

template <int COUNT, int SIZE>
class TestSegString :
    public SegStringEditor {
private:
    // Data
    char buf_[COUNT][SIZE];   // Storage
    size_t len_;    // String length

    // Private methods for local class use
    virtual bool getSegment(const char *&buf, size_t &len, size_t pos) const;
    virtual bool getSegment(char *&buf, size_t &len, size_t pos);
    virtual void setSize(size_t len);

public:
    // Constructors
    TestSegString();
    TestSegString(const StringReader &rhs);
    TestSegString(const char *str, size_t len = 0);

// StringReader virtual functions
    // String length
    virtual size_t size(void) const;

// StringEditor virtual functions
    // Buffer storage management
    virtual size_t capacity() const;
    virtual void reserve(size_t cap);
};


// Inline member functions

// Private methods for base class use
template <int COUNT, int SIZE>
inline bool TestSegString<COUNT, SIZE>::getSegment(const char *&buf, size_t &len, size_t pos) const {
    if (pos >= capacity()) return false;
    size_t off = pos % SIZE;
    buf = & buf_[COUNT - 1 - pos / SIZE][off];
    len = SIZE - off;
    return true;
}
template <int COUNT, int SIZE>
inline bool TestSegString<COUNT, SIZE>::getSegment(char *&buf, size_t &len, size_t pos) {
    if (pos >= capacity()) return false;
    size_t off = pos % SIZE;
    buf = & buf_[COUNT - 1 - pos / SIZE][off];
    len = SIZE - off;
    return true;
}
template <int COUNT, int SIZE>
inline void TestSegString<COUNT, SIZE>::setSize(size_t len) {
    assert(len <= capacity());
    len_ = len;
}

// Constructors
template <int COUNT, int SIZE>
inline TestSegString<COUNT, SIZE>::TestSegString() :
    len_(0) {}
template <int COUNT, int SIZE>
inline TestSegString<COUNT, SIZE>::TestSegString(const StringReader &rhs) :
    len_(0) {
    assign(rhs);
}
template <int COUNT, int SIZE>
inline TestSegString<COUNT, SIZE>::TestSegString(const char *str, size_t len) :
    len_(0) {
    if (!str) throw std::invalid_argument("FIXME");
    if (!len) len = strlen(str);
    assign(str, len);
}

// String length
template <int COUNT, int SIZE>
inline size_t TestSegString<COUNT, SIZE>::size(void) const {
    return len_;
}

// Buffer storage management
template <int COUNT, int SIZE>
inline size_t TestSegString<COUNT, SIZE>::capacity() const {
    return COUNT * SIZE;
}
template <int COUNT, int SIZE>
inline void TestSegString<COUNT, SIZE>::reserve(size_t cap) {
    if (cap > capacity()) throw std::length_error("Too Long");
}

#endif /* INC_TestSegString_H */
