/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#include <string>
#include <stdexcept>

#define epicsExportSharedSymbols
#include "epicsTimer.h"
#include "timerPrivate.h"

void epicsTimerNotify :: show ( unsigned /* level */ ) const {}

timerQueue :: TimerForC :: TimerForC ( 
    Guard & guard, timerQueue & queue, 
    epicsTimerCallback pCBIn, void *pPrivateIn ) :
    _pTimer ( & queue.createTimer ( guard ) ), 
    _pCallBack ( pCBIn ), 
    _pPrivate ( pPrivateIn )
{
}

timerQueue :: TimerForC :: ~TimerForC ()
{
    if ( _pTimer ) {
        _pTimer->destroy ();
    }
}

void timerQueue :: TimerForC :: cleanup ( Guard & guard )
{
    if ( _pTimer ) {
        _pTimer->destroy ( guard );
        _pTimer = 0;
    }
}

void timerQueue :: TimerForC :: show ( unsigned level ) const
{
    printf ( 
        "TimerForC: callback ptr %p private ptr %p\n", 
        _pCallBack, _pPrivate );
    if ( level > 1 && _pTimer ) {
        _pTimer->show ( level - 1 );
    }
}

epicsTimerNotify :: expireStatus 
    timerQueue :: TimerForC :: expire ( const epicsTime & )
{
    ( *_pCallBack ) ( _pPrivate );
    return noRestart;
}

epicsTimerQueueActiveForC ::
    epicsTimerQueueActiveForC ( RefMgr & refMgr, 
        bool okToShare, unsigned priority ) :
    timerQueueActive ( refMgr, okToShare, priority )
{
}

epicsTimerQueueActiveForC :: ~epicsTimerQueueActiveForC ()
{
}

void epicsTimerQueueActiveForC :: release ()
{
    _refMgr->release ( *this );
}

epicsTimerQueuePassiveForC :: epicsTimerQueuePassiveForC ( 
    epicsTimerQueueNotifyReschedule pRescheduleCallbackIn, 
    epicsTimerQueueNotifyQuantum pSleepQuantumCallbackIn,
    void * pPrivateIn ) :
        timerQueuePassive ( 
            * static_cast < epicsTimerQueueNotify * > ( this ) ), 
        pRescheduleCallback ( pRescheduleCallbackIn ), 
        pSleepQuantumCallback ( pSleepQuantumCallbackIn ),
        pPrivate ( pPrivateIn )
{
}

epicsTimerQueuePassiveForC :: ~epicsTimerQueuePassiveForC () 
{
}

void epicsTimerQueuePassiveForC :: reschedule ()
{
    (*this->pRescheduleCallback) ( this->pPrivate );
}

double epicsTimerQueuePassiveForC :: quantum ()
{
    return (*this->pSleepQuantumCallback) ( this->pPrivate );
}

void epicsTimerQueuePassiveForC :: destroy ()
{
    delete this;
}

epicsShareFunc epicsTimerNotify :: expireStatus :: expireStatus ( 
    restart_t restart ) : 
    delay ( - DBL_MAX )
{
    if ( restart != noRestart ) {
        throw std::logic_error 
            ( "timer restart was requested "
            "without specifying a delay?" );
    }
}

epicsShareFunc epicsTimerNotify :: expireStatus :: expireStatus 
    ( restart_t restartIn, const double & expireDelaySec ) :
    delay ( expireDelaySec )
{
    if ( restartIn != epicsTimerNotify::restart ) {
        throw std::logic_error 
            ( "no timer restart was requested, "
            "but a delay was specified?" );
    }
    if ( this->delay < 0.0 ) {
        throw std::logic_error 
            ( "timer restart was requested, but a "
            "negative delay was specified?" );
    }
}

epicsShareFunc bool epicsTimerNotify :: expireStatus :: 
    restart () const
{
    return this->delay >= 0.0;
}

epicsShareFunc double epicsTimerNotify :: expireStatus :: 
    expirationDelay () const
{
    if ( this->delay < 0.0 ) {
        throw std::logic_error 
            ( "no timer restart was requested, "
                "but you are asking for a restart delay?" );
    }
    return this->delay;
}

extern "C" epicsTimerQueuePassiveId epicsShareAPI
    epicsTimerQueuePassiveCreate ( 
        epicsTimerQueueNotifyReschedule pRescheduleCallbackIn, 
        epicsTimerQueueNotifyQuantum pSleepQuantumCallbackIn,
        void * pPrivateIn )
{
    try {
        return new epicsTimerQueuePassiveForC ( 
            pRescheduleCallbackIn, 
            pSleepQuantumCallbackIn,
            pPrivateIn );
    }
    catch ( ... ) {
        return 0;
    }
}

extern "C" void epicsShareAPI 
    epicsTimerQueuePassiveDestroy ( 
        epicsTimerQueuePassiveId pQueue )
{
    pQueue->destroy ();
}

extern "C" double epicsShareAPI 
    epicsTimerQueuePassiveProcess ( 
        epicsTimerQueuePassiveId pQueue )
{
    return pQueue->process ( epicsTime::getCurrent() );
}

extern "C" epicsTimerId epicsShareAPI epicsTimerQueuePassiveCreateTimer (
    epicsTimerQueuePassiveId pQueue, 
    epicsTimerCallback pCallback, void *pArg )
{
    try {
        return & pQueue->createTimerForC ( pCallback, pArg );
    }
    catch ( ... ) {
        return 0;
    }
}

extern "C" epicsShareFunc void epicsShareAPI 
    epicsTimerQueuePassiveDestroyTimer (   
        epicsTimerQueuePassiveId pQueue, epicsTimerId pTmr )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    pQueue->destroyTimerForC ( tmr );
}

extern "C" void  epicsShareAPI epicsTimerQueuePassiveShow (
    epicsTimerQueuePassiveId pQueue, unsigned int level )
{
    pQueue->show ( level );
}

extern "C" epicsTimerQueueId epicsShareAPI
    epicsTimerQueueAllocate ( int okToShare, unsigned int threadPriority )
{
    try {
        epicsSingleton < timerQueueActiveMgr > :: reference 
            ref = timerQueueMgrEPICS.getReference ();
        epicsTimerQueueActiveForC & tmr = 
            ref->allocate ( ref, okToShare ? true : false, threadPriority );
        return &tmr;
    }
    catch ( ... ) {
        return 0;
    }
}

extern "C" void epicsShareAPI epicsTimerQueueRelease ( epicsTimerQueueId pQueue )
{
    pQueue->release ();
}

extern "C" epicsTimerId epicsShareAPI 
    epicsTimerQueueCreateTimer (
        epicsTimerQueueId pQueue, 
        epicsTimerCallback pCallback, void *pArg )
{
    try {
        return & pQueue->createTimerForC ( pCallback, pArg );
    }
    catch ( ... ) {
        return 0;
    }
}

extern "C" void  epicsShareAPI 
    epicsTimerQueueShow (
        epicsTimerQueueId pQueue, unsigned int level )
{
    pQueue->show ( level );
}

extern "C" void epicsShareAPI 
    epicsTimerQueueDestroyTimer ( 
        epicsTimerQueueId pQueue, epicsTimerId pTmr )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    pQueue->destroyTimerForC ( tmr );
}

extern "C" void  epicsShareAPI 
    epicsTimerStartTime ( 
        epicsTimerId pTmr, const epicsTimeStamp *pTime )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    tmr.start ( *pTime );
}

extern "C" void  epicsShareAPI 
    epicsTimerStartDelay ( 
        epicsTimerId pTmr, double delaySeconds )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    tmr.start ( delaySeconds );
}

extern "C" void  epicsShareAPI 
    epicsTimerCancel ( epicsTimerId pTmr )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    tmr.cancel ();
}

extern "C" double  epicsShareAPI 
    epicsTimerGetExpireDelay ( epicsTimerId pTmr )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    return tmr.getExpireDelay ();
}

extern "C" void  epicsShareAPI 
    epicsTimerShow ( epicsTimerId pTmr, unsigned int level )
{
    timerQueue :: TimerForC & tmr ( 
        * reinterpret_cast 
            < timerQueue :: TimerForC * > ( pTmr ) );
    tmr.show ( level );
}

