/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/

/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 *
 */

#include <typeinfo>
#include <string>
#include <stdexcept>
#include <cstdio>

#define epicsExportSharedSymbols
#include "timerPrivate.h"
#include "errlog.h"

timerQueue :: Timer :: Timer ( timerQueue & queueIn ) :
    queue ( queueIn ), curState ( stateLimbo ), pNotify ( 0 ) 
{
}

timerQueue :: Timer :: ~Timer ()
{
    if ( this->curState == statePending || 
            this->curState == stateActive ) {
        this->cancel ();
    }
}

void timerQueue :: Timer :: destroy ( Guard & guard )
{
    guard.assertIdenticalMutex ( this->queue.mutex );
    this->cancel ( guard );
    this->queue.timerFreeList.deleteProduct ( *this );
}

void timerQueue :: Timer :: destroy () 
{
    Guard guard ( this->queue.mutex );
    this->destroy ( guard );
}

void timerQueue :: Timer :: start ( 
    epicsTimerNotify & notify, double delaySeconds )
{
    this->start ( notify, 
        epicsTime::getCurrent () + delaySeconds );
}

void timerQueue :: Timer :: start ( 
    epicsTimerNotify & notify, const epicsTime & expire )
{
    Guard locker ( this->queue.mutex );
    this->privateStart ( notify, expire );
}

void timerQueue :: Timer :: privateStart ( 
    epicsTimerNotify & notify, const epicsTime & expire )
{
    this->pNotify = & notify;
    this->exp = expire - ( this->queue.notify.quantum () / 2.0 );

    bool reschedualNeeded = false;
    if ( this->curState == stateActive ) {
        // above expire time and notify will override 
        // any restart parameters that may be returned 
        // from the timer expire callback
        return;
    }
    else if ( this->curState == statePending ) {
        this->queue.timerList.remove ( *this );
        if ( this->queue.timerList.first() == this && 
                this->queue.timerList.count() > 0 ) {
            reschedualNeeded = true;
        }
    }

#   ifdef DEBUG
        unsigned preemptCount=0u;
#   endif

    //
    // insert into the pending queue
    //
    // Finds proper time sorted location using a linear search.
    //
    // **** this should use a binary tree ????
    //
    tsDLIter < Timer > pTmr = this->queue.timerList.lastIter ();
    while ( true ) {
        if ( ! pTmr.valid () ) {
            //
            // add to the beginning of the list
            //
            this->queue.timerList.push ( *this );
            reschedualNeeded = true;
            break;
        }
        if ( pTmr->exp <= this->exp ) {
            //
            // add after the item found that expires earlier
            //
            this->queue.timerList.insertAfter ( *this, *pTmr );
            break;
        }
#       ifdef DEBUG
            preemptCount++;
#       endif
        --pTmr;
    }

    this->curState = Timer::statePending;

    if ( reschedualNeeded ) {
        this->queue.notify.reschedule ();
    }

    debugPrintf ( ("Start of \"%s\" with delay %f "
        "at %p preempting %u\n", 
        typeid ( this->notify ).name (), 
        expire - epicsTime::getCurrent (), 
        this, preemptCount ) );
}

void timerQueue :: Timer :: cancel ()
{
    Guard guard ( this->queue.mutex );
    this->cancel ( guard );
}

void timerQueue :: Timer :: cancel ( Guard & guard )
{
    guard.assertIdenticalMutex ( this->queue.mutex );
    this->pNotify = 0;
    if ( this->curState == statePending ) {
        this->queue.timerList.remove ( *this );
        this->curState = stateLimbo;
        if ( this->queue.timerList.first() == this && 
                this->queue.timerList.count() > 0 ) {
            this->queue.notify.reschedule ();
        }
    }
    else if ( this->curState == stateActive ) {
        this->queue.cancelPending = true;
        this->curState = Timer::stateLimbo;
        if ( this->queue.processThread != 
                epicsThreadGetIdSelf() ) {
            // make certain timer expire() does not 
            // run after cancel () returns,
            // but dont require that lock is applied 
            // while calling expire()
            while ( this->queue.cancelPending && 
                    this->queue.pExpireTmr == this ) {
                Guard :: Release unguard ( guard );
                this->queue.cancelBlockingEvent.wait ();
            }
            // in case other threads are waiting
            this->queue.cancelBlockingEvent.signal ();
        }
    }
}

epicsTimer :: expireInfo 
    timerQueue :: Timer :: getExpireInfo () const
{
    // taking a lock here guarantees that users will not 
    // see brief intervals when a timer isnt active because
    // it is is canceled when start is called
    Guard locker ( this->queue.mutex );
    if ( this->curState == statePending || 
            this->curState == stateActive ) {
        return expireInfo ( true, this->exp );
    }
    return expireInfo ( false, epicsTime() );
}

void timerQueue :: Timer :: show ( unsigned int level ) const
{
    Guard locker ( this->queue.mutex );
    double delay;
    if ( this->curState == statePending || 
            this->curState == stateActive ) {
        try {
            delay = this->exp - epicsTime::getCurrent();
        }
        catch ( ... ) {
            delay = - DBL_MAX;
        }
    }
    else {
        delay = -DBL_MAX;
    }
    const char *pStateName;
    if ( this->curState == statePending ) {
        pStateName = "pending";
    }
    else if ( this->curState == stateActive ) {
        pStateName = "active";
    }
    else if ( this->curState == stateLimbo ) {
        pStateName = "limbo";
    }
    else {
        pStateName = "corrupt";
    }
    printf ( "timerQueue :: Timer, state = %s, delay = %f\n",
        pStateName, delay );
    if ( level >= 1u && this->pNotify ) {
        this->pNotify->show ( level - 1u );
    }
}
