/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *
 *      Author  Jeffrey O. Hill
 *              johill@lanl.gov
 *              505 665 1831
 */

#ifndef epicsTimerPrivate_h
#define epicsTimerPrivate_h

#include <typeinfo> 

#include "tsFreeList.h"
#include "epicsSingleton.h"
#include "tsDLList.h"
#include "epicsTimer.h"
#include "compilerDependencies.h"

#ifdef DEBUG
#   define debugPrintf(ARGSINPAREN) printf ARGSINPAREN
#else
#   define debugPrintf(ARGSINPAREN)
#endif

template < class T > class epicsGuard;
typedef epics :: Mutex :: Guard Guard;

using std :: type_info; 

class timerQueue : public epicsTimerQueue {
public:

    class Timer : 
        public epicsTimer, 
        public tsDLNode < Timer > {
    public:
        Timer ( class timerQueue & );
        ~Timer ();
        void destroy ();
        void destroy ( Guard & );
        void start ( class epicsTimerNotify &, 
                const epicsTime & );
        void start ( class epicsTimerNotify &, 
                double delaySeconds );
        void cancel ();
        expireInfo getExpireInfo () const;
        void show ( unsigned int level ) const;
    protected:
        timerQueue & queue;
    private:
        enum state { 
            statePending = 45, 
            stateActive = 56, 
            stateLimbo = 78 };
        epicsTime exp; // experation time 
        state curState; // current state 
        epicsTimerNotify * pNotify; // callback
        void privateStart ( 
            epicsTimerNotify & notify, const epicsTime & );
        void cancel ( Guard & guard );
        Timer & operator = ( const Timer & );
        friend class timerQueue;
    };

    class TimerForC : 
        public epicsTimerNotify {
    public:
        TimerForC ( Guard & guard, timerQueue &, 
            epicsTimerCallback, void *pPrivateIn );
        ~TimerForC ();
        void cleanup ( Guard & guard );
        void start ( const epicsTime & );
        void start ( double delaySeconds );
        void cancel ();
        void show ( unsigned level ) const;
        double getExpireDelay ();
    private:
        Timer * _pTimer;
        epicsTimerCallback _pCallBack;
        void * _pPrivate;
        expireStatus expire ( const epicsTime & currentTime );
        TimerForC & operator = ( const TimerForC & );
    };

    timerQueue ( epicsTimerQueueNotify &notify );
    virtual ~timerQueue ();
    epicsTimer & createTimer ();
    TimerForC & createTimerForC (
        epicsTimerCallback pCallback, void *pArg );
    void destroyTimerForC ( TimerForC & );
    double process ( const epicsTime & currentTime );
    void show ( unsigned int level ) const;
private:
    mutable epics :: Mutex mutex; // init first
    tsFreeList < Timer, 0x20 > timerFreeList;
    tsFreeList < TimerForC, 0x20 > timerForCFreeList;
    epicsEvent cancelBlockingEvent;
    tsDLList < Timer > timerList;
    epicsTime exceptMsgTimeStamp; 
    epicsTimerQueueNotify & notify;
    Timer * pExpireTmr;
    epicsThreadId processThread;
    bool cancelPending;
    static const double exceptMsgMinPeriod;
    void printExceptMsg ( const char * pName,
                    const type_info & type ); 
    Timer & createTimer ( Guard & );
	timerQueue ( const timerQueue & );
    timerQueue & operator = ( const timerQueue & );
#   if __GNUC__ == 2 && _GNUC_MINOR__ <= 96
        friend class Timer;
        friend class TimerForC;
#   endif
};

class timerQueueActiveMgr;

class timerQueueActiveMgrPrivate { // X aCC 655
public:
    timerQueueActiveMgrPrivate ();
protected:
    virtual ~timerQueueActiveMgrPrivate () = 0;
private:
    unsigned referenceCount;
    friend class timerQueueActiveMgr;
};

class timerQueueActive : 
    public epicsTimerQueueActive, 
    public epicsThreadRunable, 
    public epicsTimerQueueNotify,
    public timerQueueActiveMgrPrivate {
public:
    typedef epicsSingleton < timerQueueActiveMgr > :: reference RefMgr;
    timerQueueActive ( RefMgr &, bool okToShare, unsigned priority );
    epicsTimer & createTimer ();
    timerQueue :: TimerForC & createTimerForC ( 
        epicsTimerCallback pCallback, void *pArg );
    void destroyTimerForC ( timerQueue :: TimerForC & );
    void show ( unsigned int level ) const;
    bool sharingOK () const;
    unsigned threadPriority () const;
protected:
    ~timerQueueActive ();
    RefMgr _refMgr;
private:
    timerQueue queue;
    epicsEvent rescheduleEvent;
    epicsEvent exitEvent;
    epicsThread thread;
    const double sleepQuantum;
    bool okToShare;
    bool exitFlag;
    bool terminateFlag;
    void run ();
    void reschedule ();
    double quantum ();
    epicsTimerQueue & getEpicsTimerQueue ();
	timerQueueActive ( const timerQueueActive & );
    timerQueueActive & operator = ( const timerQueueActive & );
};

class timerQueueActiveMgr {
public:
    typedef epicsSingleton < timerQueueActiveMgr > :: reference RefThis;
	timerQueueActiveMgr ();
    ~timerQueueActiveMgr ();
    epicsTimerQueueActiveForC & allocate ( RefThis &, bool okToShare, 
        unsigned threadPriority = epicsThreadPriorityMin + 10 );
    void release ( epicsTimerQueueActiveForC & );
private:
    epics :: Mutex mutex;
    tsDLList < epicsTimerQueueActiveForC > sharedQueueList;
	timerQueueActiveMgr ( const timerQueueActiveMgr & );
    timerQueueActiveMgr & operator = ( const timerQueueActiveMgr & );
};

extern epicsSingleton < timerQueueActiveMgr > timerQueueMgrEPICS;

class timerQueuePassive : public epicsTimerQueuePassive {
public:
    timerQueuePassive ( epicsTimerQueueNotify & );
    epicsTimer & createTimer ();
    timerQueue :: TimerForC & createTimerForC ( 
        epicsTimerCallback pCallback, void *pArg );
    void destroyTimerForC ( timerQueue :: TimerForC & );
    void show ( unsigned int level ) const;
    double process ( const epicsTime & currentTime );
protected:
    timerQueue queue;
    ~timerQueuePassive ();
    epicsTimerQueue & getEpicsTimerQueue ();
	timerQueuePassive ( const timerQueuePassive & );
    timerQueuePassive & operator = ( const timerQueuePassive & );
};

struct epicsTimerQueuePassiveForC : 
    public epicsTimerQueueNotify, public timerQueuePassive {
public:
    epicsTimerQueuePassiveForC ( 
        epicsTimerQueueNotifyReschedule, 
        epicsTimerQueueNotifyQuantum,
        void * pPrivate );
    void destroy ();
protected:
    ~epicsTimerQueuePassiveForC ();
private:
    epicsTimerQueueNotifyReschedule pRescheduleCallback;
    epicsTimerQueueNotifyQuantum pSleepQuantumCallback;
    void * pPrivate;
    static epicsSingleton < tsFreeList < epicsTimerQueuePassiveForC, 0x10 > > pFreeList;
    void reschedule ();
    double quantum ();
};

struct epicsTimerQueueActiveForC : public timerQueueActive, 
    public tsDLNode < epicsTimerQueueActiveForC > {
public:
    epicsTimerQueueActiveForC ( RefMgr &, bool okToShare, unsigned priority );
    void release ();
protected:
    virtual ~epicsTimerQueueActiveForC ();
private:
	epicsTimerQueueActiveForC ( const epicsTimerQueueActiveForC & );
    epicsTimerQueueActiveForC & operator = ( const epicsTimerQueueActiveForC & );
};

inline bool timerQueueActive::sharingOK () const
{
    return this->okToShare;
}

inline unsigned timerQueueActive::threadPriority () const
{
    return thread.getPriority ();
}

inline void timerQueue :: TimerForC :: 
    start ( const epicsTime & expTime )
{
    _pTimer->start ( *this, expTime );
}

inline void timerQueue :: TimerForC :: 
    start ( double delaySeconds )
{
    _pTimer->start ( *this, delaySeconds );
}

inline void timerQueue :: TimerForC :: cancel ()
{
    _pTimer->cancel ();
}

inline double timerQueue :: TimerForC :: getExpireDelay ()
{
    return _pTimer->getExpireDelay ();
}

inline void timerQueueActive :: destroyTimerForC ( 
    timerQueue :: TimerForC & tmr )
{
    queue.destroyTimerForC ( tmr );
}

inline void timerQueuePassive :: destroyTimerForC ( 
    timerQueue :: TimerForC & tmr )
{
    queue.destroyTimerForC ( tmr );
}

#endif // epicsTimerPrivate_h

